\echo Use "ALTER EXTENSION pg_profile UPDATE" to load this file. \quit
DROP FUNCTION create_server;
DROP FUNCTION create_server_partitions;
DROP FUNCTION drop_server(name);
DROP FUNCTION export_data;
DROP FUNCTION get_report_context;
DROP FUNCTION get_report_datasets;
DROP FUNCTION import_data;
DROP FUNCTION import_section_data_profile;
DROP FUNCTION init_sample;
DROP FUNCTION profile_checkavail_walstats;
DROP FUNCTION report_queries_format;
DROP FUNCTION take_sample(integer, boolean);
DROP FUNCTION settings_format;
DROP FUNCTION settings_format_diff;
CREATE FUNCTION create_server(IN server name, IN server_connstr text, IN server_enabled boolean = TRUE,
IN max_sample_age integer = NULL, IN description text = NULL) RETURNS integer SET search_path=@extschema@ AS $$
DECLARE
    server_exists     integer;
    sserver_id        integer;
BEGIN

    SELECT count(*) INTO server_exists FROM servers WHERE server_name=server;
    IF server_exists > 0 THEN
        RAISE 'Server already exists.';
    END IF;

    INSERT INTO servers(server_name,server_description,connstr,enabled,max_sample_age)
    VALUES (server,description,server_connstr,server_enabled,max_sample_age)
    RETURNING server_id INTO sserver_id;

    -- Subsample settings table entry
    INSERT INTO server_subsample(
      server_id,
      subsample_enabled)
    VALUES (
      sserver_id,
      true);

    /*
    * We might create server sections to avoid concurrency on tables
    */
    PERFORM create_server_partitions(sserver_id);

    RETURN sserver_id;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION create_server(IN server name, IN server_connstr text, IN server_enabled boolean,
IN max_sample_age integer, IN description text) IS 'Create a new server';
CREATE FUNCTION create_server_partitions(IN sserver_id integer) RETURNS integer
SET search_path=@extschema@ AS $$
DECLARE
    in_extension      boolean;
BEGIN
    -- Create last_stat_statements table partition
    EXECUTE format(
      'CREATE TABLE last_stat_statements_srv%1$s PARTITION OF last_stat_statements '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    -- PK constraint for new partition
    EXECUTE format(
      'ALTER TABLE last_stat_statements_srv%1$s '
      'ADD CONSTRAINT pk_last_stat_satements_srv%1$s PRIMARY KEY (server_id, sample_id, userid, datid, queryid, toplevel)',
      sserver_id);

    -- Create last_stat_kcache table partition
    EXECUTE format(
      'CREATE TABLE last_stat_kcache_srv%1$s PARTITION OF last_stat_kcache '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_kcache_srv%1$s '
      'ADD CONSTRAINT pk_last_stat_kcache_srv%1$s PRIMARY KEY (server_id, sample_id, datid, userid, queryid, toplevel), '
      'ADD CONSTRAINT fk_last_kcache_stmts_srv%1$s FOREIGN KEY '
        '(server_id, sample_id, datid, userid, queryid, toplevel) REFERENCES '
        'last_stat_statements_srv%1$s(server_id, sample_id, datid, userid, queryid, toplevel) '
        'ON DELETE CASCADE',
      sserver_id);

    -- Create last_stat_database table partition
    EXECUTE format(
      'CREATE TABLE last_stat_database_srv%1$s PARTITION OF last_stat_database '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_database_srv%1$s '
        'ADD CONSTRAINT pk_last_stat_database_srv%1$s PRIMARY KEY (server_id, sample_id, datid), '
        'ADD CONSTRAINT fk_last_stat_database_samples_srv%1$s '
          'FOREIGN KEY (server_id, sample_id) '
          'REFERENCES samples(server_id, sample_id) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_tablespaces table partition
    EXECUTE format(
      'CREATE TABLE last_stat_tablespaces_srv%1$s PARTITION OF last_stat_tablespaces '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_tablespaces_srv%1$s '
        'ADD CONSTRAINT pk_last_stat_tablespaces_srv%1$s PRIMARY KEY (server_id, sample_id, tablespaceid), '
        'ADD CONSTRAINT fk_last_stat_tablespaces_samples_srv%1$s '
          'FOREIGN KEY (server_id, sample_id) REFERENCES samples(server_id, sample_id) '
          'ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_tables table partition
    EXECUTE format(
      'CREATE TABLE last_stat_tables_srv%1$s PARTITION OF last_stat_tables '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_tables_srv%1$s '
        'ADD CONSTRAINT pk_last_stat_tables_srv%1$s '
          'PRIMARY KEY (server_id, sample_id, datid, relid), '
        'ADD CONSTRAINT fk_last_stat_tables_dat_srv%1$s '
          'FOREIGN KEY (server_id, sample_id, datid) '
          'REFERENCES sample_stat_database(server_id, sample_id, datid) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_indexes table partition
    EXECUTE format(
      'CREATE TABLE last_stat_indexes_srv%1$s PARTITION OF last_stat_indexes '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_indexes_srv%1$s '
        'ADD CONSTRAINT pk_last_stat_indexes_srv%1$s '
          'PRIMARY KEY (server_id, sample_id, datid, indexrelid), '
        'ADD CONSTRAINT fk_last_stat_indexes_dat_srv%1$s '
        'FOREIGN KEY (server_id, sample_id, datid) '
          'REFERENCES sample_stat_database(server_id, sample_id, datid) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_user_functions table partition
    EXECUTE format(
      'CREATE TABLE last_stat_user_functions_srv%1$s PARTITION OF last_stat_user_functions '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_user_functions_srv%1$s '
      'ADD CONSTRAINT pk_last_stat_user_functions_srv%1$s '
      'PRIMARY KEY (server_id, sample_id, datid, funcid), '
      'ADD CONSTRAINT fk_last_stat_user_functions_dat_srv%1$s '
        'FOREIGN KEY (server_id, sample_id, datid) '
        'REFERENCES sample_stat_database(server_id, sample_id, datid) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_activity table partition
    EXECUTE format(
      'CREATE TABLE last_stat_activity_srv%1$s PARTITION OF last_stat_activity '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_activity_srv%1$s '
      'ADD CONSTRAINT pk_last_stat_activity_srv%1$s '
        'PRIMARY KEY (server_id, sample_id, pid, subsample_ts), '
      'ADD CONSTRAINT fk_last_stat_activity_sample_srv%1$s '
        'FOREIGN KEY (server_id, sample_id) '
        'REFERENCES samples(server_id, sample_id) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_activity table partition
    EXECUTE format(
      'CREATE TABLE last_stat_activity_count_srv%1$s PARTITION OF last_stat_activity_count '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_activity_count_srv%1$s '
      'ADD CONSTRAINT fk_last_stat_activity_count_sample_srv%1$s '
        'FOREIGN KEY (server_id, sample_id) '
        'REFERENCES samples(server_id, sample_id) ON DELETE RESTRICT',
        sserver_id);

    /*
    * Check if partition is already in our extension. This happens when function
    * is called during CREATE EXTENSION script execution
    */
    SELECT count(*) = 1 INTO in_extension
    FROM pg_depend dep
      JOIN pg_extension ext ON (dep.refobjid = ext.oid)
      JOIN pg_class rel ON (rel.oid = dep.objid AND rel.relkind= 'r')
    WHERE ext.extname='pg_profile'
      AND rel.relname = format('last_stat_statements_srv%1$s', sserver_id);

    IF NOT in_extension THEN
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_statements_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_kcache_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_database_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_tablespaces_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_tables_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_indexes_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_user_functions_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_activity_srv%1$s',
        sserver_id);
      EXECUTE format('ALTER EXTENSION pg_profile ADD TABLE last_stat_activity_count_srv%1$s',
        sserver_id);
    END IF;

    RETURN sserver_id;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION drop_server(IN server name) RETURNS integer SET search_path=@extschema@ AS $$
DECLARE
    del_rows    integer;
    dserver_id  integer;
BEGIN
    SELECT server_id INTO STRICT dserver_id FROM servers WHERE server_name = server;
    DELETE FROM bl_samples WHERE server_id = dserver_id;
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_kcache_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_statements_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_database_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_tables_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_indexes_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_tablespaces_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_user_functions_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_activity_srv%1$s',
      dserver_id);
    EXECUTE format('ALTER EXTENSION pg_profile DROP TABLE last_stat_activity_count_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_kcache_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_statements_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_database_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_tables_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_indexes_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_tablespaces_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_user_functions_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_activity_srv%1$s',
      dserver_id);
    EXECUTE format(
      'DROP TABLE last_stat_activity_count_srv%1$s',
      dserver_id);
    DELETE FROM last_stat_cluster WHERE server_id = dserver_id;
    DELETE FROM last_stat_io WHERE server_id = dserver_id;
    DELETE FROM last_stat_slru WHERE server_id = dserver_id;
    DELETE FROM last_stat_wal WHERE server_id = dserver_id;
    DELETE FROM last_stat_archiver WHERE server_id = dserver_id;
    DELETE FROM sample_stat_tablespaces WHERE server_id = dserver_id;
    DELETE FROM tablespaces_list WHERE server_id = dserver_id;
    /*
     * We have several constraints that should be deferred to avoid
     * violation due to several cascade deletion branches
     */
    SET CONSTRAINTS
        fk_stat_indexes_indexes,
        fk_st_tablespaces_tablespaces,
        fk_st_tables_tables,
        fk_indexes_tables,
        fk_user_functions_functions,
        fk_stmt_list,
        fk_kcache_stmt_list,
        fk_statements_roles
      DEFERRED;
    DELETE FROM samples WHERE server_id = dserver_id;
    DELETE FROM indexes_list WHERE server_id = dserver_id;
    DELETE FROM tables_list WHERE server_id = dserver_id;
    SET CONSTRAINTS
        fk_stat_indexes_indexes,
        fk_st_tablespaces_tablespaces,
        fk_st_tables_tables,
        fk_indexes_tables,
        fk_user_functions_functions,
        fk_stmt_list,
        fk_kcache_stmt_list,
        fk_statements_roles
      IMMEDIATE;
    DELETE FROM servers WHERE server_id = dserver_id;
    GET DIAGNOSTICS del_rows = ROW_COUNT;
    RETURN del_rows;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION drop_server(IN server name) IS 'Drop a server';
CREATE FUNCTION set_server_subsampling(IN server name, IN subsample_enabled boolean,
  IN min_query_duration interval hour to second,
  IN min_xact_duration interval hour to second,
  IN min_xact_age integer,
  IN min_idle_xact_dur interval hour to second)
RETURNS integer SET search_path=@extschema@ AS $$
DECLARE
    upd_rows integer;
BEGIN
    INSERT INTO server_subsample(
      server_id,
      subsample_enabled,
      min_query_dur,
      min_xact_dur,
      min_xact_age,
      min_idle_xact_dur,
      min_wait_dur
    )
    SELECT
      s.server_id,
      set_server_subsampling.subsample_enabled,
      set_server_subsampling.min_query_duration,
      set_server_subsampling.min_xact_duration,
      set_server_subsampling.min_xact_age,
      set_server_subsampling.min_idle_xact_dur,
      NULL
    FROM servers s
    WHERE server_name = set_server_subsampling.server
    ON CONFLICT (server_id) DO
    UPDATE SET
      (subsample_enabled, min_query_dur, min_xact_dur, min_xact_age,
       min_idle_xact_dur, min_wait_dur) =
      (
        COALESCE(EXCLUDED.subsample_enabled,server_subsample.subsample_enabled),
        COALESCE(EXCLUDED.min_query_dur,server_subsample.min_query_dur),
        COALESCE(EXCLUDED.min_xact_dur,server_subsample.min_xact_dur),
        COALESCE(EXCLUDED.min_xact_age,server_subsample.min_xact_age),
        COALESCE(EXCLUDED.min_idle_xact_dur,server_subsample.min_idle_xact_dur),
        COALESCE(EXCLUDED.min_lock_dur,server_subsample.min_wait_dur)
      );

    GET DIAGNOSTICS upd_rows = ROW_COUNT;
    RETURN upd_rows;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION set_server_subsampling(IN name, IN boolean,
  IN interval hour to second, IN interval hour to second, IN integer,
  IN interval hour to second)
IS 'Setup subsampling for a server';
CREATE FUNCTION export_data(IN server_name name = NULL, IN min_sample_id integer = NULL,
  IN max_sample_id integer = NULL, IN obfuscate_queries boolean = FALSE)
RETURNS TABLE(
    section_id  bigint,
    row_data    json
) SET search_path=@extschema@ AS $$
DECLARE
  section_counter   bigint = 0;
  ext_version       text = NULL;
  tables_list       json = NULL;
  sserver_id        integer = NULL;
  r_result          RECORD;
BEGIN
  /*
    Exported table will contain rows of extension tables, packed in JSON
    Each row will have a section ID, defining a table in most cases
    First sections contains metadata - extension name and version, tables list
  */
  -- Extension info
  IF (SELECT count(*) = 1 FROM pg_catalog.pg_extension WHERE extname = 'pg_profile') THEN
    SELECT extversion INTO STRICT r_result FROM pg_catalog.pg_extension WHERE extname = 'pg_profile';
    ext_version := r_result.extversion;
  ELSE
    RAISE 'Export is not supported for manual installed version';
  END IF;
  RETURN QUERY EXECUTE $q$SELECT $3, row_to_json(s)
    FROM (SELECT $1 AS extension,
              $2 AS version,
              $3 + 1 AS tab_list_section
    ) s$q$
    USING 'pg_profile', ext_version, section_counter;
  section_counter := section_counter + 1;
  -- tables list
  EXECUTE $q$
    WITH RECURSIVE exp_tables (reloid, relname, inc_rels) AS (
      -- start with all independent tables
        SELECT rel.oid, rel.relname, array_agg(rel.oid) OVER()
          FROM pg_depend dep
            JOIN pg_extension ext ON (dep.refobjid = ext.oid)
            JOIN pg_class rel ON (rel.oid = dep.objid AND rel.relkind IN ('r','p'))
            LEFT OUTER JOIN fkdeps con ON (con.reloid = dep.objid)
          WHERE ext.extname = $1 AND rel.relname NOT IN
              ('import_queries', 'import_queries_version_order',
              'report', 'report_static', 'report_struct', 'last_stat_activity')
            AND NOT rel.relispartition
            AND con.reloid IS NULL
      UNION
      -- and add all tables that have resolved dependencies by previously added tables
          SELECT con.reloid as reloid, con.relname, recurse.inc_rels||array_agg(con.reloid) OVER()
          FROM
            fkdeps con JOIN
            exp_tables recurse ON
              (array_append(recurse.inc_rels,con.reloid) @> con.reldeps AND
              NOT ARRAY[con.reloid] <@ recurse.inc_rels)
    ),
    fkdeps (reloid, relname, reldeps) AS (
      -- tables with their foreign key dependencies
      SELECT rel.oid as reloid, rel.relname, array_agg(con.confrelid), array_agg(rel.oid) OVER()
      FROM pg_depend dep
        JOIN pg_extension ext ON (dep.refobjid = ext.oid)
        JOIN pg_class rel ON (rel.oid = dep.objid AND rel.relkind IN ('r','p'))
        JOIN pg_constraint con ON (con.conrelid = dep.objid AND con.contype = 'f')
      WHERE ext.extname = $1 AND rel.relname NOT IN
        ('import_queries', 'import_queries_version_order',
        'report', 'report_static', 'report_struct', 'last_stat_activity')
        AND NOT rel.relispartition
      GROUP BY rel.oid, rel.relname
    )
    SELECT json_agg(row_to_json(tl)) FROM
    (SELECT row_number() OVER() + $2 AS section_id, relname FROM exp_tables) tl ;
  $q$ INTO tables_list
  USING 'pg_profile', section_counter;
  section_id := section_counter;
  row_data := tables_list;
  RETURN NEXT;
  section_counter := section_counter + 1;
  -- Server selection
  IF export_data.server_name IS NOT NULL THEN
    sserver_id := get_server_by_name(export_data.server_name);
  END IF;
  -- Tables data
  FOR r_result IN
    SELECT json_array_elements(tables_list)->>'relname' as relname
  LOOP
    -- Tables select conditions
    CASE
      WHEN r_result.relname != 'sample_settings'
        AND (r_result.relname LIKE 'sample%' OR r_result.relname LIKE 'last%') THEN
        RETURN QUERY EXECUTE format(
            $q$SELECT $1,row_to_json(dt) FROM
              (SELECT * FROM %I WHERE ($2 IS NULL OR $2 = server_id) AND
                ($3 IS NULL OR sample_id >= $3) AND
                ($4 IS NULL OR sample_id <= $4)) dt$q$,
            r_result.relname
          )
        USING
          section_counter,
          sserver_id,
          min_sample_id,
          max_sample_id;
      WHEN r_result.relname = 'bl_samples' THEN
        RETURN QUERY EXECUTE format(
            $q$
            SELECT $1,row_to_json(dt) FROM (
              SELECT *
              FROM %I b
                JOIN (
                  SELECT bl_id
                  FROM bl_samples
                    WHERE ($2 IS NULL OR $2 = server_id)
                  GROUP BY bl_id
                  HAVING
                    ($3 IS NULL OR min(sample_id) >= $3) AND
                    ($4 IS NULL OR max(sample_id) <= $4)
                ) bl_smp USING (bl_id)
              WHERE ($2 IS NULL OR $2 = server_id)
              ) dt$q$,
            r_result.relname
          )
        USING
          section_counter,
          sserver_id,
          min_sample_id,
          max_sample_id;
      WHEN r_result.relname = 'baselines' THEN
        RETURN QUERY EXECUTE format(
            $q$
            SELECT $1,row_to_json(dt) FROM (
              SELECT b.*
              FROM %I b
              JOIN bl_samples bs USING(server_id, bl_id)
                WHERE ($2 IS NULL OR $2 = server_id)
              GROUP BY b.server_id, b.bl_id, b.bl_name, b.keep_until
              HAVING
                ($3 IS NULL OR min(sample_id) >= $3) AND
                ($4 IS NULL OR max(sample_id) <= $4)
              ) dt$q$,
            r_result.relname
          )
        USING
          section_counter,
          sserver_id,
          min_sample_id,
          max_sample_id;
      WHEN r_result.relname = 'stmt_list' THEN
        RETURN QUERY EXECUTE format(
            $sql$SELECT $1,row_to_json(dt) FROM
              (SELECT rows.server_id, rows.queryid_md5,
                CASE $5
                  WHEN TRUE THEN pg_catalog.md5(rows.query)
                  ELSE rows.query
                END AS query,
                last_sample_id
               FROM %I AS rows WHERE (server_id,queryid_md5) IN
                (SELECT server_id, queryid_md5 FROM sample_statements WHERE
                  ($2 IS NULL OR $2 = server_id) AND
                ($3 IS NULL OR sample_id >= $3) AND
                ($4 IS NULL OR sample_id <= $4))) dt$sql$,
            r_result.relname
          )
        USING
          section_counter,
          sserver_id,
          min_sample_id,
          max_sample_id,
          obfuscate_queries;
      WHEN r_result.relname = 'act_query' THEN
        RETURN QUERY EXECUTE format(
            $sql$SELECT $1,row_to_json(dt) FROM
              (SELECT rows.server_id, rows.act_query_md5,
                CASE $5
                  WHEN TRUE THEN pg_catalog.md5(rows.act_query)
                  ELSE rows.act_query
                END AS act_query,
                last_sample_id
               FROM %I AS rows WHERE (server_id, act_query_md5) IN
                (SELECT server_id, act_query_md5 FROM sample_act_statement WHERE
                  ($2 IS NULL OR $2 = server_id) AND
                ($3 IS NULL OR sample_id >= $3) AND
                ($4 IS NULL OR sample_id <= $4))) dt$sql$,
            r_result.relname
          )
        USING
          section_counter,
          sserver_id,
          min_sample_id,
          max_sample_id,
          obfuscate_queries;
      ELSE
        RETURN QUERY EXECUTE format(
            $q$SELECT $1,row_to_json(dt) FROM (SELECT * FROM %I WHERE $2 IS NULL OR $2 = server_id) dt$q$,
            r_result.relname
          )
        USING section_counter, sserver_id;
    END CASE;
    section_counter := section_counter + 1;
  END LOOP;
  RETURN;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION export_data(IN server_name name, IN min_sample_id integer,
  IN max_sample_id integer, IN obfuscate_queries boolean) IS 'Export collected data as a table';
CREATE FUNCTION import_data(data regclass, server_name_prefix text = NULL) RETURNS bigint
SET search_path=@extschema@ AS $$
DECLARE
  import_meta     jsonb;
  tables_list     jsonb;
  servers_list    jsonb; -- import servers list
  tmp_srv_map     jsonb = '{}'; -- import server_id to local server_id mapping
  dump_versions   text[];

  new_server_id   integer = null;
  s_id            integer;
  sserver_id      integer;
  import_stage    integer;
  tot_processed   bigint = 0;
  func_processed  bigint = 0;
  c_datarows      refcursor;

  servers_sect_id   integer = NULL;
  settings_sect_id  integer = NULL;

  r_result        RECORD;
BEGIN
  -- Get import metadata
  EXECUTE format('SELECT row_data::jsonb FROM %s WHERE section_id = 0',data)
  INTO STRICT import_meta;

  -- Check dump compatibility
  IF (SELECT count(*) < 1 FROM import_queries_version_order
      WHERE (extension, version) =
        (import_meta ->> 'extension', import_meta ->> 'version'))
  THEN
    RAISE 'Unsupported extension version: %', (import_meta ->> 'extension')||' '||(import_meta ->> 'version');
  END IF;

  -- Get import tables list
  EXECUTE format('SELECT row_data::jsonb FROM %s WHERE section_id = $1',data)
  USING (import_meta ->> 'tab_list_section')::integer
  INTO STRICT tables_list;

  -- Get vital section identifiers
  FOR r_result IN (
    SELECT
      section_id,
      relname
    FROM
      jsonb_to_recordset(tables_list) as tbllist(section_id integer, relname text)
    WHERE
      relname IN ('servers', 'sample_settings')
  ) LOOP
    CASE r_result.relname
      WHEN 'servers' THEN servers_sect_id := r_result.section_id;
      WHEN 'sample_settings' THEN settings_sect_id := r_result.section_id;
    END CASE;
  END LOOP;

  -- Servers processing
  -- Get import servers list
  EXECUTE format($q$SELECT
      jsonb_agg(srvjs.row_data::jsonb)
    FROM
      %1$s srvjs
    WHERE
      srvjs.section_id = $1$q$,
    data)
  USING servers_sect_id
  INTO STRICT servers_list;

  /*
   * Performing importing to local servers matching. We need to consider several cases:
   * - creation dates and system identifiers matched - we have a match
   * - creation dates and system identifiers don't match, but names matched - conflict as we can't create a new server
   * - nothing matched - a new local server is to be created
   * By the way, we'll populate tmp_srv_map table, containing
   * a mapping between local and importing servers to use on data load.
   */
  FOR r_result IN EXECUTE format($q$SELECT
      COALESCE($3,'')||
        imp_srv.server_name       imp_server_name,
      ls.server_name              local_server_name,
      imp_srv.server_created      imp_server_created,
      ls.server_created           local_server_created,
      d.row_data->>'reset_val'    imp_system_identifier,
      ls.system_identifier        local_system_identifier,
      imp_srv.server_id           imp_server_id,
      ls.server_id                local_server_id,
      imp_srv.server_description  imp_server_description,
      imp_srv.db_exclude          imp_server_db_exclude,
      imp_srv.connstr             imp_server_connstr,
      imp_srv.max_sample_age      imp_server_max_sample_age,
      imp_srv.last_sample_id      imp_server_last_sample_id,
      imp_srv.size_smp_wnd_start  imp_size_smp_wnd_start,
      imp_srv.size_smp_wnd_dur    imp_size_smp_wnd_dur,
      imp_srv.size_smp_interval   imp_size_smp_interval
    FROM
      jsonb_to_recordset($1) as
        imp_srv(
          server_id           integer,
          server_name         name,
          server_description  text,
          server_created      timestamp with time zone,
          db_exclude          name[],
          enabled             boolean,
          connstr             text,
          max_sample_age      integer,
          last_sample_id      integer,
          size_smp_wnd_start  time with time zone,
          size_smp_wnd_dur    interval hour to second,
          size_smp_interval   interval day to minute
        )
      JOIN %s d ON
        (d.section_id = $2 AND d.row_data->>'name' = 'system_identifier'
          AND (d.row_data->>'server_id')::integer = imp_srv.server_id)
      LEFT OUTER JOIN (
        SELECT
          srv.server_id,
          srv.server_name,
          srv.server_created,
          set.reset_val as system_identifier
        FROM servers srv
          LEFT OUTER JOIN sample_settings set ON (set.server_id = srv.server_id AND set.name = 'system_identifier')
        ) ls ON
        ((imp_srv.server_created = ls.server_created AND d.row_data->>'reset_val' = ls.system_identifier)
          OR COALESCE($3,'')||imp_srv.server_name = ls.server_name)
    $q$,
    data)
  USING
    servers_list,
    settings_sect_id,
    server_name_prefix
  LOOP
    IF r_result.imp_server_created = r_result.local_server_created AND
      r_result.imp_system_identifier = r_result.local_system_identifier
    THEN
      /* use this local server when matched by server creation time and system identifier */
      tmp_srv_map := jsonb_set(
        tmp_srv_map,
        ARRAY[r_result.imp_server_id::text],
        to_jsonb(r_result.local_server_id)
      );
      /* Update local server if new last_sample_id is greatest*/
      UPDATE servers
      SET
        (
          db_exclude,
          connstr,
          max_sample_age,
          last_sample_id,
          size_smp_wnd_start,
          size_smp_wnd_dur,
          size_smp_interval
        ) = (
          r_result.imp_server_db_exclude,
          r_result.imp_server_connstr,
          r_result.imp_server_max_sample_age,
          r_result.imp_server_last_sample_id,
          r_result.imp_size_smp_wnd_start,
          r_result.imp_size_smp_wnd_dur,
          r_result.imp_size_smp_interval
        )
      WHERE server_id = r_result.local_server_id
        AND last_sample_id < r_result.imp_server_last_sample_id;
    ELSIF r_result.imp_server_name = r_result.local_server_name
    THEN
      /* Names matched, but identifiers does not - we have a conflict */
      RAISE 'Local server "%" creation date or system identifier does not match imported one (try renaming local server)',
        r_result.local_server_name;
    ELSIF r_result.local_server_name IS NULL
    THEN
      /* No match at all - we are creating a new server */
      INSERT INTO servers AS srv (
        server_name,
        server_description,
        server_created,
        db_exclude,
        enabled,
        connstr,
        max_sample_age,
        last_sample_id,
        size_smp_wnd_start,
        size_smp_wnd_dur,
        size_smp_interval)
      VALUES (
        r_result.imp_server_name,
        r_result.imp_server_description,
        r_result.imp_server_created,
        r_result.imp_server_db_exclude,
        FALSE,
        r_result.imp_server_connstr,
        r_result.imp_server_max_sample_age,
        r_result.imp_server_last_sample_id,
        r_result.imp_size_smp_wnd_start,
        r_result.imp_size_smp_wnd_dur,
        r_result.imp_size_smp_interval
      )
      RETURNING server_id INTO new_server_id;
      tmp_srv_map := jsonb_set(
        tmp_srv_map,
        ARRAY[r_result.imp_server_id::text],
        to_jsonb(new_server_id)
      );
      PERFORM create_server_partitions(new_server_id);
    ELSE
      /* This shouldn't ever happen */
      RAISE 'Import and local servers matching exception';
    END IF;
  END LOOP;

  /* Version history array for checking during import
  *  This array should be ordered oldest version last, thus first
  *  entry will be the current version
  */
  WITH RECURSIVE versions AS (
      SELECT
        import_meta ->> 'extension' AS extension,
        import_meta ->> 'version' AS version,
        1 AS lvl
    UNION
      SELECT vo.parent_extension, vo.parent_version, v.lvl + 1 AS lvl
      FROM versions v JOIN import_queries_version_order vo ON
        (vo.extension, vo.version) = (v.extension, v.version)
  )
  SELECT array_agg(version || ':' || extension ORDER BY lvl DESC) INTO dump_versions
  FROM versions
  WHERE extension IS NOT NULL;

  /* Import tables data */

  SET CONSTRAINTS ALL DEFERRED;
  import_stage := 0;
  WHILE import_stage < 3 LOOP
    FOR r_result IN (
      SELECT
        section_id,
        relname
      FROM
        jsonb_to_recordset(tables_list) AS tab_list(section_id integer, relname text)
    )
    LOOP
      CASE import_stage
        WHEN 0 THEN CONTINUE WHEN r_result.relname LIKE 'last_%';
        WHEN 1 THEN CONTINUE WHEN r_result.relname NOT LIKE 'last_%' OR
          r_result.relname = 'last_stat_kcache';
        WHEN 2 THEN CONTINUE WHEN r_result.relname != 'last_stat_kcache';
      END CASE;

      OPEN c_datarows FOR EXECUTE
        format('SELECT row_data FROM %s WHERE section_id = $1', data)
      USING r_result.section_id;
        RAISE NOTICE 'Started processing table %', r_result.relname;
        func_processed = 0;

        SELECT rows_processed, COALESCE(new_import_meta, import_meta)
          INTO func_processed, import_meta
        FROM import_section_data_profile(
          c_datarows,
          r_result.relname,
          tmp_srv_map,
          import_meta,
          dump_versions);
        -- try subsample import
        IF func_processed = -1 THEN
          SELECT rows_processed, COALESCE(new_import_meta, import_meta)
          INTO func_processed, import_meta
          FROM import_section_data_subsample(
            c_datarows,
            r_result.relname,
            tmp_srv_map,
            import_meta,
            dump_versions);
        END IF;

        IF func_processed = -1 THEN
          RAISE 'No import method for table %', r_result.relname;
        END IF;

        RAISE NOTICE 'Finished processing %',
          format('table %s (%s rows)',r_result.relname, func_processed);
        tot_processed := tot_processed + func_processed;
        /*
        * In case of pgpro_pwr extension we should call two functions in series
        * First would be pg_profile function and if it will fail to process a table
        * the coursor is to be directed to pgpro_pwr import function
        */
      CLOSE c_datarows;
    END LOOP; -- over dump tables
    import_stage := import_stage + 1;
  END LOOP; -- over import stages

  -- Finalize import
  FOR r_result IN (SELECT * FROM jsonb_each(tmp_srv_map)) LOOP
    sserver_id := r_result.value;
    SELECT max(sample_id) INTO s_id FROM samples WHERE server_id = sserver_id;

    -- Updating dictionary tables setting last_sample_id
    UPDATE tablespaces_list utl SET last_sample_id = s_id - 1
    FROM tablespaces_list tl LEFT JOIN sample_stat_tablespaces cur
      ON (cur.server_id, cur.sample_id, cur.tablespaceid) =
        (sserver_id, s_id, tl.tablespaceid)
    WHERE
      tl.last_sample_id IS NULL AND
      (utl.server_id, utl.tablespaceid) = (sserver_id, tl.tablespaceid) AND
      tl.server_id = sserver_id AND cur.server_id IS NULL;

    UPDATE funcs_list ufl SET last_sample_id = s_id - 1
    FROM funcs_list fl LEFT JOIN sample_stat_user_functions cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.funcid) =
        (sserver_id, s_id, fl.datid, fl.funcid)
    WHERE
      fl.last_sample_id IS NULL AND
      fl.server_id = sserver_id AND cur.server_id IS NULL AND
      (ufl.server_id, ufl.datid, ufl.funcid) =
      (sserver_id, fl.datid, fl.funcid);

    UPDATE indexes_list uil SET last_sample_id = s_id - 1
    FROM indexes_list il LEFT JOIN sample_stat_indexes cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.indexrelid) =
        (sserver_id, s_id, il.datid, il.indexrelid)
    WHERE
      il.last_sample_id IS NULL AND
      il.server_id = sserver_id AND cur.server_id IS NULL AND
      (uil.server_id, uil.datid, uil.indexrelid) =
      (sserver_id, il.datid, il.indexrelid);

    UPDATE tables_list utl SET last_sample_id = s_id - 1
    FROM tables_list tl LEFT JOIN sample_stat_tables cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.relid) =
        (sserver_id, s_id, tl.datid, tl.relid)
    WHERE
      tl.last_sample_id IS NULL AND
      tl.server_id = sserver_id AND cur.server_id IS NULL AND
      (utl.server_id, utl.datid, utl.relid) =
      (sserver_id, tl.datid, tl.relid);

    UPDATE stmt_list slu SET last_sample_id = s_id - 1
    FROM sample_statements ss RIGHT JOIN stmt_list sl
      ON (ss.server_id, ss.sample_id, ss.queryid_md5) =
        (sserver_id, s_id, sl.queryid_md5)
    WHERE
      sl.server_id = sserver_id AND
      sl.last_sample_id IS NULL AND
      ss.server_id IS NULL AND
      (slu.server_id, slu.queryid_md5) = (sserver_id, sl.queryid_md5);

    UPDATE roles_list rlu SET last_sample_id = s_id - 1
    FROM
        sample_statements ss
      RIGHT JOIN roles_list rl
      ON (ss.server_id, ss.sample_id, ss.userid) =
        (sserver_id, s_id, rl.userid)
    WHERE
      rl.server_id = sserver_id AND
      rl.last_sample_id IS NULL AND
      ss.server_id IS NULL AND
      (rlu.server_id, rlu.userid) = (sserver_id, rl.userid);

    -- Cleanup last_ tables
    DELETE FROM last_stat_activity WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_activity_count WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_archiver WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_cluster WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_database WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_indexes WHERE server_id=sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_io WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_kcache WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_slru WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_statements WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_tables WHERE server_id=sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_tablespaces WHERE server_id = sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_user_functions WHERE server_id=sserver_id AND sample_id != s_id;
    DELETE FROM last_stat_wal WHERE server_id = sserver_id AND sample_id != s_id;
  END LOOP; --over import servers
  SET CONSTRAINTS ALL IMMEDIATE;
  RETURN tot_processed;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION import_data(regclass, text) IS
  'Import sample data from table, exported by export_data function';
CREATE FUNCTION import_section_data_profile(IN data refcursor, IN imp_table_name name, IN srv_map jsonb,
  IN import_meta jsonb, IN versions_array text[],
  OUT rows_processed bigint, OUT new_import_meta jsonb)
SET search_path=profile AS $$
DECLARE
  datarow          record;
  rowcnt           bigint = 0;
  row_proc         bigint = 0;
  section_meta     jsonb;
BEGIN
  new_import_meta := import_meta;
  CASE imp_table_name
    WHEN 'servers' THEN NULL; -- skip already imported servers table
    -- skip obsolete tables from old versions
    WHEN 'sample_stat_tables_failures' THEN NULL;
    WHEN 'sample_stat_indexes_failures' THEN NULL;
    WHEN 'samples' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO samples(server_id, sample_id, sample_time)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.sample_time
        FROM json_to_record(datarow.row_data) AS dr(
            server_id       integer,
            sample_id       integer,
            sample_time     timestamp(0) with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_samples DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'baselines' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO baselines(server_id, bl_id, bl_name, keep_until)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.bl_id,
          dr.bl_name,
          dr.keep_until
        FROM json_to_record(datarow.row_data) AS dr(
            server_id    integer,
            bl_id        integer,
            bl_name      character varying(25),
            keep_until   timestamp (0) with time zone
          );
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_settings' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_settings(server_id,first_seen,setting_scope,name,setting,
          reset_val,boot_val,unit,sourcefile,sourceline,pending_restart)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.first_seen,
          dr.setting_scope,
          dr.name,
          dr.setting,
          dr.reset_val,
          dr.boot_val,
          dr.unit,
          dr.sourcefile,
          dr.sourceline,
          dr.pending_restart
        FROM json_to_record(datarow.row_data) AS dr(
            server_id        integer,
            first_seen       timestamp(0) with time zone,
            setting_scope    smallint,
            name             text,
            setting          text,
            reset_val        text,
            boot_val         text,
            unit             text,
            sourcefile       text,
            sourceline       integer,
            pending_restart  boolean
          )
        ON CONFLICT ON CONSTRAINT pk_sample_settings DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'bl_samples' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO bl_samples(server_id, sample_id, bl_id)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.bl_id
        FROM json_to_record(datarow.row_data) AS dr(
            server_id       integer,
            sample_id       integer,
            bl_id           integer
          );
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_cluster' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_cluster(server_id,sample_id,checkpoints_timed,
          checkpoints_req,checkpoint_write_time,checkpoint_sync_time,buffers_checkpoint,
          buffers_clean,maxwritten_clean,buffers_backend,buffers_backend_fsync,
          buffers_alloc,stats_reset,wal_size,wal_lsn,in_recovery)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.checkpoints_timed,
          dr.checkpoints_req,
          dr.checkpoint_write_time,
          dr.checkpoint_sync_time,
          dr.buffers_checkpoint,
          dr.buffers_clean,
          dr.maxwritten_clean,
          dr.buffers_backend,
          dr.buffers_backend_fsync,
          dr.buffers_alloc,
          dr.stats_reset,
          dr.wal_size,
          dr.wal_lsn,
          dr.in_recovery
        FROM json_to_record(datarow.row_data) AS dr(
            server_id              integer,
            sample_id              integer,
            checkpoints_timed      bigint,
            checkpoints_req        bigint,
            checkpoint_write_time  double precision,
            checkpoint_sync_time   double precision,
            buffers_checkpoint     bigint,
            buffers_clean          bigint,
            maxwritten_clean       bigint,
            buffers_backend        bigint,
            buffers_backend_fsync  bigint,
            buffers_alloc          bigint,
            stats_reset            timestamp with time zone,
            wal_size               bigint,
            wal_lsn                pg_lsn,
            in_recovery            boolean
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_cluster DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_database' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_database(server_id, sample_id, datid,datname,
          xact_commit, xact_rollback, blks_read, blks_hit, tup_returned, tup_fetched,
          tup_inserted, tup_updated, tup_deleted, conflicts, temp_files, temp_bytes,
          deadlocks, checksum_failures, checksum_last_failure, blk_read_time,
          blk_write_time, stats_reset, datsize,
          datsize_delta, datistemplate, session_time, active_time,
          idle_in_transaction_time, sessions, sessions_abandoned, sessions_fatal,
          sessions_killed)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.datname,
          dr.xact_commit,
          dr.xact_rollback,
          dr.blks_read,
          dr.blks_hit,
          dr.tup_returned,
          dr.tup_fetched,
          dr.tup_inserted,
          dr.tup_updated,
          dr.tup_deleted,
          dr.conflicts,
          dr.temp_files,
          dr.temp_bytes,
          dr.deadlocks,
          dr.checksum_failures,
          dr.checksum_last_failure,
          dr.blk_read_time,
          dr.blk_write_time,
          dr.stats_reset,
          dr.datsize,
          dr.datsize_delta,
          dr.datistemplate,
          dr.session_time,
          dr.active_time,
          dr.idle_in_transaction_time,
          dr.sessions,
          dr.sessions_abandoned,
          dr.sessions_fatal,
          dr.sessions_killed
        FROM json_to_record(datarow.row_data) AS dr(
          server_id       integer,
          sample_id       integer,
          datid           oid,
          datname         name,
          xact_commit     bigint,
          xact_rollback   bigint,
          blks_read       bigint,
          blks_hit        bigint,
          tup_returned    bigint,
          tup_fetched     bigint,
          tup_inserted    bigint,
          tup_updated     bigint,
          tup_deleted     bigint,
          conflicts       bigint,
          temp_files      bigint,
          temp_bytes      bigint,
          deadlocks       bigint,
          blk_read_time   double precision,
          blk_write_time  double precision,
          stats_reset     timestamp with time zone,
          datsize         bigint,
          datsize_delta   bigint,
          datistemplate   boolean,
          session_time    double precision,
          active_time     double precision,
          idle_in_transaction_time  double precision,
          sessions        bigint,
          sessions_abandoned  bigint,
          sessions_fatal      bigint,
          sessions_killed     bigint,
          checksum_failures   bigint,
          checksum_last_failure timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_database DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_wal' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_wal(server_id,sample_id,wal_records,
          wal_fpi,wal_bytes,wal_buffers_full,wal_write,wal_sync,
          wal_write_time,wal_sync_time,stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.wal_records,
          dr.wal_fpi,
          dr.wal_bytes,
          dr.wal_buffers_full,
          dr.wal_write,
          dr.wal_sync,
          dr.wal_write_time,
          dr.wal_sync_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
            server_id           integer,
            sample_id           integer,
            wal_records         bigint,
            wal_fpi             bigint,
            wal_bytes           numeric,
            wal_buffers_full    bigint,
            wal_write           bigint,
            wal_sync            bigint,
            wal_write_time      double precision,
            wal_sync_time       double precision,
            stats_reset         timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_wal DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_archiver' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_archiver(server_id,sample_id,archived_count,last_archived_wal,
          last_archived_time,failed_count,last_failed_wal,last_failed_time,
          stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.archived_count,
          dr.last_archived_wal,
          dr.last_archived_time,
          dr.failed_count,
          dr.last_failed_wal,
          dr.last_failed_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
            server_id           integer,
            sample_id           integer,
            archived_count      bigint,
            last_archived_wal   text,
            last_archived_time  timestamp with time zone,
            failed_count        bigint,
            last_failed_wal     text,
            last_failed_time    timestamp with time zone,
            stats_reset         timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_archiver DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_io' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_io(server_id,sample_id,backend_type,object,context,reads,
          read_time,writes,write_time,writebacks,writeback_time,extends,extend_time,
          op_bytes,hits,evictions,reuses,fsyncs,fsync_time,stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.backend_type,
          dr.object,
          dr.context,
          dr.reads,
          dr.read_time,
          dr.writes,
          dr.write_time,
          dr.writebacks,
          dr.writeback_time,
          dr.extends,
          dr.extend_time,
          dr.op_bytes,
          dr.hits,
          dr.evictions,
          dr.reuses,
          dr.fsyncs,
          dr.fsync_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
            server_id                   integer,
            sample_id                   integer,
            backend_type                text,
            object                      text,
            context                     text,
            reads                       bigint,
            read_time                   double precision,
            writes                      bigint,
            write_time                  double precision,
            writebacks                  bigint,
            writeback_time              double precision,
            extends                     bigint,
            extend_time                 double precision,
            op_bytes                    bigint,
            hits                        bigint,
            evictions                   bigint,
            reuses                      bigint,
            fsyncs                      bigint,
            fsync_time                  double precision,
            stats_reset                 timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_io DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_slru' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_slru(server_id,sample_id,name,blks_zeroed,
          blks_hit,blks_read,blks_written,blks_exists,flushes,truncates,
          stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.name,
          dr.blks_zeroed,
          dr.blks_hit,
          dr.blks_read,
          dr.blks_written,
          dr.blks_exists,
          dr.flushes,
          dr.truncates,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            sample_id      integer,
            name           text,
            blks_zeroed    bigint,
            blks_hit       bigint,
            blks_read      bigint,
            blks_written   bigint,
            blks_exists    bigint,
            flushes        bigint,
            truncates      bigint,
            stats_reset    timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_slru DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'roles_list' THEN
      CASE
        WHEN array_position(versions_array, '0.3.5:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            INSERT INTO roles_list(server_id,last_sample_id,userid,username)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              s.sample_id,
              dr.userid,
              dr.username
            FROM json_to_record(datarow.row_data) AS dr(
                server_id      integer,
                last_sample_id integer,
                userid         oid,
                username       name
              ) LEFT JOIN samples s ON (s.server_id, s.sample_id) =
                ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
            ON CONFLICT ON CONSTRAINT pk_roles_list DO
              UPDATE SET (last_sample_id, username) =
                (EXCLUDED.last_sample_id, EXCLUDED.username)
              WHERE (roles_list.last_sample_id, roles_list.username) IS DISTINCT FROM
                (EXCLUDED.last_sample_id, EXCLUDED.username);
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
          END LOOP; -- over data rows
        ELSE
          RAISE '%', format('[import %s]: Unsupported extension version: %s %s',
            imp_table_name, import_meta ->> 'extension', import_meta ->> 'version');
      END CASE; -- over import versions
    WHEN 'tables_list' THEN
      section_meta := '{}'::jsonb;
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        /* In case of old import (before 4.5) we should collect relid - reltoastrelid
        * binding here. We'll need it in further insert into sample_stat_tables
        */
        IF datarow.row_data #> ARRAY['reltoastrelid'] IS NOT NULL THEN
          section_meta := jsonb_set(section_meta,
            ARRAY[format('%s.%s',
              datarow.row_data #>> ARRAY['server_id'],
              datarow.row_data #>> ARRAY['relid'])
            ],
            jsonb_build_object('reltoastrelid', datarow.row_data #> ARRAY['reltoastrelid'])
          );
        END IF;
        INSERT INTO tables_list(server_id,last_sample_id,datid,relid,relkind,
          schemaname,relname)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          s.sample_id,
          dr.datid,
          dr.relid,
          dr.relkind,
          dr.schemaname,
          dr.relname
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            last_sample_id integer,
            datid          oid,
            relid          oid,
            relkind        character(1),
            schemaname     name,
            relname        name
          ) LEFT JOIN samples s ON (s.server_id, s.sample_id) =
            ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
          ON CONFLICT ON CONSTRAINT pk_tables_list DO
          UPDATE SET (last_sample_id, schemaname, relname) =
            (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.relname)
          WHERE (tables_list.last_sample_id, tables_list.schemaname, tables_list.relname)
            IS DISTINCT FROM
            (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.relname);
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
      -- Return collected relations mapping
      IF section_meta != '{}'::jsonb THEN
        new_import_meta := jsonb_set(new_import_meta, '{relations_map}', section_meta);
      END IF;
    WHEN 'stmt_list' THEN
      section_meta := '{}'::jsonb;
      CASE
        WHEN array_position(versions_array, '0.3.2:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            INSERT INTO stmt_list(server_id, last_sample_id, queryid_md5, query)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              s.sample_id,
              dr.queryid_md5,
              dr.query
            FROM json_to_record(datarow.row_data) AS dr (
                server_id      integer,
                last_sample_id integer,
                queryid_md5    character(32),
                query          text
              )
              LEFT JOIN samples s ON (s.server_id, s.sample_id) =
                ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
            ON CONFLICT ON CONSTRAINT pk_stmt_list
              DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
          END LOOP; -- over data rows
        WHEN array_position(versions_array, '0.3.1:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            IF section_meta #> ARRAY[datarow.row_data->>'queryid_md5'] IS NULL THEN
              section_meta := jsonb_set(section_meta,
                ARRAY[datarow.row_data->>'queryid_md5'], '{}'::jsonb);
            END IF;
            section_meta := jsonb_set(section_meta,
              ARRAY[datarow.row_data->>'queryid_md5', datarow.row_data->>'server_id'],
              to_jsonb(md5(datarow.row_data->>'query'))
            );
            INSERT INTO stmt_list(server_id, last_sample_id, queryid_md5, query)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              NULL,
              md5(dr.query),
              dr.query
            FROM json_to_record(datarow.row_data) AS dr (
                server_id      integer,
                query          text
              )
            ON CONFLICT ON CONSTRAINT pk_stmt_list
              DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
              RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
            END IF;
          END LOOP; -- over data rows
          -- Return queryid mapping
          new_import_meta := jsonb_set(new_import_meta, '{queryid_map}', section_meta);
        ELSE
          RAISE '%', format('[import %s]: Unsupported extension version: %s %s',
            imp_table_name, import_meta ->> 'extension', import_meta ->> 'version');
      END CASE; -- over import versions
    WHEN 'funcs_list' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO funcs_list(server_id, last_sample_id,datid,funcid,schemaname,
          funcname,funcargs)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          s.sample_id,
          dr.datid,
          dr.funcid,
          dr.schemaname,
          dr.funcname,
          dr.funcargs
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            last_sample_id integer,
            datid          oid,
            funcid         oid,
            schemaname     name,
            funcname       name,
            funcargs       text
          )
          LEFT JOIN samples s ON (s.server_id, s.sample_id) =
            ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
        ON CONFLICT ON CONSTRAINT pk_funcs_list DO
        UPDATE SET (last_sample_id, schemaname, funcname, funcargs) =
          (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.funcname, EXCLUDED.funcargs)
        WHERE (funcs_list.last_sample_id, funcs_list.schemaname, funcs_list.funcname,
          funcs_list.funcargs) IS DISTINCT FROM
          (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.funcname, EXCLUDED.funcargs);
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_timings' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_timings(server_id, sample_id, event, time_spent)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.event,
          dr.time_spent
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            sample_id      integer,
            event          text,
            time_spent     interval minute to second(2)
          )
        ON CONFLICT ON CONSTRAINT pk_sample_timings
          DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'tablespaces_list' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO tablespaces_list(server_id, last_sample_id,
          tablespaceid, tablespacename, tablespacepath)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          s.sample_id,
          dr.tablespaceid,
          dr.tablespacename,
          dr.tablespacepath
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            last_sample_id integer,
            tablespaceid   oid,
            tablespacename name,
            tablespacepath text
          )
          LEFT JOIN samples s ON (s.server_id, s.sample_id) =
            ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
        ON CONFLICT ON CONSTRAINT pk_tablespace_list DO
        UPDATE SET (last_sample_id, tablespacename, tablespacepath) =
          (EXCLUDED.last_sample_id, EXCLUDED.tablespacename, EXCLUDED.tablespacepath)
        WHERE (tablespaces_list.last_sample_id, tablespaces_list.tablespacename,
            tablespaces_list.tablespacepath) IS DISTINCT FROM
          (EXCLUDED.last_sample_id, EXCLUDED.tablespacename, EXCLUDED.tablespacepath);
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_tablespaces' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_tablespaces(server_id, sample_id,
          tablespaceid, size, size_delta)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.tablespaceid,
          dr.size,
          dr.size_delta
        FROM json_to_record(datarow.row_data) AS dr(
            server_id     integer,
            sample_id     integer,
            tablespaceid  oid,
            size          bigint,
            size_delta    bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_tablespaces
          DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'wait_sampling_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO wait_sampling_total(server_id, sample_id, sample_wevnt_id,
          event_type, event, tot_waited, stmt_waited)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.sample_wevnt_id,
          dr.event_type,
          dr.event,
          dr.tot_waited,
          dr.stmt_waited
        FROM json_to_record(datarow.row_data) AS dr(
            server_id           integer,
            sample_id           integer,
            sample_wevnt_id     integer,
            event_type          text,
            event               text,
            tot_waited          bigint,
            stmt_waited         bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_weid
          DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'indexes_list' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO indexes_list(server_id, last_sample_id, datid, indexrelid, relid,
          schemaname, indexrelname)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          s.sample_id,
          dr.datid,
          dr.indexrelid,
          dr.relid,
          dr.schemaname,
          dr.indexrelname
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            last_sample_id integer,
            datid          oid,
            indexrelid     oid,
            relid          oid,
            schemaname     name,
            indexrelname   name
          )
          LEFT JOIN samples s ON (s.server_id, s.sample_id) =
            ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
        ON CONFLICT ON CONSTRAINT pk_indexes_list DO
        UPDATE SET (last_sample_id, schemaname, indexrelname) =
          (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.indexrelname)
        WHERE (indexes_list.last_sample_id, indexes_list.schemaname, indexes_list.indexrelname)
          IS DISTINCT FROM
          (EXCLUDED.last_sample_id, EXCLUDED.schemaname, EXCLUDED.indexrelname);
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_statements' THEN
      CASE -- Import version selector
        WHEN array_position(versions_array, '4.0:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            INSERT INTO sample_statements(server_id, sample_id, userid, datid, queryid, queryid_md5,
              plans, total_plan_time, min_plan_time, max_plan_time, mean_plan_time,
              stddev_plan_time, calls, total_exec_time, min_exec_time, max_exec_time, mean_exec_time,
              stddev_exec_time, rows, shared_blks_hit, shared_blks_read, shared_blks_dirtied,
              shared_blks_written, local_blks_hit, local_blks_read, local_blks_dirtied,
              local_blks_written, temp_blks_read, temp_blks_written, blk_read_time, blk_write_time,
              wal_records, wal_fpi, wal_bytes, toplevel
              , jit_functions, jit_generation_time, jit_inlining_count, jit_inlining_time,
              jit_optimization_count, jit_optimization_time, jit_emission_count, jit_emission_time,
              temp_blk_read_time, temp_blk_write_time)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              dr.sample_id,
              dr.userid,
              dr.datid,
              dr.queryid,
              dr.queryid_md5,
              dr.plans,
              dr.total_plan_time,
              dr.min_plan_time,
              dr.max_plan_time,
              dr.mean_plan_time,
              dr.stddev_plan_time,
              dr.calls,
              dr.total_exec_time,
              dr.min_exec_time,
              dr.max_exec_time,
              dr.mean_exec_time,
              dr.stddev_exec_time,
              dr.rows,
              dr.shared_blks_hit,
              dr.shared_blks_read,
              dr.shared_blks_dirtied,
              dr.shared_blks_written,
              dr.local_blks_hit,
              dr.local_blks_read,
              dr.local_blks_dirtied,
              dr.local_blks_written,
              dr.temp_blks_read,
              dr.temp_blks_written,
              dr.blk_read_time,
              dr.blk_write_time,
              dr.wal_records,
              dr.wal_fpi,
              dr.wal_bytes,
              COALESCE(dr.toplevel, true),
              dr.jit_functions,
              dr.jit_generation_time,
              dr.jit_inlining_count,
              dr.jit_inlining_time,
              dr.jit_optimization_count,
              dr.jit_optimization_time,
              dr.jit_emission_count,
              dr.jit_emission_time,
              dr.temp_blk_read_time,
              dr.temp_blk_write_time
            FROM json_to_record(datarow.row_data) AS dr(
              server_id            integer,
              sample_id            integer,
              userid               oid,
              datid                oid,
              queryid              bigint,
              queryid_md5          character(32),
              plans                bigint,
              total_plan_time      double precision,
              min_plan_time        double precision,
              max_plan_time        double precision,
              mean_plan_time       double precision,
              stddev_plan_time     double precision,
              calls                bigint,
              total_exec_time      double precision,
              min_exec_time        double precision,
              max_exec_time        double precision,
              mean_exec_time       double precision,
              stddev_exec_time     double precision,
              rows                 bigint,
              shared_blks_hit      bigint,
              shared_blks_read     bigint,
              shared_blks_dirtied  bigint,
              shared_blks_written  bigint,
              local_blks_hit       bigint,
              local_blks_read      bigint,
              local_blks_dirtied   bigint,
              local_blks_written   bigint,
              temp_blks_read       bigint,
              temp_blks_written    bigint,
              blk_read_time        double precision,
              blk_write_time       double precision,
              wal_records          bigint,
              wal_fpi              bigint,
              wal_bytes            numeric,
              toplevel             boolean,
              jit_functions        bigint,
              jit_generation_time  double precision,
              jit_inlining_count   bigint,
              jit_inlining_time    double precision,
              jit_optimization_count  bigint,
              jit_optimization_time   double precision,
              jit_emission_count   bigint,
              jit_emission_time    double precision,
              temp_blk_read_time   double precision,
              temp_blk_write_time  double precision
              )
            ON CONFLICT ON CONSTRAINT pk_sample_statements_n DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
              RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
            END IF;
          END LOOP; -- over data rows
        WHEN array_position(versions_array, '0.3.1:pg_profile') IS NOT NULL THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;

            INSERT INTO roles_list (server_id,userid,username)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              dr.userid,
              '_unknown_'
            FROM json_to_record(datarow.row_data) AS dr(
                  server_id            integer,
                  userid               oid
              )
            ON CONFLICT ON CONSTRAINT pk_roles_list DO
              UPDATE SET (last_sample_id, username) =
                (EXCLUDED.last_sample_id, EXCLUDED.username);

            INSERT INTO sample_statements(server_id, sample_id, userid, datid, queryid, queryid_md5,
              plans, total_plan_time, min_plan_time, max_plan_time, mean_plan_time,
              stddev_plan_time, calls, total_exec_time, min_exec_time, max_exec_time, mean_exec_time,
              stddev_exec_time, rows, shared_blks_hit, shared_blks_read, shared_blks_dirtied,
              shared_blks_written, local_blks_hit, local_blks_read, local_blks_dirtied,
              local_blks_written, temp_blks_read, temp_blks_written, blk_read_time, blk_write_time,
              wal_records, wal_fpi, wal_bytes, toplevel
              , jit_functions, jit_generation_time, jit_inlining_count, jit_inlining_time,
              jit_optimization_count, jit_optimization_time, jit_emission_count, jit_emission_time,
              temp_blk_read_time, temp_blk_write_time)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              dr.sample_id,
              dr.userid,
              dr.datid,
              dr.queryid,
              CASE WHEN starts_with(versions_array[1], '0.3.1:') THEN
                import_meta #>> ARRAY['queryid_map', dr.queryid_md5, dr.server_id::text]
              ELSE
                dr.queryid_md5
              END,
              dr.plans,
              dr.total_plan_time,
              dr.min_plan_time,
              dr.max_plan_time,
              dr.mean_plan_time,
              dr.stddev_plan_time,
              dr.calls,
              dr.total_exec_time,
              dr.min_exec_time,
              dr.max_exec_time,
              dr.mean_exec_time,
              dr.stddev_exec_time,
              dr.rows,
              dr.shared_blks_hit,
              dr.shared_blks_read,
              dr.shared_blks_dirtied,
              dr.shared_blks_written,
              dr.local_blks_hit,
              dr.local_blks_read,
              dr.local_blks_dirtied,
              dr.local_blks_written,
              dr.temp_blks_read,
              dr.temp_blks_written,
              dr.blk_read_time,
              dr.blk_write_time,
              dr.wal_records,
              dr.wal_fpi,
              dr.wal_bytes,
              COALESCE(dr.toplevel, true),
              dr.jit_functions,
              dr.jit_generation_time,
              dr.jit_inlining_count,
              dr.jit_inlining_time,
              dr.jit_optimization_count,
              dr.jit_optimization_time,
              dr.jit_emission_count,
              dr.jit_emission_time,
              dr.temp_blk_read_time,
              dr.temp_blk_write_time
            FROM json_to_record(datarow.row_data) AS dr(
              server_id            integer,
              sample_id            integer,
              userid               oid,
              datid                oid,
              queryid              bigint,
              queryid_md5          character(32),
              plans                bigint,
              total_plan_time      double precision,
              min_plan_time        double precision,
              max_plan_time        double precision,
              mean_plan_time       double precision,
              stddev_plan_time     double precision,
              calls                bigint,
              total_exec_time      double precision,
              min_exec_time        double precision,
              max_exec_time        double precision,
              mean_exec_time       double precision,
              stddev_exec_time     double precision,
              rows                 bigint,
              shared_blks_hit      bigint,
              shared_blks_read     bigint,
              shared_blks_dirtied  bigint,
              shared_blks_written  bigint,
              local_blks_hit       bigint,
              local_blks_read      bigint,
              local_blks_dirtied   bigint,
              local_blks_written   bigint,
              temp_blks_read       bigint,
              temp_blks_written    bigint,
              blk_read_time        double precision,
              blk_write_time       double precision,
              wal_records          bigint,
              wal_fpi              bigint,
              wal_bytes            numeric,
              toplevel             boolean,
              jit_functions        bigint,
              jit_generation_time  double precision,
              jit_inlining_count   bigint,
              jit_inlining_time    double precision,
              jit_optimization_count  bigint,
              jit_optimization_time   double precision,
              jit_emission_count   bigint,
              jit_emission_time    double precision,
              temp_blk_read_time   double precision,
              temp_blk_write_time  double precision
              )
            ON CONFLICT ON CONSTRAINT pk_sample_statements_n DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
              RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
            END IF;
          END LOOP; -- over data rows
        ELSE
          RAISE '%', format('[import %s]: Unsupported extension version: %s %s',
            imp_table_name, import_meta ->> 'extension', import_meta ->> 'version');
      END CASE; -- over import versions
    WHEN 'sample_statements_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_statements_total(server_id, sample_id, datid, plans, total_plan_time,
          calls, total_exec_time, rows, shared_blks_hit, shared_blks_read,
          shared_blks_dirtied, shared_blks_written, local_blks_hit, local_blks_read,
          local_blks_dirtied, local_blks_written, temp_blks_read, temp_blks_written,
          blk_read_time, blk_write_time, wal_records, wal_fpi, wal_bytes, statements
          , jit_functions, jit_generation_time, jit_inlining_count, jit_inlining_time,
          jit_optimization_count, jit_optimization_time, jit_emission_count, jit_emission_time,
          temp_blk_read_time, temp_blk_write_time)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.plans,
          dr.total_plan_time,
          dr.calls,
          dr.total_exec_time,
          dr.rows,
          dr.shared_blks_hit,
          dr.shared_blks_read,
          dr.shared_blks_dirtied,
          dr.shared_blks_written,
          dr.local_blks_hit,
          dr.local_blks_read,
          dr.local_blks_dirtied,
          dr.local_blks_written,
          dr.temp_blks_read,
          dr.temp_blks_written,
          dr.blk_read_time,
          dr.blk_write_time,
          dr.wal_records,
          dr.wal_fpi,
          dr.wal_bytes,
          dr.statements,
          dr.jit_functions,
          dr.jit_generation_time,
          dr.jit_inlining_count,
          dr.jit_inlining_time,
          dr.jit_optimization_count,
          dr.jit_optimization_time,
          dr.jit_emission_count,
          dr.jit_emission_time,
          dr.temp_blk_read_time,
          dr.temp_blk_write_time
        FROM json_to_record(datarow.row_data) AS dr(
            server_id            integer,
            sample_id            integer,
            datid                oid,
            plans                bigint,
            total_plan_time      double precision,
            calls                bigint,
            total_exec_time      double precision,
            rows                 bigint,
            shared_blks_hit      bigint,
            shared_blks_read     bigint,
            shared_blks_dirtied  bigint,
            shared_blks_written  bigint,
            local_blks_hit       bigint,
            local_blks_read      bigint,
            local_blks_dirtied   bigint,
            local_blks_written   bigint,
            temp_blks_read       bigint,
            temp_blks_written    bigint,
            blk_read_time        double precision,
            blk_write_time       double precision,
            wal_records          bigint,
            wal_fpi              bigint,
            wal_bytes            numeric,
            statements           bigint,
            jit_functions        bigint,
            jit_generation_time  double precision,
            jit_inlining_count   bigint,
            jit_inlining_time    double precision,
            jit_optimization_count  bigint,
            jit_optimization_time   double precision,
            jit_emission_count   bigint,
            jit_emission_time    double precision,
            temp_blk_read_time   double precision,
            temp_blk_write_time  double precision
          )
        ON CONFLICT ON CONSTRAINT pk_sample_statements_total DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_kcache_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_kcache_total(server_id, sample_id, datid, plan_user_time,
          plan_system_time, plan_minflts, plan_majflts, plan_nswaps, plan_reads, plan_writes,
          plan_msgsnds, plan_msgrcvs, plan_nsignals, plan_nvcsws, plan_nivcsws,
          exec_user_time, exec_system_time, exec_minflts, exec_majflts, exec_nswaps,
          exec_reads, exec_writes, exec_msgsnds, exec_msgrcvs, exec_nsignals, exec_nvcsws,
          exec_nivcsws, statements)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.plan_user_time,
          dr.plan_system_time,
          dr.plan_minflts,
          dr.plan_majflts,
          dr.plan_nswaps,
          dr.plan_reads,
          dr.plan_writes,
          dr.plan_msgsnds,
          dr.plan_msgrcvs,
          dr.plan_nsignals,
          dr.plan_nvcsws,
          dr.plan_nivcsws,
          dr.exec_user_time,
          dr.exec_system_time,
          dr.exec_minflts,
          dr.exec_majflts,
          dr.exec_nswaps,
          dr.exec_reads,
          dr.exec_writes,
          dr.exec_msgsnds,
          dr.exec_msgrcvs,
          dr.exec_nsignals,
          dr.exec_nvcsws,
          dr.exec_nivcsws,
          dr.statements
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            datid             oid,
            plan_user_time    double precision,
            plan_system_time  double precision,
            plan_minflts      bigint,
            plan_majflts      bigint,
            plan_nswaps       bigint,
            plan_reads        bigint,
            plan_writes       bigint,
            plan_msgsnds      bigint,
            plan_msgrcvs      bigint,
            plan_nsignals     bigint,
            plan_nvcsws       bigint,
            plan_nivcsws      bigint,
            exec_user_time    double precision,
            exec_system_time  double precision,
            exec_minflts      bigint,
            exec_majflts      bigint,
            exec_nswaps       bigint,
            exec_reads        bigint,
            exec_writes       bigint,
            exec_msgsnds      bigint,
            exec_msgrcvs      bigint,
            exec_nsignals     bigint,
            exec_nvcsws       bigint,
            exec_nivcsws      bigint,
            statements        bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_kcache_total DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_user_functions' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_user_functions(server_id, sample_id, datid, funcid,
          calls, total_time, self_time, trg_fn)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.funcid,
          dr.calls,
          dr.total_time,
          dr.self_time,
          dr.trg_fn
        FROM json_to_record(datarow.row_data) AS dr(
            server_id   integer,
            sample_id   integer,
            datid       oid,
            funcid      oid,
            calls       bigint,
            total_time  double precision,
            self_time   double precision,
            trg_fn      boolean
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_user_functions DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_user_func_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_user_func_total(server_id, sample_id, datid, calls,
          total_time, trg_fn)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.calls,
          dr.total_time,
          dr.trg_fn
        FROM json_to_record(datarow.row_data) AS dr(
            server_id   integer,
            sample_id   integer,
            datid       oid,
            calls       bigint,
            total_time  double precision,
            trg_fn      boolean
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_user_func_total DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_tables' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_tables(server_id,sample_id,
          datid,relid,reltoastrelid,tablespaceid,seq_scan,
          seq_tup_read,idx_scan,idx_tup_fetch,n_tup_ins,n_tup_upd,n_tup_del,n_tup_hot_upd,
          n_live_tup,n_dead_tup,n_mod_since_analyze,n_ins_since_vacuum,last_vacuum,
          last_autovacuum,last_analyze,last_autoanalyze,vacuum_count,autovacuum_count,
          analyze_count,autoanalyze_count,heap_blks_read,heap_blks_hit,idx_blks_read,
          idx_blks_hit,toast_blks_read,toast_blks_hit,tidx_blks_read,tidx_blks_hit,
          relsize,relsize_diff,relpages_bytes,relpages_bytes_diff,last_seq_scan,
          last_idx_scan,n_tup_newpage_upd)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.relid,
          COALESCE(
            dr.reltoastrelid,
            (import_meta #>> ARRAY[
                'relations_map',
                format('%s.%s', dr.server_id::text, dr.relid::text),
                'reltoastrelid'
              ]
            )::oid
          ) AS reltoastrelid,
          dr.tablespaceid,
          dr.seq_scan,
          dr.seq_tup_read,
          dr.idx_scan,
          dr.idx_tup_fetch,
          dr.n_tup_ins,
          dr.n_tup_upd,
          dr.n_tup_del,
          dr.n_tup_hot_upd,
          dr.n_live_tup,
          dr.n_dead_tup,
          dr.n_mod_since_analyze,
          dr.n_ins_since_vacuum,
          dr.last_vacuum,
          dr.last_autovacuum,
          dr.last_analyze,
          dr.last_autoanalyze,
          dr.vacuum_count,
          dr.autovacuum_count,
          dr.analyze_count,
          dr.autoanalyze_count,
          dr.heap_blks_read,
          dr.heap_blks_hit,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.toast_blks_read,
          dr.toast_blks_hit,
          dr.tidx_blks_read,
          dr.tidx_blks_hit,
          dr.relsize,
          dr.relsize_diff,
          dr.relpages_bytes,
          dr.relpages_bytes_diff,
          dr.last_seq_scan,
          dr.last_idx_scan,
          dr.n_tup_newpage_upd
        FROM json_to_record(datarow.row_data) AS dr(
          server_id            integer,
          sample_id            integer,
          datid                oid,
          relid                oid,
          reltoastrelid        oid,
          tablespaceid         oid,
          seq_scan             bigint,
          seq_tup_read         bigint,
          idx_scan             bigint,
          idx_tup_fetch        bigint,
          n_tup_ins            bigint,
          n_tup_upd            bigint,
          n_tup_del            bigint,
          n_tup_hot_upd        bigint,
          n_live_tup           bigint,
          n_dead_tup           bigint,
          n_mod_since_analyze  bigint,
          n_ins_since_vacuum   bigint,
          last_vacuum          timestamp with time zone,
          last_autovacuum      timestamp with time zone,
          last_analyze         timestamp with time zone,
          last_autoanalyze     timestamp with time zone,
          vacuum_count         bigint,
          autovacuum_count     bigint,
          analyze_count        bigint,
          autoanalyze_count    bigint,
          heap_blks_read       bigint,
          heap_blks_hit        bigint,
          idx_blks_read        bigint,
          idx_blks_hit         bigint,
          toast_blks_read      bigint,
          toast_blks_hit       bigint,
          tidx_blks_read       bigint,
          tidx_blks_hit        bigint,
          relsize              bigint,
          relsize_diff         bigint,
          relpages_bytes       bigint,
          relpages_bytes_diff  bigint,
          last_seq_scan        timestamp with time zone,
          last_idx_scan        timestamp with time zone,
          n_tup_newpage_upd    bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_tables DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_tables_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_tables_total(server_id,sample_id,datid,tablespaceid,relkind,
          seq_scan,seq_tup_read,idx_scan,idx_tup_fetch,n_tup_ins,n_tup_upd,n_tup_del,
          n_tup_hot_upd,vacuum_count,autovacuum_count,analyze_count,autoanalyze_count,
          heap_blks_read,heap_blks_hit,idx_blks_read,idx_blks_hit,toast_blks_read,
          toast_blks_hit,tidx_blks_read,tidx_blks_hit,relsize_diff,n_tup_newpage_upd)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.tablespaceid,
          dr.relkind,
          dr.seq_scan,
          dr.seq_tup_read,
          dr.idx_scan,
          dr.idx_tup_fetch,
          dr.n_tup_ins,
          dr.n_tup_upd,
          dr.n_tup_del,
          dr.n_tup_hot_upd,
          dr.vacuum_count,
          dr.autovacuum_count,
          dr.analyze_count,
          dr.autoanalyze_count,
          dr.heap_blks_read,
          dr.heap_blks_hit,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.toast_blks_read,
          dr.toast_blks_hit,
          dr.tidx_blks_read,
          dr.tidx_blks_hit,
          dr.relsize_diff,
          dr.n_tup_newpage_upd
        FROM json_to_record(datarow.row_data) AS dr(
          server_id          integer,
          sample_id          integer,
          datid              oid,
          tablespaceid       oid,
          relkind            character(1),
          seq_scan           bigint,
          seq_tup_read       bigint,
          idx_scan           bigint,
          idx_tup_fetch      bigint,
          n_tup_ins          bigint,
          n_tup_upd          bigint,
          n_tup_del          bigint,
          n_tup_hot_upd      bigint,
          vacuum_count       bigint,
          autovacuum_count   bigint,
          analyze_count      bigint,
          autoanalyze_count  bigint,
          heap_blks_read     bigint,
          heap_blks_hit      bigint,
          idx_blks_read      bigint,
          idx_blks_hit       bigint,
          toast_blks_read    bigint,
          toast_blks_hit     bigint,
          tidx_blks_read     bigint,
          tidx_blks_hit      bigint,
          relsize_diff       bigint,
          n_tup_newpage_upd  bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_tables_tot DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_indexes' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_indexes(server_id,sample_id,datid,indexrelid,tablespaceid,
          idx_scan,idx_tup_read,idx_tup_fetch,idx_blks_read,idx_blks_hit,relsize,
          relsize_diff,indisunique,relpages_bytes,relpages_bytes_diff,last_idx_scan)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.indexrelid,
          dr.tablespaceid,
          dr.idx_scan,
          dr.idx_tup_read,
          dr.idx_tup_fetch,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.relsize,
          dr.relsize_diff,
          dr.indisunique,
          dr.relpages_bytes,
          dr.relpages_bytes_diff,
          dr.last_idx_scan
        FROM json_to_record(datarow.row_data) AS dr(
          server_id      integer,
          sample_id      integer,
          datid          oid,
          indexrelid     oid,
          tablespaceid   oid,
          idx_scan       bigint,
          idx_tup_read   bigint,
          idx_tup_fetch  bigint,
          idx_blks_read  bigint,
          idx_blks_hit   bigint,
          relsize        bigint,
          relsize_diff   bigint,
          indisunique    boolean,
          relpages_bytes bigint,
          relpages_bytes_diff bigint,
          last_idx_scan  timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_indexes DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_indexes_total' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_indexes_total (server_id,sample_id,datid,tablespaceid,idx_scan,
          idx_tup_read,idx_tup_fetch,idx_blks_read,idx_blks_hit,relsize_diff)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.tablespaceid,
          dr.idx_scan,
          dr.idx_tup_read,
          dr.idx_tup_fetch,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.relsize_diff
        FROM json_to_record(datarow.row_data) AS dr(
          server_id      integer,
          sample_id      integer,
          datid          oid,
          tablespaceid   oid,
          idx_scan       bigint,
          idx_tup_read   bigint,
          idx_tup_fetch  bigint,
          idx_blks_read  bigint,
          idx_blks_hit   bigint,
          relsize_diff   bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_indexes_tot DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_kcache' THEN
      CASE -- Import version selector
        WHEN array_position(versions_array, '0.3.2:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            INSERT INTO sample_kcache(server_id,sample_id,userid,datid,queryid,queryid_md5,
              plan_user_time,plan_system_time,plan_minflts,plan_majflts,
              plan_nswaps,plan_reads,plan_writes,plan_msgsnds,plan_msgrcvs,plan_nsignals,
              plan_nvcsws,plan_nivcsws,exec_user_time,exec_system_time,exec_minflts,
              exec_majflts,exec_nswaps,exec_reads,exec_writes,exec_msgsnds,exec_msgrcvs,
              exec_nsignals,exec_nvcsws,exec_nivcsws,toplevel)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              dr.sample_id,
              dr.userid,
              dr.datid,
              dr.queryid,
              dr.queryid_md5,
              dr.plan_user_time,
              dr.plan_system_time,
              dr.plan_minflts,
              dr.plan_majflts,
              dr.plan_nswaps,
              dr.plan_reads,
              dr.plan_writes,
              dr.plan_msgsnds,
              dr.plan_msgrcvs,
              dr.plan_nsignals,
              dr.plan_nvcsws,
              dr.plan_nivcsws,
              dr.exec_user_time,
              dr.exec_system_time,
              dr.exec_minflts,
              dr.exec_majflts,
              dr.exec_nswaps,
              dr.exec_reads,
              dr.exec_writes,
              dr.exec_msgsnds,
              dr.exec_msgrcvs,
              dr.exec_nsignals,
              dr.exec_nvcsws,
              dr.exec_nivcsws,
              coalesce(dr.toplevel, true)
            FROM json_to_record(datarow.row_data) AS dr(
              server_id         integer,
              sample_id         integer,
              userid            oid,
              datid             oid,
              queryid           bigint,
              queryid_md5       character(32),
              plan_user_time    double precision,
              plan_system_time  double precision,
              plan_minflts      bigint,
              plan_majflts      bigint,
              plan_nswaps       bigint,
              plan_reads        bigint,
              plan_writes       bigint,
              plan_msgsnds      bigint,
              plan_msgrcvs      bigint,
              plan_nsignals     bigint,
              plan_nvcsws       bigint,
              plan_nivcsws      bigint,
              exec_user_time    double precision,
              exec_system_time  double precision,
              exec_minflts      bigint,
              exec_majflts      bigint,
              exec_nswaps       bigint,
              exec_reads        bigint,
              exec_writes       bigint,
              exec_msgsnds      bigint,
              exec_msgrcvs      bigint,
              exec_nsignals     bigint,
              exec_nvcsws       bigint,
              exec_nivcsws      bigint,
              toplevel          boolean
              )
            ON CONFLICT ON CONSTRAINT pk_sample_kcache_n DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
          END LOOP; -- over data rows
        WHEN array_position(versions_array, '0.3.1:pg_profile') >= 1 THEN
          LOOP
            FETCH data INTO datarow;
            EXIT WHEN NOT FOUND;
            INSERT INTO sample_kcache(server_id,sample_id,userid,datid,queryid,queryid_md5,
              plan_user_time,plan_system_time,plan_minflts,plan_majflts,
              plan_nswaps,plan_reads,plan_writes,plan_msgsnds,plan_msgrcvs,plan_nsignals,
              plan_nvcsws,plan_nivcsws,exec_user_time,exec_system_time,exec_minflts,
              exec_majflts,exec_nswaps,exec_reads,exec_writes,exec_msgsnds,exec_msgrcvs,
              exec_nsignals,exec_nvcsws,exec_nivcsws,toplevel)
            SELECT
              (srv_map ->> dr.server_id::text)::integer,
              dr.sample_id,
              dr.userid,
              dr.datid,
              dr.queryid,
              CASE WHEN starts_with(versions_array[1], '0.3.1:') THEN
                import_meta #>> ARRAY[dr.queryid_md5, dr.server_id::text]
              ELSE
                dr.queryid_md5
              END,
              dr.plan_user_time,
              dr.plan_system_time,
              dr.plan_minflts,
              dr.plan_majflts,
              dr.plan_nswaps,
              dr.plan_reads,
              dr.plan_writes,
              dr.plan_msgsnds,
              dr.plan_msgrcvs,
              dr.plan_nsignals,
              dr.plan_nvcsws,
              dr.plan_nivcsws,
              dr.exec_user_time,
              dr.exec_system_time,
              dr.exec_minflts,
              dr.exec_majflts,
              dr.exec_nswaps,
              dr.exec_reads,
              dr.exec_writes,
              dr.exec_msgsnds,
              dr.exec_msgrcvs,
              dr.exec_nsignals,
              dr.exec_nvcsws,
              dr.exec_nivcsws,
              coalesce(dr.toplevel, true)
            FROM json_to_record(datarow.row_data) AS dr(
              server_id         integer,
              sample_id         integer,
              userid            oid,
              datid             oid,
              queryid           bigint,
              queryid_md5       character(32),
              plan_user_time    double precision,
              plan_system_time  double precision,
              plan_minflts      bigint,
              plan_majflts      bigint,
              plan_nswaps       bigint,
              plan_reads        bigint,
              plan_writes       bigint,
              plan_msgsnds      bigint,
              plan_msgrcvs      bigint,
              plan_nsignals     bigint,
              plan_nvcsws       bigint,
              plan_nivcsws      bigint,
              exec_user_time    double precision,
              exec_system_time  double precision,
              exec_minflts      bigint,
              exec_majflts      bigint,
              exec_nswaps       bigint,
              exec_reads        bigint,
              exec_writes       bigint,
              exec_msgsnds      bigint,
              exec_msgrcvs      bigint,
              exec_nsignals     bigint,
              exec_nvcsws       bigint,
              exec_nivcsws      bigint,
              toplevel          boolean
              )
            ON CONFLICT ON CONSTRAINT pk_sample_kcache_n DO NOTHING;
            GET DIAGNOSTICS row_proc = ROW_COUNT;
            rowcnt := rowcnt + row_proc;
            IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
          END LOOP; -- over data rows
        ELSE
          RAISE '%', format('[import %s]: Unsupported extension version: %s %s',
            imp_table_name, import_meta ->> 'extension', import_meta ->> 'version');
      END CASE; -- over import versions
    WHEN 'last_stat_database' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_database (server_id,sample_id,datid,datname,xact_commit,
          xact_rollback,blks_read,blks_hit,tup_returned,tup_fetched,tup_inserted,
          tup_updated,tup_deleted,conflicts,temp_files,temp_bytes,deadlocks,
          checksum_failures,checksum_last_failure,
          blk_read_time,blk_write_time,stats_reset,datsize,datsize_delta,datistemplate,
          session_time,active_time,
          idle_in_transaction_time,sessions,sessions_abandoned,sessions_fatal,
          sessions_killed)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.datname,
          dr.xact_commit,
          dr.xact_rollback,
          dr.blks_read,
          dr.blks_hit,
          dr.tup_returned,
          dr.tup_fetched,
          dr.tup_inserted,
          dr.tup_updated,
          dr.tup_deleted,
          dr.conflicts,
          dr.temp_files,
          dr.temp_bytes,
          dr.deadlocks,
          dr.checksum_failures,
          dr.checksum_last_failure,
          dr.blk_read_time,
          dr.blk_write_time,
          dr.stats_reset,
          dr.datsize,
          dr.datsize_delta,
          dr.datistemplate,
          dr.session_time,
          dr.active_time,
          dr.idle_in_transaction_time,
          dr.sessions,
          dr.sessions_abandoned,
          dr.sessions_fatal,
          dr.sessions_killed
        FROM json_to_record(datarow.row_data) AS dr(
          server_id       integer,
          sample_id       integer,
          datid           oid,
          datname         name,
          xact_commit     bigint,
          xact_rollback   bigint,
          blks_read       bigint,
          blks_hit        bigint,
          tup_returned    bigint,
          tup_fetched     bigint,
          tup_inserted    bigint,
          tup_updated     bigint,
          tup_deleted     bigint,
          conflicts       bigint,
          temp_files      bigint,
          temp_bytes      bigint,
          deadlocks       bigint,
          blk_read_time   double precision,
          blk_write_time  double precision,
          stats_reset     timestamp with time zone,
          datsize         bigint,
          datsize_delta   bigint,
          datistemplate   boolean,
          session_time    double precision,
          active_time     double precision,
          idle_in_transaction_time  double precision,
          sessions        bigint,
          sessions_abandoned  bigint,
          sessions_fatal      bigint,
          sessions_killed     bigint,
          checksum_failures   bigint,
          checksum_last_failure timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_tablespaces' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_tablespaces (server_id,sample_id,tablespaceid,tablespacename,
          tablespacepath,size,size_delta)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.tablespaceid,
          dr.tablespacename,
          dr.tablespacepath,
          dr.size,
          dr.size_delta
        FROM json_to_record(datarow.row_data) AS dr(
          server_id       integer,
          sample_id       integer,
          tablespaceid    oid,
          tablespacename  name,
          tablespacepath  text,
          size            bigint,
          size_delta      bigint
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_cluster' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_cluster (server_id,sample_id,checkpoints_timed,
          checkpoints_req,checkpoint_write_time,checkpoint_sync_time,
          buffers_checkpoint,buffers_clean,maxwritten_clean,buffers_backend,
          buffers_backend_fsync,buffers_alloc,stats_reset,wal_size)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.checkpoints_timed,
          dr.checkpoints_req,
          dr.checkpoint_write_time,
          dr.checkpoint_sync_time,
          dr.buffers_checkpoint,
          dr.buffers_clean,
          dr.maxwritten_clean,
          dr.buffers_backend,
          dr.buffers_backend_fsync,
          dr.buffers_alloc,
          dr.stats_reset,
          dr.wal_size
        FROM json_to_record(datarow.row_data) AS dr(
          server_id              integer,
          sample_id              integer,
          checkpoints_timed      bigint,
          checkpoints_req        bigint,
          checkpoint_write_time  double precision,
          checkpoint_sync_time   double precision,
          buffers_checkpoint     bigint,
          buffers_clean          bigint,
          maxwritten_clean       bigint,
          buffers_backend        bigint,
          buffers_backend_fsync  bigint,
          buffers_alloc          bigint,
          stats_reset            timestamp with time zone,
          wal_size               bigint
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_archiver' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_archiver (server_id,sample_id,archived_count,last_archived_wal,
          last_archived_time,failed_count,last_failed_wal,last_failed_time,stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.archived_count,
          dr.last_archived_wal,
          dr.last_archived_time,
          dr.failed_count,
          dr.last_failed_wal,
          dr.last_failed_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
          server_id           integer,
          sample_id           integer,
          archived_count      bigint,
          last_archived_wal   text,
          last_archived_time  timestamp with time zone,
          failed_count        bigint,
          last_failed_wal     text,
          last_failed_time    timestamp with time zone,
          stats_reset         timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_tables' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_tables (server_id,sample_id,datid,relid,schemaname,relname,
          seq_scan,seq_tup_read,idx_scan,idx_tup_fetch,n_tup_ins,n_tup_upd,n_tup_del,
          n_tup_hot_upd,n_live_tup,n_dead_tup,n_mod_since_analyze,n_ins_since_vacuum,
          last_vacuum,last_autovacuum,last_analyze,last_autoanalyze,vacuum_count,
          autovacuum_count,analyze_count,autoanalyze_count,heap_blks_read,heap_blks_hit,
          idx_blks_read,idx_blks_hit,toast_blks_read,toast_blks_hit,tidx_blks_read,
          tidx_blks_hit,relsize,relsize_diff,tablespaceid,reltoastrelid,relkind,in_sample,
          relpages_bytes,relpages_bytes_diff,last_seq_scan,last_idx_scan,n_tup_newpage_upd)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.relid,
          dr.schemaname,
          dr.relname,
          dr.seq_scan,
          dr.seq_tup_read,
          dr.idx_scan,
          dr.idx_tup_fetch,
          dr.n_tup_ins,
          dr.n_tup_upd,
          dr.n_tup_del,
          dr.n_tup_hot_upd,
          dr.n_live_tup,
          dr.n_dead_tup,
          dr.n_mod_since_analyze,
          dr.n_ins_since_vacuum,
          dr.last_vacuum,
          dr.last_autovacuum,
          dr.last_analyze,
          dr.last_autoanalyze,
          dr.vacuum_count,
          dr.autovacuum_count,
          dr.analyze_count,
          dr.autoanalyze_count,
          dr.heap_blks_read,
          dr.heap_blks_hit,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.toast_blks_read,
          dr.toast_blks_hit,
          dr.tidx_blks_read,
          dr.tidx_blks_hit,
          dr.relsize,
          dr.relsize_diff,
          dr.tablespaceid,
          dr.reltoastrelid,
          dr.relkind,
          COALESCE(dr.in_sample, false),
          dr.relpages_bytes,
          dr.relpages_bytes_diff,
          dr.last_seq_scan,
          dr.last_idx_scan,
          dr.n_tup_newpage_upd
        FROM json_to_record(datarow.row_data) AS dr(
          server_id            integer,
          sample_id            integer,
          datid                oid,
          relid                oid,
          schemaname           name,
          relname              name,
          seq_scan             bigint,
          seq_tup_read         bigint,
          idx_scan             bigint,
          idx_tup_fetch        bigint,
          n_tup_ins            bigint,
          n_tup_upd            bigint,
          n_tup_del            bigint,
          n_tup_hot_upd        bigint,
          n_live_tup           bigint,
          n_dead_tup           bigint,
          n_mod_since_analyze  bigint,
          n_ins_since_vacuum   bigint,
          last_vacuum          timestamp with time zone,
          last_autovacuum      timestamp with time zone,
          last_analyze         timestamp with time zone,
          last_autoanalyze     timestamp with time zone,
          vacuum_count         bigint,
          autovacuum_count     bigint,
          analyze_count        bigint,
          autoanalyze_count    bigint,
          heap_blks_read       bigint,
          heap_blks_hit        bigint,
          idx_blks_read        bigint,
          idx_blks_hit         bigint,
          toast_blks_read      bigint,
          toast_blks_hit       bigint,
          tidx_blks_read       bigint,
          tidx_blks_hit        bigint,
          relsize              bigint,
          relsize_diff         bigint,
          tablespaceid         oid,
          reltoastrelid        oid,
          relkind              character(1),
          in_sample            boolean,
          relpages_bytes       bigint,
          relpages_bytes_diff  bigint,
          last_seq_scan        timestamp with time zone,
          last_idx_scan        timestamp with time zone,
          n_tup_newpage_upd    bigint
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_indexes' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_indexes (server_id,sample_id,datid,relid,indexrelid,
          schemaname,relname,indexrelname,idx_scan,idx_tup_read,idx_tup_fetch,
          idx_blks_read,idx_blks_hit,relsize,relsize_diff,tablespaceid,indisunique,
          in_sample,relpages_bytes,relpages_bytes_diff,last_idx_scan)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.relid,
          dr.indexrelid,
          dr.schemaname,
          dr.relname,
          dr.indexrelname,
          dr.idx_scan,
          dr.idx_tup_read,
          dr.idx_tup_fetch,
          dr.idx_blks_read,
          dr.idx_blks_hit,
          dr.relsize,
          dr.relsize_diff,
          dr.tablespaceid,
          dr.indisunique,
          COALESCE(dr.in_sample, false),
          dr.relpages_bytes,
          dr.relpages_bytes_diff,
          dr.last_idx_scan
        FROM json_to_record(datarow.row_data) AS dr(
          server_id      integer,
          sample_id      integer,
          datid          oid,
          relid          oid,
          indexrelid     oid,
          schemaname     name,
          relname        name,
          indexrelname   name,
          idx_scan       bigint,
          idx_tup_read   bigint,
          idx_tup_fetch  bigint,
          idx_blks_read  bigint,
          idx_blks_hit   bigint,
          relsize        bigint,
          relsize_diff   bigint,
          tablespaceid   oid,
          indisunique    boolean,
          in_sample      boolean,
          relpages_bytes bigint,
          relpages_bytes_diff bigint,
          last_idx_scan  timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_user_functions' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_user_functions (server_id,sample_id,datid,funcid,schemaname,
          funcname,funcargs,calls,total_time,self_time,trg_fn,in_sample)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.datid,
          dr.funcid,
          dr.schemaname,
          dr.funcname,
          dr.funcargs,
          dr.calls,
          dr.total_time,
          dr.self_time,
          dr.trg_fn,
          COALESCE(dr.in_sample, false)
        FROM json_to_record(datarow.row_data) AS dr(
          server_id   integer,
          sample_id   integer,
          datid       oid,
          funcid      oid,
          schemaname  name,
          funcname    name,
          funcargs    text,
          calls       bigint,
          total_time  double precision,
          self_time   double precision,
          trg_fn      boolean,
          in_sample   boolean
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_wal' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_wal (server_id,sample_id,wal_records,
          wal_fpi,wal_bytes,wal_buffers_full,wal_write,wal_sync,
          wal_write_time,wal_sync_time,stats_reset)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.wal_records,
          dr.wal_fpi,
          dr.wal_bytes,
          dr.wal_buffers_full,
          dr.wal_write,
          dr.wal_sync,
          dr.wal_write_time,
          dr.wal_sync_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
          server_id           integer,
          sample_id           integer,
          wal_records         bigint,
          wal_fpi             bigint,
          wal_bytes           numeric,
          wal_buffers_full    bigint,
          wal_write           bigint,
          wal_sync            bigint,
          wal_write_time      double precision,
          wal_sync_time       double precision,
          stats_reset         timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_kcache' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_kcache (server_id,sample_id,userid,datid,toplevel,queryid,
          plan_user_time,plan_system_time,plan_minflts,plan_majflts,
          plan_nswaps,plan_reads,plan_writes,plan_msgsnds,plan_msgrcvs,plan_nsignals,
          plan_nvcsws,plan_nivcsws,exec_user_time,exec_system_time,exec_minflts,
          exec_majflts,exec_nswaps,exec_reads,exec_writes,exec_msgsnds,exec_msgrcvs,
          exec_nsignals,exec_nvcsws,exec_nivcsws
        )
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.userid,
          dr.datid,
          dr.toplevel,
          dr.queryid,
          dr.plan_user_time,
          dr.plan_system_time,
          dr.plan_minflts,
          dr.plan_majflts,
          dr.plan_nswaps,
          dr.plan_reads,
          dr.plan_writes,
          dr.plan_msgsnds,
          dr.plan_msgrcvs,
          dr.plan_nsignals,
          dr.plan_nvcsws,
          dr.plan_nivcsws,
          dr.exec_user_time,
          dr.exec_system_time,
          dr.exec_minflts,
          dr.exec_majflts,
          dr.exec_nswaps,
          dr.exec_reads,
          dr.exec_writes,
          dr.exec_msgsnds,
          dr.exec_msgrcvs,
          dr.exec_nsignals,
          dr.exec_nvcsws,
          dr.exec_nivcsws
        FROM json_to_record(datarow.row_data) AS dr(
          server_id         integer,
          sample_id         integer,
          userid            oid,
          datid             oid,
          toplevel          boolean,
          queryid           bigint,
          plan_user_time    double precision,
          plan_system_time  double precision,
          plan_minflts      bigint,
          plan_majflts      bigint,
          plan_nswaps       bigint,
          plan_reads        bigint,
          plan_writes       bigint,
          plan_msgsnds      bigint,
          plan_msgrcvs      bigint,
          plan_nsignals     bigint,
          plan_nvcsws       bigint,
          plan_nivcsws      bigint,
          exec_user_time    double precision,
          exec_system_time  double precision,
          exec_minflts      bigint,
          exec_majflts      bigint,
          exec_nswaps       bigint,
          exec_reads        bigint,
          exec_writes       bigint,
          exec_msgsnds      bigint,
          exec_msgrcvs      bigint,
          exec_nsignals     bigint,
          exec_nvcsws       bigint,
          exec_nivcsws      bigint
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_statements' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_statements (server_id,sample_id,userid,username,datid,queryid,
          queryid_md5,plans,total_plan_time,min_plan_time,max_plan_time,mean_plan_time,
          stddev_plan_time,calls,total_exec_time,min_exec_time,max_exec_time,mean_exec_time,
          stddev_exec_time,rows,shared_blks_hit,shared_blks_read,shared_blks_dirtied,
          shared_blks_written,local_blks_hit,local_blks_read,local_blks_dirtied,
          local_blks_written,temp_blks_read,temp_blks_written,blk_read_time,blk_write_time,
          wal_records,wal_fpi,wal_bytes,toplevel,in_sample
          ,jit_functions,jit_generation_time,jit_inlining_count,jit_inlining_time,
          jit_optimization_count,jit_optimization_time,jit_emission_count,jit_emission_time,
          temp_blk_read_time,temp_blk_write_time
          )
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.userid,
          dr.username,
          dr.datid,
          dr.queryid,
          dr.queryid_md5,
          dr.plans,
          dr.total_plan_time,
          dr.min_plan_time,
          dr.max_plan_time,
          dr.mean_plan_time,
          dr.stddev_plan_time,
          dr.calls,
          dr.total_exec_time,
          dr.min_exec_time,
          dr.max_exec_time,
          dr.mean_exec_time,
          dr.stddev_exec_time,
          dr.rows,
          dr.shared_blks_hit,
          dr.shared_blks_read,
          dr.shared_blks_dirtied,
          dr.shared_blks_written,
          dr.local_blks_hit,
          dr.local_blks_read,
          dr.local_blks_dirtied,
          dr.local_blks_written,
          dr.temp_blks_read,
          dr.temp_blks_written,
          dr.blk_read_time,
          dr.blk_write_time,
          dr.wal_records,
          dr.wal_fpi,
          dr.wal_bytes,
          dr.toplevel,
          dr.in_sample,
          dr.jit_functions,
          dr.jit_generation_time,
          dr.jit_inlining_count,
          dr.jit_inlining_time,
          dr.jit_optimization_count,
          dr.jit_optimization_time,
          dr.jit_emission_count,
          dr.jit_emission_time,
          dr.temp_blk_read_time,
          dr.temp_blk_write_time
        FROM json_to_record(datarow.row_data) AS dr(
          server_id            integer,
          sample_id            integer,
          userid               oid,
          username             name,
          datid                oid,
          queryid              bigint,
          queryid_md5          character(32),
          plans                bigint,
          total_plan_time      double precision,
          min_plan_time        double precision,
          max_plan_time        double precision,
          mean_plan_time       double precision,
          stddev_plan_time     double precision,
          calls                bigint,
          total_exec_time      double precision,
          min_exec_time        double precision,
          max_exec_time        double precision,
          mean_exec_time       double precision,
          stddev_exec_time     double precision,
          rows                 bigint,
          shared_blks_hit      bigint,
          shared_blks_read     bigint,
          shared_blks_dirtied  bigint,
          shared_blks_written  bigint,
          local_blks_hit       bigint,
          local_blks_read      bigint,
          local_blks_dirtied   bigint,
          local_blks_written   bigint,
          temp_blks_read       bigint,
          temp_blks_written    bigint,
          blk_read_time        double precision,
          blk_write_time       double precision,
          wal_records          bigint,
          wal_fpi              bigint,
          wal_bytes            numeric,
          toplevel             boolean,
          in_sample            boolean,
          jit_functions        bigint,
          jit_generation_time  double precision,
          jit_inlining_count   bigint,
          jit_inlining_time    double precision,
          jit_optimization_count  bigint,
          jit_optimization_time   double precision,
          jit_emission_count   bigint,
          jit_emission_time    double precision,
          temp_blk_read_time   double precision,
          temp_blk_write_time  double precision
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_io' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_io (server_id,sample_id,backend_type,object,context,reads,
          read_time,writes,write_time,writebacks,writeback_time,extends,extend_time,
          op_bytes,hits,evictions,reuses,fsyncs,fsync_time,stats_reset
          )
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.backend_type,
          dr.object,
          dr.context,
          dr.reads,
          dr.read_time,
          dr.writes,
          dr.write_time,
          dr.writebacks,
          dr.writeback_time,
          dr.extends,
          dr.extend_time,
          dr.op_bytes,
          dr.hits,
          dr.evictions,
          dr.reuses,
          dr.fsyncs,
          dr.fsync_time,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
          server_id         integer,
          sample_id         integer,
          backend_type      text,
          object            text,
          context           text,
          reads             bigint,
          read_time         double precision,
          writes            bigint,
          write_time        double precision,
          writebacks        bigint,
          writeback_time    double precision,
          extends           bigint,
          extend_time       double precision,
          op_bytes          bigint,
          hits              bigint,
          evictions         bigint,
          reuses            bigint,
          fsyncs            bigint,
          fsync_time        double precision,
          stats_reset       timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_slru' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_slru (server_id,sample_id,name,blks_zeroed,
          blks_hit,blks_read,blks_written,blks_exists,flushes,truncates,
          stats_reset
          )
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.name,
          dr.blks_zeroed,
          dr.blks_hit,
          dr.blks_read,
          dr.blks_written,
          dr.blks_exists,
          dr.flushes,
          dr.truncates,
          dr.stats_reset
        FROM json_to_record(datarow.row_data) AS dr(
          server_id      integer,
          sample_id      integer,
          name           text,
          blks_zeroed    bigint,
          blks_hit       bigint,
          blks_read      bigint,
          blks_written   bigint,
          blks_exists    bigint,
          flushes        bigint,
          truncates      bigint,
          stats_reset    timestamp with time zone
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'act_query' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO act_query(server_id, act_query_md5, act_query, last_sample_id)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.act_query_md5,
          dr.act_query,
          s.sample_id
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            act_query_md5  char(32),
            act_query      text,
            last_sample_id integer
          )
          LEFT JOIN samples s ON (s.server_id, s.sample_id) =
            ((srv_map ->> dr.server_id::text)::integer, dr.last_sample_id)
        ON CONFLICT ON CONSTRAINT pk_act_query DO
          NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'server_subsample' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO server_subsample(server_id, subsample_enabled, min_query_dur, min_xact_dur,
          min_xact_age, min_idle_xact_dur)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.subsample_enabled,
          dr.min_query_dur,
          dr.min_xact_dur,
          dr.min_xact_age,
          dr.min_idle_xact_dur
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            subsample_enabled boolean,
            min_query_dur     interval hour to second,
            min_xact_dur      interval hour to second,
            min_xact_age      bigint,
            min_idle_xact_dur interval hour to second
          )
        ON CONFLICT ON CONSTRAINT pk_server_subsample DO
        UPDATE SET (subsample_enabled, min_query_dur, min_xact_dur,
          min_xact_age, min_idle_xact_dur) =
          (EXCLUDED.subsample_enabled, EXCLUDED.min_query_dur, EXCLUDED.min_xact_dur,
          EXCLUDED.min_xact_age, EXCLUDED.min_idle_xact_dur)
        WHERE (server_subsample.subsample_enabled, server_subsample.min_query_dur,
          server_subsample.min_xact_dur, server_subsample.min_xact_age,
          server_subsample.min_idle_xact_dur)
          IS DISTINCT FROM
          (EXCLUDED.subsample_enabled, EXCLUDED.min_query_dur, EXCLUDED.min_xact_dur,
          EXCLUDED.min_xact_age, EXCLUDED.min_idle_xact_dur);
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_backend' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_backend(server_id, sample_id, pid,
          backend_start, datid, datname, usesysid, usename,
          client_addr, client_hostname, client_port, backend_type,
          backend_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.datid,
          dr.datname,
          dr.usesysid,
          dr.usename,
          dr.client_addr,
          dr.client_hostname,
          dr.client_port,
          dr.backend_type,
          dr.backend_last_ts
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            datid             oid,
            datname           name,
            usesysid          oid,
            usename           name,
            application_name  text,
            client_addr       inet,
            client_hostname   text,
            client_port       integer,
            backend_type      text,
            backend_last_ts   timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_backends DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_xact' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_xact(server_id, sample_id, pid, backend_start, xact_start,
          backend_xid, xact_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.xact_start,
          dr.backend_xid,
          dr.xact_last_ts
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            xact_start        timestamp with time zone,
            backend_xid       xid,
            xact_last_ts      timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_xact DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_backend_state' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_backend_state(server_id, sample_id,
          pid, backend_start, application_name, state_code,
          state_change, state_last_ts, xact_start, backend_xmin,
          backend_xmin_age)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.application_name,
          dr.state_code,
          dr.state_change,
          dr.state_last_ts,
          dr.xact_start,
          dr.backend_xmin,
          dr.backend_xmin_age
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            application_name  text,
            state_code        integer,
            state_change      timestamp with time zone,
            state_last_ts     timestamp with time zone,
            xact_start        timestamp with time zone,
            backend_xmin      xid,
            backend_xmin_age  bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_bk_state DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_statement' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_statement(server_id, sample_id, pid, leader_pid, state_change,
          query_start, query_id, act_query_md5, stmt_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.leader_pid,
          dr.state_change,
          dr.query_start,
          dr.query_id,
          dr.act_query_md5,
          dr.stmt_last_ts
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            leader_pid        integer,
            state_change      timestamp with time zone,
            query_start       timestamp with time zone,
            query_id          bigint,
            act_query_md5     char(32),
            stmt_last_ts      timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_act_stmt DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    ELSE
      rows_processed := -1;
      RETURN; -- table not found
  END CASE; -- over table name
  rows_processed := rowcnt;
  RETURN;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION import_section_data_subsample(IN data refcursor, IN imp_table_name name, IN srv_map jsonb,
  IN import_meta jsonb, IN versions_array text[],
  OUT rows_processed bigint, OUT new_import_meta jsonb)
SET search_path=profile AS $$
DECLARE
  datarow          record;
  rowcnt           bigint = 0;
  row_proc         bigint = 0;
  section_meta     jsonb;
BEGIN
  new_import_meta := import_meta;
  CASE imp_table_name
    WHEN 'last_stat_activity_count' THEN NULL; -- skip table withoup PK
    WHEN 'server_subsample' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO server_subsample(server_id, subsample_enabled, min_query_dur,
          min_xact_dur, min_xact_age, min_idle_xact_dur)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.subsample_enabled,
          dr.min_query_dur,
          dr.min_xact_dur,
          dr.min_xact_age,
          dr.min_idle_xact_dur
        FROM json_to_record(datarow.row_data) AS dr(
            server_id           integer,
            subsample_enabled   boolean,
            min_query_dur       interval hour to second,
            min_xact_dur        interval hour to second,
            min_xact_age        bigint,
            min_idle_xact_dur   interval hour to second
          )
        ON CONFLICT ON CONSTRAINT pk_server_subsample DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_backend' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_backend(server_id, sample_id, pid, backend_start, datid,
          datname, usesysid, usename, application_name, client_addr, client_hostname,
          client_port, backend_type, backend_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.datid,
          dr.datname,
          dr.usesysid,
          dr.usename,
          dr.application_name,
          dr.client_addr,
          dr.client_hostname,
          dr.client_port,
          dr.backend_type,
          dr.backend_last_ts,
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            datid             oid,
            datname           name,
            usesysid          oid,
            usename           name,
            application_name  text,
            client_addr       inet,
            client_hostname   text,
            client_port       integer,
            backend_type      text,
            backend_last_ts   timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_backends DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_xact' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_xact(server_id, sample_id, pid, backend_start,xact_start,
          backend_xid,xact_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.xact_start,
          dr.backend_xid,
          dr.xact_last_ts
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            xact_start        timestamp with time zone,
            backend_xid       xid,
            xact_last_ts      timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_xact DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_backend_state' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_backend_state(server_id, sample_id,
          pid, backend_start, state_code, state_last_ts, xact_start,
          backend_xmin, backend_xmin_age)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.backend_start,
          dr.state_code,
          dr.state_change,
          dr.state_last_ts,
          dr.xact_start,
          dr.backend_xmin,
          dr.backend_xmin_age
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            backend_start     timestamp with time zone,
            state_code        integer,
            state_change      timestamp with time zone,
            state_last_ts     timestamp with time zone,
            xact_start        timestamp with time zone,
            backend_xmin      xid,
            backend_xmin_age  bigint
          )
        ON CONFLICT ON CONSTRAINT pk_sample_bk_state DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'act_query' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO act_query(server_id, act_query_md5, act_query, last_sample_id)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          act_query_md5,
          act_query,
          NULL
        FROM json_to_record(datarow.row_data) AS dr(
            server_id      integer,
            act_query_md5  char(32),
            act_query      text,
            last_sample_id integer
          )
        ON CONFLICT ON CONSTRAINT pk_act_query DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_act_statement' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_act_statement(server_id, sample_id, pid,
          leader_pid, state_change, query_start, query_id,
          act_query_md5, stmt_last_ts)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.pid,
          dr.leader_pid,
          dr.state_change,
          dr.query_start,
          dr.query_id,
          dr.act_query_md5,
          dr.stmt_last_ts
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            pid               integer,
            leader_pid        integer,
            state_change      timestamp with time zone,
            query_start       timestamp with time zone,
            query_id          bigint,
            act_query_md5     char(32),
            stmt_last_ts      timestamp with time zone
          )
        ON CONFLICT ON CONSTRAINT pk_sample_act_stmt DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_activity' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_activity(server_id, sample_id,
          subsample_ts, datid, datname, pid, leader_pid, usesysid,
          usename, application_name, client_addr, client_hostname,
          client_port, backend_start, xact_start, query_start,
          state_change, state, backend_xid, backend_xmin, query_id,
          query, backend_type, backend_xmin_age
        )
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id         integer,
          dr.subsample_ts      timestamp with time zone,
          dr.datid             oid,
          dr.datname           name,
          dr.pid               integer,
          dr.leader_pid        integer,
          dr.usesysid          oid,
          dr.usename           name,
          dr.application_name  text,
          dr.client_addr       inet,
          dr.client_hostname   text,
          dr.client_port       integer,
          dr.backend_start     timestamp with time zone,
          dr.xact_start        timestamp with time zone,
          dr.query_start       timestamp with time zone,
          dr.state_change      timestamp with time zone,
          dr.state             text,
          dr.backend_xid       xid,
          dr.backend_xmin      xid,
          dr.query_id          bigint,
          dr.query             text,
          dr.backend_type      text,
          dr.backend_xmin_age  bigint
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            subsample_ts      timestamp with time zone,
            datid             oid,
            datname           name,
            pid               integer,
            leader_pid        integer,
            usesysid          oid,
            usename           name,
            application_name  text,
            client_addr       inet,
            client_hostname   text,
            client_port       integer,
            backend_start     timestamp with time zone,
            xact_start        timestamp with time zone,
            query_start       timestamp with time zone,
            state_change      timestamp with time zone,
            state             text,
            backend_xid       xid,
            backend_xmin      xid,
            query_id          bigint,
            query             text,
            backend_type      text,
            backend_xmin_age  bigint
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'session_attr' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO session_attr(server_id, sess_attr_id,
          backend_type, datid, datname, usesysid, usename,
          application_name, client_addr, last_sample_id)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sess_attr_id      serial,
          dr.backend_type      text,
          dr.datid             oid,
          dr.datname           name,
          dr.usesysid          oid,
          dr.usename           name,
          dr.application_name  text,
          dr.client_addr       inet,
          dr.last_sample_id    integer
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sess_attr_id      integer,
            backend_type      text,
            datid             oid,
            datname           name,
            usesysid          oid,
            usename           name,
            application_name  text,
            client_addr       inet,
            last_sample_id    integer
          )
        ON CONFLICT ON CONSTRAINT pk_subsample_sa DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'sample_stat_activity_cnt' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO sample_stat_activity_cnt(server_id, sample_id,
          subsample_ts, sess_attr_id, total, active, idle, idle_t,
          idle_ta, state_null, lwlock, lock, bufferpin, activity,
          extension, client, ipc, timeout, io)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.subsample_ts,
          dr.sess_attr_id,
          dr.total,
          dr.active,
          dr.idle,
          dr.idle_t,
          dr.idle_ta,
          dr.state_null,
          dr.lwlock,
          dr.lock,
          dr.bufferpin,
          dr.activity,
          dr.extension,
          dr.client,
          dr.ipc,
          dr.timeout,
          dr.io
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            subsample_ts      timestamp with time zone,
            sess_attr_id      integer,
            total             integer,
            active            integer,
            idle              integer,
            idle_t            integer,
            idle_ta           integer,
            state_null        integer,
            lwlock            integer,
            lock              integer,
            bufferpin         integer,
            activity          integer,
            extension         integer,
            client            integer,
            ipc               integer,
            timeout           integer,
            io                integer
          )
        ON CONFLICT ON CONSTRAINT pk_sample_stat_activity_cnt DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    WHEN 'last_stat_activity_count' THEN
      LOOP
        FETCH data INTO datarow;
        EXIT WHEN NOT FOUND;
        INSERT INTO last_stat_activity_count(server_id, sample_id,
          subsample_ts, backend_type, datid, datname, usesysid,
          usename, application_name, client_addr, total, active, idle,
          idle_t, idle_ta, state_null, lwlock, lock, bufferpin,
          activity, extension, client, ipc, timeout, io)
        SELECT
          (srv_map ->> dr.server_id::text)::integer,
          dr.sample_id,
          dr.subsample_ts,
          dr.backend_type,
          dr.datid,
          dr.datname,
          dr.usesysid,
          dr.usename,
          dr.application_name,
          dr.client_addr,
          dr.total,
          dr.active,
          dr.idle,
          dr.idle_t,
          dr.idle_ta,
          dr.state_null,
          dr.lwlock,
          dr.lock,
          dr.bufferpin,
          dr.activity,
          dr.extension,
          dr.client,
          dr.ipc,
          dr.timeout,
          dr.io
        FROM json_to_record(datarow.row_data) AS dr(
            server_id         integer,
            sample_id         integer,
            subsample_ts      timestamp with time zone,
            backend_type      text,
            datid             oid,
            datname           name,
            usesysid          oid,
            usename           name,
            application_name  text,
            client_addr       inet,
            total             integer,
            active            integer,
            idle              integer,
            idle_t            integer,
            idle_ta           integer,
            state_null        integer,
            lwlock            integer,
            lock              integer,
            bufferpin         integer,
            activity          integer,
            extension         integer,
            client            integer,
            ipc               integer,
            timeout           integer,
            io                integer
          )
        ON CONFLICT DO NOTHING;
        GET DIAGNOSTICS row_proc = ROW_COUNT;
        rowcnt := rowcnt + row_proc;
        IF (rowcnt > 0 AND rowcnt % 1000 = 0) THEN
          RAISE NOTICE '%', format('Table %s processed: %s rows', imp_table_name, rowcnt);
        END IF;
      END LOOP; -- over data rows
    ELSE
      rows_processed := -1;
      RETURN; -- table not found
  END CASE; -- over table name
  rows_processed := rowcnt;
  RETURN;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION init_sample(IN sserver_id integer
) RETURNS jsonb SET search_path=@extschema@ SET lock_timeout=300000 AS $$
DECLARE
    server_properties jsonb = '{"extensions":[],"settings":[],"timings":{},"properties":{}}'; -- version, extensions, etc.
    qres              record;
    qres_subsample    record;
    server_connstr    text;

    server_query      text;
    server_host       text = NULL;
BEGIN
    -- Get server connstr
    SELECT properties INTO server_properties FROM get_connstr(sserver_id, server_properties);

    -- Getting timing collection setting
    BEGIN
        SELECT current_setting('pg_profile.track_sample_timings')::boolean AS collect_timings
          INTO qres;
        server_properties := jsonb_set(server_properties,
          '{collect_timings}',
          to_jsonb(qres.collect_timings)
        );
    EXCEPTION
        WHEN OTHERS THEN
          server_properties := jsonb_set(server_properties,
            '{collect_timings}',
            to_jsonb(false)
          );
    END;

    -- Getting TopN setting
    BEGIN
        SELECT current_setting('pg_profile.topn')::integer AS topn INTO qres;
        server_properties := jsonb_set(server_properties,'{properties,topn}',to_jsonb(qres.topn));
    EXCEPTION
        WHEN OTHERS THEN
          server_properties := jsonb_set(server_properties,
            '{properties,topn}',
            to_jsonb(20)
          );
    END;

    -- Adding dblink extension schema to search_path if it does not already there
    IF (SELECT count(*) = 0 FROM pg_catalog.pg_extension WHERE extname = 'dblink') THEN
      RAISE 'dblink extension must be installed';
    END IF;

    SELECT extnamespace::regnamespace AS dblink_schema INTO STRICT qres FROM pg_catalog.pg_extension WHERE extname = 'dblink';
    IF NOT string_to_array(current_setting('search_path'),', ') @> ARRAY[qres.dblink_schema::text] THEN
      EXECUTE 'SET LOCAL search_path TO ' || current_setting('search_path')||','|| qres.dblink_schema;
    END IF;

    IF dblink_get_connections() @> ARRAY['server_connection'] THEN
        PERFORM dblink_disconnect('server_connection');
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,connect}',jsonb_build_object('start',clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,total}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Server connection
    PERFORM dblink_connect('server_connection', server_properties #>> '{properties,server_connstr}');
    -- Transaction
    PERFORM dblink('server_connection','BEGIN TRANSACTION ISOLATION LEVEL REPEATABLE READ READ ONLY');
    -- Setting application name
    PERFORM dblink('server_connection','SET application_name=''pg_profile''');
    -- Setting lock_timout prevents hanging of take_sample() call due to DDL in long transaction
    PERFORM dblink('server_connection','SET lock_timeout=3000');
    -- Reset search_path for security reasons
    PERFORM dblink('server_connection','SET search_path=''''');

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,connect,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,get server environment}',jsonb_build_object('start',clock_timestamp()));
    END IF;
    -- Get settings values for the server
    FOR qres IN
      SELECT * FROM dblink('server_connection',
          'SELECT name, '
          'reset_val, '
          'unit, '
          'pending_restart '
          'FROM pg_catalog.pg_settings '
          'WHERE name IN ('
            '''server_version_num'''
          ')')
        AS dbl(name text, reset_val text, unit text, pending_restart boolean)
    LOOP
      server_properties := jsonb_insert(server_properties,'{"settings",0}',to_jsonb(qres));
    END LOOP;

    -- Is it PostgresPro?
    IF (SELECT pgpro_fxs = 3
        FROM dblink('server_connection',
          'select count(1) as pgpro_fxs '
          'from pg_catalog.pg_proc '
          'where proname IN (''pgpro_build'',''pgpro_edition'',''pgpro_version'')'
        ) AS pgpro (pgpro_fxs integer))
    THEN
      server_properties := jsonb_set(server_properties,'{properties,pgpro}',to_jsonb(true));
    ELSE
      server_properties := jsonb_set(server_properties,'{properties,pgpro}',to_jsonb(false));
    END IF;

    -- Get extensions, that we need to perform statements stats collection
    FOR qres IN
      SELECT * FROM dblink('server_connection',
          'SELECT extname, '
          'extnamespace::regnamespace::name AS extnamespace, '
          'extversion '
          'FROM pg_catalog.pg_extension '
          'WHERE extname IN ('
            '''pg_stat_statements'','
            '''pg_wait_sampling'','
            '''pg_stat_kcache'''
          ')')
        AS dbl(extname name, extnamespace name, extversion text)
    LOOP
      server_properties := jsonb_insert(server_properties,'{"extensions",0}',to_jsonb(qres));
    END LOOP;

    -- Check system identifier
    WITH remote AS (
      SELECT
        dbl.system_identifier
      FROM dblink('server_connection',
        'SELECT system_identifier '
        'FROM pg_catalog.pg_control_system()'
      ) AS dbl (system_identifier bigint)
    )
    SELECT min(reset_val::bigint) != (
        SELECT
          system_identifier
        FROM remote
      ) AS sysid_changed,
      (
        SELECT
          s.server_name = 'local' AND cs.system_identifier != r.system_identifier
        FROM
          pg_catalog.pg_control_system() cs
          CROSS JOIN remote r
          JOIN servers s ON (s.server_id = sserver_id)
      ) AS local_missmatch
      INTO STRICT qres
    FROM sample_settings
    WHERE server_id = sserver_id AND name = 'system_identifier';
    IF qres.sysid_changed THEN
      RAISE 'Server system_identifier has changed! '
        'Ensure server connection string is correct. '
        'Consider creating a new server for this cluster.';
    END IF;
    IF qres.local_missmatch THEN
      RAISE 'Local system_identifier does not match '
        'with server specified by connection string of '
        '"local" server';
    END IF;

    -- Subsample settings collection
    -- Get last base sample identifier of a server
    SELECT
      last_sample_id,
      subsample_enabled,
      min_query_dur,
      min_xact_dur,
      min_xact_age,
      min_idle_xact_dur
      INTO STRICT qres_subsample
    FROM servers JOIN server_subsample USING (server_id)
    WHERE server_id = sserver_id;

    server_properties := jsonb_set(server_properties,
      '{properties,last_sample_id}',
      to_jsonb(qres_subsample.last_sample_id)
    );

    /* Getting subsample GUC thresholds used as defaults*/
    BEGIN
        SELECT current_setting('pg_profile.subsample_enabled')::boolean AS subsample_enabled INTO qres;
        server_properties := jsonb_set(
          server_properties,
          '{properties,subsample_enabled}',
          to_jsonb(COALESCE(qres_subsample.subsample_enabled, qres.subsample_enabled))
        );
    EXCEPTION
        WHEN OTHERS THEN
          server_properties := jsonb_set(server_properties,
            '{properties,subsample_enabled}',
            to_jsonb(COALESCE(qres_subsample.subsample_enabled, true))
          );
    END;

    -- Setup subsample settings when they are enabled
    IF (server_properties #>> '{properties,subsample_enabled}')::boolean THEN
      BEGIN
          SELECT current_setting('pg_profile.min_query_duration')::interval AS min_query_dur INTO qres;
          server_properties := jsonb_set(
            server_properties,
            '{properties,min_query_dur}',
            to_jsonb(COALESCE(qres_subsample.min_query_dur, qres.min_query_dur))
          );
      EXCEPTION
          WHEN OTHERS THEN
            server_properties := jsonb_set(server_properties,
              '{properties,min_query_dur}',
              COALESCE (
                to_jsonb(qres_subsample.min_query_dur)
                , 'null'::jsonb
              )
            );
      END;

      BEGIN
          SELECT current_setting('pg_profile.min_xact_duration')::interval AS min_xact_dur INTO qres;
          server_properties := jsonb_set(
            server_properties,
            '{properties,min_xact_dur}',
            to_jsonb(COALESCE(qres_subsample.min_xact_dur, qres.min_xact_dur))
          );
      EXCEPTION
          WHEN OTHERS THEN
            server_properties := jsonb_set(server_properties,
              '{properties,min_xact_dur}',
              COALESCE (
                to_jsonb(qres_subsample.min_xact_dur)
                , 'null'::jsonb
              )
            );
      END;

      BEGIN
          SELECT current_setting('pg_profile.min_xact_age')::integer AS min_xact_age INTO qres;
          server_properties := jsonb_set(
            server_properties,
            '{properties,min_xact_age}',
            to_jsonb(COALESCE(qres_subsample.min_xact_age, qres.min_xact_age))
          );
      EXCEPTION
          WHEN OTHERS THEN
            server_properties := jsonb_set(server_properties,
              '{properties,min_xact_age}',
              COALESCE (
                to_jsonb(qres_subsample.min_xact_age)
                , 'null'::jsonb
              )
            );
      END;

      BEGIN
          SELECT current_setting('pg_profile.min_idle_xact_duration')::interval AS min_idle_xact_dur INTO qres;
          server_properties := jsonb_set(
            server_properties,
            '{properties,min_idle_xact_dur}',
            to_jsonb(COALESCE(qres_subsample.min_idle_xact_dur, qres.min_idle_xact_dur))
          );
      EXCEPTION
          WHEN OTHERS THEN
            server_properties := jsonb_set(server_properties,
              '{properties,min_idle_xact_dur}',
              COALESCE (
                to_jsonb(qres_subsample.min_idle_xact_dur)
                , 'null'::jsonb
              )
            );
      END;
    END IF; -- when subsamples enabled
    RETURN server_properties;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION take_sample(IN sserver_id integer, IN skip_sizes boolean
) RETURNS integer SET search_path=@extschema@ AS $$
DECLARE
    s_id              integer;
    topn              integer;
    ret               integer;
    server_properties jsonb;
    qres              record;
    settings_refresh  boolean = true;
    collect_timings   boolean = false;

    server_query      text;
BEGIN
    -- Adding dblink extension schema to search_path if it does not already there
    IF (SELECT count(*) = 0 FROM pg_catalog.pg_extension WHERE extname = 'dblink') THEN
      RAISE 'dblink extension must be installed';
    END IF;

    SELECT extnamespace::regnamespace AS dblink_schema INTO STRICT qres FROM pg_catalog.pg_extension WHERE extname = 'dblink';
    IF NOT string_to_array(current_setting('search_path'),', ') @> ARRAY[qres.dblink_schema::text] THEN
      EXECUTE 'SET LOCAL search_path TO ' || current_setting('search_path')||','|| qres.dblink_schema;
    END IF;

    -- Only one running take_sample() function allowed per server!
    -- Explicitly lock server in servers table
    BEGIN
        SELECT * INTO qres FROM servers WHERE server_id = sserver_id FOR UPDATE NOWAIT;
    EXCEPTION
        WHEN OTHERS THEN RAISE 'Can''t get lock on server. Is there another take_sample() function running on this server?';
    END;

    -- Initialize sample
    server_properties := init_sample(sserver_id);
    ASSERT server_properties IS NOT NULL, 'lost properties';

    topn := (server_properties #>> '{properties,topn}')::integer;

    -- Creating a new sample record
    UPDATE servers SET last_sample_id = last_sample_id + 1 WHERE server_id = sserver_id
      RETURNING last_sample_id INTO s_id;
    INSERT INTO samples(sample_time,server_id,sample_id)
      VALUES (now(),sserver_id,s_id);

    -- Once the new sample is created it becomes last one
    server_properties := jsonb_set(
      server_properties,
      '{properties,last_sample_id}',
      to_jsonb(s_id)
    );

    -- Getting max_sample_age setting
    BEGIN
        ret := COALESCE(current_setting('pg_profile.max_sample_age')::integer);
    EXCEPTION
        WHEN OTHERS THEN ret := 7;
    END;
    -- Applying skip sizes policy
    IF skip_sizes IS NULL THEN
      IF num_nulls(qres.size_smp_wnd_start, qres.size_smp_wnd_dur, qres.size_smp_interval) > 0 THEN
        skip_sizes := false;
      ELSE
        /*
        Skip sizes collection if there was a sample with sizes recently
        or if we are not in size collection time window
        */
        SELECT
          count(*) > 0 OR
          NOT
          CASE WHEN timezone('UTC',current_time) > timezone('UTC',qres.size_smp_wnd_start) THEN
            timezone('UTC',now()) <=
            timezone('UTC',now())::date +
            timezone('UTC',qres.size_smp_wnd_start) +
            qres.size_smp_wnd_dur
          ELSE
            timezone('UTC',now()) <=
            timezone('UTC',now() - interval '1 day')::date +
            timezone('UTC',qres.size_smp_wnd_start) +
            qres.size_smp_wnd_dur
          END
            INTO STRICT skip_sizes
        FROM
          sample_stat_tables_total st
          JOIN samples s USING (server_id, sample_id)
        WHERE
          server_id = sserver_id
          AND st.relsize_diff IS NOT NULL
          AND sample_time > now() - qres.size_smp_interval;
      END IF;
    END IF;

    -- Collecting postgres parameters
    /* We might refresh all parameters if version() was changed
    * This is needed for deleting obsolete parameters, not appearing in new
    * Postgres version.
    */
    SELECT ss.setting != dblver.version INTO settings_refresh
    FROM v_sample_settings ss, dblink('server_connection','SELECT version() as version') AS dblver (version text)
    WHERE ss.server_id = sserver_id AND ss.sample_id = s_id AND ss.name='version' AND ss.setting_scope = 2;
    settings_refresh := COALESCE(settings_refresh,true);

    -- Constructing server sql query for settings
    server_query := 'SELECT 1 as setting_scope,name,setting,reset_val,boot_val,unit,sourcefile,sourceline,pending_restart '
      'FROM pg_catalog.pg_settings '
      'UNION ALL SELECT 2 as setting_scope,''version'',version(),version(),NULL,NULL,NULL,NULL,False '
      'UNION ALL SELECT 2 as setting_scope,''pg_postmaster_start_time'','
      'pg_catalog.pg_postmaster_start_time()::text,'
      'pg_catalog.pg_postmaster_start_time()::text,NULL,NULL,NULL,NULL,False '
      'UNION ALL SELECT 2 as setting_scope,''pg_conf_load_time'','
      'pg_catalog.pg_conf_load_time()::text,pg_catalog.pg_conf_load_time()::text,NULL,NULL,NULL,NULL,False '
      'UNION ALL SELECT 2 as setting_scope,''system_identifier'','
      'system_identifier::text,system_identifier::text,system_identifier::text,'
      'NULL,NULL,NULL,False FROM pg_catalog.pg_control_system()';

    -- Is it PostgresPro?
    IF (server_properties #>> '{properties,pgpro}')::boolean THEN
      server_query := concat_ws(' ',server_query,
          'UNION ALL SELECT 2 as setting_scope,''pgpro_build'',pgpro_build(),pgpro_build(),NULL,NULL,NULL,NULL,False '
          'UNION ALL SELECT 2 as setting_scope,''pgpro_edition'',pgpro_edition(),pgpro_edition(),NULL,NULL,NULL,NULL,False '
          'UNION ALL SELECT 2 as setting_scope,''pgpro_version'',pgpro_version(),pgpro_version(),NULL,NULL,NULL,NULL,False '
      );
    END IF;

    INSERT INTO sample_settings(
      server_id,
      first_seen,
      setting_scope,
      name,
      setting,
      reset_val,
      boot_val,
      unit,
      sourcefile,
      sourceline,
      pending_restart
    )
    SELECT
      s.server_id as server_id,
      s.sample_time as first_seen,
      cur.setting_scope,
      cur.name,
      cur.setting,
      cur.reset_val,
      cur.boot_val,
      cur.unit,
      cur.sourcefile,
      cur.sourceline,
      cur.pending_restart
    FROM
      sample_settings lst JOIN (
        -- Getting last versions of settings
        SELECT server_id, name, max(first_seen) as first_seen
        FROM sample_settings
        WHERE server_id = sserver_id AND (
          NOT settings_refresh
          -- system identifier shouldn't have a duplicate in case of version change
          -- this breaks export/import procedures, as those are related to this ID
          OR name = 'system_identifier'
        )
        GROUP BY server_id, name
      ) lst_times
      USING (server_id, name, first_seen)
      -- Getting current settings values
      RIGHT OUTER JOIN dblink('server_connection',server_query
          ) AS cur (
            setting_scope smallint,
            name text,
            setting text,
            reset_val text,
            boot_val text,
            unit text,
            sourcefile text,
            sourceline integer,
            pending_restart boolean
          )
        USING (setting_scope, name)
      JOIN samples s ON (s.server_id = sserver_id AND s.sample_id = s_id)
    WHERE
      cur.reset_val IS NOT NULL AND (
        lst.name IS NULL
        OR cur.reset_val != lst.reset_val
        OR cur.pending_restart != lst.pending_restart
        OR lst.sourcefile != cur.sourcefile
        OR lst.sourceline != cur.sourceline
        OR lst.unit != cur.unit
      );

    INSERT INTO sample_settings(
      server_id,
      first_seen,
      setting_scope,
      name,
      setting,
      reset_val,
      boot_val,
      unit,
      sourcefile,
      sourceline,
      pending_restart
    )
    SELECT
      s.server_id,
      s.sample_time,
      1 as setting_scope,
      'pg_profile.topn',
      topn,
      topn,
      topn,
      null,
      null,
      null,
      false
    FROM samples s LEFT OUTER JOIN  v_sample_settings prm ON
      (s.server_id = prm.server_id AND s.sample_id = prm.sample_id AND prm.name = 'pg_profile.topn' AND prm.setting_scope = 1)
    WHERE s.server_id = sserver_id AND s.sample_id = s_id AND (prm.setting IS NULL OR prm.setting::integer != topn);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,get server environment,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,collect database stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Construct pg_stat_database query
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 140000
      )
      THEN
        server_query := 'SELECT '
            'dbs.datid, '
            'dbs.datname, '
            'dbs.xact_commit, '
            'dbs.xact_rollback, '
            'dbs.blks_read, '
            'dbs.blks_hit, '
            'dbs.tup_returned, '
            'dbs.tup_fetched, '
            'dbs.tup_inserted, '
            'dbs.tup_updated, '
            'dbs.tup_deleted, '
            'dbs.conflicts, '
            'dbs.temp_files, '
            'dbs.temp_bytes, '
            'dbs.deadlocks, '
            'dbs.checksum_failures, '
            'dbs.checksum_last_failure, '
            'dbs.blk_read_time, '
            'dbs.blk_write_time, '
            'dbs.session_time, '
            'dbs.active_time, '
            'dbs.idle_in_transaction_time, '
            'dbs.sessions, '
            'dbs.sessions_abandoned, '
            'dbs.sessions_fatal, '
            'dbs.sessions_killed, '
            'dbs.stats_reset, '
            'pg_database_size(dbs.datid) as datsize, '
            '0 as datsize_delta, '
            'db.datistemplate, '
            'db.dattablespace, '
            'db.datallowconn '
          'FROM pg_catalog.pg_stat_database dbs '
          'JOIN pg_catalog.pg_database db ON (dbs.datid = db.oid) '
          'WHERE dbs.datname IS NOT NULL';
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 120000
      )
      THEN
        server_query := 'SELECT '
            'dbs.datid, '
            'dbs.datname, '
            'dbs.xact_commit, '
            'dbs.xact_rollback, '
            'dbs.blks_read, '
            'dbs.blks_hit, '
            'dbs.tup_returned, '
            'dbs.tup_fetched, '
            'dbs.tup_inserted, '
            'dbs.tup_updated, '
            'dbs.tup_deleted, '
            'dbs.conflicts, '
            'dbs.temp_files, '
            'dbs.temp_bytes, '
            'dbs.deadlocks, '
            'dbs.checksum_failures, '
            'dbs.checksum_last_failure, '
            'dbs.blk_read_time, '
            'dbs.blk_write_time, '
            'NULL as session_time, '
            'NULL as active_time, '
            'NULL as idle_in_transaction_time, '
            'NULL as sessions, '
            'NULL as sessions_abandoned, '
            'NULL as sessions_fatal, '
            'NULL as sessions_killed, '
            'dbs.stats_reset, '
            'pg_database_size(dbs.datid) as datsize, '
            '0 as datsize_delta, '
            'db.datistemplate, '
            'db.dattablespace, '
            'db.datallowconn '
          'FROM pg_catalog.pg_stat_database dbs '
          'JOIN pg_catalog.pg_database db ON (dbs.datid = db.oid) '
          'WHERE dbs.datname IS NOT NULL';
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer < 120000
      )
      THEN
        server_query := 'SELECT '
            'dbs.datid, '
            'dbs.datname, '
            'dbs.xact_commit, '
            'dbs.xact_rollback, '
            'dbs.blks_read, '
            'dbs.blks_hit, '
            'dbs.tup_returned, '
            'dbs.tup_fetched, '
            'dbs.tup_inserted, '
            'dbs.tup_updated, '
            'dbs.tup_deleted, '
            'dbs.conflicts, '
            'dbs.temp_files, '
            'dbs.temp_bytes, '
            'dbs.deadlocks, '
            'NULL as checksum_failures, '
            'NULL as checksum_last_failure, '
            'dbs.blk_read_time, '
            'dbs.blk_write_time, '
            'NULL as session_time, '
            'NULL as active_time, '
            'NULL as idle_in_transaction_time, '
            'NULL as sessions, '
            'NULL as sessions_abandoned, '
            'NULL as sessions_fatal, '
            'NULL as sessions_killed, '
            'dbs.stats_reset, '
            'pg_database_size(dbs.datid) as datsize, '
            '0 as datsize_delta, '
            'db.datistemplate, '
            'db.dattablespace, '
            'db.datallowconn '
          'FROM pg_catalog.pg_stat_database dbs '
          'JOIN pg_catalog.pg_database db ON (dbs.datid = db.oid) '
          'WHERE dbs.datname IS NOT NULL';
    END CASE;

    -- pg_stat_database data
    INSERT INTO last_stat_database (
        server_id,
        sample_id,
        datid,
        datname,
        xact_commit,
        xact_rollback,
        blks_read,
        blks_hit,
        tup_returned,
        tup_fetched,
        tup_inserted,
        tup_updated,
        tup_deleted,
        conflicts,
        temp_files,
        temp_bytes,
        deadlocks,
        checksum_failures,
        checksum_last_failure,
        blk_read_time,
        blk_write_time,
        session_time,
        active_time,
        idle_in_transaction_time,
        sessions,
        sessions_abandoned,
        sessions_fatal,
        sessions_killed,
        stats_reset,
        datsize,
        datsize_delta,
        datistemplate,
        dattablespace,
        datallowconn)
    SELECT
        sserver_id,
        s_id,
        datid,
        datname,
        xact_commit AS xact_commit,
        xact_rollback AS xact_rollback,
        blks_read AS blks_read,
        blks_hit AS blks_hit,
        tup_returned AS tup_returned,
        tup_fetched AS tup_fetched,
        tup_inserted AS tup_inserted,
        tup_updated AS tup_updated,
        tup_deleted AS tup_deleted,
        conflicts AS conflicts,
        temp_files AS temp_files,
        temp_bytes AS temp_bytes,
        deadlocks AS deadlocks,
        checksum_failures as checksum_failures,
        checksum_last_failure as checksum_failures,
        blk_read_time AS blk_read_time,
        blk_write_time AS blk_write_time,
        session_time AS session_time,
        active_time AS active_time,
        idle_in_transaction_time AS idle_in_transaction_time,
        sessions AS sessions,
        sessions_abandoned AS sessions_abandoned,
        sessions_fatal AS sessions_fatal,
        sessions_killed AS sessions_killed,
        stats_reset,
        datsize AS datsize,
        datsize_delta AS datsize_delta,
        datistemplate AS datistemplate,
        dattablespace AS dattablespace,
        datallowconn AS datallowconn
    FROM dblink('server_connection',server_query) AS rs (
        datid oid,
        datname name,
        xact_commit bigint,
        xact_rollback bigint,
        blks_read bigint,
        blks_hit bigint,
        tup_returned bigint,
        tup_fetched bigint,
        tup_inserted bigint,
        tup_updated bigint,
        tup_deleted bigint,
        conflicts bigint,
        temp_files bigint,
        temp_bytes bigint,
        deadlocks bigint,
        checksum_failures bigint,
        checksum_last_failure timestamp with time zone,
        blk_read_time double precision,
        blk_write_time double precision,
        session_time double precision,
        active_time double precision,
        idle_in_transaction_time double precision,
        sessions bigint,
        sessions_abandoned bigint,
        sessions_fatal bigint,
        sessions_killed bigint,
        stats_reset timestamp with time zone,
        datsize bigint,
        datsize_delta bigint,
        datistemplate boolean,
        dattablespace oid,
        datallowconn boolean
        );

    EXECUTE format('ANALYZE last_stat_database_srv%1$s',
      sserver_id);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,collect database stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate database stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;
    -- Calc stat_database diff
    INSERT INTO sample_stat_database(
      server_id,
      sample_id,
      datid,
      datname,
      xact_commit,
      xact_rollback,
      blks_read,
      blks_hit,
      tup_returned,
      tup_fetched,
      tup_inserted,
      tup_updated,
      tup_deleted,
      conflicts,
      temp_files,
      temp_bytes,
      deadlocks,
      checksum_failures,
      checksum_last_failure,
      blk_read_time,
      blk_write_time,
      session_time,
      active_time,
      idle_in_transaction_time,
      sessions,
      sessions_abandoned,
      sessions_fatal,
      sessions_killed,
      stats_reset,
      datsize,
      datsize_delta,
      datistemplate
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.datid,
        cur.datname,
        cur.xact_commit - COALESCE(lst.xact_commit,0),
        cur.xact_rollback - COALESCE(lst.xact_rollback,0),
        cur.blks_read - COALESCE(lst.blks_read,0),
        cur.blks_hit - COALESCE(lst.blks_hit,0),
        cur.tup_returned - COALESCE(lst.tup_returned,0),
        cur.tup_fetched - COALESCE(lst.tup_fetched,0),
        cur.tup_inserted - COALESCE(lst.tup_inserted,0),
        cur.tup_updated - COALESCE(lst.tup_updated,0),
        cur.tup_deleted - COALESCE(lst.tup_deleted,0),
        cur.conflicts - COALESCE(lst.conflicts,0),
        cur.temp_files - COALESCE(lst.temp_files,0),
        cur.temp_bytes - COALESCE(lst.temp_bytes,0),
        cur.deadlocks - COALESCE(lst.deadlocks,0),
        cur.checksum_failures - COALESCE(lst.checksum_failures,0),
        cur.checksum_last_failure,
        cur.blk_read_time - COALESCE(lst.blk_read_time,0),
        cur.blk_write_time - COALESCE(lst.blk_write_time,0),
        cur.session_time - COALESCE(lst.session_time,0),
        cur.active_time - COALESCE(lst.active_time,0),
        cur.idle_in_transaction_time - COALESCE(lst.idle_in_transaction_time,0),
        cur.sessions - COALESCE(lst.sessions,0),
        cur.sessions_abandoned - COALESCE(lst.sessions_abandoned,0),
        cur.sessions_fatal - COALESCE(lst.sessions_fatal,0),
        cur.sessions_killed - COALESCE(lst.sessions_killed,0),
        cur.stats_reset,
        cur.datsize as datsize,
        cur.datsize - COALESCE(lst.datsize,0) as datsize_delta,
        cur.datistemplate
    FROM last_stat_database cur
      LEFT OUTER JOIN last_stat_database lst ON
        (lst.server_id, lst.sample_id, lst.datid, lst.datname) =
        (sserver_id, s_id - 1, cur.datid, cur.datname)
        AND lst.stats_reset IS NOT DISTINCT FROM cur.stats_reset
    WHERE
      (cur.server_id, cur.sample_id) = (sserver_id, s_id);

    /*
    * In case of statistics reset full database size, and checksum checksum_failures
    * is incorrectly considered as increment by previous query.
    * So, we need to update it with correct value
    */
    UPDATE sample_stat_database sdb
    SET
      datsize_delta = cur.datsize - lst.datsize,
      checksum_failures = cur.checksum_failures - lst.checksum_failures,
      checksum_last_failure = cur.checksum_last_failure
    FROM
      last_stat_database cur
      JOIN last_stat_database lst ON
        (lst.server_id, lst.sample_id, lst.datid, lst.datname) =
        (sserver_id, s_id - 1, cur.datid, cur.datname)
    WHERE cur.stats_reset IS DISTINCT FROM lst.stats_reset AND
      (cur.server_id, cur.sample_id) = (sserver_id, s_id) AND
      (sdb.server_id, sdb.sample_id, sdb.datid, sdb.datname) =
      (cur.server_id, cur.sample_id, cur.datid, cur.datname);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate database stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,collect tablespace stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Construct tablespace stats query
    server_query := 'SELECT '
        'oid as tablespaceid,'
        'spcname as tablespacename,'
        'pg_catalog.pg_tablespace_location(oid) as tablespacepath,'
        'pg_catalog.pg_tablespace_size(oid) as size,'
        '0 as size_delta '
        'FROM pg_catalog.pg_tablespace ';

    -- Get tablespace stats
    INSERT INTO last_stat_tablespaces(
      server_id,
      sample_id,
      tablespaceid,
      tablespacename,
      tablespacepath,
      size,
      size_delta
    )
    SELECT
      sserver_id,
      s_id,
      dbl.tablespaceid,
      dbl.tablespacename,
      dbl.tablespacepath,
      dbl.size AS size,
      dbl.size_delta AS size_delta
    FROM dblink('server_connection', server_query)
    AS dbl (
        tablespaceid            oid,
        tablespacename          name,
        tablespacepath          text,
        size                    bigint,
        size_delta              bigint
    );

    EXECUTE format('ANALYZE last_stat_tablespaces_srv%1$s',
      sserver_id);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,collect tablespace stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,collect statement stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Search for statements statistics extension
    CASE
      -- pg_stat_statements statistics collection
      WHEN (
        SELECT count(*) = 1
        FROM jsonb_to_recordset(server_properties #> '{extensions}') AS ext(extname text)
        WHERE extname = 'pg_stat_statements'
      ) THEN
        PERFORM collect_pg_stat_statements_stats(server_properties, sserver_id, s_id, topn);
      ELSE
        NULL;
    END CASE;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,collect statement stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,collect wait sampling stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Search for wait sampling extension
    CASE
      -- pg_wait_sampling statistics collection
      WHEN (
        SELECT count(*) = 1
        FROM jsonb_to_recordset(server_properties #> '{extensions}') AS ext(extname text)
        WHERE extname = 'pg_wait_sampling'
      ) THEN
        PERFORM collect_pg_wait_sampling_stats(server_properties, sserver_id, s_id, topn);
      ELSE
        NULL;
    END CASE;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,collect wait sampling stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_bgwriter}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- pg_stat_bgwriter data
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer < 100000
      )
      THEN
        server_query := 'SELECT '
          'checkpoints_timed,'
          'checkpoints_req,'
          'checkpoint_write_time,'
          'checkpoint_sync_time,'
          'buffers_checkpoint,'
          'buffers_clean,'
          'maxwritten_clean,'
          'buffers_backend,'
          'buffers_backend_fsync,'
          'buffers_alloc,'
          'stats_reset,'
          'CASE WHEN pg_catalog.pg_is_in_recovery() '
          'THEN pg_catalog.pg_xlog_location_diff(pg_catalog.pg_last_xlog_replay_location(),''0/00000000'') '
          'ELSE pg_catalog.pg_xlog_location_diff(pg_catalog.pg_current_xlog_location(),''0/00000000'') '
          'END AS wal_size,'
          'CASE WHEN pg_catalog.pg_is_in_recovery() '
          'THEN pg_catalog.pg_last_xlog_replay_location() '
          'ELSE pg_catalog.pg_current_xlog_location() '
          'END AS wal_lsn,'
          'pg_is_in_recovery() AS in_recovery '
          'FROM pg_catalog.pg_stat_bgwriter';
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 100000
      )
      THEN
        server_query := 'SELECT '
          'checkpoints_timed,'
          'checkpoints_req,'
          'checkpoint_write_time,'
          'checkpoint_sync_time,'
          'buffers_checkpoint,'
          'buffers_clean,'
          'maxwritten_clean,'
          'buffers_backend,'
          'buffers_backend_fsync,'
          'buffers_alloc,'
          'stats_reset,'
          'CASE WHEN pg_catalog.pg_is_in_recovery() '
            'THEN pg_catalog.pg_wal_lsn_diff(pg_catalog.pg_last_wal_replay_lsn(),''0/00000000'') '
            'ELSE pg_catalog.pg_wal_lsn_diff(pg_catalog.pg_current_wal_lsn(),''0/00000000'') '
          'END AS wal_size,'
          'CASE WHEN pg_catalog.pg_is_in_recovery() '
            'THEN pg_catalog.pg_last_wal_replay_lsn() '
            'ELSE pg_catalog.pg_current_wal_lsn() '
          'END AS wal_lsn,'
          'pg_catalog.pg_is_in_recovery() as in_recovery '
        'FROM pg_catalog.pg_stat_bgwriter';
    END CASE;

    IF server_query IS NOT NULL THEN
      INSERT INTO last_stat_cluster (
        server_id,
        sample_id,
        checkpoints_timed,
        checkpoints_req,
        checkpoint_write_time,
        checkpoint_sync_time,
        buffers_checkpoint,
        buffers_clean,
        maxwritten_clean,
        buffers_backend,
        buffers_backend_fsync,
        buffers_alloc,
        stats_reset,
        wal_size,
        wal_lsn,
        in_recovery)
      SELECT
        sserver_id,
        s_id,
        checkpoints_timed,
        checkpoints_req,
        checkpoint_write_time,
        checkpoint_sync_time,
        buffers_checkpoint,
        buffers_clean,
        maxwritten_clean,
        buffers_backend,
        buffers_backend_fsync,
        buffers_alloc,
        stats_reset,
        wal_size,
        wal_lsn,
        in_recovery
      FROM dblink('server_connection',server_query) AS rs (
        checkpoints_timed bigint,
        checkpoints_req bigint,
        checkpoint_write_time double precision,
        checkpoint_sync_time double precision,
        buffers_checkpoint bigint,
        buffers_clean bigint,
        maxwritten_clean bigint,
        buffers_backend bigint,
        buffers_backend_fsync bigint,
        buffers_alloc bigint,
        stats_reset timestamp with time zone,
        wal_size bigint,
        wal_lsn pg_lsn,
        in_recovery boolean);
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_bgwriter,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_wal}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- pg_stat_wal data
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 140000
      )
      THEN
        server_query := 'SELECT '
          'wal_records,'
          'wal_fpi,'
          'wal_bytes,'
          'wal_buffers_full,'
          'wal_write,'
          'wal_sync,'
          'wal_write_time,'
          'wal_sync_time,'
          'stats_reset '
          'FROM pg_catalog.pg_stat_wal';
      ELSE
        server_query := NULL;
    END CASE;

    IF server_query IS NOT NULL THEN
      INSERT INTO last_stat_wal (
        server_id,
        sample_id,
        wal_records,
        wal_fpi,
        wal_bytes,
        wal_buffers_full,
        wal_write,
        wal_sync,
        wal_write_time,
        wal_sync_time,
        stats_reset
      )
      SELECT
        sserver_id,
        s_id,
        wal_records,
        wal_fpi,
        wal_bytes,
        wal_buffers_full,
        wal_write,
        wal_sync,
        wal_write_time,
        wal_sync_time,
        stats_reset
      FROM dblink('server_connection',server_query) AS rs (
        wal_records         bigint,
        wal_fpi             bigint,
        wal_bytes           numeric,
        wal_buffers_full    bigint,
        wal_write           bigint,
        wal_sync            bigint,
        wal_write_time      double precision,
        wal_sync_time       double precision,
        stats_reset         timestamp with time zone);
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_wal,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_io}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- pg_stat_io data
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 160000
      )
      THEN
        server_query := 'SELECT '
          'backend_type,'
          'object,'
          'context,'
          'reads,'
          'read_time,'
          'writes,'
          'write_time,'
          'writebacks,'
          'writeback_time,'
          'extends,'
          'extend_time,'
          'op_bytes,'
          'hits,'
          'evictions,'
          'reuses,'
          'fsyncs,'
          'fsync_time,'
          'stats_reset '
          'FROM pg_catalog.pg_stat_io '
          'WHERE greatest('
              'reads,'
              'writes,'
              'writebacks,'
              'extends,'
              'hits,'
              'evictions,'
              'reuses,'
              'fsyncs'
            ') > 0'
          ;
      ELSE
        server_query := NULL;
    END CASE;

    IF server_query IS NOT NULL THEN
      INSERT INTO last_stat_io (
        server_id,
        sample_id,
        backend_type,
        object,
        context,
        reads,
        read_time,
        writes,
        write_time,
        writebacks,
        writeback_time,
        extends,
        extend_time,
        op_bytes,
        hits,
        evictions,
        reuses,
        fsyncs,
        fsync_time,
        stats_reset
      )
      SELECT
        sserver_id,
        s_id,
        backend_type,
        object,
        context,
        reads,
        read_time,
        writes,
        write_time,
        writebacks,
        writeback_time,
        extends,
        extend_time,
        op_bytes,
        hits,
        evictions,
        reuses,
        fsyncs,
        fsync_time,
        stats_reset
      FROM dblink('server_connection',server_query) AS rs (
        backend_type      text,
        object            text,
        context           text,
        reads             bigint,
        read_time         double precision,
        writes            bigint,
        write_time        double precision,
        writebacks        bigint,
        writeback_time    double precision,
        extends           bigint,
        extend_time       double precision,
        op_bytes          bigint,
        hits              bigint,
        evictions         bigint,
        reuses            bigint,
        fsyncs            bigint,
        fsync_time        double precision,
        stats_reset       timestamp with time zone
      );
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_io,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_slru}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- pg_stat_slru data
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer >= 130000
      )
      THEN
        server_query := 'SELECT '
          'name,'
          'blks_zeroed,'
          'blks_hit,'
          'blks_read,'
          'blks_written,'
          'blks_exists,'
          'flushes,'
          'truncates,'
          'stats_reset '
          'FROM pg_catalog.pg_stat_slru '
          'WHERE greatest('
              'blks_zeroed,'
              'blks_hit,'
              'blks_read,'
              'blks_written,'
              'blks_exists,'
              'flushes,'
              'truncates'
            ') > 0'
          ;
      ELSE
        server_query := NULL;
    END CASE;

    IF server_query IS NOT NULL THEN
      INSERT INTO last_stat_slru (
        server_id,
        sample_id,
        name,
        blks_zeroed,
        blks_hit,
        blks_read,
        blks_written,
        blks_exists,
        flushes,
        truncates,
        stats_reset
      )
      SELECT
        sserver_id,
        s_id,
        name,
        blks_zeroed,
        blks_hit,
        blks_read,
        blks_written,
        blks_exists,
        flushes,
        truncates,
        stats_reset
      FROM dblink('server_connection',server_query) AS rs (
        name          text,
        blks_zeroed   bigint,
        blks_hit      bigint,
        blks_read     bigint,
        blks_written  bigint,
        blks_exists   bigint,
        flushes       bigint,
        truncates     bigint,
        stats_reset   timestamp with time zone
      );
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_slru,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_archiver}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- pg_stat_archiver data
    CASE
      WHEN (
        SELECT count(*) = 1 FROM jsonb_to_recordset(server_properties #> '{settings}')
          AS x(name text, reset_val text)
        WHERE name = 'server_version_num'
          AND reset_val::integer > 90500
      )
      THEN
        server_query := 'SELECT '
          'archived_count,'
          'last_archived_wal,'
          'last_archived_time,'
          'failed_count,'
          'last_failed_wal,'
          'last_failed_time,'
          'stats_reset '
          'FROM pg_catalog.pg_stat_archiver';
    END CASE;

    IF server_query IS NOT NULL THEN
      INSERT INTO last_stat_archiver (
        server_id,
        sample_id,
        archived_count,
        last_archived_wal,
        last_archived_time,
        failed_count,
        last_failed_wal,
        last_failed_time,
        stats_reset)
      SELECT
        sserver_id,
        s_id,
        archived_count as archived_count,
        last_archived_wal as last_archived_wal,
        last_archived_time as last_archived_time,
        failed_count as failed_count,
        last_failed_wal as last_failed_wal,
        last_failed_time as last_failed_time,
        stats_reset as stats_reset
      FROM dblink('server_connection',server_query) AS rs (
        archived_count              bigint,
        last_archived_wal           text,
        last_archived_time          timestamp with time zone,
        failed_count                bigint,
        last_failed_wal             text,
        last_failed_time            timestamp with time zone,
        stats_reset                 timestamp with time zone
      );
    END IF;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,query pg_stat_archiver,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,collect object stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Collecting stat info for objects of all databases
    server_properties := collect_obj_stats(server_properties, sserver_id, s_id, skip_sizes);
    ASSERT server_properties IS NOT NULL, 'lost properties';

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,collect object stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,processing subsamples}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    /*
     We must get a lock on subsample before taking a subsample to avoid
     sample failure due to lock held in concurrent take_subsample() call.
     take_subsample() function acquires a lock in NOWAIT mode to avoid long
     waits in a subsample. But we should wait here in sample because sample
     must be taken anyway and we need to avoid subsample interfere.
    */
    PERFORM
    FROM server_subsample
    WHERE server_id = sserver_id
    FOR UPDATE;

    server_properties := take_subsample(sserver_id, server_properties);
    server_properties := collect_subsamples(sserver_id, s_id, server_properties);
    ASSERT server_properties IS NOT NULL, 'lost properties';

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,processing subsamples,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,disconnect}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    PERFORM dblink('server_connection', 'COMMIT');
    PERFORM dblink_disconnect('server_connection');

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,disconnect,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,maintain repository}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Updating dictionary table in case of object renaming:
    -- Databases
    UPDATE sample_stat_database AS db
    SET datname = lst.datname
    FROM last_stat_database AS lst
    WHERE
      (db.server_id, lst.server_id, lst.sample_id, db.datid) =
      (sserver_id, sserver_id, s_id, lst.datid)
      AND db.datname != lst.datname;
    -- Tables
    UPDATE tables_list AS tl
    SET (schemaname, relname) = (lst.schemaname, lst.relname)
    FROM last_stat_tables AS lst
    WHERE (tl.server_id, lst.server_id, lst.sample_id, tl.datid, tl.relid, tl.relkind) =
        (sserver_id, sserver_id, s_id, lst.datid, lst.relid, lst.relkind)
      AND (tl.schemaname, tl.relname) != (lst.schemaname, lst.relname);
    -- Functions
    UPDATE funcs_list AS fl
    SET (schemaname, funcname, funcargs) =
      (lst.schemaname, lst.funcname, lst.funcargs)
    FROM last_stat_user_functions AS lst
    WHERE (fl.server_id, lst.server_id, lst.sample_id, fl.datid, fl.funcid) =
        (sserver_id, sserver_id, s_id, lst.datid, lst.funcid)
      AND (fl.schemaname, fl.funcname, fl.funcargs) !=
        (lst.schemaname, lst.funcname, lst.funcargs);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,maintain repository,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate tablespace stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    INSERT INTO tablespaces_list AS itl (
        server_id,
        last_sample_id,
        tablespaceid,
        tablespacename,
        tablespacepath
      )
    SELECT
      cur.server_id,
      NULL,
      cur.tablespaceid,
      cur.tablespacename,
      cur.tablespacepath
    FROM
      last_stat_tablespaces cur
    WHERE
      (cur.server_id, cur.sample_id) = (sserver_id, s_id)
    ON CONFLICT ON CONSTRAINT pk_tablespace_list DO
    UPDATE SET
        (last_sample_id, tablespacename, tablespacepath) =
        (EXCLUDED.last_sample_id, EXCLUDED.tablespacename, EXCLUDED.tablespacepath)
      WHERE
        (itl.last_sample_id, itl.tablespacename, itl.tablespacepath) IS DISTINCT FROM
        (EXCLUDED.last_sample_id, EXCLUDED.tablespacename, EXCLUDED.tablespacepath);

    -- Calculate diffs for tablespaces
    INSERT INTO sample_stat_tablespaces(
      server_id,
      sample_id,
      tablespaceid,
      size,
      size_delta
    )
    SELECT
      cur.server_id as server_id,
      cur.sample_id as sample_id,
      cur.tablespaceid as tablespaceid,
      cur.size as size,
      cur.size - COALESCE(lst.size, 0) AS size_delta
    FROM last_stat_tablespaces cur
      LEFT OUTER JOIN last_stat_tablespaces lst ON
        (lst.server_id, lst.sample_id, cur.tablespaceid) =
        (sserver_id, s_id - 1, lst.tablespaceid)
    WHERE (cur.server_id, cur.sample_id) = ( sserver_id, s_id);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate tablespace stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate object stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- collect databases objects stats
    server_properties := sample_dbobj_delta(server_properties,sserver_id,s_id,topn,skip_sizes);
    ASSERT server_properties IS NOT NULL, 'lost properties';

    DELETE FROM last_stat_tablespaces WHERE server_id = sserver_id AND sample_id != s_id;

    DELETE FROM last_stat_database WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate object stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate cluster stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Calc stat cluster diff
    INSERT INTO sample_stat_cluster(
      server_id,
      sample_id,
      checkpoints_timed,
      checkpoints_req,
      checkpoint_write_time,
      checkpoint_sync_time,
      buffers_checkpoint,
      buffers_clean,
      maxwritten_clean,
      buffers_backend,
      buffers_backend_fsync,
      buffers_alloc,
      stats_reset,
      wal_size,
      wal_lsn,
      in_recovery
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.checkpoints_timed - COALESCE(lst.checkpoints_timed,0),
        cur.checkpoints_req - COALESCE(lst.checkpoints_req,0),
        cur.checkpoint_write_time - COALESCE(lst.checkpoint_write_time,0),
        cur.checkpoint_sync_time - COALESCE(lst.checkpoint_sync_time,0),
        cur.buffers_checkpoint - COALESCE(lst.buffers_checkpoint,0),
        cur.buffers_clean - COALESCE(lst.buffers_clean,0),
        cur.maxwritten_clean - COALESCE(lst.maxwritten_clean,0),
        cur.buffers_backend - COALESCE(lst.buffers_backend,0),
        cur.buffers_backend_fsync - COALESCE(lst.buffers_backend_fsync,0),
        cur.buffers_alloc - COALESCE(lst.buffers_alloc,0),
        cur.stats_reset,
        cur.wal_size - COALESCE(lst.wal_size,0),
        cur.wal_lsn,
        cur.in_recovery
        /* We will overwrite this value in case of stats reset
         * (see below)
         */
    FROM last_stat_cluster cur
      LEFT OUTER JOIN last_stat_cluster lst ON
        (lst.server_id, lst.sample_id) =
        (sserver_id, s_id - 1)
        AND cur.stats_reset IS NOT DISTINCT FROM lst.stats_reset
    WHERE
      (cur.server_id, cur.sample_id) = (sserver_id, s_id);

    /* wal_size is calculated since 0 to current value when stats reset happened
     * so, we need to update it
     */
    UPDATE sample_stat_cluster ssc
    SET wal_size = cur.wal_size - lst.wal_size
    FROM last_stat_cluster cur
      JOIN last_stat_cluster lst ON
        (lst.server_id, lst.sample_id) =
        (sserver_id, s_id - 1)
    WHERE
      (ssc.server_id, ssc.sample_id) = (sserver_id, s_id) AND
      (cur.server_id, cur.sample_id) = (sserver_id, s_id) AND
      cur.stats_reset IS DISTINCT FROM lst.stats_reset;

    DELETE FROM last_stat_cluster WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate cluster stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate IO stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Calc I/O stat diff
    INSERT INTO sample_stat_io(
        server_id,
        sample_id,
        backend_type,
        object,
        context,
        reads,
        read_time,
        writes,
        write_time,
        writebacks,
        writeback_time,
        extends,
        extend_time,
        op_bytes,
        hits,
        evictions,
        reuses,
        fsyncs,
        fsync_time,
        stats_reset
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.backend_type,
        cur.object,
        cur.context,
        cur.reads - COALESCE(lst.reads, 0),
        cur.read_time - COALESCE(lst.read_time, 0),
        cur.writes - COALESCE(lst.writes, 0),
        cur.write_time - COALESCE(lst.write_time, 0),
        cur.writebacks - COALESCE(lst.writebacks, 0),
        cur.writeback_time - COALESCE(lst.writeback_time, 0),
        cur.extends - COALESCE(lst.extends, 0),
        cur.extend_time - COALESCE(lst.extend_time, 0),
        cur.op_bytes,
        cur.hits - COALESCE(lst.hits, 0),
        cur.evictions - COALESCE(lst.evictions, 0),
        cur.reuses - COALESCE(lst.reuses, 0),
        cur.fsyncs - COALESCE(lst.fsyncs, 0),
        cur.fsync_time - COALESCE(lst.fsync_time, 0),
        cur.stats_reset
    FROM last_stat_io cur
    LEFT OUTER JOIN last_stat_io lst ON
      (lst.server_id, lst.sample_id, lst.backend_type, lst.object, lst.context) =
      (sserver_id, s_id - 1, cur.backend_type, cur.object, cur.context)
      AND (cur.op_bytes,cur.stats_reset) IS NOT DISTINCT FROM (lst.op_bytes,lst.stats_reset)
    WHERE
      (cur.server_id, cur.sample_id) = (sserver_id, s_id) AND
      GREATEST(
        cur.reads - COALESCE(lst.reads, 0),
        cur.writes - COALESCE(lst.writes, 0),
        cur.writebacks - COALESCE(lst.writebacks, 0),
        cur.extends - COALESCE(lst.extends, 0),
        cur.hits - COALESCE(lst.hits, 0),
        cur.evictions - COALESCE(lst.evictions, 0),
        cur.reuses - COALESCE(lst.reuses, 0),
        cur.fsyncs - COALESCE(lst.fsyncs, 0)
      ) > 0;

    DELETE FROM last_stat_io WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate IO stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate SLRU stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Calc SLRU stat diff
    INSERT INTO sample_stat_slru(
        server_id,
        sample_id,
        name,
        blks_zeroed,
        blks_hit,
        blks_read,
        blks_written,
        blks_exists,
        flushes,
        truncates,
        stats_reset
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.name,
        cur.blks_zeroed - COALESCE(lst.blks_zeroed, 0),
        cur.blks_hit - COALESCE(lst.blks_hit, 0),
        cur.blks_read - COALESCE(lst.blks_read, 0),
        cur.blks_written - COALESCE(lst.blks_written, 0),
        cur.blks_exists - COALESCE(lst.blks_exists, 0),
        cur.flushes - COALESCE(lst.flushes, 0),
        cur.truncates - COALESCE(lst.truncates, 0),
        cur.stats_reset
    FROM last_stat_slru cur
    LEFT OUTER JOIN last_stat_slru lst ON
      (lst.server_id, lst.sample_id, lst.name) =
      (sserver_id, s_id - 1, cur.name)
      AND cur.stats_reset IS NOT DISTINCT FROM lst.stats_reset
    WHERE
      (cur.server_id, cur.sample_id) = (sserver_id, s_id) AND
      GREATEST(
        cur.blks_zeroed - COALESCE(lst.blks_zeroed, 0),
        cur.blks_hit - COALESCE(lst.blks_hit, 0),
        cur.blks_read - COALESCE(lst.blks_read, 0),
        cur.blks_written - COALESCE(lst.blks_written, 0),
        cur.blks_exists - COALESCE(lst.blks_exists, 0),
        cur.flushes - COALESCE(lst.flushes, 0),
        cur.truncates - COALESCE(lst.truncates, 0)
      ) > 0;

    DELETE FROM last_stat_slru WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate SLRU stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate WAL stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Calc WAL stat diff
    INSERT INTO sample_stat_wal(
      server_id,
      sample_id,
      wal_records,
      wal_fpi,
      wal_bytes,
      wal_buffers_full,
      wal_write,
      wal_sync,
      wal_write_time,
      wal_sync_time,
      stats_reset
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.wal_records - COALESCE(lst.wal_records,0),
        cur.wal_fpi - COALESCE(lst.wal_fpi,0),
        cur.wal_bytes - COALESCE(lst.wal_bytes,0),
        cur.wal_buffers_full - COALESCE(lst.wal_buffers_full,0),
        cur.wal_write - COALESCE(lst.wal_write,0),
        cur.wal_sync - COALESCE(lst.wal_sync,0),
        cur.wal_write_time - COALESCE(lst.wal_write_time,0),
        cur.wal_sync_time - COALESCE(lst.wal_sync_time,0),
        cur.stats_reset
    FROM last_stat_wal cur
    LEFT OUTER JOIN last_stat_wal lst ON
      (lst.server_id, lst.sample_id) = (sserver_id, s_id - 1)
      AND cur.stats_reset IS NOT DISTINCT FROM lst.stats_reset
    WHERE (cur.server_id, cur.sample_id) = (sserver_id, s_id);

    DELETE FROM last_stat_wal WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate WAL stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,calculate archiver stats}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Calc stat archiver diff
    INSERT INTO sample_stat_archiver(
      server_id,
      sample_id,
      archived_count,
      last_archived_wal,
      last_archived_time,
      failed_count,
      last_failed_wal,
      last_failed_time,
      stats_reset
    )
    SELECT
        cur.server_id,
        cur.sample_id,
        cur.archived_count - COALESCE(lst.archived_count,0),
        cur.last_archived_wal,
        cur.last_archived_time,
        cur.failed_count - COALESCE(lst.failed_count,0),
        cur.last_failed_wal,
        cur.last_failed_time,
        cur.stats_reset
    FROM last_stat_archiver cur
    LEFT OUTER JOIN last_stat_archiver lst ON
      (lst.server_id, lst.sample_id) =
      (cur.server_id, cur.sample_id - 1)
      AND cur.stats_reset IS NOT DISTINCT FROM lst.stats_reset
    WHERE cur.sample_id = s_id AND cur.server_id = sserver_id;

    DELETE FROM last_stat_archiver WHERE server_id = sserver_id AND sample_id != s_id;

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,calculate archiver stats,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,delete obsolete samples}',jsonb_build_object('start',clock_timestamp()));
    END IF;

    -- Updating dictionary tables setting last_sample_id
    UPDATE tablespaces_list utl SET last_sample_id = s_id - 1
    FROM tablespaces_list tl LEFT JOIN sample_stat_tablespaces cur
      ON (cur.server_id, cur.sample_id, cur.tablespaceid) =
        (sserver_id, s_id, tl.tablespaceid)
    WHERE
      tl.last_sample_id IS NULL AND
      (utl.server_id, utl.tablespaceid) = (sserver_id, tl.tablespaceid) AND
      tl.server_id = sserver_id AND cur.server_id IS NULL;

    UPDATE funcs_list ufl SET last_sample_id = s_id - 1
    FROM funcs_list fl LEFT JOIN sample_stat_user_functions cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.funcid) =
        (sserver_id, s_id, fl.datid, fl.funcid)
    WHERE
      fl.last_sample_id IS NULL AND
      fl.server_id = sserver_id AND cur.server_id IS NULL AND
      (ufl.server_id, ufl.datid, ufl.funcid) =
      (sserver_id, fl.datid, fl.funcid);

    UPDATE indexes_list uil SET last_sample_id = s_id - 1
    FROM indexes_list il LEFT JOIN sample_stat_indexes cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.indexrelid) =
        (sserver_id, s_id, il.datid, il.indexrelid)
    WHERE
      il.last_sample_id IS NULL AND
      il.server_id = sserver_id AND cur.server_id IS NULL AND
      (uil.server_id, uil.datid, uil.indexrelid) =
      (sserver_id, il.datid, il.indexrelid);

    UPDATE tables_list utl SET last_sample_id = s_id - 1
    FROM tables_list tl LEFT JOIN sample_stat_tables cur
      ON (cur.server_id, cur.sample_id, cur.datid, cur.relid) =
        (sserver_id, s_id, tl.datid, tl.relid)
    WHERE
      tl.last_sample_id IS NULL AND
      tl.server_id = sserver_id AND cur.server_id IS NULL AND
      (utl.server_id, utl.datid, utl.relid) =
      (sserver_id, tl.datid, tl.relid);

    UPDATE stmt_list slu SET last_sample_id = s_id - 1
    FROM sample_statements ss RIGHT JOIN stmt_list sl
      ON (ss.server_id, ss.sample_id, ss.queryid_md5) =
        (sserver_id, s_id, sl.queryid_md5)
    WHERE
      sl.server_id = sserver_id AND
      sl.last_sample_id IS NULL AND
      ss.server_id IS NULL AND
      (slu.server_id, slu.queryid_md5) = (sserver_id, sl.queryid_md5);

    UPDATE roles_list rlu SET last_sample_id = s_id - 1
    FROM
        sample_statements ss
      RIGHT JOIN roles_list rl
      ON (ss.server_id, ss.sample_id, ss.userid) =
        (sserver_id, s_id, rl.userid)
    WHERE
      rl.server_id = sserver_id AND
      rl.last_sample_id IS NULL AND
      ss.server_id IS NULL AND
      (rlu.server_id, rlu.userid) = (sserver_id, rl.userid);

    -- Deleting obsolete baselines
    DELETE FROM baselines
    WHERE keep_until < now()
      AND server_id = sserver_id;

    -- Deleting obsolete samples
    PERFORM num_nulls(min(s.sample_id),max(s.sample_id)) > 0 OR
      delete_samples(sserver_id, min(s.sample_id), max(s.sample_id)) > 0
    FROM samples s JOIN
      servers n USING (server_id)
    WHERE s.server_id = sserver_id
        AND s.sample_time < now() - (COALESCE(n.max_sample_age,ret) || ' days')::interval
        AND (s.server_id,s.sample_id) NOT IN (SELECT server_id,sample_id FROM bl_samples WHERE server_id = sserver_id);

    IF (server_properties #>> '{collect_timings}')::boolean THEN
      server_properties := jsonb_set(server_properties,'{timings,delete obsolete samples,end}',to_jsonb(clock_timestamp()));
      server_properties := jsonb_set(server_properties,'{timings,total,end}',to_jsonb(clock_timestamp()));
      -- Save timing statistics of sample
      INSERT INTO sample_timings
      SELECT sserver_id, s_id, key,(value::jsonb #>> '{end}')::timestamp with time zone - (value::jsonb #>> '{start}')::timestamp with time zone as time_spent
      FROM jsonb_each_text(server_properties #> '{timings}');
    END IF;
    ASSERT server_properties IS NOT NULL, 'lost properties';

    RETURN 0;
END;
$$ LANGUAGE plpgsql;
COMMENT ON FUNCTION take_sample(IN sserver_id integer, IN skip_sizes boolean) IS
  'Statistics sample creation function (by server_id)';
CREATE FUNCTION settings_format(IN sserver_id integer, IN start_id integer, IN end_id integer)
RETURNS TABLE (
    klass       text,
    name        text,
    reset_val   text,
    unit        text,
    source      text,
    notes       text,
    default_val boolean,
    defined_val boolean,
    h_ord       integer -- report header ordering
  )
SET search_path=@extschema@ AS $$
  SELECT
    CASE WHEN changed THEN 'new' ELSE 'init' END AS klass,
    name,
    reset_val,
    unit,
    concat_ws(':', sourcefile, sourceline) AS source,
    concat_ws(', ',
      CASE WHEN changed THEN first_seen ELSE NULL END,
      CASE WHEN pending_restart THEN 'Pending restart' ELSE NULL END
    ) AS notes,
    default_val,
    NOT default_val,
    CASE name
      WHEN 'version' THEN 10
      WHEN 'pgpro_version' THEN 21
      WHEN 'pgpro_edition' THEN 22
      WHEN 'pgpro_build' THEN 23
      ELSE NULL
    END AS h_ord
  FROM
    settings_and_changes(sserver_id, start_id, end_id)
  ORDER BY
    name,setting_scope,first_seen,pending_restart ASC NULLS FIRST
$$ LANGUAGE sql;
CREATE FUNCTION settings_format_diff(IN sserver_id integer,
  IN start1_id integer, IN end1_id integer,
  IN start2_id integer, IN end2_id integer)
RETURNS TABLE (
    klass       text,
    name        text,
    reset_val   text,
    unit        text,
    source      text,
    notes       text,
    default_val boolean,
    defined_val boolean,
    h_ord       integer -- report header ordering
  )
SET search_path=@extschema@ AS $$
  SELECT
    concat_ws('_',
      CASE WHEN changed THEN 'new' ELSE 'init' END,
      CASE WHEN s1.name IS NULL THEN 'i2'
           WHEN s2.name IS NULL THEN 'i1'
           ELSE NULL
      END
    ) AS klass,
    name,
    reset_val,
    COALESCE(s1.unit,s2.unit) as unit,
    concat_ws(':',
      COALESCE(s1.sourcefile,s2.sourcefile),
      COALESCE(s1.sourceline,s2.sourceline)
    ) AS source,
    concat_ws(', ',
      CASE WHEN changed THEN first_seen ELSE NULL END,
      CASE WHEN pending_restart THEN 'Pending restart' ELSE NULL END
    ) AS notes,
    default_val,
    NOT default_val,
    CASE name
      WHEN 'version' THEN 10
      WHEN 'pgpro_version' THEN 21
      WHEN 'pgpro_edition' THEN 22
      WHEN 'pgpro_build' THEN 23
      ELSE NULL
    END AS h_ord
  FROM
    settings_and_changes(sserver_id, start1_id, end1_id) s1
    FULL OUTER JOIN
    settings_and_changes(sserver_id, start2_id, end2_id) s2
    USING(first_seen, setting_scope, name, setting, reset_val, pending_restart, changed, default_val)
  ORDER BY
    name,setting_scope,first_seen,pending_restart ASC NULLS FIRST
$$ LANGUAGE sql;
CREATE FUNCTION report_queries_format(IN report_context jsonb, IN sserver_id integer, IN queries_list jsonb,
  IN start1_id integer, IN end1_id integer, IN start2_id integer, IN end2_id integer)
RETURNS TABLE(
  hexqueryid  text,
  query_texts jsonb
)
SET search_path=@extschema@ AS $$
DECLARE
    c_queries CURSOR(lim integer)
    FOR
    SELECT
      queryid,
      ord,
      rowspan,
      query
    FROM (
      SELECT
      queryid,
      row_number() OVER (PARTITION BY queryid
        ORDER BY
          last_sample_id DESC NULLS FIRST,
          queryid_md5 DESC NULLS FIRST
        ) ord,
      -- Calculate a value for statement rowspan attribute
      least(count(*) OVER (PARTITION BY queryid),lim) rowspan,
      query
      FROM (
        SELECT DISTINCT
          server_id,
          queryid,
          queryid_md5
        FROM
          jsonb_to_recordset(queries_list) ql(
            userid   bigint,
            datid    bigint,
            queryid  bigint
          )
          JOIN sample_statements ss USING (datid, userid, queryid)
        WHERE
          ss.server_id = sserver_id
          AND (
            sample_id BETWEEN start1_id AND end1_id
            OR sample_id BETWEEN start2_id AND end2_id
          )
      ) queryids
      JOIN stmt_list USING (server_id, queryid_md5)
      WHERE query IS NOT NULL
    ) ord_stmt_v
    WHERE ord <= lim
    ORDER BY
      queryid ASC,
      ord ASC;

    qr_result         RECORD;
    qlen_limit        integer;
    lim               CONSTANT integer := 3;
BEGIN
    IF NOT has_column_privilege('stmt_list', 'query', 'SELECT') THEN
      -- Return empty set when permissions denied to see query text
      hexqueryid := '';
      query_texts := '[]'::jsonb;
      query_texts := jsonb_insert(query_texts, '{-1}',
        to_jsonb('You must be a member of pg_read_all_stats to access query texts'::text));
      RETURN NEXT;
      RETURN;
    END IF;
    qlen_limit := (report_context #>> '{report_properties,max_query_length}')::integer;
    FOR qr_result IN c_queries(lim)
    LOOP
        IF qr_result.ord = 1 THEN
          hexqueryid := to_hex(qr_result.queryid);
          query_texts := '[]'::jsonb;
        END IF;
        query_texts := jsonb_insert(query_texts, '{-1}',
          to_jsonb(left(qr_result.query, qlen_limit)));
        IF qr_result.ord = qr_result.rowspan THEN
          RETURN NEXT;
        END IF;
    END LOOP;
    RETURN;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION profile_checkavail_walstats(IN sserver_id integer, IN start_id integer, IN end_id integer)
RETURNS BOOLEAN
SET search_path=@extschema@ AS $$
-- Check if there is WAL stats collected
  SELECT
    count(wal_bytes) > 0
  FROM sample_stat_wal
  WHERE
    server_id = sserver_id
    AND sample_id BETWEEN start_id + 1 AND end_id
$$ LANGUAGE sql;
CREATE FUNCTION get_report_context(IN sserver_id integer, IN start1_id integer, IN end1_id integer,
  IN description text = NULL,
  IN start2_id integer = NULL, IN end2_id integer = NULL)
RETURNS jsonb SET search_path=@extschema@ AS $$
DECLARE
  report_context  jsonb;
  r_result    RECORD;

  qlen_limit  integer;
  topn        integer;

  start1_time text;
  end1_time   text;
  start2_time text;
  end2_time   text;
BEGIN
    ASSERT num_nulls(start1_id, end1_id) = 0, 'At least first interval bounds is necessary';

    -- Getting query length limit setting
    BEGIN
        qlen_limit := current_setting('pg_profile.max_query_length')::integer;
    EXCEPTION
        WHEN OTHERS THEN qlen_limit := 20000;
    END;

    -- Getting TopN setting
    BEGIN
        topn := current_setting('pg_profile.topn')::integer;
    EXCEPTION
        WHEN OTHERS THEN topn := 20;
    END;

    -- Populate report settings
    -- Check if all samples of requested interval are available
    IF (
      SELECT count(*) != end1_id - start1_id + 1 FROM samples
      WHERE server_id = sserver_id AND sample_id BETWEEN start1_id AND end1_id
    ) THEN
      RAISE 'Not enough samples between %',
        format('%s AND %s', start1_id, end1_id);
    END IF;

    -- Get report times
    SELECT sample_time::text INTO STRICT start1_time FROM samples
    WHERE (server_id,sample_id) = (sserver_id,start1_id);
    SELECT sample_time::text INTO STRICT end1_time FROM samples
    WHERE (server_id,sample_id) = (sserver_id,end1_id);

    IF num_nulls(start2_id, end2_id) = 2 THEN
      report_context := jsonb_build_object(
      'report_features',jsonb_build_object(
        'dbstats_reset', profile_checkavail_dbstats_reset(sserver_id, start1_id, end1_id),
        'stmt_cnt_range', profile_checkavail_stmt_cnt(sserver_id, start1_id, end1_id),
        'stmt_cnt_all', profile_checkavail_stmt_cnt(sserver_id, 0, 0),
        'cluster_stats_reset', profile_checkavail_cluster_stats_reset(sserver_id, start1_id, end1_id),
        'wal_stats_reset', profile_checkavail_wal_stats_reset(sserver_id, start1_id, end1_id),
        'statstatements',profile_checkavail_statstatements(sserver_id, start1_id, end1_id),
        'planning_times',profile_checkavail_planning_times(sserver_id, start1_id, end1_id),
        'wait_sampling_tot',profile_checkavail_wait_sampling_total(sserver_id, start1_id, end1_id),
        'io_times',profile_checkavail_io_times(sserver_id, start1_id, end1_id),
        'statement_wal_bytes',profile_checkavail_stmt_wal_bytes(sserver_id, start1_id, end1_id),
        'statements_top_temp', profile_checkavail_top_temp(sserver_id, start1_id, end1_id),
        'statements_temp_io_times', profile_checkavail_statements_temp_io_times(sserver_id, start1_id, end1_id),
        'wal_stats',profile_checkavail_walstats(sserver_id, start1_id, end1_id),
        'sess_stats',profile_checkavail_sessionstats(sserver_id, start1_id, end1_id),
        'function_stats',profile_checkavail_functions(sserver_id, start1_id, end1_id),
        'trigger_function_stats',profile_checkavail_trg_functions(sserver_id, start1_id, end1_id),
        'kcachestatements',profile_checkavail_rusage(sserver_id,start1_id,end1_id),
        'rusage_planstats',profile_checkavail_rusage_planstats(sserver_id,start1_id,end1_id),
        'statements_jit_stats',profile_checkavail_statements_jit_stats(sserver_id, start1_id, end1_id),
        'top_tables_dead', profile_checkavail_tbl_top_dead(sserver_id,start1_id,end1_id),
        'top_tables_mods', profile_checkavail_tbl_top_mods(sserver_id,start1_id,end1_id),
        'table_new_page_updates', (
          SELECT COALESCE(sum(n_tup_newpage_upd), 0) > 0
          FROM sample_stat_tables_total
          WHERE server_id = sserver_id AND sample_id BETWEEN start1_id + 1 AND end1_id
        ),
        'stat_io', (
          SELECT COUNT(*) > 0 FROM (
            SELECT backend_type
            FROM sample_stat_io
            WHERE server_id = sserver_id AND
              sample_id BETWEEN start1_id + 1 AND end1_id LIMIT 1
            ) c
        ),
        'stat_io_reset', (
          -- We should include both ends here to detect resets performed
          SELECT bool_or(group_reset)
          FROM (
            SELECT COUNT(DISTINCT stats_reset) > 1 AS group_reset
            FROM sample_stat_io
            WHERE server_id = sserver_id AND sample_id BETWEEN start1_id AND end1_id
            GROUP BY backend_type, object, context
          ) gr
        ),
        'stat_slru', (
          SELECT COUNT(*) > 0 FROM (
            SELECT name
            FROM sample_stat_slru
            WHERE server_id = sserver_id AND
              sample_id BETWEEN start1_id + 1 AND end1_id LIMIT 1
            ) c
        ),
        'stat_slru_reset', (
          -- We should include both ends here to detect resets performed
          SELECT bool_or(group_reset)
          FROM (
            SELECT COUNT(DISTINCT stats_reset) > 1 AS group_reset
            FROM sample_stat_slru
            WHERE server_id = sserver_id AND sample_id BETWEEN start1_id AND end1_id
            GROUP BY name
          ) gr
        ),
        'checksum_fail_detected', COALESCE((
          SELECT sum(checksum_failures) > 0
          FROM sample_stat_database
          WHERE server_id = sserver_id AND sample_id BETWEEN start1_id + 1 AND end1_id
          ), false)
        ),
      'report_properties',jsonb_build_object(
        'interval_duration_sec',
          (SELECT extract(epoch FROM e.sample_time - s.sample_time)
          FROM samples s JOIN samples e USING (server_id)
          WHERE e.sample_id=end1_id and s.sample_id=start1_id
            AND server_id = sserver_id),
        'topn', topn,
        'max_query_length', qlen_limit,
        'start1_id', start1_id,
        'end1_id', end1_id,
        'report_start1', start1_time,
        'report_end1', end1_time
        )
      );

      -- stat_activity features
      IF has_table_privilege('sample_act_backend', 'SELECT') THEN
        report_context := jsonb_set(report_context, '{report_features,act_backend}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend
              WHERE server_id = sserver_id AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_ix}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state
              WHERE server_id = sserver_id AND
                state_code = 1 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_ixa}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state
              WHERE server_id = sserver_id AND
                state_code = 2 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_active}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state JOIN sample_act_statement
                USING (server_id, sample_id, pid, state_change)
              WHERE server_id = sserver_id AND
                state_code = 3 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id
                )
              LIMIT 1
            ) c
          )
        );
      END IF;
    ELSIF num_nulls(start2_id, end2_id) = 0 THEN
      -- Get report times
      SELECT sample_time::text INTO STRICT start2_time FROM samples
      WHERE (server_id,sample_id) = (sserver_id,start2_id);
      SELECT sample_time::text INTO STRICT end2_time FROM samples
      WHERE (server_id,sample_id) = (sserver_id,end2_id);
      -- Check if all samples of requested interval are available
      IF (
        SELECT count(*) != end2_id - start2_id + 1 FROM samples
        WHERE server_id = sserver_id AND sample_id BETWEEN start2_id AND end2_id
      ) THEN
        RAISE 'Not enough samples between %',
          format('%s AND %s', start2_id, end2_id);
      END IF;
      report_context := jsonb_build_object(
      'report_features',jsonb_build_object(
        'dbstats_reset', profile_checkavail_dbstats_reset(sserver_id, start1_id, end1_id) OR
          profile_checkavail_dbstats_reset(sserver_id, start2_id, end2_id),
        'stmt_cnt_range', profile_checkavail_stmt_cnt(sserver_id, start1_id, end1_id) OR
          profile_checkavail_stmt_cnt(sserver_id, start2_id, end2_id),
        'stmt_cnt_all', profile_checkavail_stmt_cnt(sserver_id, 0, 0),
        'cluster_stats_reset', profile_checkavail_cluster_stats_reset(sserver_id, start1_id, end1_id) OR
          profile_checkavail_cluster_stats_reset(sserver_id, start2_id, end2_id),
        'wal_stats_reset', profile_checkavail_wal_stats_reset(sserver_id, start1_id, end1_id) OR
          profile_checkavail_wal_stats_reset(sserver_id, start2_id, end2_id),
        'statstatements',profile_checkavail_statstatements(sserver_id, start1_id, end1_id) OR
          profile_checkavail_statstatements(sserver_id, start2_id, end2_id),
        'planning_times',profile_checkavail_planning_times(sserver_id, start1_id, end1_id) OR
          profile_checkavail_planning_times(sserver_id, start2_id, end2_id),
        'wait_sampling_tot',profile_checkavail_wait_sampling_total(sserver_id, start1_id, end1_id) OR
          profile_checkavail_wait_sampling_total(sserver_id, start2_id, end2_id),
        'io_times',profile_checkavail_io_times(sserver_id, start1_id, end1_id) OR
          profile_checkavail_io_times(sserver_id, start2_id, end2_id),
        'statement_wal_bytes',profile_checkavail_stmt_wal_bytes(sserver_id, start1_id, end1_id) OR
          profile_checkavail_stmt_wal_bytes(sserver_id, start2_id, end2_id),
        'statements_top_temp', profile_checkavail_top_temp(sserver_id, start1_id, end1_id) OR
            profile_checkavail_top_temp(sserver_id, start2_id, end2_id),
        'statements_temp_io_times', profile_checkavail_statements_temp_io_times(sserver_id, start1_id, end1_id) OR
            profile_checkavail_statements_temp_io_times(sserver_id, start2_id, end2_id),
        'wal_stats',profile_checkavail_walstats(sserver_id, start1_id, end1_id) OR
          profile_checkavail_walstats(sserver_id, start2_id, end2_id),
        'sess_stats',profile_checkavail_sessionstats(sserver_id, start1_id, end1_id) OR
          profile_checkavail_sessionstats(sserver_id, start2_id, end2_id),
        'function_stats',profile_checkavail_functions(sserver_id, start1_id, end1_id) OR
          profile_checkavail_functions(sserver_id, start2_id, end2_id),
        'trigger_function_stats',profile_checkavail_trg_functions(sserver_id, start1_id, end1_id) OR
          profile_checkavail_trg_functions(sserver_id, start2_id, end2_id),
        'kcachestatements',profile_checkavail_rusage(sserver_id, start1_id, end1_id) OR
          profile_checkavail_rusage(sserver_id, start2_id, end2_id),
        'rusage_planstats',profile_checkavail_rusage_planstats(sserver_id, start1_id, end1_id) OR
          profile_checkavail_rusage_planstats(sserver_id, start2_id, end2_id),
        'statements_jit_stats',profile_checkavail_statements_jit_stats(sserver_id, start1_id, end1_id) OR
          profile_checkavail_statements_jit_stats(sserver_id, start2_id, end2_id),
        'table_new_page_updates', (
          SELECT COALESCE(sum(n_tup_newpage_upd), 0) > 0
          FROM sample_stat_tables_total
          WHERE server_id = sserver_id AND
            (sample_id BETWEEN start1_id + 1 AND end1_id OR
            sample_id BETWEEN start2_id + 1 AND end2_id)
        ),
        'stat_io', (
          SELECT COUNT(*) > 0 FROM (
            SELECT backend_type
            FROM sample_stat_io
            WHERE server_id = sserver_id AND (
              sample_id BETWEEN start1_id + 1 AND end1_id OR
              sample_id BETWEEN start2_id + 1 AND end2_id
              ) LIMIT 1
            ) c
        ),
        'stat_io_reset', (
          -- We should include both ends here to detect resets performed
          SELECT bool_or(group_reset)
          FROM (
            SELECT COUNT(DISTINCT stats_reset) > 1 AS group_reset
            FROM sample_stat_io
            WHERE server_id = sserver_id AND (
              sample_id BETWEEN start1_id AND end1_id OR
              sample_id BETWEEN start2_id AND end2_id
              )
            GROUP BY backend_type, object, context
          ) gr
        ),
        'stat_slru', (
          SELECT COUNT(*) > 0 FROM (
            SELECT name
            FROM sample_stat_slru
            WHERE server_id = sserver_id AND (
              sample_id BETWEEN start1_id + 1 AND end1_id OR
              sample_id BETWEEN start2_id + 1 AND end2_id
              ) LIMIT 1
            ) c
        ),
        'stat_slru_reset', (
          -- We should include both ends here to detect resets performed
          SELECT bool_or(group_reset)
          FROM (
            SELECT COUNT(DISTINCT stats_reset) > 1 AS group_reset
            FROM sample_stat_slru
            WHERE server_id = sserver_id AND (
              sample_id BETWEEN start1_id AND end1_id OR
              sample_id BETWEEN start2_id AND end2_id
            )
            GROUP BY name
          ) gr
        ),
        'checksum_fail_detected', COALESCE((
          SELECT sum(checksum_failures) > 0
          FROM sample_stat_database
          WHERE server_id = sserver_id AND
            (sample_id BETWEEN start1_id + 1 AND end1_id OR
            sample_id BETWEEN start2_id + 1 AND end2_id)
          ), false)
        ),
      'report_properties',jsonb_build_object(
        'interval1_duration_sec',
          (SELECT extract(epoch FROM e.sample_time - s.sample_time)
          FROM samples s JOIN samples e USING (server_id)
          WHERE e.sample_id=end1_id and s.sample_id=start1_id
            AND server_id = sserver_id),
        'interval2_duration_sec',
          (SELECT extract(epoch FROM e.sample_time - s.sample_time)
          FROM samples s JOIN samples e USING (server_id)
          WHERE e.sample_id=end2_id and s.sample_id=start2_id
            AND server_id = sserver_id),

        'topn', topn,
        'max_query_length', qlen_limit,

        'start1_id', start1_id,
        'end1_id', end1_id,
        'report_start1', start1_time,
        'report_end1', end1_time,

        'start2_id', start2_id,
        'end2_id', end2_id,
        'report_start2', start2_time,
        'report_end2', end2_time
        )
      );

      -- stat_activity features
      IF has_table_privilege('sample_act_backend', 'SELECT') THEN
        report_context := jsonb_set(report_context, '{report_features,act_backend}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend
              WHERE server_id = sserver_id AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id OR
                  sample_id BETWEEN start2_id + 1 AND end2_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_ix}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state
              WHERE server_id = sserver_id AND
                state_code = 1 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id OR
                  sample_id BETWEEN start2_id + 1 AND end2_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_ixa}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state
              WHERE server_id = sserver_id AND
                state_code = 2 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id OR
                  sample_id BETWEEN start2_id + 1 AND end2_id
                )
              LIMIT 1
            ) c
          )
        );
        report_context := jsonb_set(report_context, '{report_features,act_active}',
          (
          SELECT
            to_jsonb(COUNT(*) > 0) FROM (
              SELECT 1
              FROM sample_act_backend_state
              WHERE server_id = sserver_id AND
                state_code = 3 AND (
                  sample_id BETWEEN start1_id + 1 AND end1_id OR
                  sample_id BETWEEN start2_id + 1 AND end2_id
                )
              LIMIT 1
            ) c
          )
        );
      END IF;
    ELSE
      RAISE 'Two bounds must be specified for second interval';
    END IF;

    -- Server name and description
    SELECT server_name, server_description INTO STRICT r_result
    FROM servers WHERE server_id = sserver_id;
    report_context := jsonb_set(report_context, '{report_properties,server_name}',
      to_jsonb(r_result.server_name)
    );
    IF r_result.server_description IS NOT NULL AND r_result.server_description != ''
    THEN
      report_context := jsonb_set(report_context, '{report_properties,server_description}',
        to_jsonb(format(
          '<p>%s</p>',
          r_result.server_description
        ))
      );
    ELSE
      report_context := jsonb_set(report_context, '{report_properties,server_description}',to_jsonb(''::text));
    END IF;
    -- Report description
    IF description IS NOT NULL AND description != '' THEN
      report_context := jsonb_set(report_context, '{report_properties,description}',
        to_jsonb(format(
          '<h2>Report description</h2><p>%s</p>',
          description
        ))
      );
    ELSE
      report_context := jsonb_set(report_context, '{report_properties,description}',to_jsonb(''::text));
    END IF;
    -- Version substitution
    IF (SELECT count(*) = 1 FROM pg_catalog.pg_extension WHERE extname = 'pg_profile') THEN
      SELECT extversion INTO STRICT r_result FROM pg_catalog.pg_extension WHERE extname = 'pg_profile';
      report_context := jsonb_set(report_context, '{report_properties,pgprofile_version}',
        to_jsonb(r_result.extversion)
      );
    END IF;
  RETURN report_context;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION get_report_datasets(IN report_context jsonb, IN sserver_id integer)
RETURNS jsonb SET search_path=@extschema@ AS $$
DECLARE
  start1_id   integer = (report_context #>> '{report_properties,start1_id}')::integer;
  start2_id   integer = (report_context #>> '{report_properties,start2_id}')::integer;
  end1_id     integer = (report_context #>> '{report_properties,end1_id}')::integer;
  end2_id     integer = (report_context #>> '{report_properties,end2_id}')::integer;

  datasets         jsonb = '{}';
  dataset          jsonb;
  queries_set      jsonb = '[]';
  act_queries_set  jsonb = '[]';
  r_result         RECORD;
  r_dataset        text;
BEGIN
  IF num_nulls(start1_id, end1_id) = 0 AND num_nulls(start2_id, end2_id) > 0 THEN
    -- Regular report
    -- report properties dataset
    dataset := '[]'::jsonb;
    dataset := dataset || (report_context #>> '{report_properties}')::jsonb;
    datasets := jsonb_set(datasets, '{properties}', dataset);

    -- database statistics dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM dbstats_format(sserver_id, start1_id, end1_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{dbstat}', dataset);

    IF (report_context #>> '{report_features,dbstats_reset}')::boolean THEN
      -- dbstats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM dbstats_reset_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{dbstats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stat_io}')::boolean THEN
      -- stat_io dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stat_io_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stat_io}', dataset);
      IF (report_context #>> '{report_features,stat_io_reset}')::boolean THEN
        -- IO reset dataset
        dataset := '[]'::jsonb;
        FOR r_result IN (
            SELECT *
            FROM cluster_stat_io_reset_format(sserver_id, start1_id, end1_id)
          ) LOOP
          dataset := dataset || to_jsonb(r_result);
        END LOOP;
        datasets := jsonb_set(datasets, '{stat_io_reset}', dataset);
      END IF;
    END IF;

    IF (report_context #>> '{report_features,stat_slru}')::boolean THEN
      -- stat_slru dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stat_slru_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stat_slru}', dataset);
      IF (report_context #>> '{report_features,stat_slru_reset}')::boolean THEN
        -- SLRU reset dataset
        dataset := '[]'::jsonb;
        FOR r_result IN (
            SELECT *
            FROM cluster_stat_slru_reset_format(sserver_id, start1_id, end1_id)
          ) LOOP
          dataset := dataset || to_jsonb(r_result);
        END LOOP;
        datasets := jsonb_set(datasets, '{stat_slru_reset}', dataset);
      END IF;
    END IF;

    IF (report_context #>> '{report_features,statstatements}')::boolean THEN
      -- statements by database dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM statements_dbstats_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{statements_dbstats}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stmt_cnt_range}')::boolean THEN
      -- statements count of max for interval
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stmt_cnt_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stmt_cnt_range}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stmt_cnt_all}')::boolean THEN
      -- statements count of max for all samples
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stmt_cnt_format(sserver_id, 0, 0)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stmt_cnt_all}', dataset);
    END IF;

    -- cluster statistics dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM cluster_stats_format(sserver_id, start1_id, end1_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{cluster_stats}', dataset);

    IF (report_context #>> '{report_features,cluster_stats_reset}')::boolean THEN
      -- cluster stats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stats_reset_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{cluster_stats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,wal_stats_reset}')::boolean THEN
      -- WAL stats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wal_stats_reset_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wal_stats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,wal_stats}')::boolean THEN
      -- WAL stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wal_stats_format(sserver_id, start1_id, end1_id,
            (report_context #>> '{report_properties,interval_duration_sec}')::numeric)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wal_stats}', dataset);
    END IF;

    -- Tablespace stats dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM tablespace_stats_format(sserver_id, start1_id, end1_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{tablespace_stats}', dataset);

    IF (report_context #>> '{report_features,wait_sampling_tot}')::boolean THEN
      -- Wait totals dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wait_sampling_total_stats_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wait_sampling_total_stats}', dataset);
      -- Wait events dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_wait_sampling_events_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wait_sampling_events}', dataset);
    END IF;

    IF (report_context #>> '{report_features,statstatements}')::boolean THEN
      -- Statement stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_statements_format(sserver_id, start1_id, end1_id)
          WHERE least(
              ord_total_time,
              ord_plan_time,
              ord_exec_time,
              ord_calls,
              ord_io_time,
              ord_shared_blocks_fetched,
              ord_shared_blocks_read,
              ord_shared_blocks_dirt,
              ord_shared_blocks_written,
              ord_wal,
              ord_temp,
              ord_jit
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{top_statements}', dataset);
    END IF;

    IF (report_context #>> '{report_features,kcachestatements}')::boolean THEN
      -- Statement rusage stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_rusage_statements_format(sserver_id, start1_id, end1_id)
          WHERE least(
              ord_cpu_time,
              ord_io_bytes
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{top_rusage_statements}', dataset);
    END IF;

    IF (report_context #>> '{report_features,act_backend}')::boolean THEN
      -- Session states and queries cartured by subsamples
      -- Database aggregated data
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stat_activity_agg_format(sserver_id, start1_id, end1_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{db_activity_agg}', dataset);
      -- Top states dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stat_activity_states_format(sserver_id, start1_id, end1_id)
          WHERE least(
              ord_dur,
              ord_age
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{act_top_states}', dataset);
    END IF;

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_tables_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_dml,
          ord_seq_scan,
          ord_upd,
          ord_growth,
          ord_vac,
          ord_anl
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_tables}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_io_tables_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_read,
          ord_fetch
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_io_tables}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_io_indexes_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_read,
          ord_fetch
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_io_indexes}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_indexes_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_growth,
          ord_unused,
          ord_vac
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_indexes}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_functions_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_time,
          ord_calls,
          ord_trgtime
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_functions}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_tbl_last_sample_format(sserver_id, start1_id, end1_id)
        WHERE least(
          ord_dead,
          ord_mod
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_tbl_last_sample}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM settings_format(sserver_id, start1_id, end1_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{settings}', dataset);

    -- Now we need to collect queries over datasets
    FOR r_dataset IN (SELECT jsonb_object_keys(datasets)) LOOP
      -- skip datasets without queries
      CONTINUE WHEN NOT
        (datasets #> ARRAY[r_dataset, '0']) ?| ARRAY['queryid'];
      FOR r_result IN (
        SELECT
          userid,
          datid,
          queryid
        FROM
          jsonb_to_recordset(datasets #> ARRAY[r_dataset]) AS
            entry(
              userid  oid,
              datid   oid,
              queryid bigint
            )
        )
      LOOP
        queries_set := queries_set || jsonb_build_object(
          'userid', r_result.userid,
          'datid', r_result.datid,
          'queryid', r_result.queryid
        );
      END LOOP; -- over dataset entries
    END LOOP; -- over datasets

    -- Collect stat_activity queries over datasets
    FOR r_dataset IN (SELECT jsonb_object_keys(datasets)) LOOP
      -- skip datasets without queries
      CONTINUE WHEN NOT
        (datasets #> ARRAY[r_dataset, '0']) ?| ARRAY['act_query_md5'];
      FOR r_result IN (
        SELECT
          act_query_md5
        FROM
          jsonb_to_recordset(datasets #> ARRAY[r_dataset]) AS
            entry(
              act_query_md5  text
            )
        )
      LOOP
        act_queries_set := act_queries_set || jsonb_build_object(
          'act_query_md5', r_result.act_query_md5
        );
      END LOOP; -- over dataset entries
    END LOOP; -- over datasets

    -- Query texts dataset should be formed the last
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM report_queries_format(report_context, sserver_id, queries_set, start1_id, end1_id, NULL, NULL)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{queries}', dataset);

    -- stat_activity queries list construction
    IF (report_context #>> '{report_features,act_backend}')::boolean THEN
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM report_active_queries_format(report_context, sserver_id, act_queries_set)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{act_queries}', dataset);
    END IF;

  ELSIF num_nulls(start1_id, end1_id, start2_id, end2_id) = 0 THEN
    -- Differential report
    -- report properties dataset
    dataset := '[]'::jsonb;
    dataset := dataset || (report_context #>> '{report_properties}')::jsonb;
    datasets := jsonb_set(datasets, '{properties}', dataset);

    -- database statistics dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM dbstats_format_diff(sserver_id, start1_id, end1_id,
          start2_id, end2_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{dbstat}', dataset);

    IF (report_context #>> '{report_features,dbstats_reset}')::boolean THEN
      -- dbstats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM dbstats_reset_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{dbstats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stat_io}')::boolean THEN
      -- stat_io dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stat_io_format(sserver_id, start1_id, end1_id,
                 start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stat_io}', dataset);
      IF (report_context #>> '{report_features,stat_io_reset}')::boolean THEN
        -- SLRU reset dataset
        dataset := '[]'::jsonb;
        FOR r_result IN (
            SELECT *
            FROM cluster_stat_io_reset_format(sserver_id,
              start1_id, end1_id, start2_id, end2_id)
          ) LOOP
          dataset := dataset || to_jsonb(r_result);
        END LOOP;
        datasets := jsonb_set(datasets, '{stat_io_reset}', dataset);
      END IF;
    END IF;

    IF (report_context #>> '{report_features,stat_slru}')::boolean THEN
      -- stat_slru dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stat_slru_format(sserver_id, start1_id, end1_id,
                 start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stat_slru}', dataset);
      IF (report_context #>> '{report_features,stat_slru_reset}')::boolean THEN
        -- SLRU reset dataset
        dataset := '[]'::jsonb;
        FOR r_result IN (
            SELECT *
            FROM cluster_stat_slru_reset_format(sserver_id,
              start1_id, end1_id, start2_id, end2_id)
          ) LOOP
          dataset := dataset || to_jsonb(r_result);
        END LOOP;
        datasets := jsonb_set(datasets, '{stat_slru_reset}', dataset);
      END IF;
    END IF;

    IF (report_context #>> '{report_features,statstatements}')::boolean THEN
      -- statements by database dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM statements_dbstats_format_diff(sserver_id, start1_id, end1_id,
                 start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{statements_dbstats}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stmt_cnt_range}')::boolean THEN
      -- statements count of max for interval
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stmt_cnt_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stmt_cnt_range}', dataset);
    END IF;

    IF (report_context #>> '{report_features,stmt_cnt_all}')::boolean THEN
      -- statements count of max for all samples
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stmt_cnt_format(sserver_id, 0, 0)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{stmt_cnt_all}', dataset);
    END IF;

    -- cluster statistics dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM cluster_stats_format_diff(sserver_id, start1_id, end1_id,
                 start2_id, end2_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{cluster_stats}', dataset);

    IF (report_context #>> '{report_features,cluster_stats_reset}')::boolean THEN
      -- cluster stats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM cluster_stats_reset_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{cluster_stats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,wal_stats_reset}')::boolean THEN
      -- WAL stats reset dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wal_stats_reset_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wal_stats_reset}', dataset);
    END IF;

    IF (report_context #>> '{report_features,wal_stats}')::boolean THEN
      -- WAL stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wal_stats_format(sserver_id,
            start1_id, end1_id, start2_id, end2_id,
            (report_context #>> '{report_properties,interval1_duration_sec}')::numeric,
            (report_context #>> '{report_properties,interval2_duration_sec}')::numeric)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wal_stats}', dataset);
    END IF;

    -- Tablespace stats dataset
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM tablespace_stats_format_diff(sserver_id, start1_id, end1_id, start2_id, end2_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{tablespace_stats}', dataset);

    IF (report_context #>> '{report_features,wait_sampling_tot}')::boolean THEN
      -- Wait totals dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM wait_sampling_total_stats_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wait_sampling_total_stats}', dataset);
      -- Wait events dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_wait_sampling_events_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{wait_sampling_events}', dataset);
    END IF;

    IF (report_context #>> '{report_features,statstatements}')::boolean THEN
      -- Statement stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_statements_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
          WHERE least(
              ord_total_time,
              ord_plan_time,
              ord_exec_time,
              ord_calls,
              ord_io_time,
              ord_shared_blocks_fetched,
              ord_shared_blocks_read,
              ord_shared_blocks_dirt,
              ord_shared_blocks_written,
              ord_wal,
              ord_temp,
              ord_jit
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{top_statements}', dataset);
    END IF;

    IF (report_context #>> '{report_features,kcachestatements}')::boolean THEN
      -- Statement rusage stats dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM top_rusage_statements_format_diff(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
          WHERE least(
              ord_cpu_time,
              ord_io_bytes
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{top_rusage_statements}', dataset);
    END IF;

    IF (report_context #>> '{report_features,act_backend}')::boolean THEN
      -- Session states and queries cartured by subsamples
      -- Database aggregated data
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stat_activity_agg_format(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{db_activity_agg}', dataset);
      -- Top states dataset
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM stat_activity_states_format(sserver_id, start1_id, end1_id,
            start2_id, end2_id)
          WHERE least(
              ord_dur,
              ord_age
            ) <= (report_context #>> '{report_properties,topn}')::numeric
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{act_top_states}', dataset);
    END IF;

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_tables_format_diff(sserver_id, start1_id, end1_id, start2_id, end2_id)
        WHERE least(
          ord_dml,
          ord_seq_scan,
          ord_upd,
          ord_growth,
          ord_vac,
          ord_anl
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_tables}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_io_tables_format_diff(sserver_id, start1_id, end1_id, start2_id, end2_id)
        WHERE least(
          ord_read,
          ord_fetch
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_io_tables}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_io_indexes_format_diff(sserver_id, start1_id, end1_id, start2_id, end2_id)
        WHERE least(
          ord_read,
          ord_fetch
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_io_indexes}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_indexes_format_diff(sserver_id,
          start1_id, end1_id, start2_id, end2_id)
        WHERE least(
          ord_growth,
          ord_vac
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_indexes}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM top_functions_format_diff(sserver_id,
          start1_id, end1_id, start2_id, end2_id)
        WHERE least(
          ord_time,
          ord_calls,
          ord_trgtime
        ) <= (report_context #>> '{report_properties,topn}')::numeric
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{top_functions}', dataset);

    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM settings_format_diff(sserver_id,
          start1_id, end1_id, start2_id, end2_id)
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{settings}', dataset);

    -- Now we need to collect queries over datasets
    FOR r_dataset IN (SELECT jsonb_object_keys(datasets)) LOOP
      -- skip datasets without queries
      CONTINUE WHEN NOT
        (datasets #> ARRAY[r_dataset, '0']) ?| ARRAY['queryid'];
      FOR r_result IN (
        SELECT
          userid,
          datid,
          queryid
        FROM
          jsonb_to_recordset(datasets #> ARRAY[r_dataset]) AS
            entry(
              userid  oid,
              datid   oid,
              queryid bigint
            )
        )
      LOOP
        queries_set := queries_set || jsonb_build_object(
          'userid', r_result.userid,
          'datid', r_result.datid,
          'queryid', r_result.queryid
        );
      END LOOP; -- over dataset entries
    END LOOP; -- over datasets

    -- Collect stat_activity queries over datasets
    FOR r_dataset IN (SELECT jsonb_object_keys(datasets)) LOOP
      -- skip datasets without queries
      CONTINUE WHEN NOT
        (datasets #> ARRAY[r_dataset, '0']) ?| ARRAY['act_query_md5'];
      FOR r_result IN (
        SELECT
          act_query_md5
        FROM
          jsonb_to_recordset(datasets #> ARRAY[r_dataset]) AS
            entry(
              act_query_md5  text
            )
        )
      LOOP
        act_queries_set := act_queries_set || jsonb_build_object(
          'act_query_md5', r_result.act_query_md5
        );
      END LOOP; -- over dataset entries
    END LOOP; -- over datasets

    -- Query texts dataset should be formed the last
    dataset := '[]'::jsonb;
    FOR r_result IN (
        SELECT *
        FROM report_queries_format(report_context, sserver_id, queries_set,
          start1_id, end1_id, start2_id, end2_id
        )
      ) LOOP
      dataset := dataset || to_jsonb(r_result);
    END LOOP;
    datasets := jsonb_set(datasets, '{queries}', dataset);

    -- stat_activity queries list construction
    IF (report_context #>> '{report_features,act_backend}')::boolean THEN
      dataset := '[]'::jsonb;
      FOR r_result IN (
          SELECT *
          FROM report_active_queries_format(report_context, sserver_id, act_queries_set)
        ) LOOP
        dataset := dataset || to_jsonb(r_result);
      END LOOP;
      datasets := jsonb_set(datasets, '{act_queries}', dataset);
    END IF;

  END IF;
  RETURN datasets;
END;
$$ LANGUAGE plpgsql;
/* Subsample objects storage */

CREATE TABLE server_subsample (
    server_id           integer,
    subsample_enabled   boolean DEFAULT true NOT NULL,
    min_query_dur       interval hour to second,
    min_xact_dur        interval hour to second,
    min_xact_age        bigint,
    min_idle_xact_dur   interval hour to second,
    CONSTRAINT pk_server_subsample PRIMARY KEY (server_id),
    CONSTRAINT fk_server_subsample_server FOREIGN KEY (server_id)
      REFERENCES servers(server_id) ON DELETE CASCADE
);
COMMENT ON TABLE server_subsample IS 'Server subsampling settings';

CREATE TABLE sample_act_backend (
    server_id         integer,
    sample_id         integer,
    pid               integer,
    backend_start     timestamp with time zone,
    datid             oid,
    datname           name,
    usesysid          oid,
    usename           name,
    client_addr       inet,
    client_hostname   text,
    client_port       integer,
    backend_type      text,
    backend_last_ts   timestamp with time zone,
    CONSTRAINT pk_sample_backends PRIMARY KEY (server_id, sample_id, pid, backend_start),
    CONSTRAINT fk_backends_sample FOREIGN KEY (server_id, sample_id)
      REFERENCES samples (server_id, sample_id) ON DELETE CASCADE
);

CREATE TABLE sample_act_xact (
    server_id         integer,
    sample_id         integer,
    pid               integer,
    backend_start     timestamp with time zone,
    xact_start        timestamp with time zone,
    backend_xid       xid,
    xact_last_ts      timestamp with time zone,
    CONSTRAINT pk_sample_xact PRIMARY KEY (server_id, sample_id, pid, xact_start),
    CONSTRAINT fk_xact_act_backend FOREIGN KEY (server_id, sample_id, pid, backend_start)
      REFERENCES sample_act_backend(server_id, sample_id, pid, backend_start)
      ON DELETE CASCADE
);

CREATE TABLE sample_act_backend_state (
    server_id         integer,
    sample_id         integer,
    pid               integer,
    backend_start     timestamp with time zone,
    application_name  text,
    state_code        integer, -- 1 idle in xact, 2 idle in xact aborted, 3 active
    state_change      timestamp with time zone,
    state_last_ts     timestamp with time zone,
    xact_start        timestamp with time zone,
    backend_xmin      xid,
    backend_xmin_age  bigint,
    CONSTRAINT pk_sample_bk_state PRIMARY KEY (server_id, sample_id, pid, state_change),
    CONSTRAINT fk_bk_state_bk FOREIGN KEY (server_id, sample_id, pid, backend_start)
      REFERENCES sample_act_backend (server_id, sample_id, pid, backend_start)
      ON DELETE CASCADE,
    CONSTRAINT fk_bk_state_xact FOREIGN KEY (server_id, sample_id, pid, xact_start)
      REFERENCES sample_act_xact (server_id, sample_id, pid, xact_start)
      ON DELETE CASCADE
);

CREATE TABLE act_query(
    server_id      integer NOT NULL REFERENCES servers(server_id) ON DELETE CASCADE
      DEFERRABLE INITIALLY IMMEDIATE,
    act_query_md5  char(32),
    act_query      text,
    last_sample_id integer,
    CONSTRAINT pk_act_query PRIMARY KEY (server_id, act_query_md5),
    CONSTRAINT fk_act_query_samples FOREIGN KEY (server_id, last_sample_id)
      REFERENCES samples (server_id, sample_id) ON DELETE CASCADE
      DEFERRABLE INITIALLY IMMEDIATE
);
CREATE INDEX ix_act_query_smp ON act_query(server_id, last_sample_id);
COMMENT ON TABLE act_query IS 'Statements, captured in subsamples from pg_stat_activity';

CREATE TABLE sample_act_statement (
    server_id         integer,
    sample_id         integer,
    pid               integer,
    leader_pid        integer,
    state_change      timestamp with time zone,
    query_start       timestamp with time zone,
    query_id          bigint,
    act_query_md5     char(32),
    stmt_last_ts      timestamp with time zone,
    CONSTRAINT pk_sample_act_stmt PRIMARY KEY (server_id, sample_id, pid, query_start),
    CONSTRAINT fk_act_stmt_bk_state FOREIGN KEY (server_id, sample_id, pid, state_change)
      REFERENCES sample_act_backend_state (server_id, sample_id, pid, state_change)
      ON DELETE CASCADE,
    CONSTRAINT fk_act_stmt_query FOREIGN KEY (server_id, act_query_md5)
      REFERENCES act_query(server_id, act_query_md5)
      ON DELETE CASCADE
);

CREATE TABLE last_stat_activity (
    server_id         integer,
    sample_id         integer,
    subsample_ts      timestamp with time zone,
    datid             oid,
    datname           name,
    pid               integer,
    leader_pid        integer,
    usesysid          oid,
    usename           name,
    application_name  text,
    client_addr       inet,
    client_hostname   text,
    client_port       integer,
    backend_start     timestamp with time zone,
    xact_start        timestamp with time zone,
    query_start       timestamp with time zone,
    state_change      timestamp with time zone,
    state             text,
    backend_xid       xid,
    backend_xmin      xid,
    query_id          bigint,
    query             text,
    backend_type      text,
    backend_xmin_age  bigint
)
PARTITION BY LIST (server_id);
COMMENT ON TABLE last_stat_activity IS 'Subsample table for interesting stat_activity entries';

CREATE TABLE session_attr(
    server_id         integer,
    sess_attr_id      serial,
    backend_type      text,
    datid             oid,
    datname           name,
    usesysid          oid,
    usename           name,
    application_name  text,
    client_addr       inet,
    last_sample_id    integer,
    CONSTRAINT pk_subsample_sa PRIMARY KEY (server_id, sess_attr_id),
    CONSTRAINT uk_subsample_sa UNIQUE (server_id, backend_type, datid,
      datname, usesysid, usename, application_name, client_addr),
    CONSTRAINT fk_session_attr_samples FOREIGN KEY (server_id, last_sample_id)
      REFERENCES samples (server_id, sample_id) ON DELETE CASCADE
      DEFERRABLE INITIALLY IMMEDIATE
);
--ALTER SEQUENCE session_attr_sess_attr_id CYCLE;

CREATE TABLE sample_stat_activity_cnt(
    server_id         integer,
    sample_id         integer,
    subsample_ts      timestamp with time zone,

    sess_attr_id      integer,

    total             integer,
    active            integer,
    idle              integer,
    idle_t            integer,
    idle_ta           integer,
    state_null        integer,
    lwlock            integer,
    lock              integer,
    bufferpin         integer,
    activity          integer,
    extension         integer,
    client            integer,
    ipc               integer,
    timeout           integer,
    io                integer,

    CONSTRAINT pk_sample_stat_activity_cnt PRIMARY KEY (
      server_id, sample_id, subsample_ts, sess_attr_id
    ),
    CONSTRAINT subsample_attrs FOREIGN KEY (server_id, sess_attr_id)
      REFERENCES session_attr(server_id, sess_attr_id) ON DELETE CASCADE
      DEFERRABLE INITIALLY IMMEDIATE
);

CREATE TABLE last_stat_activity_count(
    server_id         integer,
    sample_id         integer,
    subsample_ts      timestamp with time zone,
    backend_type      text,
    datid             oid,
    datname           name,
    usesysid          oid,
    usename           name,
    application_name  text,
    client_addr       inet,
    total             integer,
    active            integer,
    idle              integer,
    idle_t            integer,
    idle_ta           integer,
    state_null        integer,
    lwlock            integer,
    lock              integer,
    bufferpin         integer,
    activity          integer,
    extension         integer,
    client            integer,
    ipc               integer,
    timeout           integer,
    io                integer
)
PARTITION BY LIST (server_id);

GRANT SELECT ON sample_act_backend TO pg_read_all_stats;
GRANT SELECT ON sample_act_xact TO pg_read_all_stats;
GRANT SELECT ON sample_act_backend_state TO pg_read_all_stats;
GRANT SELECT ON act_query TO pg_read_all_stats;
GRANT SELECT ON sample_act_statement TO pg_read_all_stats;
GRANT SELECT ON sample_stat_activity_cnt TO pg_read_all_stats;
GRANT SELECT ON last_stat_activity_count TO pg_read_all_stats;
GRANT SELECT ON session_attr TO pg_read_all_stats;
/* Subsampling functions */
CREATE FUNCTION take_subsample(IN sserver_id integer, IN properties jsonb = NULL)
RETURNS jsonb SET search_path=@extschema@ AS $$
DECLARE
  server_query  text;

  session_rows  integer;
  qres          record;

  s_id          integer;  -- last base sample identifier
  srv_version   integer;

  guc_min_query_dur       interval hour to second;
  guc_min_xact_dur        interval hour to second;
  guc_min_xact_age        integer;
  guc_min_idle_xact_dur   interval hour to second;

BEGIN
  -- Adding dblink extension schema to search_path if it does not already there
  IF (SELECT count(*) = 0 FROM pg_catalog.pg_extension WHERE extname = 'dblink') THEN
    RAISE 'dblink extension must be installed';
  END IF;

  SELECT extnamespace::regnamespace AS dblink_schema INTO STRICT qres FROM pg_catalog.pg_extension WHERE extname = 'dblink';
  IF NOT string_to_array(current_setting('search_path'),',') @> ARRAY[qres.dblink_schema::text] THEN
    EXECUTE 'SET LOCAL search_path TO ' || current_setting('search_path')||','|| qres.dblink_schema;
  END IF;

  IF (properties IS NULL) THEN
    -- Initialization is not done yet
    properties := init_sample(sserver_id);
  END IF; -- empty properties

  -- Skip subsampling if it is disabled
  IF (NOT (properties #>> '{settings,subsample_enabled}')::boolean) THEN
    RETURN properties;
  END IF;

  -- Only one running take_subsample() function allowed per server!
  BEGIN
    PERFORM
    FROM server_subsample
    WHERE server_id = sserver_id
    FOR UPDATE NOWAIT;
  EXCEPTION
    WHEN OTHERS THEN
      RAISE 'Can''t get lock on server. Is there another '
        'take_subsample() function running on this server?';
  END;

  s_id := (properties #>> '{properties,last_sample_id}')::integer;

  SELECT reset_val::integer INTO STRICT srv_version FROM jsonb_to_recordset(properties #> '{settings}')
    AS x(name text, reset_val text)
  WHERE name = 'server_version_num';

  -- Current session states collection
  PERFORM dblink_exec('server_connection', 'BEGIN TRANSACTION READ ONLY');
  -- collect sessions and their states
  CASE
    WHEN srv_version >= 140000 THEN
      server_query :=
        'SELECT subsample_ts, datid, datname, pid, leader_pid, usesysid,'
          'usename, application_name, client_addr, client_hostname,'
          'client_port, backend_start, xact_start, query_start, state_change,'
          'state, backend_xid, backend_xmin, query_id, query, backend_type,'
          'backend_xmin_age '
        'FROM ('
            'SELECT '
              'clock_timestamp() as subsample_ts,'
              'datid,'
              'datname,'
              'pid,'
              'leader_pid,'
              'usesysid,'
              'usename,'
              'application_name,'
              'client_addr,'
              'client_hostname,'
              'client_port,'
              'backend_start,'
              'xact_start,'
              'query_start,'
              'state_change,'
              'state,'
              'backend_xid,'
              'backend_xmin,'
              'query_id,'
              'query,'
              'backend_type,'
              'age(backend_xmin) as backend_xmin_age,'
              'CASE '
                'WHEN xact_start <= now() - %1$s::interval THEN '
                  'row_number() OVER (ORDER BY xact_start ASC) '
                'ELSE NULL '
              'END as xact_ord, '
              'CASE '
                'WHEN query_start <= now() - %3$s::interval '
                  'AND state IN (''active'',''fastpath function call'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''active'',''fastpath function call'') ORDER BY query_start ASC) '
                'ELSE NULL '
              'END as query_ord, '
              'CASE '
                'WHEN state_change <= now() - %4$s::interval '
                'AND state IN (''idle in transaction'',''idle in transaction (aborted)'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''idle in transaction'',''idle in transaction (aborted)'') ORDER BY state_change ASC) '
                'ELSE NULL '
              'END as state_ord, '
              'CASE '
                'WHEN age(backend_xmin) >= %2$s THEN '
                  'row_number() OVER (ORDER BY age(backend_xmin) DESC) '
                'ELSE NULL '
              'END  as age_ord '
            'FROM pg_stat_activity '
            'WHERE state NOT IN (''idle'',''disabled'') '
          ') stat_activity '
        'WHERE least('
          'xact_ord,'
          'query_ord,'
          'state_ord,'
          'age_ord'
        ') <= %5$s';
    WHEN srv_version >= 130000 THEN
      server_query :=
        'SELECT subsample_ts, datid, datname, pid, leader_pid, usesysid,'
          'usename, application_name, client_addr, client_hostname,'
          'client_port, backend_start, xact_start, query_start, state_change,'
          'state, backend_xid, backend_xmin, query_id, query, backend_type,'
          'backend_xmin_age '
        'FROM ('
            'SELECT '
              'clock_timestamp() as subsample_ts,'
              'datid,'
              'datname,'
              'pid,'
              'leader_pid,'
              'usesysid,'
              'usename,'
              'application_name,'
              'client_addr,'
              'client_hostname,'
              'client_port,'
              'backend_start,'
              'xact_start,'
              'query_start,'
              'state_change,'
              'state,'
              'backend_xid,'
              'backend_xmin,'
              'NULL AS query_id,'
              'query,'
              'backend_type,'
              'age(backend_xmin) as backend_xmin_age,'
              'CASE '
                'WHEN xact_start <= now() - %1$s::interval THEN '
                  'row_number() OVER (ORDER BY xact_start ASC) '
                'ELSE NULL '
              'END as xact_ord, '
              'CASE '
                'WHEN query_start <= now() - %3$s::interval '
                  'AND state IN (''active'',''fastpath function call'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''active'',''fastpath function call'') ORDER BY query_start ASC) '
                'ELSE NULL '
              'END as query_ord, '
              'CASE '
                'WHEN state_change <= now() - %4$s::interval '
                'AND state IN (''idle in transaction'',''idle in transaction (aborted)'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''idle in transaction'',''idle in transaction (aborted)'') ORDER BY state_change ASC) '
                'ELSE NULL '
              'END as state_ord, '
              'CASE '
                'WHEN age(backend_xmin) >= %2$s THEN '
                  'row_number() OVER (ORDER BY age(backend_xmin) DESC) '
                'ELSE NULL '
              'END  as age_ord '
            'FROM pg_stat_activity '
            'WHERE state NOT IN (''idle'',''disabled'')'
          ') stat_activity '
        'WHERE least('
          'xact_ord,'
          'query_ord,'
          'state_ord,'
          'age_ord'
        ') <= %5$s';
    WHEN srv_version >= 100000 THEN
      server_query :=
        'SELECT subsample_ts, datid, datname, pid, leader_pid, usesysid,'
          'usename, application_name, client_addr, client_hostname,'
          'client_port, backend_start, xact_start, query_start, state_change,'
          'state, backend_xid, backend_xmin, query_id, query, backend_type,'
          'backend_xmin_age '
        'FROM ('
            'SELECT '
              'clock_timestamp() as subsample_ts,'
              'datid,'
              'datname,'
              'pid,'
              'NULL AS leader_pid,'
              'usesysid,'
              'usename,'
              'application_name,'
              'client_addr,'
              'client_hostname,'
              'client_port,'
              'backend_start,'
              'xact_start,'
              'query_start,'
              'state_change,'
              'state,'
              'backend_xid,'
              'backend_xmin,'
              'NULL AS query_id,'
              'query,'
              'backend_type,'
              'age(backend_xmin) as backend_xmin_age,'
              'CASE '
                'WHEN xact_start <= now() - %1$s::interval THEN '
                  'row_number() OVER (ORDER BY xact_start ASC) '
                'ELSE NULL '
              'END as xact_ord, '
              'CASE '
                'WHEN query_start <= now() - %3$s::interval '
                  'AND state IN (''active'',''fastpath function call'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''active'',''fastpath function call'') ORDER BY query_start ASC) '
                'ELSE NULL '
              'END as query_ord, '
              'CASE '
                'WHEN state_change <= now() - %4$s::interval '
                'AND state IN (''idle in transaction'',''idle in transaction (aborted)'') THEN '
                  'row_number() OVER (PARTITION BY state IN (''idle in transaction'',''idle in transaction (aborted)'') ORDER BY state_change ASC) '
                'ELSE NULL '
              'END as state_ord, '
              'CASE '
                'WHEN age(backend_xmin) >= %2$s THEN '
                  'row_number() OVER (ORDER BY age(backend_xmin) DESC) '
                'ELSE NULL '
              'END  as age_ord '
            'FROM pg_stat_activity '
            'WHERE state NOT IN (''idle'',''disabled'')'
          ') stat_activity '
        'WHERE least('
          'xact_ord,'
          'query_ord,'
          'state_ord,'
          'age_ord'
        ') <= %5$s';
    ELSE
      RAISE 'Unsupported postgres version';
  END CASE;

  server_query := format(
      server_query,
      COALESCE(''''||(properties #>> '{properties,min_xact_dur}')||'''', 'NULL'),
      COALESCE(''''||(properties #>> '{properties,min_xact_age}')||'''', 'NULL'),
      COALESCE(''''||(properties #>> '{properties,min_query_dur}')||'''', 'NULL'),
      COALESCE(''''||(properties #>> '{properties,min_idle_xact_dur}')||'''', 'NULL'),
      properties #>> '{properties,topn}'
  );

  -- Save the current state of captured sessions satisfying thresholds
  INSERT INTO last_stat_activity
  SELECT
      sserver_id,
      s_id,
      dbl.subsample_ts,
      dbl.datid,
      dbl.datname,
      dbl.pid,
      dbl.leader_pid,
      dbl.usesysid,
      dbl.usename,
      dbl.application_name,
      dbl.client_addr,
      dbl.client_hostname,
      dbl.client_port,
      dbl.backend_start,
      dbl.xact_start,
      dbl.query_start,
      dbl.state_change,
      dbl.state,
      dbl.backend_xid,
      dbl.backend_xmin,
      dbl.query_id,
      dbl.query,
      dbl.backend_type,
      dbl.backend_xmin_age
  FROM
      dblink('server_connection', server_query) AS dbl(
          subsample_ts      timestamp with time zone,
          datid             oid,
          datname           name,
          pid               integer,
          leader_pid        integer,
          usesysid          oid,
          usename           name,
          application_name  text,
          client_addr       inet,
          client_hostname   text,
          client_port       integer,
          backend_start     timestamp with time zone,
          xact_start        timestamp with time zone,
          query_start       timestamp with time zone,
          state_change      timestamp with time zone,
          state             text,
          backend_xid       xid,
          backend_xmin      xid,
          query_id          bigint,
          query             text,
          backend_type      text,
          backend_xmin_age      bigint
      );
  GET DIAGNOSTICS session_rows = ROW_COUNT;

  IF session_rows > 0 THEN
    /*
      We should also old states
      We have four thresholds probably defined for subsamples, so
      we'll delete the previous captured state when we don't need it
      anymore for all of them.
    */
    DELETE FROM last_stat_activity dlsa
    USING
      (
        SELECT pid, xact_start, max(subsample_ts) as subsample_ts
        FROM last_stat_activity
        WHERE (server_id, sample_id) = (sserver_id, s_id)
        GROUP BY pid, xact_start
      ) last_xact_state
      JOIN
      last_stat_activity lxs ON
        (sserver_id, s_id, last_xact_state.pid, last_xact_state.subsample_ts) =
        (lxs.server_id, lxs.sample_id, lxs.pid, lxs.subsample_ts)
    WHERE (dlsa.server_id, dlsa.sample_id, dlsa.pid, dlsa.xact_start) =
      (sserver_id, s_id, last_xact_state.pid, last_xact_state.xact_start)
      AND dlsa.subsample_ts < last_xact_state.subsample_ts
      AND
      /*
        As we are observing the same xact here (pid, xact_start)
        min_xact_dur threshold can't apply any limitation on deleting
        the old entry.
      */
      -- Can we delete dlsa state due to min_xact_age threshold?
        ((dlsa.backend_xmin IS NOT DISTINCT FROM lxs.backend_xmin)
        OR coalesce(
          dlsa.backend_xmin_age < (properties #>> '{properties,min_xact_age}')::integer,
          true)
        )
      AND
      -- Can we delete dlsa state due to min_query_dur threshold?
      CASE
        WHEN dlsa.state IN ('active', 'fastpath function call') THEN
        (
          ((dlsa.query_start, dlsa.state)
            IS NOT DISTINCT FROM
            (lxs.query_start, lxs.state)
          )
          OR coalesce(
            dlsa.subsample_ts - dlsa.query_start <
              (properties #>> '{properties,min_query_dur}')::interval,
            true)
        )
        ELSE true
      END
      AND
      -- Can we delete dlsa state due to min_idle_xact_dur threshold?
      CASE
        WHEN dlsa.state IN ('idle in transaction','idle in transaction (aborted)') THEN
        (
          ((dlsa.state_change)
            IS NOT DISTINCT FROM
            (lxs.state_change)
          )
          OR coalesce(
            dlsa.subsample_ts - dlsa.state_change <
              (properties #>> '{properties,min_idle_xact_dur}')::interval,
            true)
        )
        ELSE true
      END
    ;
  END IF;

  GET DIAGNOSTICS session_rows = ROW_COUNT;
  EXECUTE format('ANALYZE last_stat_activity_srv%1$s',
    sserver_id);

  -- Collect sessions count by states and waits
  server_query :=
    'SELECT '
      'now() as subsample_ts,'
      'backend_type,'
      'datid,'
      'datname,'
      'usesysid,'
      'usename,'
      'application_name,'
      'client_addr,'
      'count(*) as total,'
      'count(*) FILTER (WHERE state = ''active'') as active,'
      'count(*) FILTER (WHERE state = ''idle'') as idle,'
      'count(*) FILTER (WHERE state = ''idle in transaction'') as idle_t,'
      'count(*) FILTER (WHERE state = ''idle in transaction (aborted)'') as idle_ta,'
      'count(*) FILTER (WHERE state IS NULL) as state_null,'
      'count(*) FILTER (WHERE wait_event_type = ''LWLock'') as lwlock,'
      'count(*) FILTER (WHERE wait_event_type = ''Lock'') as lock,'
      'count(*) FILTER (WHERE wait_event_type = ''BufferPin'') as bufferpin,'
      'count(*) FILTER (WHERE wait_event_type = ''Activity'') as activity,'
      'count(*) FILTER (WHERE wait_event_type = ''Extension'') as extension,'
      'count(*) FILTER (WHERE wait_event_type = ''Client'') as client,'
      'count(*) FILTER (WHERE wait_event_type = ''IPC'') as ipc,'
      'count(*) FILTER (WHERE wait_event_type = ''Timeout'') as timeout,'
      'count(*) FILTER (WHERE wait_event_type = ''IO'') as io '
    'FROM pg_stat_activity '
    'GROUP BY backend_type, datid, datname, usesysid, usename, application_name, client_addr';

  -- Save the current state of captured sessions satisfying thresholds
  INSERT INTO last_stat_activity_count
  SELECT
      sserver_id,
      s_id,
      dbl.subsample_ts,
      dbl.backend_type,
      dbl.datid,
      dbl.datname,
      dbl.usesysid,
      dbl.usename,
      dbl.application_name,
      dbl.client_addr,

      dbl.total,
      dbl.active,
      dbl.idle,
      dbl.idle_t,
      dbl.idle_ta,
      dbl.state_null,
      dbl.lwlock,
      dbl.lock,
      dbl.bufferpin,
      dbl.activity,
      dbl.extension,
      dbl.client,
      dbl.ipc,
      dbl.timeout,
      dbl.io
  FROM
      dblink('server_connection', server_query) AS dbl(
          subsample_ts      timestamp with time zone,
          backend_type      text,
          datid             oid,
          datname           name,
          usesysid          oid,
          usename           name,
          application_name  text,
          client_addr       inet,

          total             integer,
          active            integer,
          idle              integer,
          idle_t            integer,
          idle_ta           integer,
          state_null        integer,
          lwlock            integer,
          lock              integer,
          bufferpin         integer,
          activity          integer,
          extension         integer,
          client            integer,
          ipc               integer,
          timeout           integer,
          io                integer
      );

  PERFORM dblink_exec('server_connection', 'COMMIT');

  RETURN properties;
END
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION take_subsample(IN integer, IN jsonb) IS
  'Take a sub-sample for a server by server_id';

CREATE FUNCTION take_subsample(IN server name) RETURNS integer SET search_path=@extschema@ AS $$
DECLARE
    sserver_id    integer;
BEGIN
    SELECT server_id INTO sserver_id FROM servers WHERE server_name = server;
    IF sserver_id IS NULL THEN
        RAISE 'Server not found';
    ELSE
        PERFORM take_subsample(sserver_id);
        RETURN 0;
    END IF;
END;
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION take_subsample(IN name) IS
  'Statistics sub-sample taking function (by server name)';

CREATE FUNCTION take_subsample_subset(IN sets_cnt integer = 1, IN current_set integer = 0)
RETURNS TABLE (
    server      name,
    result      text,
    elapsed     interval day to second (2)
)
SET search_path=@extschema@ AS $$
DECLARE
    c_servers CURSOR FOR
      SELECT server_id,server_name FROM (
        SELECT server_id,server_name, row_number() OVER () AS srv_rn
        FROM servers JOIN server_subsample USING (server_id)
        WHERE enabled AND subsample_enabled
        ) AS t1
      WHERE srv_rn % sets_cnt = current_set;
    etext               text := '';
    edetail             text := '';
    econtext            text := '';

    qres          RECORD;
    start_clock   timestamp (2) with time zone;
BEGIN
    IF sets_cnt IS NULL OR sets_cnt < 1 THEN
      RAISE 'sets_cnt value is invalid. Must be positive';
    END IF;
    IF current_set IS NULL OR current_set < 0 OR current_set > sets_cnt - 1 THEN
      RAISE 'current_cnt value is invalid. Must be between 0 and sets_cnt - 1';
    END IF;
    FOR qres IN c_servers LOOP
        BEGIN
            start_clock := clock_timestamp()::timestamp (2) with time zone;
            server := qres.server_name;
            PERFORM take_subsample(qres.server_id);
            elapsed := clock_timestamp()::timestamp (2) with time zone - start_clock;
            result := 'OK';
            RETURN NEXT;
        EXCEPTION
            WHEN OTHERS THEN
                BEGIN
                    GET STACKED DIAGNOSTICS etext = MESSAGE_TEXT,
                        edetail = PG_EXCEPTION_DETAIL,
                        econtext = PG_EXCEPTION_CONTEXT;
                    result := format (E'%s\n%s\n%s', etext, econtext, edetail);
                    elapsed := clock_timestamp()::timestamp (2) with time zone - start_clock;
                    RETURN NEXT;
                END;
        END;
    END LOOP;
    RETURN;
END;
$$ LANGUAGE plpgsql;

COMMENT ON FUNCTION take_subsample_subset(IN sets_cnt integer, IN current_set integer) IS
  'Statistics sub-sample taking function (for subset of enabled servers). Used for simplification of parallel sub-sample collection.';

CREATE FUNCTION take_subsample() RETURNS TABLE (
    server      name,
    result      text,
    elapsed     interval day to second (2)
)
SET search_path=@extschema@ AS $$
  SELECT * FROM take_subsample_subset(1,0);
$$ LANGUAGE sql;

COMMENT ON FUNCTION take_subsample() IS 'Subsample taking function (for all enabled servers).';

CREATE FUNCTION collect_subsamples(IN sserver_id integer, IN s_id integer, IN properties jsonb = NULL)
RETURNS jsonb SET search_path=@extschema@ AS $$
DECLARE
    qres  RECORD;

    c_statements CURSOR FOR
    WITH last_stmt_state AS (
      SELECT server_id, sample_id, pid, query_start, max(subsample_ts) as subsample_ts
      FROM last_stat_activity
      WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
      GROUP BY server_id, sample_id, pid, query_start
    )
    SELECT
      pid,
      leader_pid,
      state_change,
      query_start,
      query_id,
      query,
      subsample_ts
    FROM
      last_stat_activity
      JOIN last_stmt_state USING (server_id, sample_id, pid, query_start, subsample_ts)
    WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
    ;
BEGIN
    IF (SELECT NOT subsample_enabled FROM server_subsample WHERE server_id = sserver_id) THEN
      RETURN properties;
    END IF;

    INSERT INTO sample_act_backend (
      server_id,
      sample_id,
      pid,
      backend_start,
      datid,
      datname,
      usesysid,
      usename,
      client_addr,
      client_hostname,
      client_port,
      backend_type,
      backend_last_ts
    )
    WITH last_backend_state AS (
      SELECT server_id, sample_id, pid, backend_start, max(subsample_ts) as subsample_ts
      FROM last_stat_activity
      WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
      GROUP BY server_id, sample_id, pid, backend_start
    )
    SELECT
      server_id,
      s_id as sample_id,
      pid,
      backend_start,
      datid,
      datname,
      usesysid,
      usename,
      client_addr,
      client_hostname,
      client_port,
      backend_type,
      subsample_ts
    FROM
      last_stat_activity
      JOIN last_backend_state
        USING (server_id, sample_id, pid, backend_start, subsample_ts)
    WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
    ;

    INSERT INTO sample_act_xact (
      server_id,
      sample_id,
      pid,
      backend_start,
      xact_start,
      backend_xid,
      xact_last_ts
    )
    WITH last_xact_state AS (
      SELECT server_id, sample_id, pid, xact_start, max(subsample_ts) as subsample_ts
      FROM last_stat_activity
      WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
      GROUP BY server_id, sample_id, pid, xact_start
    )
    SELECT
      server_id,
      s_id AS sample_id,
      pid,
      backend_start,
      xact_start,
      backend_xid,
      subsample_ts AS xact_last_ts
    FROM
      last_stat_activity
      JOIN last_xact_state
        USING (server_id, sample_id, pid, xact_start, subsample_ts)
    WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
    ;

    INSERT INTO sample_act_backend_state (
      server_id,
      sample_id,
      pid,
      backend_start,
      application_name,
      state_code,
      state_change,
      state_last_ts,
      xact_start,
      backend_xmin,
      backend_xmin_age
    )
    WITH last_backend_state AS (
      SELECT server_id, sample_id, pid, state_change, max(subsample_ts) as subsample_ts
      FROM last_stat_activity
      WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
        AND state IN ('idle in transaction', 'idle in transaction (aborted)', 'active')
      GROUP BY server_id, sample_id, pid, state_change
    )
    SELECT
      server_id,
      s_id AS sample_id,
      pid,
      backend_start,
      application_name,
      CASE state
        WHEN 'idle in transaction' THEN 1
        WHEN 'idle in transaction (aborted)' THEN 2
        WHEN 'active' THEN 3
        ELSE 0
      END state_code,
      state_change,
      subsample_ts AS state_last_ts,
      xact_start,
      backend_xmin,
      backend_xmin_age
    FROM
      last_stat_activity
      JOIN last_backend_state
        USING (server_id, sample_id, pid, state_change, subsample_ts)
    WHERE (server_id, sample_id) = (sserver_id, s_id - 1)
    ;

    FOR qres IN c_statements
    LOOP
      INSERT INTO act_query (server_id, act_query_md5, act_query, last_sample_id)
      VALUES (sserver_id, md5(qres.query), qres.query, NULL)
      ON CONFLICT ON CONSTRAINT pk_act_query
      DO UPDATE SET last_sample_id = NULL;

      INSERT INTO sample_act_statement(
        server_id,
        sample_id,
        pid,
        leader_pid,
        state_change,
        query_start,
        query_id,
        act_query_md5,
        stmt_last_ts
      ) VALUES (
        sserver_id,
        s_id,
        qres.pid,
        qres.leader_pid,
        qres.state_change,
        qres.query_start,
        qres.query_id,
        md5(qres.query),
        qres.subsample_ts
      );
    END LOOP;

    -- Save session counters
    -- Insert new values of session attributes
    INSERT INTO session_attr AS isa (
      server_id,
      backend_type,
      datid,
      datname,
      usesysid,
      usename,
      application_name,
      client_addr
    )
    SELECT DISTINCT
      ls.server_id,
      ls.backend_type,
      ls.datid,
      ls.datname,
      ls.usesysid,
      ls.usename,
      ls.application_name,
      ls.client_addr
    FROM
      last_stat_activity_count ls LEFT JOIN session_attr sa ON
        -- ensure partition pruning
        (sa.server_id = sserver_id) AND
        (ls.server_id, ls.backend_type, ls.datid, ls.datname, ls.usesysid,
          ls.usename, ls.application_name, ls.client_addr)
        IS NOT DISTINCT FROM
        (sa.server_id, sa.backend_type, sa.datid, sa.datname, sa.usesysid,
          sa.usename, sa.application_name, sa.client_addr)
    WHERE
      (ls.server_id, ls.sample_id) = (sserver_id, s_id - 1)
      AND sa.server_id IS NULL
    ;

    INSERT INTO sample_stat_activity_cnt (
      server_id,
      sample_id,
      subsample_ts,

      sess_attr_id,

      total,
      active,
      idle,
      idle_t,
      idle_ta,
      state_null,
      lwlock,
      lock,
      bufferpin,
      activity,
      extension,
      client,
      ipc,
      timeout,
      io
    )
    SELECT
      sserver_id,
      s_id,
      l.subsample_ts,

      s.sess_attr_id,

      l.total,
      l.active,
      l.idle,
      l.idle_t,
      l.idle_ta,
      l.state_null,
      l.lwlock,
      l.lock,
      l.bufferpin,
      l.activity,
      l.extension,
      l.client,
      l.ipc,
      l.timeout,
      l.io
    FROM
      last_stat_activity_count l JOIN session_attr s ON
        (l.server_id, s.server_id, s.backend_type) = (sserver_id, sserver_id, l.backend_type) AND
        (s.datid, s.datname, s.usesysid, s.usename, s.application_name, s.client_addr)
          IS NOT DISTINCT FROM
        (l.datid, l.datname, l.usesysid, l.usename, l.application_name, l.client_addr)
    WHERE l.sample_id = s_id - 1;

    -- Mark disappered and returned entries
    UPDATE session_attr sau
    SET last_sample_id =
      CASE
        WHEN num_nulls(ss.server_id, sa.last_sample_id) = 2 THEN s_id - 1
        WHEN num_nulls(ss.server_id, sa.last_sample_id) = 0 THEN NULL
      END
    FROM sample_stat_activity_cnt ss RIGHT JOIN session_attr sa
      ON (sa.server_id, ss.server_id, ss.sample_id, ss.sess_attr_id) =
        (sserver_id, sserver_id, s_id, sa.sess_attr_id)
    WHERE
      num_nulls(ss.server_id, sa.last_sample_id) IN (0,2) AND
      (sau.server_id, sau.sess_attr_id) = (sserver_id, sa.sess_attr_id);

    UPDATE act_query squ SET last_sample_id = s_id - 1
    FROM sample_act_statement ss RIGHT JOIN act_query sq
      ON (ss.server_id, ss.sample_id, ss.act_query_md5) =
        (sserver_id, s_id, sq.act_query_md5)
    WHERE
      sq.server_id = sserver_id AND
      sq.last_sample_id IS NULL AND
      ss.server_id IS NULL AND
      (squ.server_id, squ.act_query_md5) = (sserver_id, sq.act_query_md5);

    EXECUTE format('TRUNCATE TABLE last_stat_activity_srv%1$s',
      sserver_id);
    EXECUTE format('TRUNCATE TABLE last_stat_activity_count_srv%1$s',
      sserver_id);

    RETURN properties;
END;
$$ LANGUAGE plpgsql;
CREATE FUNCTION stat_activity_states(IN sserver_id integer, IN start_id integer, IN end_id integer)
RETURNS TABLE(
    server_id         integer,
    datid             oid,
    datname           name,
    leader_pid        integer,
    pid               integer,
    usesysid          oid,
    usename           name,
    application_name  text,
    client_addr       inet,
    client_hostname   text,
    client_port       integer,
    backend_start     timestamp with time zone,
    backend_last_ts   timestamp with time zone,
    xact_start        timestamp with time zone,
    xact_last_ts      timestamp with time zone,
    state_code        integer,
    state             text,
    state_change      timestamp with time zone,
    state_last_ts     timestamp with time zone,
    backend_xid       xid,
    backend_xmin      xid,
    backend_xmin_age  bigint,
    backend_type      text,
    query_start       timestamp with time zone,
    stmt_last_ts      timestamp with time zone,
    query_id          bigint,
    act_query_md5     char(32),
    backend_duration  interval,
    xact_duration     interval,
    state_duration    interval,
    query_duration    interval
) SET search_path=@extschema@ AS $$
  SELECT
    server_id,
    datid,
    datname,
    leader_pid,
    pid,
    usesysid,
    usename,
    application_name,
    client_addr,
    client_hostname,
    client_port,
    backend_start,
    backend_last_ts,
    xact_start,
    xact_last_ts,
    state_code,
    CASE state_code
      WHEN 1 THEN 'idle in transaction'
      WHEN 2 THEN 'idle in transaction (aborted)'
      WHEN 3 THEN 'active'
      ELSE NULL
    END state,
    state_change,
    state_last_ts,
    backend_xid,
    backend_xmin,
    backend_xmin_age,
    backend_type,
    query_start,
    stmt_last_ts,
    query_id,
    act_query_md5,
    backend_last_ts - backend_start as backend_duration,
    xact_last_ts - xact_start as xact_duration,
    state_last_ts - state_change as state_duration,
    CASE
      WHEN state_code = 3 THEN stmt_last_ts - query_start
      ELSE state_change - query_start
    END as query_duration
  FROM
    (sample_act_backend b
    JOIN (
        SELECT server_id, pid, backend_start, max(backend_last_ts) as backend_last_ts
        FROM sample_act_backend
        WHERE server_id = sserver_id
          AND sample_id BETWEEN start_id + 1 AND end_id
        GROUP BY server_id, pid, backend_start
      ) as act_backend USING (server_id, pid, backend_start, backend_last_ts))
    JOIN
    (sample_act_backend_state s JOIN (
        SELECT server_id, pid, state_change, max(state_last_ts) as state_last_ts
        FROM sample_act_backend_state
        WHERE server_id = sserver_id
          AND sample_id BETWEEN start_id + 1 AND end_id
        GROUP BY server_id, pid, state_change
      ) as backend_state
      USING (server_id, pid, state_change, state_last_ts))
    USING (server_id, pid, backend_start)
    LEFT JOIN
    (sample_act_statement q JOIN (
        SELECT server_id, pid, query_start, max(stmt_last_ts) as stmt_last_ts
        FROM sample_act_statement
        WHERE server_id = sserver_id
          AND sample_id BETWEEN start_id + 1 AND end_id
        GROUP BY server_id, pid, query_start
      ) as act_stmt
      USING (server_id, pid, query_start, stmt_last_ts))
    USING (server_id, pid, state_change)
    LEFT JOIN
    (sample_act_xact x JOIN (
        SELECT server_id, pid, xact_start, max(xact_last_ts) as xact_last_ts
        FROM sample_act_xact
        WHERE server_id = sserver_id
          AND sample_id BETWEEN start_id + 1 AND end_id
        GROUP BY server_id, pid, xact_start
      ) as act_xact USING (server_id, pid, xact_start, xact_last_ts))
    USING (server_id, pid, backend_start, xact_start)
  WHERE
    server_id = sserver_id
$$ LANGUAGE sql;

CREATE FUNCTION stat_activity_agg(IN sserver_id integer, IN start_id integer, IN end_id integer)
RETURNS TABLE(
    server_id         integer,
    datid             oid,
    datname           name,
    idle_xact_dur     interval,
    idle_xact_a_dur   interval,
    active_dur        interval,
    max_idle_xact_d   interval,
    max_idle_xact_a_d interval,
    max_active_dur    interval,
    max_xact_age      bigint
) SET search_path=@extschema@ AS $$
  SELECT
    server_id,
    datid,
    datname,
    sum(state_last_ts - state_change) FILTER (WHERE state = 'idle in transaction') as idle_xact_dur,
    sum(state_last_ts - state_change) FILTER (WHERE state = 'idle in transaction (aborted)') as idle_xact_a_dur,
    sum(state_last_ts - state_change) FILTER (WHERE state = 'active') as active_dur,
    max(state_last_ts - state_change) FILTER (WHERE state = 'idle in transaction') as max_idle_xact_d,
    max(state_last_ts - state_change) FILTER (WHERE state = 'idle in transaction (aborted)') as max_idle_xact_a_d,
    max(state_last_ts - state_change) FILTER (WHERE state = 'active') as max_active_dur,
    max(backend_xmin_age) as max_xact_age
  FROM
    stat_activity_states(sserver_id, start_id, end_id)
  GROUP BY server_id, datid, datname
$$ LANGUAGE sql;

CREATE FUNCTION stat_activity_agg_format(IN sserver_id integer, IN start_id integer, IN end_id integer)
RETURNS TABLE(
    datid             oid,
    dbname            name,
    idle_xact_dur     text,
    idle_xact_a_dur   text,
    active_dur        text,
    max_idle_xact_d   text,
    max_idle_xact_a_d text,
    max_active_dur    text,
    max_xact_age      bigint,
    ord_db            integer
) SET search_path=@extschema@ AS $$
    SELECT
        st.datid,
        COALESCE(st.datname, 'Total') as dbname,
        NULLIF(date_trunc('second', sum(st.idle_xact_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_dur,
        NULLIF(date_trunc('second', sum(st.idle_xact_a_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_a_dur,
        NULLIF(date_trunc('second', sum(st.active_dur) + '0.5 second'),
          '0 second'::interval)::text as active_dur,
        NULLIF(date_trunc('second', max(st.max_idle_xact_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_d,
        NULLIF(date_trunc('second', max(st.max_idle_xact_a_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_a_d,
        NULLIF(date_trunc('second', max(st.max_active_dur) + '0.5 second'),
          '0 second'::interval)::text as max_active_dur,
        NULLIF(max(st.max_xact_age), 0) as max_xact_age,
        -- ordering fields
        row_number() OVER (ORDER BY st.datname NULLS LAST)::integer AS ord_db
    FROM
        stat_activity_agg(sserver_id, start_id, end_id) st
    GROUP BY GROUPING SETS ((st.datid, st.datname), ())
$$ LANGUAGE sql;

CREATE FUNCTION stat_activity_agg_format(IN sserver_id integer, IN start1_id integer, IN end1_id integer,
  IN start2_id integer, IN end2_id integer)
RETURNS TABLE(
    datid               oid,
    dbname              name,
    idle_xact_dur1      text,
    idle_xact_a_dur1    text,
    active_dur1         text,
    max_idle_xact_d1    text,
    max_idle_xact_a_d1  text,
    max_active_dur1     text,
    max_xact_age1       bigint,
    idle_xact_dur2      text,
    idle_xact_a_dur2    text,
    active_dur2         text,
    max_idle_xact_d2    text,
    max_idle_xact_a_d2  text,
    max_active_dur2     text,
    max_xact_age2       bigint,
    ord_db              integer
) SET search_path=@extschema@ AS $$
    SELECT
        COALESCE(st1.datid,st2.datid) AS datid,
        COALESCE(COALESCE(st1.datname,st2.datname),'Total') AS dbname,
        NULLIF(date_trunc('second', sum(st1.idle_xact_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_dur1,
        NULLIF(date_trunc('second', sum(st1.idle_xact_a_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_a_dur1,
        NULLIF(date_trunc('second', sum(st1.active_dur) + '0.5 second'),
          '0 second'::interval)::text as active_dur1,
        NULLIF(date_trunc('second', max(st1.max_idle_xact_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_d1,
        NULLIF(date_trunc('second', max(st1.max_idle_xact_a_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_a_d1,
        NULLIF(date_trunc('second', max(st1.active_dur) + '0.5 second'),
          '0 second'::interval)::text as max_active_dur1,
        NULLIF(max(st1.max_xact_age), 0) as max_xact_age1,
        NULLIF(date_trunc('second', sum(st2.idle_xact_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_dur2,
        NULLIF(date_trunc('second', sum(st2.idle_xact_a_dur) + '0.5 second'),
          '0 second'::interval)::text as idle_xact_a_dur2,
        NULLIF(date_trunc('second', sum(st2.active_dur) + '0.5 second'),
          '0 second'::interval)::text as active_dur2,
        NULLIF(date_trunc('second', max(st2.max_idle_xact_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_d2,
        NULLIF(date_trunc('second', max(st2.max_idle_xact_a_d) + '0.5 second'),
          '0 second'::interval)::text as max_idle_xact_a_d2,
        NULLIF(date_trunc('second', max(st2.active_dur) + '0.5 second'),
          '0 second'::interval)::text as max_active_dur2,
        NULLIF(max(st2.max_xact_age), 0) as max_xact_age2,
        -- ordering fields
        row_number() OVER (ORDER BY COALESCE(st1.datname,st2.datname) NULLS LAST)::integer AS ord_db
    FROM stat_activity_agg(sserver_id, start1_id, end1_id) st1
      FULL OUTER JOIN stat_activity_agg(sserver_id, start2_id, end2_id) st2
        USING (server_id, datid)
    GROUP BY GROUPING SETS ((COALESCE(st1.datid,st2.datid), COALESCE(st1.datname,st2.datname)),
      ())
$$ LANGUAGE sql;

CREATE FUNCTION stat_activity_states_format(IN sserver_id integer,
  IN start_id integer, IN end_id integer)
RETURNS TABLE(
    datid             oid,
    dbname            name,
    pid               integer,
    userid            oid,
    username          name,
    client_addr       inet,
    client_hostname   text,
    client_port       integer,
    application_name  text,

    backend_start     timestamp with time zone,
    xact_start        timestamp with time zone,
    state_change      timestamp with time zone,

    backend_start_format  timestamp with time zone,
    xact_start_format     timestamp with time zone,
    state_change_format   timestamp with time zone,

    flt_state_code        integer,
    state_code            integer,
    state                 text,
    state_duration_format interval,
    xact_duration_format  interval,
    xmin_age              bigint,

    query_id              bigint,
    act_query_md5         char(32),

    ord_dur               integer,
    ord_age               integer,
    ord_xact              integer,
    flt_age               boolean
) SET search_path=@extschema@ AS $$
    SELECT
      datid,
      datname,
      pid,
      usesysid,
      usename,
      client_addr,
      client_hostname,
      client_port,
      application_name,

      backend_start AS backend_start,
      xact_start AS xact_start,
      state_change AS state_change,

      date_trunc('second',backend_start) AS backend_start_format,
      date_trunc('second',xact_start) AS xact_start_format,
      date_trunc('second',state_change) AS state_change_format,

      CASE
        WHEN state_code IN (1, 2, 3) AND act_query_md5 IS NOT NULL THEN state_code
        WHEN state_code NOT IN (1, 2, 3) THEN state_code
        ELSE NULL
      END AS flt_state_code,
      state_code,
      state,
      date_trunc('second',state_duration) AS state_duration_format,
      date_trunc('second',xact_duration) AS xact_duration_format,
      backend_xmin_age,

      query_id,
      act_query_md5,

      row_number() OVER (PARTITION BY state_code ORDER BY state_duration DESC)::integer AS ord_dur,
      row_number() OVER (ORDER BY backend_xmin_age DESC)::integer AS ord_age,
      row_number() OVER (ORDER BY xact_duration DESC)::integer AS ord_xact,
      coalesce(xact_last_ts = state_last_ts, false) AND backend_xmin_age > 0 as flt_age
    FROM
      stat_activity_states(sserver_id, start_id, end_id)
$$ LANGUAGE sql;

CREATE FUNCTION stat_activity_states_format(IN sserver_id integer,
  IN start1_id integer, IN end1_id integer,
  IN start2_id integer, IN end2_id integer)
RETURNS TABLE(
    datid             oid,
    dbname            name,
    pid               integer,
    userid            oid,
    username          name,
    client_addr       inet,
    client_hostname   text,
    client_port       integer,
    application_name  text,

    backend_start     timestamp with time zone,
    xact_start        timestamp with time zone,
    state_change      timestamp with time zone,

    backend_start_format  timestamp with time zone,
    xact_start_format     timestamp with time zone,
    state_change_format   timestamp with time zone,

    flt_state_code        integer,
    state_code            integer,
    state                 text,
    state_duration_format interval,
    xact_duration_format  interval,
    xmin_age              bigint,

    query_id              bigint,
    act_query_md5         char(32),

    ord_dur               integer,
    ord_age               integer,
    ord_xact              integer,
    flt_age               boolean
) SET search_path=@extschema@ AS $$
    SELECT
      datid,
      datname,
      pid,
      usesysid,
      usename,
      client_addr,
      client_hostname,
      client_port,
      application_name,

      backend_start AS backend_start,
      xact_start AS xact_start,
      state_change AS state_change,

      date_trunc('second',backend_start) AS backend_start_format,
      date_trunc('second',xact_start) AS xact_start_format,
      date_trunc('second',state_change) AS state_change_format,

      CASE
        WHEN state_code IN (1, 2, 3) AND act_query_md5 IS NOT NULL THEN state_code
        WHEN state_code NOT IN (1, 2, 3) THEN state_code
        ELSE NULL
      END AS flt_state_code,
      state_code,
      state,
      date_trunc('second',state_duration) AS state_duration_format,
      date_trunc('second',xact_duration) AS xact_duration_format,
      backend_xmin_age,

      query_id,
      act_query_md5,

      row_number() OVER (PARTITION BY state_code ORDER BY state_duration DESC)::integer AS ord_dur,
      row_number() OVER (ORDER BY backend_xmin_age DESC)::integer AS ord_age,
      row_number() OVER (ORDER BY xact_duration DESC)::integer AS ord_xact,
      coalesce(xact_last_ts = state_last_ts, false) AND backend_xmin_age > 0 as flt_age
    FROM (
      SELECT * FROM stat_activity_states(sserver_id, start1_id, end1_id)
      UNION
      SELECT * FROM stat_activity_states(sserver_id, start2_id, end2_id)
    ) st
$$ LANGUAGE sql;

CREATE FUNCTION report_active_queries_format(IN report_context jsonb, IN sserver_id integer,
  IN act_queries_list jsonb)
RETURNS TABLE (
  act_query_md5   text,
  query_texts     jsonb
)
SET search_path=@extschema@ AS $$
  SELECT
    act_query_md5,
    jsonb_build_array(act_query)
  FROM
    jsonb_to_recordset(act_queries_list) AS ql (
      act_query_md5    text
    )
    JOIN act_query USING (act_query_md5)
  WHERE
    server_id = sserver_id
$$ LANGUAGE sql;
INSERT INTO import_queries_version_order VALUES
('pg_profile','4.6','pg_profile','4.5')
;

DELETE FROM report_struct;
DELETE FROM report;
DELETE FROM report_static;

DO $$
DECLARE
  sserver_id    integer;
BEGIN
  FOR sserver_id IN (SELECT * FROM servers) LOOP
    -- Create last_stat_activity table partition
    EXECUTE format(
      'CREATE TABLE last_stat_activity_srv%1$s PARTITION OF last_stat_activity '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_activity_srv%1$s '
      'ADD CONSTRAINT pk_last_stat_activity_srv%1$s '
        'PRIMARY KEY (server_id, sample_id, pid, subsample_ts), '
      'ADD CONSTRAINT fk_last_stat_activity_sample_srv%1$s '
        'FOREIGN KEY (server_id, sample_id) '
        'REFERENCES samples(server_id, sample_id) ON DELETE RESTRICT',
        sserver_id);

    -- Create last_stat_activity table partition
    EXECUTE format(
      'CREATE TABLE last_stat_activity_count_srv%1$s PARTITION OF last_stat_activity_count '
      'FOR VALUES IN (%1$s)',
      sserver_id);
    EXECUTE format(
      'ALTER TABLE last_stat_activity_count_srv%1$s '
      'ADD CONSTRAINT fk_last_stat_activity_count_sample_srv%1$s '
        'FOREIGN KEY (server_id, sample_id) '
        'REFERENCES samples(server_id, sample_id) ON DELETE RESTRICT',
        sserver_id);
  END LOOP;
END;
$$ LANGUAGE plpgsql;
/* === report_static table data === */
INSERT INTO report_static(static_name, static_text)
VALUES
('css', $css$
/* ----------- Common styles ----------- */
:root {
    --main-bg-color: #68B8F9;
    --main-font-color: black;
    --secondary-bg-color: #a4d4fb;
    --main-bottom: -20%;
    --main-right: 2%;
    --main-opacity: 1;
    --main-border-radius: 5px;
    --main-box-shadow: rgba(0, 0, 0, 0.35) 0 5px 15px;
    --main-position: fixed;
    --main-width: 500px;
    --main-height: 100px;
    --main-transition-property: top;
    --main-transition-delay: 500ms;
}

#container {
    position: absolute;
    top: 40px;
    left: 25%;
    overflow: auto;
    margin-left: 10px;
}

#container .no-menu {
    left: 0;
}
html {
    scroll-behavior: smooth;
    scroll-padding-left: 30%;
    margin-top: 40px;
}

h3 {
    scroll-margin-top: 60px;
}

td {
    scroll-margin-top: 50px;
}

a:hover {
    text-decoration: none;
}

#navigator {
    position: fixed;
    background-color: white;
    box-shadow: rgba(0, 0, 0, 0.35) 0 5px 15px;
    top: 40px;
    left: 0;
    bottom: 0;
    width: 25%;
    overflow: auto;
    display: flex;
    flex-direction: row;
    flex-wrap: nowrap;
}

#navigator.hidden {
    width: 0;
}
#navigator ul.active {
    display: block;
    font-family: Monospace, serif;
    width: fit-content;
    height: fit-content;
}

#navigator ul.hidden {
    display: none;
}

#navigator li.current {
    border: 2px solid black;
    border-radius: 3px;
}
#navigator li.hidden {
    display: none;
}
#navigator li.active {
    background-color: var(--secondary-bg-color);
    border-radius: 3px;
}

select {
    padding: 6px;
    margin-top: 8px;
    margin-left: 4px;
    border: 0;
}

#commonstat {
    display: flex;
    flex-wrap: wrap;
}

#commonstat div {
    margin-right: 10px;
}

table {
    margin-bottom: 10px;
}

table, th, td {
    border: 1px solid black;
    border-collapse: collapse;
    padding: 4px;
}

table tr td.table_obj_value, table tr td.mono {
    font-family: Monospace, serif;
}

table tr td.table_obj_value {
    text-align: right;
}

table tr td.fullScreen {
    width: 85%
}

table tr td.halfScreen {
    width: 42%;
}

table p {
    margin: 0.2em;
}

table tr.new td.switch_bold {
    font-weight: bold;
}

table th {
    color: black;
    background-color: #ffcc99;
}

table tr:target, td:target {
    border: medium solid limegreen;
}

table tr:target td:first-of-type, table td:target {
    font-weight: bold;
}

table tr.active td {
    background-color: #CCF1FF;
}

table tr.active td {
    background-color: #CCF1FF;
}

table tr.active td {
    background-color: #CCF1FF;
}

table tr.active td {
    background-color: #CCF1FF;
}

table tr.active td:not(.hdr) {
    background-color: #CCF1FF;
}

div.warning {
    display: inline-flex;
    padding: 10px;
    max-width: 300px;
    border: 1px solid black;
    border-radius: 5px;
    margin-bottom: 10px;
    background-color: pink;
    font-size: 12px;
    font-family: Monospace, serif;
}

div.notice {
    display: inline-flex;
    padding: 10px;
    max-width: 300px;
    border: 1px solid black;
    border-radius: 5px;
    margin-bottom: 10px;
    background-color: #CCF1FF;
    font-size: 12px;
    font-family: Monospace, serif;
}

div.popup {
    background-color: var(--main-bg-color);
    bottom: var(--main-bottom);
    color: var(--main-font-color);
    right: var(--main-right);
    opacity: var(--main-opacity);
    border-radius: var(--main-border-radius);
    position: var(--main-position);
    max-width: var(--main-width);
    transition: bottom 250ms linear;
    z-index: 500 !important;
    padding: 10px 20px;
}

svg rect:hover, svg circle:hover, svg path:hover, a.copyButton svg:hover > rect {
    stroke: limegreen;
    cursor: pointer;
}

/* Add a black background color to the top navigation bar */
.topnav {
    position: fixed;
    top: 0;
    width: 100%;
    background-color: #e9e9e9;
    z-index: 999;
}

/* Style the links inside the navigation bar */
.topnav a {
    float: left;
    display: block;
    background-color: #2196F3;
    color: black;
    text-align: center;
    padding: 14px 16px;
    text-decoration: none;
    cursor: pointer;
    font-size: 14px;
    font-family: Monospace, serif;
}

/* Change the color of links on hover */
.topnav a:hover {
    background-color: #ddd;
}

/* Style the "active" element to highlight the current page */
.topnav a.active {
    color: white;
}

/* Style the search box inside the navigation bar */
.topnav input[type=search] {
    float: left;
    border: none;
    margin-top: 8px;
    margin-left: 200px;
    font-size: 14px;
    font-family: Monospace, serif;
    -webkit-appearance: searchfield-cancel-button;
}

/* When the screen is less than 600px wide, stack the links and the search field vertically instead of horizontally */
@media screen and (max-width: 600px) {
    .topnav a, .topnav input[type=search] {
        float: none;
        display: block;
        text-align: left;
        width: 100%;
        margin: 0;
        padding: 14px;
    }

    .topnav input[type=search] {
        border: 1px solid #ccc;
    }
}

/* ----------- Differential report styles ----------- */

td.int1, .int1 td:not(.hdr), table tr.new_i1 {
    background-color: #FFEEEE;
}

td.int2, .int2 td:not(.hdr), table tr.new_i2 {
    background-color: #EEEEFF;
}

table tr.int1:not(.active) td:not(.hdr) {
    background-color: #FFEEEE;
}

table tr.int2:not(.active) td:not(.hdr) {
    background-color: #EEEEFF;
}

table tr.int2 td {
    border-top: hidden;
}

table tr.queryRow {
    background-color: white;
}

table tr.grey {
    background-color: #eee;
}
table tr.queryRow {
    background-color: white;
}

table tr.queryRow p, table tr.queryRow a, table tr.planRow p, table tr.planRow a {
    display: inline-block;
}

table tr.queryRow p, table tr.planRow p {
    width: 95%;
}

table tr.queryRow a, table tr.planRow a {
    float: right;
}

table tr td.queryText {
    font-size: 13px;
}

table tr.new_i1 td.switch_bold, table tr.new_i2 td.switch_bold, .new td.switch_bold {
    font-weight: bold;
}

table th {
    color: black;
    background-color: #ffcc99;
}

.label {
    color: grey;
}

table tr:target, td:target {
    border: medium solid limegreen;
}

table tr:target td:first-of-type, table td:target {
    font-weight: bold;
}

table tr.parent td {
    background-color: #D8E8C2;
}

table tr.child td {
    background-color: #BBDD97;
    border-top-style: hidden;
}

table tr.active td {
    background-color: #CCF1FF;
}{static:css_post}
$css$
),
(
 'script_js', $js$
class Utilities {
    static sort(data, key, direction) {
        return structuredClone(data.sort((a, b) => {
            if (a[key] < b[key]) {
                return -1 * direction;
            } else if (a[key] > b[key]) {
                return direction;
            } else {
                return 0;
            }
        }))
    }
    static sum(data, key) {
        return data.reduce((partialSum, a) => partialSum + a[key], 0);
    }
    static filter(data, key) {
        if (key.type === "exists") {
            if (data.every(obj => key["field"] in obj)) {
                return structuredClone(data.filter(obj => obj[key["field"]]));
            }
        } else if (key.type === "equal") {
            if (data.every(obj => key["field"] in obj)) {
                return structuredClone(data.filter(obj => obj[key["field"]] === key["value"]));
            }
        }
        return data;
    }
    static find(data, key, value) {
        return structuredClone(data.filter(obj => obj[key] === value));
    }
    static limit(data, num) {
        if (num > 0) {
            return structuredClone(data.slice(0, num));
        }
        return data;
    }
    static getInputField() {
        return document.getElementById('searchField');
    }
    static cancelSearchResults(rowsForSearch) {
        rowsForSearch.forEach(row => {
            row.style.display = '';
        })
        this.getInputField().value = '';
    }
    static searchQueryWithStatistics(rowsForSearch, keyword) {
        let foundQueries = Utilities.searchQueryText(keyword);
        rowsForSearch.forEach(row => {
            if (row.dataset["hexqueryid"]
                && foundQueries[row.dataset["hexqueryid"]]) {
                row.style.display = '';
            } else {
                row.style.display = 'none';
                if (row.nextSibling && row.nextSibling.classList.contains('queryRow')) {
                    row.nextSibling.style.display = 'none';
                }
            }
        })
    }
    static searchQueryText(keyword) {
        let foundQueries = {};
        data.datasets.queries.forEach(query => {
            Object.keys(query).forEach(key => {
                query.query_texts.forEach(query_text => {
                    if (query_text && query_text.toLowerCase().includes(keyword) && !foundQueries[query["hexqueryid"]]) {
                        foundQueries[query["hexqueryid"]] = true;
                    }
                })
            })
            if (query.plans) {
                query.plans.forEach(plan => {
                    if (plan.plan_text.toLowerCase().includes(keyword) && !foundQueries[query["hexqueryid"]]) {
                        foundQueries[query["hexqueryid"]] = true;
                    }
                })
            }
        })
        return foundQueries;
    }
    static searchWithParam(rowsForSearch, keyword, searchParam) {
        let foundQueries = {};
        if (searchParam === 'all') {
            foundQueries = Utilities.searchQueryText(keyword)
        }
        rowsForSearch.forEach(row => {
            if (row.dataset[searchParam] && row.dataset[searchParam].toLowerCase().includes(keyword)) {
                row.style.display = '';
                if (row.dataset["hexqueryid"]) {
                    foundQueries[row.dataset["hexqueryid"]] = true;
                }
            } else {
                row.style.display = 'none';
                if (row.nextSibling && row.nextSibling.classList.contains('queryRow')) {
                    row.nextSibling.style.display = 'none';
                }
            }
        })
        rowsForSearch.forEach(row => {
            if (row.parentNode.id === 'sqllist_t' || searchParam === 'all') {
                if (foundQueries[row.dataset["hexqueryid"]]) {
                    row.style.display = '';
                }
            } else {
                if (foundQueries[row.dataset["hexqueryid"]] && row.dataset[searchParam].toLowerCase().includes(keyword)) {
                    row.style.display = '';
                }
            }
        })
    }
    static searchParams = {
        'all': 'Everywhere',
        'dbname': 'Database',
        'username': 'User',
        'relname': 'Table',
        'indexrelname': 'Index',
        'querytext': 'Query'
    }
    static search(rowsForSearch, searchParam, keyword) {
        keyword = keyword.toLowerCase();
        if (!keyword) {
            Utilities.cancelSearchResults(rowsForSearch);
        } else if (searchParam === 'querytext') {
            Utilities.searchQueryWithStatistics(rowsForSearch, keyword);
        } else {
            Utilities.searchWithParam(rowsForSearch, keyword, searchParam);
        }
    }
}
class BaseChart {
    static drawIntoTable(cls, newRow, column, data) {
        let key = column.ordering[0] === '-' ? column.ordering.substring(1) : column.ordering;
        let value = column.id;
        let direction = column.ordering[0] === '-' ? -1 : 1;
        let newCell = newRow.insertCell(-1);
        newCell.setAttribute('class', column.class);
        if (Utilities.sum(data, value) > 0) {
            let orderedData = Utilities.sort(data, key, direction);
            let svg = cls.drawSVG(orderedData, value, key);
            newCell.appendChild(svg);
        }
    }
}
class PipeChart extends BaseChart {
    static drawSVG(orderedData, value, key) {
        let x = 0; // Start position of nested svg
        let svg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        svg.setAttribute('width', '100%');
        svg.setAttribute('height', '2em');
        orderedData.forEach(elem => {
            let width = Math.floor(elem[value]);
            let nestedSvg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
            nestedSvg.setAttribute('x', `${x}%`);
            nestedSvg.setAttribute('height', '2em');
            nestedSvg.setAttribute('width', `${width}%`);
            let title = document.createElementNS('http://www.w3.org/2000/svg', 'title');
            title.innerHTML = `${elem.objname}: ${elem[value]}`;
            let text = document.createElementNS('http://www.w3.org/2000/svg', 'text');
            text.setAttribute('y', '70%');
            text.setAttribute('x', '0.3em');
            text.innerHTML = `${elem.objname}: ${elem[key]}`;
            let rect = document.createElementNS('http://www.w3.org/2000/svg', 'rect');
            rect.setAttribute('height', '90%');
            rect.setAttribute('x', '0%');
            rect.setAttribute('y', '10%');
            rect.setAttribute('ry', '15%');
            rect.setAttribute('stroke', 'black');
            rect.setAttribute('stroke-width', '1px');
            rect.setAttribute('width', '100%');
            rect.setAttribute('fill', `#${elem.objcolor}`);
            nestedSvg.appendChild(title);
            nestedSvg.appendChild(rect);
            svg.appendChild(nestedSvg);
            nestedSvg.appendChild(text);
            x += width;
        })
        return svg;
    }
    static drawIntoTable(newRow, column, data) {
        BaseChart.drawIntoTable(PipeChart, newRow, column, data);
        return true;
    }
}
class PieChart extends BaseChart {
    static drawPieSlice(startRad, diffRad, radius, center, elem, key) {
        let nestedSvg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        let title = document.createElementNS('http://www.w3.org/2000/svg', 'title');
        title.innerHTML = `${elem.objname}: ${elem[key]}`;
        if (diffRad >= Math.PI * 1.999) {
            let circle = document.createElementNS('http://www.w3.org/2000/svg', 'circle');
            circle.setAttribute('cx', center[0]);
            circle.setAttribute('cy', center[1]);
            circle.setAttribute('r', radius);
            circle.setAttribute('fill', `#${elem.objcolor}`);
            circle.setAttribute('stroke', 'black');
            circle.setAttribute('stroke-width', '1px');
            nestedSvg.appendChild(title);
            nestedSvg.appendChild(circle);
            return nestedSvg;
        }
        const hoverOffset = 5;
        let startPointX = center[0] + radius * Math.cos(startRad);
        let startPointY = center[1] + radius * Math.sin(startRad);
        let startPoint = `M ${startPointX} ${startPointY}`;
        let arcFinishX = center[0] + radius * Math.cos(startRad + diffRad);// 150;
        let arcFinishY = center[1] + radius * Math.sin(startRad + diffRad); // 0
        let arcAngle = 0;
        let arcType = diffRad <= Math.PI ? 0 : 1;
        let arcClockwise = 1;
        let arc = `A ${radius} ${radius} ${arcAngle} ${arcType} ${arcClockwise} ${arcFinishX} ${arcFinishY}`;
        let lineOne = `L ${center[0]} ${center[1]}`;
        let lineTwo = `L ${startPointX} ${startPointY}`;
        let d = `${startPoint} ${arc} ${lineOne} ${lineTwo} Z`;
        let slice = document.createElementNS('http://www.w3.org/2000/svg', 'path');
        slice.setAttribute('d', d);
        slice.setAttribute('fill', `#${elem.objcolor}`);
        slice.setAttribute('stroke', 'black');
        slice.setAttribute('stroke-width', '1px');
        slice.addEventListener('mouseover', function (){
            let centerOne = center[0] + hoverOffset * Math.cos(startRad + diffRad / 2);
            let centerTwo = center[1] + hoverOffset * Math.sin(startRad + diffRad / 2);
            startPointX = centerOne + radius * Math.cos(startRad);
            startPointY = centerTwo + radius * Math.sin(startRad);
            startPoint = `M ${startPointX} ${startPointY}`;
            arcFinishX = centerOne + radius * Math.cos(startRad + diffRad);// 150;
            arcFinishY = centerTwo + radius * Math.sin(startRad + diffRad); // 0
            arc = `A ${radius} ${radius} ${arcAngle} ${arcType} ${arcClockwise} ${arcFinishX} ${arcFinishY}`;
            lineOne = `L ${centerOne} ${centerTwo}`;
            lineTwo = `L ${startPointX} ${startPointY}`;
            let dHover = `${startPoint} ${arc} ${lineOne} ${lineTwo} Z`;
            slice.setAttribute('d', dHover);
        })
        slice.addEventListener('mouseout', function () {
            slice.setAttribute('d', d);
        })
        nestedSvg.appendChild(title);
        nestedSvg.appendChild(slice);
        return nestedSvg;
    }
    static drawLegendItem(num, elem, key) {
        let legendItem = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        let square = document.createElementNS('http://www.w3.org/2000/svg', 'rect');
        let y = num*20 + 20;
        square.setAttribute('x', '350');
        square.setAttribute('y', `${y}`);
        square.setAttribute('height', '15');
        square.setAttribute('width', '15');
        square.setAttribute('stroke', 'black');
        square.setAttribute('stroke-width', '1px');
        square.setAttribute('fill', `#${elem.objcolor}`);
        let text = document.createElementNS('http://www.w3.org/2000/svg', 'text');
        text.setAttribute('y', `${y+12}`);
        text.setAttribute('x', '370');
        text.setAttribute('font-size', '10');
        text.innerHTML = `${elem.objname}: ${elem[key]}`;
        legendItem.appendChild(square);
        legendItem.appendChild(text);
        return legendItem;
    }
    static drawSVG(orderedData, value, key) {
        let svg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        let legend = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        svg.setAttribute('width', '100%');
        svg.setAttribute('height', '400');
        const center = [150, 150];
        const radius = 145;
        let startRad = 0;
        orderedData.forEach((elem, num) => {
            let diffRad = (elem[value] / 100) * Math.PI * 2;
            let slice = PieChart.drawPieSlice(startRad, diffRad, radius, center, elem, key);
            let legendItem = PieChart.drawLegendItem(num, elem, key);
            svg.appendChild(slice);
            legend.appendChild(legendItem);
            startRad += diffRad;
        })
        svg.appendChild(legend);
        return svg;
    }
    static drawIntoTable(newRow, column, data) {
        BaseChart.drawIntoTable(PieChart, newRow, column, data);
        return true;
    }
}
class BaseSection {
    constructor(section) {
        this.section = section;
        this.sectionHasContent = ('content' in section);
        this.sectionHasBlocks = ('header' in section);
        this.sectionHasTitle = ('tbl_cap' in section);
    }
    static buildTitle(section) {
        let title = document.createElement('h3');
        title.innerHTML = section.tbl_cap;
        if (section.href) {
            title.id = section.href;
        }
        return title;
    }
    init() {
        let div = document.createElement('div');
        if (this.section.sect_id) {
            div.setAttribute('id', this.section.sect_id);
        }
        if (this.sectionHasTitle) {
            div.appendChild(BaseSection.buildTitle(this.section));
        }
        if (this.sectionHasContent) {
            let contentClass = this.section.content.class;
            let contentText = this.section.content.text;
            let contentDiv = document.createElement('div');
            contentDiv.classList.add(contentClass);
            contentDiv.innerHTML = contentText;
            div.appendChild(contentDiv);
        }
        if (this.sectionHasBlocks) {
            let table = document.createElement('table');
            table.setAttribute('id', `${this.section.sect_id}_t`);
            for (let i = 0; i < this.section.header.length; i++) {
                let newBlock = structuredClone(this.section);
                newBlock.header = this.section.header[i];
                if (newBlock.header.source) {
                    newBlock.data = data.datasets[newBlock.header.source];
                    if (newBlock.header.filter) {
                        newBlock.data = Utilities.filter(newBlock.data, newBlock.header.filter);
                    }
                    if (newBlock.header.ordering) {
                        let direction = 1;
                        let key = newBlock.header.ordering;
                        if (newBlock.header.ordering.startsWith('-')) {
                            direction = -1;
                            key = newBlock.header.ordering.slice(1);
                        }
                        newBlock.data = Utilities.sort(newBlock.data, key, direction);
                    }
                    if (newBlock.header.limit) {
                        newBlock.data = Utilities.limit(newBlock.data, Number.parseInt(data.properties[newBlock.header.limit]));
                    }
                } else {
                    newBlock.data = this.section.data[i];
                }
                table.setAttribute('class', newBlock.header.class);
                if (newBlock.header["highlight"]) {
                    table.classList.add('highlight');
                }
                if (newBlock.header["preview"]) {
                    table.classList.add('preview');
                }
                if (newBlock.header.type === 'row_table') {
                    div.appendChild(new HorizontalTable(newBlock, table).init());
                } else if (newBlock.header.type === 'column_table') {
                    div.appendChild(new VerticalTable(newBlock, table).init());
                }
            }
        }
        return div;
    }
}
class BaseTable extends BaseSection {
    static uniqueHeaders = {
        'waitEventsBlock': BaseTable.buildWaitEventsBlockHeader
    }
    static uniqueCells = {
        'queryId': BaseTable.buildQueryIdCell,
        'jitCellId': BaseTable.buildJitCellId,
        'jitTimeCell': BaseTable.buildJitTimeCell,
        'interval': BaseTable.buildIntervalCell,
        'queryTextId': BaseTable.buildQueryTextIdCell,
        'planId': BaseTable.buildPlanIdCell,
        'queryText': BaseTable.buildQueryTextCell,
        'waitEvent': BaseTable.buildWaitEventDetailsCell,
        'pipeChart': PipeChart.drawIntoTable,
        'pieChart': PieChart.drawIntoTable
    }
    static properties = {
        'topn': data.properties.topn,
        'end1_id': data.properties.end1_id,
        'end2_id': data.properties.end2_id,
        'start1_id': data.properties.start1_id,
        'start2_id': data.properties.start2_id,
        'description': data.properties.description,
        'report_end1': data.properties.report_end1,
        'report_end2': data.properties.report_end2,
        'server_name': data.properties.server_name,
        'report_start1': data.properties.report_start1,
        'report_start2': data.properties.report_start2,
        'max_query_length': data.properties.max_query_length,
        'pgprofile_version': data.properties.pgprofile_version,
        'server_description': data.properties.server_description,
        'checksum_fail_detected': data.properties.checksum_fail_detected,
        'interval1_duration_sec': data.properties.interval1_duration_sec,
        'interval2_duration_sec': data.properties.interval2_duration_sec
    }
    constructor(section, table) {
        super(section);
        this.table = table;
    }
    static buildWaitEventsBlockHeader(table, header) {
        let tr0 = document.createElement('tr');
        let tr1 = document.createElement('tr');
        let tr2 = document.createElement('tr');
        tr0.classList.add('header');
        tr1.classList.add('header');
        tr2.classList.add('header');
        let th0 = document.createElement('th');
        let mainColumn = header.columns[0];
        th0.innerHTML = mainColumn.caption;
        th0.setAttribute('colspan', '100%');
        tr0.appendChild(th0);
        mainColumn.columns.forEach(column => {
            let th1 = document.createElement('th');
            th1.innerHTML = column.caption;
            if (column.columns) {
                th1.setAttribute('colspan', column.columns.length);
                column.columns.forEach(column => {
                    let th2 = document.createElement('th');
                    th2.innerHTML = column.caption;
                    if (column.title) {
                        th2.setAttribute('title', column.title);
                    }
                    tr2.appendChild(th2);
                })
            } else {
                th1.setAttribute('rowspan', '2');
            }
            tr1.appendChild(th1);
        })
        table.appendChild(tr0);
        table.appendChild(tr1);
        table.appendChild(tr2);
    }
    static buildJitTimeCell(newRow, column, row) {
        let newCell = newRow.insertCell(-1);
        if (row[column.id]) {
            newCell.setAttribute('class', 'table_obj_value');
            let p = document.createElement('p');
            let a = document.createElement('a');
            a.href = `#jit_${row.hexqueryid}_${row.datid}_${row.userid}_${row.toplevel}`;
            a.innerHTML = row[column.id];
            p.appendChild(a);
            newCell.appendChild(p);
        }
        return !!row[column.id];
    }
    static buildQueryIdCell(newRow, column, row) {
        let newCell = newRow.insertCell(-1);
        newCell.setAttribute('class', column.class);
        if (column.rowspan) {
            newCell.setAttribute('rowspan', '2');
        }
        let p1 = document.createElement('p');
        let a1 = document.createElement('a');
        a1.href = `#${row.hexqueryid}`;
        a1.innerHTML = row.hexqueryid;
        p1.appendChild(a1);
        let p2 = document.createElement('p');
        let small = document.createElement('small');
        p2.appendChild(small);
        small.innerHTML = `[${row.hashed_ids}]`;
        if (!row.toplevel) {
            small = document.createElement('small');
            p2.appendChild(small);
            small.innerHTML = '(N)';
            small.title = 'Nested level';
        }
        newCell.appendChild(p1);
        newCell.appendChild(p2);
        let button = Copier.drawButton();
        button.classList.add('copyQueryId');
        p1.appendChild(button);
        return !!row.hexqueryid;
    }
    static buildJitCellId(newRow, column, row) {
        BaseTable.buildQueryIdCell(newRow, column, row);
        newRow.firstChild.setAttribute('id', `jit_${row.hexqueryid}_${row.datid}_${row.userid}_${row.toplevel}`);
    }
    static buildPlanIdCell(newRow, column, row) {
        let newCell = newRow.insertCell(-1);
        newCell.setAttribute('class', column.class);
        if (column.rowspan) {
            newCell.setAttribute('rowspan', '2');
        }
        let p1 = document.createElement('p');
        let a1 = document.createElement('a');
        a1.setAttribute('href', `#${row.hexqueryid}_${row.hexplanid}`);
        a1.innerHTML = row.hexplanid;
        p1.appendChild(a1);
        newCell.appendChild(p1);
        return !!row.hexplanid;
    }
    static buildQueryTextIdCell(newRow, column, row) {
        let cell = newRow.insertCell(-1);
        let columnId = row[column.id];
        newRow.classList.add('statement');
        cell.setAttribute('id', columnId);
        let newText = document.createTextNode(columnId);
        cell.setAttribute('class', column.class);
        const queryTextsLength = row.query_texts.length;
        cell.setAttribute('rowspan', `${queryTextsLength}`);
        cell.appendChild(newText);
        return !!columnId;
    }
    static buildQueryTextCell(newRow, column, row) {
        for (let i = 0; i < row[column.id].length; i++) {
            if (i > 0) {
                newRow = newRow.parentNode.insertRow(-1);
                newRow.setAttribute('data-hexqueryid', row.hexqueryid);
                newRow.setAttribute('data-all', row.hexqueryid);
            }
            let newCell = newRow.insertCell(-1);
            let text = row[column.id][i];
            newCell.setAttribute('class', column.class);
            let newText = document.createTextNode(text);
            newCell.appendChild(newText);
        }
        if (row.plans && row.plans.length) {
            for (let i = 0; i < row.plans.length; i++) {
                newRow = newRow.parentNode.insertRow(-1);
                newRow.setAttribute('data-hexqueryid', row.hexqueryid);
                newRow.setAttribute('data-hexplanid', row.plans[i].hexplanid);
                newRow.setAttribute('data-all', `${row.hexqueryid} ${row.plans[i].hexplanid}`);
                newRow.classList.add('plantext');
                let newCell = newRow.insertCell(-1);
                let text = row.plans[i].hexplanid;
                let newText = document.createTextNode(text);
                newCell.appendChild(newText);
                newCell.setAttribute('class', "mono queryTextId");
                newCell.setAttribute('id', `${row.hexqueryid}_${row.plans[i].hexplanid}`);
                newCell = newRow.insertCell(-1);
                text = `<pre>${row.plans[i].plan_text}</pre>`;
                newCell.insertAdjacentHTML('afterbegin', text);
            }
        }
        return !!row[column.id];
    }
    static buildIntervalCell(newRow, column, row) {
        let cell = newRow.insertCell(-1);
        let newText = document.createTextNode(column.id);
        if (column.class) {
            cell.setAttribute('class', 'table_obj_value');
        }
        if (column.rowspan) {
            cell.setAttribute('rowspan', '2');
        }
        if (column.title) {
            cell.setAttribute('title', BaseTable.getTagTitle(column.title));
        }
        cell.appendChild(newText);
        return true;
    }
    static buildWaitEventDetailsCell(newRow, column, row) {
        let newCell = newRow.insertCell(-1);
        newRow.setAttribute('id', `${row.event_type}_${row.hexqueryid}_${row.hexplanid}`);
        if (column.rowspan) {
            newCell.setAttribute('rowspan', row.rowspan);
        }
        if (column.class) {
            newCell.setAttribute('class', column.class);
            newCell.classList.add('table_obj_value');
        }
        let dataExists = Boolean(row[column.id]);
        if (dataExists) {
            for (let i = 0; i < row[column.id].length; i++) {
                let details = row[column.id][i];
                if (details.event) {
                    let p = document.createElement('p');
                    let a = document.createElement('a');
                    if (row.event_type === 'Total') {
                        a.href = `#${details.event}_${row.hexqueryid}_${row.hexplanid}`;
                    }
                    a.innerHTML = details.event;
                    p.appendChild(a);
                    let colons = document.createTextNode(': ');
                    p.appendChild(colons);
                    let strong = document.createElement('strong');
                    strong.innerHTML = details.wait;
                    p.appendChild(strong);
                    newCell.appendChild(p);
                }
            }
        }
        return true;
    }
    static bifurcateObject(column) {
        let objects = [];
        let keys = Object.keys(column);
        for (let i = 0; i < column.id.length; i++) {
            let newObj = structuredClone(column);
            for (let j = 0; j < keys.length; j++) {
                if (typeof newObj[keys[j]] === 'object') {
                    newObj[keys[j]] = column[keys[j]][i];
                }
            }
            objects.push(newObj);
        }
        return objects;
    }
    static getTagTitle(title) {
        const TITLES = {
            'properties.timePeriod1': `(${BaseTable.properties.report_start1} - ${BaseTable.properties.report_end1})`,
            'properties.timePeriod2': `(${BaseTable.properties.report_start2} - ${BaseTable.properties.report_end2})`,
            'properties.timePeriod1,properties.timePeriod2': 'Sample\'s time period'
        }
        if (TITLES[title] !== undefined) {
            return TITLES[title];
        }
        return title;
    }
    static hasSpecialClass(column) {
        if (column.class) {
            let classList = column.class.split(' ');
            for (let i = 0; i < classList.length; i++) {
                let klass = classList[i].trim();
                if (BaseTable.uniqueCells[klass]) {
                    return klass;
                }
            }
        }
        return false;
    }
    static collectHeader(header, deep, resultMatrix) {
        if (resultMatrix === null) {
            resultMatrix = [];
        }
        if (header.caption) {
            if (!resultMatrix[deep]) {
                resultMatrix.push([]);
            }
            let sumCols = 0;
            if (header.columns) {
                deep++;
                header.columns.forEach(column => {
                    BaseTable.collectHeader(column, deep, resultMatrix);
                    if (column.columns) {
                        sumCols += column.columns.length - 1;
                    }
                })
                deep--;
            }
            let th = {
                'caption': header.caption,
                'colspan': header.columns ? header.columns.length + sumCols : 1,
                'rowspan': deep === 0 && !('columns' in header) ? 2 : 1
            };
            if (header.id) {
                th['id'] = header.id
            }
            if (header.title) {
                th['title'] = BaseTable.getTagTitle(header.title)
            }
            resultMatrix[deep].push(th);
        } else {
            if (header.columns) {
                header.columns.forEach(column => {
                    BaseTable.collectHeader(column, deep, resultMatrix);
                })
            }
        }
        return resultMatrix;
    }
    buildHeader() {
        let headerMatrix = BaseTable.collectHeader(this.section.header, 0, null);
        headerMatrix.forEach(row => {
            let tr = document.createElement('tr');
            tr.classList.add('header');
            row.forEach(column => {
                let th = document.createElement('th');
                th.innerHTML = column.caption;
                th.setAttribute('rowspan', headerMatrix.length === 1 ? 1 : column.rowspan);
                th.setAttribute('colspan', column.colspan);
                if (column.title) {
                    th.setAttribute('title', column.title);
                }
                if (column.id) {
                    th.setAttribute('id', column.id);
                }
                tr.appendChild(th);
            });
            this.table.appendChild(tr);
        });
        if (headerMatrix.length === 1) {
            let emptyRow = document.createElement('tr');
            this.table.appendChild(emptyRow);
        }
    }
    init() {
        this.buildHeader();
        this.insertRows();
        return this.table;
    }
}
class HorizontalTable extends BaseTable {
    static buildCell(newRow, column, row) {
        let specialClass = BaseTable.hasSpecialClass(column);
        if (specialClass) {
            let notEmpty = BaseTable.uniqueCells[specialClass](newRow, column, row);
            return !notEmpty;
        }
        let newCell = newRow.insertCell(-1);
        if (column.rowspan) {
            newCell.setAttribute('rowspan', '2');
        }
        if (column.class) {
            newCell.setAttribute('class', column.class);
        }
        if (row[column.id]) {
            let newText = document.createTextNode(row[column.id]);
            newCell.appendChild(newText);
            return false;
        }
        return true;
    }
    static collectColumns(header, array, matrix) {
        if (array === null) {
            array = [];
        }
        if (matrix === null) {
            matrix = [];
        }
        if ('columns' in header) {
            for (let i = 0; i < header.columns.length; i++) {
                let column = header.columns[i];
                HorizontalTable.collectColumns(column, array, matrix);
            }
        } else {
            if (typeof header.id === 'string') {
                array.push(header);
            } else if (typeof header.id === 'object') {
                matrix.push(BaseTable.bifurcateObject(header));
            }
        }
        if (matrix.length) {
            matrix = matrix[0].map((_, colIndex) => matrix.map(row => row[colIndex]));
        }
        return [array, matrix];
    }
    static getColumns(columns) {
        let matrixCollection = HorizontalTable.collectColumns(columns, null, null);
        let singeRowColumns = matrixCollection[0];
        let matrixWithVectors = matrixCollection[1];
        let newMatrix = Array.from(matrixCollection[1]);
        if (matrixWithVectors[0]) {
            newMatrix[0] = singeRowColumns.concat(matrixWithVectors[0]);
        } else {
            newMatrix[0] = singeRowColumns;
        }
        return newMatrix;
    }
    static setDataAttrs(newRow, row, dataAttrs) {
        let searchArray = [];
        if (dataAttrs) {
            dataAttrs.forEach(attr => {
                newRow.setAttribute(`data-${attr.id}`, row[attr.id]);
                searchArray.push(row[attr.id]);
            })
            let searchString = Object.values(searchArray).join(' ').replace(/\s\s+/g, ' ');
            newRow.setAttribute('data-all', searchString);
        }
    }
    insertRows() {
        let columns = HorizontalTable.getColumns(this.section.header);  // Getting columns matrix
        let highlightRowAttrs = this.section.header.highlight;
        let previewQueryAttrs = this.section.header.preview;
        let rows = this.section.data;  // Getting json array with data
        if (highlightRowAttrs) {
            let result = JSON.stringify(highlightRowAttrs);
            this.table.setAttribute('data-highlight', result);
        }
        if (previewQueryAttrs) {
            let result = JSON.stringify(previewQueryAttrs);
            this.table.setAttribute('data-preview', result);
        }
        for (let i = 0; i < rows.length; i++) {
            let row = rows[i];
            for (let j = 0; j < columns.length; j++) {
                let newRow = this.table.insertRow(-1);
                if (columns.length > 1) {
                    newRow.classList.add(`int${j + 1}`);
                }
                if (row.klass) {
                    newRow.classList.add(row.klass);
                }
                if (i % 2 !== 0) {
                    newRow.classList.add('grey');
                }
                if (highlightRowAttrs) {
                    HorizontalTable.setDataAttrs(newRow, row, highlightRowAttrs);
                }
                if (previewQueryAttrs) {
                    HorizontalTable.setDataAttrs(newRow, row, previewQueryAttrs);
                }
                let isEmpty = [];
                for (let k = 0; k < columns[j].length; k++) {
                    let column = columns[j][k];
                    isEmpty.push(HorizontalTable.buildCell(newRow, column, row));
                }
                if (isEmpty.every(Boolean)) {
                    newRow.style.visibility = 'collapse';
                    let prevSib = newRow.previousSibling;
                    prevSib.style.verticalAlign = 'baseline';
                }
            }
        }
    }
}
class VerticalTable extends BaseTable {
    static getColumns(section) {
        let columns = [];
        let rows = section.header.rows;
        for (let i = 0; i < rows.length; i++) {
            if (typeof rows[i].id === 'object') {
                columns.push(BaseTable.bifurcateObject(rows[i]));
            } else {
                columns.push(rows[i]);
            }
        }
        return columns;
    }
    static buildCell(newRow, column, row) {
        let specialClass = BaseTable.hasSpecialClass(column);
        if (specialClass) {
            BaseTable.uniqueCells[specialClass](newRow, column, row);
            return;
        }
        let newCell = newRow.insertCell(-1);
        if ('caption' in column) {
            let newText = document.createTextNode(column['caption']);
            newCell.appendChild(newText);
        }
        if ('title' in column) {
            newCell.setAttribute('title', column.title);
        }
        if ('id' in column) {
            newCell.innerHTML = row[0][column.id];
        }
        if ('class' in column) {
            newCell.setAttribute('class', column.class);
        }
    }
    insertRows() {
        let columns = VerticalTable.getColumns(this.section);
        let rows = this.section.data;
        for (let i = 0; i < columns.length; i++) {
            let newRow = this.table.insertRow(-1);
            if (i % 2 !== 0) {
                newRow.classList.add('grey');
            }
            VerticalTable.buildCell(newRow, columns[i], rows);
            columns[i].cells.forEach(cell => {
                VerticalTable.buildCell(newRow, cell, rows);
            });
        }
    }
}
class Highlighter {
    static transition = 'background-color 70ms';
    static toggleClass(tr, allRows) {
        Highlighter.cleanAllActiveClasses(allRows);
        if (Object.keys(tr.dataset).length && !tr.classList.contains('active')) {
            allRows.forEach((row) => {
                let isEqual = Highlighter.isDatasetEqual(tr, row);
                if (isEqual) {
                    Highlighter.setBackgroundColorToRow(row, '', this.transition);
                    row.classList.add('active');
                    let navId = this.getClosestTag(row, 0, 'div').firstChild.id;
                    if (navId) {
                        let navLi = document.getElementById(`navigator_${navId}`);
                        if (navLi && !navLi.classList.contains('active')) {
                            navLi.classList.add('active');
                        }
                    }
                }
            });
        }
    }
    static getAllRows() {
        return document.querySelectorAll('tr');
    }
    static getHighlightableRows() {
        return document.querySelectorAll('table.highlight tr:not(.queryRow)');
    }
    static getClosestTag(target, curDeep, targetTag) {
        let tooDeep = curDeep >= 5;
        let headOfTable = target.tagName.toLowerCase() === 'th';
        let stillNotRow = target.tagName.toLowerCase() !== targetTag;
        if (tooDeep) {
            return false;
        } else if (headOfTable) {
            return false;
        } else if (stillNotRow) {
            curDeep++;
            return Highlighter.getClosestTag(target.parentNode, curDeep, targetTag);
        } else {
            return target;
        }
    }
    static cleanAllActiveClasses(rows) {
        rows.forEach(elem => {
            if (elem.classList.contains('active')) {
                elem.classList.remove('active');
            }
        })
        let navigator = document.getElementById('navigator');
        if (navigator) {
            let allItems = document.querySelectorAll('li');
            allItems.forEach(item => {
                if (item.classList.contains('active')) {
                    item.classList.remove('active');
                }
            })
        }
    }
    static getTargetDS(tr) {
        let fieldsList = JSON.parse(tr.closest('table').dataset.highlight);
        let targetDS = {};
        fieldsList.forEach(field => {
            targetDS[field.id] = tr.dataset[field.id]
        })
        return targetDS;
    }
    static isDatasetEqual(tr, row) {
        let targetDataset = Highlighter.getTargetDS(tr);
        let rowDataset = row.dataset;
        let trIsSqlList = Highlighter.getClosestTag(tr, 0, 'table').id === 'sqllist_t';
        let rowIsSqlList = Highlighter.getClosestTag(row, 0, 'table').id === 'sqllist_t';
        let isSameQuery = targetDataset.hexqueryid !== undefined
            && targetDataset.hexqueryid === rowDataset.hexqueryid
            && targetDataset.planid === rowDataset.planid;
        if ((trIsSqlList || rowIsSqlList) && isSameQuery) {
            return true;
        }
        for (let data in targetDataset) {
            if (targetDataset[data] === '*' && rowDataset[data] !== undefined) {
                continue;
            }
            if (data === 'all') {
                continue;
            }
            if (targetDataset[data] !== rowDataset[data]) {
                return false;
            }
        }
        return true;
    }
    static setBackgroundColorToRow(tr, hoverColor, transition) {
        tr.querySelectorAll('td').forEach(td => {
            td.style.backgroundColor = hoverColor;
            td.style.transition = transition;
        })
        let siblings = null;
        if (tr.classList.contains('int1')) {
            siblings = tr.nextSibling.querySelectorAll('td');
        } else if (tr.classList.contains('int2')) {
            siblings = tr.previousSibling.querySelectorAll('td');
        }
        if (siblings) {
            siblings.forEach(elem => {
                elem.style.backgroundColor = hoverColor;
                elem.style.transition = transition;
            })
        }
    }
    static highlight(event, allRows) {
        if (event.target.tagName.toLowerCase() !== 'a') {
            let tr = Highlighter.getClosestTag(event.target, 0, 'tr');
            if (tr && Object.keys(tr.dataset).length) {
                Highlighter.toggleClass(tr, allRows);
            }
        }
    }
    static smartHover(eventType, event, transition) {
        let hoverColor = '#D9FFCC';
        let tr = Highlighter.getClosestTag(event.target, 0, 'tr');
        if (tr && !tr.classList.contains('active') && eventType === 'mouseover') {
            Highlighter.setBackgroundColorToRow(tr, hoverColor, transition);
        } else if (tr && eventType === 'mouseout') {
            Highlighter.setBackgroundColorToRow(tr, '', transition);
        }
    }
    static init() {
        const ALL_ROWS = Highlighter.getAllRows();
        const HIGHLIGHTABLE_ROWS = Highlighter.getHighlightableRows();
        HIGHLIGHTABLE_ROWS.forEach((elem) => {
            elem.addEventListener('click', (event) => {
                Highlighter.highlight(event, HIGHLIGHTABLE_ROWS);
            });
        })
        ALL_ROWS.forEach((elem) => {
            ['mouseover', 'mouseout'].forEach(eventType => {
                elem.addEventListener(eventType, (event) => {
                    Highlighter.smartHover(eventType, event, this.transition);
                });
            })
        })
    }
}
class ReportNavigator {
    static buildReportNavigator(NAVIGATOR) {
        window.onscroll = function () {
            document.querySelectorAll('h3').forEach(title => {
                let position = title.getBoundingClientRect().top;
                if (position >= 0 && position < 100) {
                    let li = `[href*=${title.id}]`;
                    let elem = NAVIGATOR.querySelector(li).closest('li');
                    if (!elem.classList.contains('current')) {
                        NAVIGATOR.querySelectorAll('li').forEach(item => {
                            item.classList.remove('current');
                        });
                        elem.classList.add('current');
                    }
                }
            })
        }
    }
    static buildPageContent(data, parentNode, visibility) {
        data.sections.forEach(section => {
            let hasTableCap = ('tbl_cap' in section);
            let hasNestedSections = ('sections' in section);
            let ul = document.createElement('ul');
            let li = document.createElement('li');
            li.classList.add(visibility);
            if (hasTableCap) {
                let a = document.createElement('a');
                a.innerHTML = section.tbl_cap;
                a.href = `#${section.href}`;
                a.classList.add('anchor');
                li.setAttribute('id', `navigator_${section.href}`);
                li.appendChild(a);
                parentNode.appendChild(li);
            } else {
                ul = li;
            }
            if (hasNestedSections) {
                parentNode.appendChild(this.buildPageContent(section, ul, "visible"));
            }
            li.addEventListener("click", ev => {
                if (ev.target.parentNode.nextSibling.tagName === "UL") {
                    ev.target.parentNode.nextSibling.childNodes.forEach(el => {
                        if (el.classList.contains("hidden")) {
                            el.classList.remove("hidden");
                            el.classList.add("visible");
                        } else {
                            el.classList.remove("visible");
                            el.classList.add("hidden")
                        }
                    })
                }
            })
        })
        return parentNode;
    }
    static init() {
        const TOPNAV = document.getElementById('topnav');
        const MAIN_REPORT = document.getElementById('container');
        MAIN_REPORT.setAttribute('class', 'with-menu')
        let menu = document.createElement('a');
        menu.setAttribute('class', 'active');
        for (let i = 0; i < 3; i++) {
            let burger = document.createElement('div');
            burger.style.width = "10px";
            burger.style.height = "2px";
            burger.style.background = "white";
            burger.style.marginBottom = "2px";
            menu.appendChild(burger);
        }
        let input = document.createElement('input');
        input.setAttribute('id', 'searchField');
        input.setAttribute('type', 'search');
        input.setAttribute('placeholder', 'Filter...');
        let select = document.createElement('select');
        select.setAttribute('id', 'searchParam');
        Object.entries(Utilities.searchParams).forEach(([key, value]) => {
            let choiceElement = document.createElement('option');
            choiceElement.setAttribute('value', key);
            choiceElement.innerHTML = value;
            select.appendChild(choiceElement);
        })
        let rowsForSearch = document.querySelectorAll('tr[data-all]');
        input.addEventListener('input', ev => {
            let keyword = ev.target.value.trim();
            let searchParam = document.getElementById('searchParam').value;
            Utilities.search(rowsForSearch, searchParam, keyword);
        })
        select.addEventListener('change', ev => {
            let searchParam = ev.target.value;
            let keyword = Utilities.getInputField().value;
            if (keyword) {
                Utilities.search(rowsForSearch, searchParam, keyword);
            }
        })
        document.addEventListener('keydown', evt => {
            if (evt.key === 'Escape') {
                Utilities.cancelSearchResults(rowsForSearch);
            }
        });
        TOPNAV.appendChild(menu);
        TOPNAV.appendChild(input);
        TOPNAV.appendChild(select);
        const NAVIGATOR = document.createElement('div');
        NAVIGATOR.setAttribute('id', 'navigator');
        let ul = document.createElement('ul');
        ul.setAttribute('class', 'active');
        NAVIGATOR.appendChild(ul);
        menu.addEventListener('click', event => {
            if (menu.classList.contains('active')) {
                menu.setAttribute('class', '')
                MAIN_REPORT.style.left = "0";
                NAVIGATOR.style.width = "0";
            } else {
                menu.setAttribute('class', 'active')
                MAIN_REPORT.style.left = "25%";
                NAVIGATOR.style.width = "25%";
            }
        })
        document.querySelector('body').appendChild(NAVIGATOR);
        ReportNavigator.buildPageContent(data, ul, "visible");
        ReportNavigator.buildReportNavigator(NAVIGATOR);
    }
}
class Previewer {
    static getParentRows() {
        return document.querySelectorAll("table.preview tr:not(.header)");
    }
    static preprocessQueryString(queryString) {
        let etc = '';
        queryString = queryString.split(',').join(', ');
        queryString = queryString.split('+').join(' + ');
        queryString = queryString.split('/').join(' / ');
        if (queryString.length > 1000) {
            queryString = queryString.substring(0, 1000);
            etc = ' ...'
        }
        return `${queryString}${etc}`
    }
    static queryTextPreviewer(queryCell, queryRow, newRow, queryString) {
        queryCell.style.width = `${Math.floor(newRow.offsetWidth * 0.95)}px`;
        queryCell.style.fontFamily = 'Monospace';
        queryRow.style.display = '';
        queryRow.setAttribute("data-hexqueryid", newRow.dataset.hexqueryid);
        if (queryCell.firstChild && queryCell.firstChild.tagName.toLowerCase() !== 'p') {
            let preprocessedText = Previewer.preprocessQueryString(queryString);
            queryCell.insertAdjacentHTML('afterbegin', `<p><i>${preprocessedText}</i></p>`);
        }
    }
    static findQuery(datasetName, dataID, hexQueryId) {
        for (let i = 0; i < data.datasets[datasetName].length; i++) {
            if (data.datasets[datasetName][i][dataID] === hexQueryId) {
                return i
            }
        }
        return -1
    }
    static init() {
        const PARENT_ROWS = Previewer.getParentRows();
        PARENT_ROWS.forEach(parentRow => {
            let queryCell = document.createElement("td");
            queryCell.setAttribute("colspan", "100");
            let queryRow = document.createElement("tr");
            queryRow.classList.add("queryRow");
            queryRow.style.display = "none";
            let preview = JSON.parse(parentRow.closest('table').dataset["preview"])[0]
            let datasetName = preview.dataset;
            let dataID = preview.id;
            let copyQueryTextButton = Copier.drawButton();
            copyQueryTextButton.setAttribute("class", "copyQueryTextButton");
            queryCell.appendChild(copyQueryTextButton);
            queryRow.appendChild(queryCell);
            if (!parentRow.classList.contains("int1")) {
                parentRow.insertAdjacentElement("afterend", queryRow);
            }
            parentRow.addEventListener("click", event => {
                if (parentRow.classList.contains('int1')) {
                    queryRow = parentRow.nextSibling.nextSibling;
                    queryCell = queryRow.firstChild;
                }
                if (event.target.tagName.toLowerCase() !== 'a' && event.target.tagName.toLowerCase() !== 'rect') {
                    if (queryRow.style.display === 'none') {
                        let queryIndex = Previewer.findQuery(datasetName, dataID, parentRow.dataset[dataID]);
                        if (queryIndex >= 0) {
                            let queryText = data.datasets[datasetName][queryIndex].query_texts[0];
                            Previewer.queryTextPreviewer(queryCell, queryRow, parentRow, queryText);
                        }
                    } else {
                        queryRow.style.display = 'none';
                    }
                }
            })
        })
    }
}
class Copier {
    static drawButton() {
        let button = document.createElement('a');
        button.setAttribute('class', 'copyButton');
        button.setAttribute('title', 'Copy to clipboard');
        let svg = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
        svg.style.marginLeft = '10px';
        svg.setAttribute('height', '14px');
        svg.setAttribute('width', '12px');
        let rect = document.createElementNS('http://www.w3.org/2000/svg', 'rect');
        rect.setAttribute('x', '2');
        rect.setAttribute('y', '2');
        rect.setAttribute('height', '12px');
        rect.setAttribute('width', '10px');
        rect.setAttribute('rx', '4');
        rect.setAttribute('stroke', 'grey');
        rect.setAttribute('fill', 'transparent');
        let replica = document.createElementNS('http://www.w3.org/2000/svg', 'rect');
        replica.setAttribute('x', '0');
        replica.setAttribute('y', '0');
        replica.setAttribute('height', '12px');
        replica.setAttribute('width', '10px');
        replica.setAttribute('rx', '4');
        replica.setAttribute('stroke', 'grey');
        replica.setAttribute('fill', 'transparent');
        svg.appendChild(rect);
        svg.appendChild(replica);
        button.appendChild(svg);
        return button;
    }
    static copyQueryId(ev) {
        let text = ev.target.closest('tr').dataset.queryid;
        navigator.clipboard.writeText(text).then(r => console.log('Copy: ', text));
    }
    static init() {
        document.querySelectorAll('.copyQueryId').forEach(button => {
            button.addEventListener('click', function () {
                let text = button.closest('tr').dataset.queryid;
                navigator.clipboard.writeText(text).then(r => console.log('Copy query ID'));
            })
        })
        document.querySelectorAll('.copyQueryTextButton').forEach(button => {
            button.addEventListener('click', function () {
                let hexQueryId = button.closest('tr').dataset.hexqueryid;
                data.datasets.queries.forEach(query => {
                    if (query.hexqueryid === hexQueryId) {
                        let text = query.query_texts[0];
                        navigator.clipboard.writeText(text).then(r => console.log('Copy query text'));
                    }
                })
            })
        })
        document.querySelectorAll('.copyPlanTextButton').forEach(button => {
            button.addEventListener('click', function () {
                let hexQueryId = button.closest('tr').dataset.hexqueryid;
                let hexPlanId = button.closest('tr').dataset.hexplanid;
                data.datasets.queries.forEach(query => {
                    if (query.hexqueryid === hexQueryId) {
                        query.plans.forEach(plan => {
                            if (plan.hexplanid === hexPlanId) {
                                let text = plan.plan_text;
                                navigator.clipboard.writeText(text).then(r => console.log('Copy plan text'));
                            }
                        })
                    }
                })
            })
        })
    }
}
function buildReport(data, parentNode) {
    data.sections.forEach(section => {
        let sectionHasNestedSections = ('sections' in section);
        let newSection = new BaseSection(section).init();
        if (sectionHasNestedSections) {
            buildReport(section, newSection);
        }
        parentNode.appendChild(newSection);
    })
    return parentNode;
}
function main() {
    const CONTAINER = document.getElementById('container');
    buildReport(data, CONTAINER);
    Highlighter.init();
    Previewer.init();
    Copier.init();
    ReportNavigator.init();
}
main();$js$
),
('report',
  '<!DOCTYPE html>'
  '<html lang="en"><head>'
  '<style>{static:css}</style>'
  '<script>const data={dynamic:data1}</script>'
  '<title>Postgres profile report ({properties:start1_id} -'
  ' {properties:end1_id})</title></head><body>'
  '<div id="topnav" class="topnav"></div>'
  '<div id="container">'
  '<H1>Postgres profile report</H1>'
  '</div>'
  '<script>{static:script_js}</script>'
  '</body></html>'),
('diffreport',
  '<!DOCTYPE html>'
  '<html lang="en"><head>'
  '<style>{static:css}</style>'
  '<script>const data={dynamic:data1}</script>'
  '<title>Postgres profile differential report (1): ({properties:start1_id} -'
  ' {properties:end1_id}) with (2): ({properties:start2_id} -'
  ' {properties:end2_id})</title></head><body>'
  '<ul id="content"></ul>'
  '<div id="topnav" class="topnav"></div>'
  '<div id="container">'
  '<H1>Postgres profile differential report</H1>'
  '</div>'
  '<script>{static:script_js}</script>'
  '</body></html>')
;
/* === report table data === */
INSERT INTO report(report_id, report_name, report_description, template)
VALUES
(1, 'report', 'Regular single interval report', 'report'),
(2, 'diffreport', 'Differential report on two intervals', 'diffreport')
;
/* === report_struct table data === */
-- Regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'rep_details', NULL, 10, 'Report details', 'Report details', NULL, NULL, 'rep_details', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "properties",'
    '"columns": ['
        '{"id": "pgprofile_version", "class": "table_obj_value", "caption": "Version"},'
        '{"id": "server_name", "class": "table_obj_value", "caption": "Server name"},'
        '{"caption": "Interval (sample)", "class": "table_obj_value", "columns": ['
          '{"id": "start1_id", "class": "table_obj_value", "caption": "start"},'
          '{"id": "end1_id", "class": "table_obj_value", "caption": "end"}'
        ']},'
        '{"caption": "Interval (time)", "class": "table_obj_value", "columns": ['
          '{"id": "report_start1", "class": "table_obj_value", "caption": "start"},'
          '{"id": "report_end1", "class": "table_obj_value", "caption": "end"}'
        ']}'
    ']}'
  ']'::jsonb),
(1, 'rep_settings', NULL, 11, NULL, NULL, NULL, NULL, 'rep_settings', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "h_ord"},'
    '"ordering": "h_ord",'
    '"columns": ['
      '{"caption": "Setting", "id": "name", "class": "table_obj_name"},'
      '{"caption": "Value", "id": "reset_val", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'stmt_cmt1', NULL, 100, NULL, NULL, 'stmt_cnt_range', NULL, NULL,
  '{"class": "warning", "text": "This interval contains sample(s) with captured statements count more than 90% of pg_stat_statements.max parameter. '
  'Consider increasing pg_stat_statements.max parameter"}',
  '[{'
  '"type": "row_table", '
  '"source": "stmt_cnt_range",'
  '"ordering": "sample_id",'
  '"columns": ['
      '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample ID"}, '
      '{"id": "sample_time", "class": "table_obj_value", "caption": "Sample Time"}, '
      '{"id": "stmt_cnt", "class": "table_obj_value", "caption": "Stmts Captured"}, '
      '{"id": "max_cnt", "class": "table_obj_value", "caption": "pg_stat_statements.max"}'
  ']}]'::jsonb),
(1, 'srvstat', NULL, 200, 'Server statistics', 'Server statistics', NULL, NULL, 'cl_stat', NULL, NULL),
(1, 'actsesshdr', NULL, 250, 'Session states', 'Session states observed by subsamples', 'act_backend', NULL, 'stat_act',
  '{"class": "notice", "text": "Statistics about session states exceeding capturing thresholds."}', NULL),
(1, 'sqlsthdr', NULL, 300, 'SQL query statistics', 'SQL query statistics', 'statstatements', NULL, 'sql_stat', NULL, NULL),
(1, 'objects', NULL, 400, 'Schema object statistics', 'Schema object statistics', NULL, NULL, 'schema_stat', NULL, NULL),
(1, 'funchdr', NULL, 500, 'User function statistics', 'User function statistics', 'function_stats', NULL, 'func_stat', NULL, NULL),
(1, 'vachdr', NULL, 600, 'Vacuum-related statistics', 'Vacuum-related statistics', NULL, NULL, 'vacuum_stats', NULL, NULL),
(1, 'settings', NULL, 700, 'Cluster settings during the report interval', 'Cluster settings during the report interval', NULL, NULL, 'pg_settings', NULL, NULL),
(1, 'stmt_warn', NULL, 800, NULL, 'Warning!', 'stmt_cnt_all', NULL, 'stmt_warn', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "stmt_cnt_all",'
    '"ordering": "sample_id",'
    '"columns": ['
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample ID"}, '
        '{"id": "sample_time", "class": "table_obj_value", "caption": "Sample Time"}, '
        '{"id": "stmt_cnt", "class": "table_obj_value", "caption": "Stmts Captured"}, '
        '{"id": "max_cnt", "class": "table_obj_value", "caption": "pg_stat_statements.max"}'
    ']}]'::jsonb)
;

-- Server section of regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'dbstat', 'srvstat', 100, 'Database statistics', 'Database statistics', NULL, NULL, 'db_stat', NULL, NULL),
(1, 'dbstatreset', 'dbstat', 200, NULL, NULL, 'dbstats_reset', NULL, NULL,
  '{"class": "warning", "text": "Database statistics reset detected during report interval! Statistics for listed databases and contained objects might be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "dbstats_reset",'
    '"ordering": "sample_id",'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "Sample", "id": "sample_id", "class": "table_obj_value"},'
      '{"caption": "Reset time", "id": "stats_reset", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'dbstatmain', 'dbstat', 300, NULL, NULL, NULL, NULL, NULL, NULL,
'[{'
    '"type": "row_table",'
    '"source": "dbstat",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "Transactions", "columns": ['
        '{"caption": "Commits", "id": "xact_commit", "class": "table_obj_value", '
          '"title": "Number of transactions in this database that have been committed"},'
        '{"caption": "Rollbacks", "id": "xact_rollback", "class": "table_obj_value", '
          '"title": "Number of transactions in this database that have been rolled back"},'
        '{"caption": "Deadlocks", "id": "deadlocks", "class": "table_obj_value", '
          '"title": "Number of deadlocks detected in this database"}'
     ']},'
      '{"caption": "Checksums", "condition": "checksum_fail_detected", "columns": ['
        '{"caption": "Failures", "id": "checksum_failures", "class": "table_obj_value", '
          '"title": "Number of block checksum failures detected"},'
        '{"caption": "Last", "id": "checksum_last_failure", "class": "table_obj_value", '
          '"title": "Last checksum failure detected"}'
     ']},'
      '{"caption": "Block statistics", "columns": ['
        '{"caption": "Hit(%)", "id": "blks_hit_pct", "class": "table_obj_value", '
          '"title": "Buffer cache hit ratio"},'
        '{"caption": "Read", "id": "blks_read", "class": "table_obj_value", '
          '"title": "Number of disk blocks read in this database"},'
        '{"caption": "Hit", "id": "blks_hit", "class": "table_obj_value", '
          '"title": "Number of times disk blocks were found already in the buffer cache"}'
     ']},'
      '{"caption": "Block I/O times", "condition": "io_times", "columns": ['
        '{"caption": "Read", "id": "blk_read_time", "class": "table_obj_value", '
          '"title": "Time spent reading data file blocks by backends, in seconds"},'
        '{"caption": "Write", "id": "blk_write_time", "class": "table_obj_value", '
          '"title": "Time spent writing data file blocks by backends, in seconds"}'
     ']},'
      '{"caption": "Tuples", "columns": ['
        '{"caption": "Ret", "id": "tup_returned", "class": "table_obj_value", '
          '"title": "Number of rows returned by queries in this database"},'
        '{"caption": "Fet", "id": "tup_fetched", "class": "table_obj_value", '
          '"title": "Number of rows fetched by queries in this database"},'
        '{"caption": "Ins", "id": "tup_inserted", "class": "table_obj_value", '
          '"title": "Number of rows inserted by queries in this database"},'
        '{"caption": "Upd", "id": "tup_updated", "class": "table_obj_value", '
          '"title": "Number of rows updated by queries in this database"},'
        '{"caption": "Del", "id": "tup_deleted", "class": "table_obj_value", '
          '"title": "Number of rows deleted"}'
     ']},'
      '{"caption": "Temp files", "columns": ['
        '{"caption": "Size", "id": "temp_bytes", "class": "table_obj_value", '
          '"title": "Total amount of data written to temporary files by queries in this database"},'
        '{"caption": "Files", "id": "temp_files", "class": "table_obj_value", '
          '"title": "Number of temporary files created by queries in this database"}'
     ']},'
     '{"caption": "Size", "id": "datsize", "class": "table_obj_value", '
       '"title": "Database size as is was at the moment of last sample in report interval"},'
     '{"caption": "Growth", "id": "datsize_delta", "class": "table_obj_value", '
       '"title": "Database size increment during report interval"}'
    ']'
  '}]'::jsonb),
(1, 'iostat', 'srvstat', 328, 'Cluster I/O statistics', 'Cluster I/O statistics', 'stat_io', NULL, 'stat_io', NULL, NULL),
(1, 'iostatrst', 'srvstat', 329, NULL, NULL, 'stat_io_reset', NULL, 'stat_io_reset',
  '{"class": "warning", "text": "IO stats reset was detected during report interval. Statistic values may be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "stat_io_reset",'
    '"columns": ['
      '{"caption": "Sample ID", "id": "sample_id", "class": "table_obj_name", '
        '"title": "Sample identifier with detected reset"},'
      '{"caption": "Object", "id": "object", "class": "table_obj_name", '
        '"title": "Target object of an I/O operation"},'
      '{"caption": "Backend", "id": "backend_type", "class": "table_obj_name", '
        '"title": "Type of backend (see stat_activity)"},'
      '{"caption": "Context", "id": "context", "class": "table_obj_name", '
        '"title": "The context of an I/O operation"},'
      '{"caption": "Reset time", "id": "stats_reset", "class": "table_obj_value", '
        '"title": "Date and time of the last reset performed in sample"}'
    ']'
    '}]'::jsonb),
(1, 'iostatmain', 'srvstat', 330, NULL, NULL, 'stat_io', NULL, 'stat_io', NULL,
'[{'
    '"type": "row_table",'
    '"source": "stat_io",'
    '"columns": ['
      '{"caption": "Object", "id": "object", "class": "table_obj_name", '
        '"title": "Target object of an I/O operation"},'
      '{"caption": "Backend", "id": "backend_type", "class": "table_obj_name", '
        '"title": "Type of backend (see stat_activity)"},'
      '{"caption": "Context", "id": "context", "class": "table_obj_name", '
        '"title": "The context of an I/O operation"},'
      '{"caption": "Reads", "columns": ['
        '{"caption": "Count", "id": "reads", "class": "table_obj_value", '
          '"title": "Number of read operations"},'
        '{"caption": "Bytes", "id": "read_sz", "class": "table_obj_value", '
          '"title": "Read data amount"},'
        '{"caption": "Time", "id": "read_time", "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in reading operation (seconds)"}'
     ']},'
      '{"caption": "Writes", "columns": ['
        '{"caption": "Count", "id": "writes", "class": "table_obj_value", '
          '"title": "Number of write operations"},'
        '{"caption": "Bytes", "id": "write_sz", "class": "table_obj_value", '
          '"title": "Written data amount"},'
        '{"caption": "Time", "id": "write_time", "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in writing operations (seconds)"}'
     ']},'
      '{"caption": "Writebacks", "columns": ['
        '{"caption": "Count", "id": "writebacks", "class": "table_obj_value", '
          '"title": "Number of blocks which the process requested the kernel write out to permanent storage"},'
        '{"caption": "Bytes", "id": "writeback_sz", "class": "table_obj_value", '
          '"title": "The amount of data requested for write out to permanent storage"},'
        '{"caption": "Time", "id": "writeback_time", "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in writeback operations (seconds)"}'
     ']},'
      '{"caption": "Extends", "columns": ['
        '{"caption": "Count", "id": "extends", "class": "table_obj_value", '
          '"title": "Number of relation extend operations"},'
        '{"caption": "Bytes", "id": "extend_sz", "class": "table_obj_value", '
          '"title": "The amount of space used by extend operations"},'
        '{"caption": "Time", "id": "extend_time", "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in extend operations (seconds)"}'
     ']},'
     '{"caption": "Hits", "id": "hits", "class": "table_obj_value", '
       '"title": "The number of times a desired block was found in a shared buffer"},'
     '{"caption": "Evictions", "id": "evictions", "class": "table_obj_value", '
       '"title": "Number of times a block has been written out from a shared or local buffer in order to make it available for another use"},'
     '{"caption": "Reuses", "id": "reuses", "class": "table_obj_value", '
       '"title": "The number of times an existing buffer in a size-limited ring buffer outside of shared buffers was reused as part of an I/O operation in the bulkread, bulkwrite, or vacuum contexts"},'
      '{"caption": "Fsyncs", "columns": ['
        '{"caption": "Count", "id": "fsyncs", "class": "table_obj_value", '
          '"title": "Number of fsync calls. These are only tracked in context normal"},'
        '{"caption": "Time", "id": "fsync_time", "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in fsync operations (seconds)"}'
     ']}'
    ']'
    '}]'::jsonb),
(1, 'slrustat', 'srvstat', 358, 'Cluster SLRU statistics', 'Cluster SLRU statistics', 'stat_slru', NULL, 'stat_slru', NULL, NULL),
(1, 'slrustatrst', 'slrustat', 359, NULL, NULL, 'stat_slru_reset', NULL, 'stat_slru_reset',
    '{"class": "warning", "text": "SLRU stats reset was detected during report interval. Statistic values may be affected"}',
    '[{'
    '"type": "row_table",'
    '"source": "stat_slru_reset",'
    '"columns": ['
      '{"caption": "Sample ID", "id": "sample_id", "class": "table_obj_name", '
        '"title": "Sample identifier with detected reset"},'
      '{"caption": "Name", "id": "name", "class": "table_obj_name", '
        '"title": "Name of the SLRU"},'
      '{"caption": "Reset time", "id": "stats_reset", "class": "table_obj_value", '
        '"title": "Date and time of the last reset performed in sample"}'
    ']'
    '}]'::jsonb),
(1, 'slrustatmain', 'slrustat', 360, NULL, NULL, 'stat_slru', NULL, 'stat_slru', NULL,
'[{'
    '"type": "row_table",'
    '"source": "stat_slru",'
    '"columns": ['
      '{"caption": "Name", "id": "name", "class": "table_obj_name", '
        '"title": "Name of the SLRU"},'
      '{"caption": "Zeroed", "id": "blks_zeroed", "class": "table_obj_value", '
        '"title": "Number of blocks zeroed during initializations"},'
      '{"caption": "Hits", "id": "blks_hit", "class": "table_obj_value", '
        '"title": "Number of times disk blocks were found already in the SLRU, so that a '
        'read was not necessary (this only includes hits in the SLRU, not the operating '
        'system''s file system cache)"},'
      '{"caption": "Reads", "id": "blks_read", "class": "table_obj_value", '
        '"title": "Number of disk blocks read for this SLRU"},'
      '{"caption": "%Hit", "id": "hit_pct", "class": "table_obj_value", '
        '"title": "Number of disk blocks hits for this SLRU as a percentage of reads + hits"},'
      '{"caption": "Writes", "id": "blks_written", "class": "table_obj_value", '
        '"title": "Number of disk blocks written for this SLRU"},'
      '{"caption": "Checked", "id": "blks_exists", "class": "table_obj_value", '
        '"title": "Number of blocks checked for existence for this SLRU (blks_exists field)"},'
      '{"caption": "Flushes", "id": "flushes", "class": "table_obj_value", '
        '"title": "Number of flushes of dirty data for this SLRU"},'
      '{"caption": "Truncates", "id": "truncates", "class": "table_obj_value", '
        '"title": "Number of truncates for this SLRU"}'
    ']'
    '}]'::jsonb),
(1, 'sesstat', 'srvstat', 400, 'Session statistics by database', 'Session statistics by database', 'sess_stats', NULL, 'db_stat_sessions', NULL,
'[{'
    '"type": "row_table",'
    '"source": "dbstat",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "Timings", "columns": ['
        '{"caption": "Total", "id": "session_time", "class": "table_obj_value", '
          '"title": "Time spent by database sessions in this database (note that statistic are only updated when the state of a session changes, so if sessions have been idle for a long time, this idle time will not be included)"},'
        '{"caption": "Active", "id": "active_time", "class": "table_obj_value", '
          '"title": "Time spent executing SQL statements in this database (this corresponds to the states active and fastpath function call in pg_stat_activity)"},'
        '{"caption": "Idle", "id": "idle_in_transaction_time", "class": "table_obj_value", '
          '"title": "Time spent idling while in a transaction in this database (this corresponds to the states idle in transaction and idle in transaction (aborted) in pg_stat_activity)"}'
     ']},'
     '{"caption": "Sessions", "columns": ['
        '{"caption": "Established", "id": "sessions", "class": "table_obj_value", '
          '"title": "Total number of sessions established to this database"},'
        '{"caption": "Abondoned", "id": "sessions_abandoned", "class": "table_obj_value", '
          '"title": "Number of database sessions to this database that were terminated because connection to the client was lost"},'
        '{"caption": "Fatal", "id": "sessions_fatal", "class": "table_obj_value", '
          '"title": "Number of database sessions to this database that were terminated by fatal errors"},'
        '{"caption": "Killed", "id": "sessions_killed", "class": "table_obj_value", '
          '"title": "Number of database sessions to this database that were terminated by operator intervention"}'
     ']}'
    ']'
    '}]'::jsonb),
(1, 'stmtstat', 'srvstat', 500, 'Statement statistics by database', 'Statement statistics by database', 'statstatements', NULL, 'st_stat', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "statements_dbstats",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
        '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
        '{"id": "calls", "class": "table_obj_value", "title": "Number of query executions", "caption": "Calls"}, '
        '{"caption": "Time (s)", "columns": ['
            '{"id": "total_plan_time", "class": "table_obj_value", "title": "Time spent planning queries", "caption": "Plan", "condition": "planning_times"}, '
            '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent executing queries", "caption": "Exec"}, '
            '{"id": "blk_read_time", "class": "table_obj_value", "title": "Time spent reading blocks", "caption": "Read"}, '
            '{"id": "blk_write_time", "class": "table_obj_value", "title": "Time spent writing blocks", "caption": "Write"}, '
            '{"id": "trg_fn_total_time", "class": "table_obj_value", "title": "Time spent in trigger functions", "caption": "Trg"}'
            ']}, '
        '{"caption": "Temp I/O Time", "condition": "statements_temp_io_times", "columns": ['
            '{"id": "temp_blk_read_time", "class": "table_obj_value", "title": "Time spent reading temporary file blocks, in seconds", "caption": "Read"}, '
            '{"id": "temp_blk_write_time", "class": "table_obj_value", "title": "Time spent writing temporary file blocks, in seconds", "caption": "Write"} '
            ']}, '
        '{"title": "Number of blocks fetched (hit + read)", "caption": "Fetched (blk)", "columns": ['
            '{"id": "shared_gets", "class": "table_obj_value", "caption": "Shared"}, '
            '{"id": "local_gets", "class": "table_obj_value", "caption": "Local"}'
            ']}, '
        '{"title": "Number of blocks dirtied", "caption": "Dirtied (blk)", "columns": ['
            '{"id": "shared_blks_dirtied", "class": "table_obj_value", "caption": "Shared"}, '
            '{"id": "local_blks_dirtied", "class": "table_obj_value", "caption": "Local"}'
            ']}, '
        '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
            '{"id": "temp_blks_read", "class": "table_obj_value", "caption": "Read"}, '
            '{"id": "temp_blks_written", "class": "table_obj_value", "caption": "Write"}'
            ']}, '
        '{"title": "Number of blocks, used for temporary tables", "caption": "Local (blk)", "columns": ['
            '{"id": "local_blks_read", "class": "table_obj_value", "caption": "Read"}, '
            '{"id": "local_blks_written", "class": "table_obj_value", "caption": "Write"}'
            ']}, '
        '{"id": "statements", "class": "table_obj_value", "caption": "Statements"}, '
        '{"id": "wal_bytes_fmt", "class": "table_obj_value", "caption": "WAL size", "condition": "statement_wal_bytes"}'
    ']}]'::jsonb),
(1, 'dbjitstat', 'srvstat', 550, 'JIT statistics by database', 'JIT statistics by database', 'statements_jit_stats', NULL, 'dbagg_jit_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "statements_dbstats",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of query executions", "caption": "Calls"}, '
      '{"caption": "Time", "columns": ['
          '{"id": "total_plan_time", "class": "table_obj_value", "title": "Time spent planning queries", "caption": "Plan"}, '
          '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent executing queries", "caption": "Exec"}'
          ']}, '
      '{"caption": "Generation", "columns": ['
          '{"id": "jit_functions", "class": "table_obj_value", "caption": "Count"}, '
          '{"id": "jit_generation_time", "class": "table_obj_value", "caption": "Gen. time"}'
          ']}, '
      '{"caption": "Inlining", "columns": ['
          '{"id": "jit_inlining_count", "class": "table_obj_value", "caption": "Count"}, '
          '{"id": "jit_inlining_time", "class": "table_obj_value", "caption": "Time"}'
          ']}, '
      '{"caption": "Optimization", "columns": ['
          '{"id": "jit_optimization_count", "class": "table_obj_value", "caption": "Count"}, '
          '{"id": "jit_optimization_time", "class": "table_obj_value", "caption": "Time"}'
          ']}, '
      '{"caption": "Emission", "columns": ['
          '{"id": "jit_emission_count", "class": "table_obj_value", "caption": "Count"}, '
          '{"id": "jit_emission_time", "class": "table_obj_value", "caption": "Time"}'
          ']}'
      ']'
    '}]'::jsonb),
(1, 'commonstat', 'srvstat', 600, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(1, 'clusterstat', 'commonstat', 650, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(1, 'clusthdr', 'clusterstat', 700, 'Cluster statistics', 'Cluster statistics', NULL, NULL, 'clu_stat', NULL, NULL),
(1, 'clustrst', 'clusterstat', 800, NULL, NULL, 'cluster_stats_reset', NULL, NULL,
  '{"class": "warning", "text": "Cluster statistics reset detected during report interval! Cluster statistics might be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "cluster_stats_reset",'
    '"columns": ['
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample"}, '
        '{"id": "bgwriter_stats_reset", "class": "table_obj_value", "caption": "BGWriter reset time"}, '
        '{"id": "archiver_stats_reset", "class": "table_obj_value", "caption": "Archiver reset time"}'
        ']'
    '}]'::jsonb),
(1, 'clust', 'clusterstat', 900, NULL, NULL, NULL, NULL, NULL, NULL,
  '[{'
    '"type": "column_table", '
    '"source": "cluster_stats",'
    '"columns": ['
        '{"caption": "Metric"}, '
        '{"caption": "Value"}'
    '],'
    '"rows": ['
        '{"caption": "Scheduled checkpoints", "cells": ['
            '{"id": "checkpoints_timed", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Requested checkpoints", "cells": ['
            '{"id": "checkpoints_req", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Checkpoint write time (s)", "cells": ['
            '{"id": "checkpoint_write_time", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Checkpoint sync time (s)", "cells": ['
            '{"id": "checkpoint_sync_time", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Checkpoint buffers written", "cells": ['
            '{"id": "buffers_checkpoint", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Background buffers written", "cells": ['
            '{"id": "buffers_clean", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Backend buffers written", "cells": ['
            '{"id": "buffers_backend", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Backend fsync count", "cells": ['
            '{"id": "buffers_backend_fsync", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Bgwriter interrupts (too many buffers)", "cells": ['
            '{"id": "maxwritten_clean", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Number of buffers allocated", "cells": ['
            '{"id": "buffers_alloc", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "WAL generated", "cells": ['
            '{"id": "wal_size_pretty", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "Start LSN", "cells": ['
            '{"id": "start_lsn", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "End LSN", "cells": ['
            '{"id": "end_lsn", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "WAL segments archived", "cells": ['
            '{"id": "archived_count", "class": "table_obj_value"}'
        ']}, '
        '{"caption": "WAL segments archive failed", "cells": ['
            '{"id": "failed_count", "class": "table_obj_value"}'
        ']}'
    ']'
  '}]'::jsonb),
(1, 'walstat', 'commonstat', 950, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(1, 'walsthdr', 'walstat', 1000, 'WAL statistics', 'WAL statistics', 'wal_stats', NULL, 'wal_stat', NULL, NULL),
(1, 'walstrst', 'walstat', 1100, NULL, NULL, 'wal_stats_reset', NULL, NULL,
  '{"class": "warning", "text": "WAL statistics reset detected during report interval! WAL statistics might be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "wal_stats_reset",'
    '"columns": ['
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample"},'
        '{"id": "wal_stats_reset", "class": "table_obj_value", "caption": "WAL stats reset time"}'
    ']}]'::jsonb),
(1, 'walst', 'walstat', 1200, NULL, NULL, 'wal_stats', NULL, NULL, NULL,
  '[{'
    '"type": "column_table",'
    '"source": "wal_stats",'
    '"columns": ['
        '{"caption": "Metric"},'
        '{"caption": "Value"}'
    '],'
    '"rows": ['
        '{"caption": "WAL generated", "title": "Total amount of WAL generated", "cells": ['
            '{"id": "wal_bytes_text", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL per second", "title": "Average amount of WAL generated per second", "cells": ['
            '{"id": "wal_bytes_per_sec", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL records", "title": "Total number of WAL records generated", "cells": ['
            '{"id": "wal_records", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL FPI", "title": "Total number of WAL full page images generated", "cells": ['
            '{"id": "wal_fpi", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL buffers full", "title": "Number of times WAL data was written to disk because WAL buffers became full", "cells": ['
            '{"id": "wal_buffers_full", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL writes", "title": "Number of times WAL buffers were written out to disk via XLogWrite request", "cells": ['
            '{"id": "wal_write", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL writes per second", "title": "Average number of times WAL buffers were written out to disk via XLogWrite request per second", "cells": ['
            '{"id": "wal_write_per_sec", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL sync", "title": "Number of times WAL files were synced to disk via issue_xlog_fsync request (if fsync is on and wal_sync_method is either fdatasync, fsync or fsync_writethrough, otherwise zero)", "cells": ['
            '{"id": "wal_sync", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL syncs per second", "title": "Average number of times WAL files were synced to disk via issue_xlog_fsync request per second", "cells": ['
            '{"id": "wal_sync_per_sec", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL write time (s)", "title": "Total amount of time spent writing WAL buffers to disk via XLogWrite request, in milliseconds (if track_wal_io_timing is enabled, otherwise zero). This includes the sync time when wal_sync_method is either open_datasync or open_sync", "cells": ['
            '{"id": "wal_write_time", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL write duty", "title": "WAL write time as a percentage of the report duration time", "cells": ['
            '{"id": "wal_write_time_per_sec", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL sync time (s)", "title": "Total amount of time spent syncing WAL files to disk via issue_xlog_fsync request, in milliseconds (if track_wal_io_timing is enabled, fsync is on, and wal_sync_method is either fdatasync, fsync or fsync_writethrough, otherwise zero)", "cells": ['
            '{"id": "wal_sync_time", "class": "table_obj_value"}'
        ']},'
        '{"caption": "WAL sync duty", "title": "WAL sync time as a percentage of the report duration time", "cells": ['
            '{"id": "wal_sync_time_per_sec", "class": "table_obj_value"}'
        ']}'
    ']'
  '}]'::jsonb),
(1, 'tbspst', 'srvstat', 1400, 'Tablespace statistics', 'Tablespace statistics', NULL, NULL, 'tablespace_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "tablespace_stats",'
    '"ordering": "tablespacename",'
    '"columns": ['
      '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
      '{"caption": "Path", "id": "tablespacepath", "class": "table_obj_value"},'
      '{"caption": "Size", "id": "size", "class": "table_obj_value", '
      '"title": "Tablespace size as it was at the moment of last sample in report interval"},'
      '{"caption": "Growth", "id": "size_delta", "class": "table_obj_value", '
      '"title": "Tablespace size increment during report interval"}'
    ']'
  '}]'::jsonb),
(1, 'wait_sampling_srvstats', 'srvstat', 1500, 'Wait sampling', 'Wait sampling', 'wait_sampling_tot', NULL, 'wait_sampling', NULL, NULL),
(1, 'wait_sampling_total', 'wait_sampling_srvstats', 100, 'Wait events types', 'Wait events types', 'wait_sampling_tot', NULL, 'wait_sampling_total', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "wait_sampling_total_stats",'
    '"ordering": "event_type_order",'
    '"columns": ['
      '{"id": "event_type", "class": "table_obj_name", "caption": "Wait event type"}, '
      '{"id": "stmt_waited", "class": "table_obj_value", "title": "Time, waited in events of wait event type executing statements in seconds", "caption": "Statements Waited (s)"}, '
      '{"id": "stmt_waited_pct", "class": "table_obj_value", "title": "Time, waited in events of wait event type as a percentage of total time waited in a cluster executing statements", "caption": "%Total"}, '
      '{"id": "tot_waited", "class": "table_obj_value", "title": "Time, waited in events of wait event type by all backends (including background activity) in seconds", "caption": "All Waited (s)"}, '
      '{"id": "tot_waited_pct", "class": "table_obj_value", "title": "Time, waited in events of wait event type as a percentage of total time waited in a cluster by all backends (including background activity)", "caption": "%Total"}'
    ']}]'::jsonb),
(1, 'wait_sampling_statements', 'wait_sampling_srvstats', 200, 'Top wait events (statements)', 'Top wait events (statements)', 'wait_sampling_tot', NULL, 'wt_smp_stmt',
 '{"class": "notice", "text": "Top wait events detected in statements execution"}',
  '[{'
    '"type": "row_table", '
    '"source": "wait_sampling_events",'
    '"filter": {"type": "exists", "field": "stmt_filter"},'
    '"ordering": "-stmt_waited",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "event_type", "class": "table_obj_name", "caption": "Wait event type"}, '
      '{"id": "event", "class": "table_obj_name", "caption": "Wait event"}, '
      '{"id": "stmt_waited", "class": "table_obj_value", "title": "Time, waited in event executing statements in seconds", "caption": "Waited (s)"}, '
      '{"id": "stmt_waited_pct", "class": "table_obj_value", "title": "Time, waited in event as a percentage of total time waited in a cluster executing statements", "caption": "%Total"}'
  ']}]'::jsonb),
(1, 'wait_sampling_all', 'wait_sampling_srvstats', 300, 'Top wait events (All)', 'Top wait events (All)', 'wait_sampling_tot', NULL, 'wt_smp_all',
 '{"class": "notice", "text": "Top wait events detected in all backends"}',
  '[{'
    '"type": "row_table", '
    '"source": "wait_sampling_events",'
    '"filter": {"type": "exists", "field": "total_filter"},'
    '"ordering": "-tot_waited",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "event_type", "class": "table_obj_name", "caption": "Wait event type"}, '
      '{"id": "event", "class": "table_obj_name", "caption": "Wait event"}, '
      '{"id": "tot_waited", "class": "table_obj_value", "title": "Time, waited in event by all backends (including background activity) in seconds", "caption": "Waited (s)"}, '
      '{"id": "tot_waited_pct", "class": "table_obj_value", "title": "Time, waited in event by all backends as a percentage of total time waited in a cluster by all backends (including background activity)", "caption": "%Total"}'
  ']}]'::jsonb)
;

-- Subsamples-based section of regular reports
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'db_activity_agg', 'actsesshdr', 100, 'Session state statistics by database', 'Session state statistics by database', 'act_backend', NULL, 'db_activity_agg',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "db_activity_agg",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "Summary", "columns": ['
        '{"caption": "Active", "id": "active_dur", "class": "table_obj_value", '
          '"title": "Summary time in ''active'' state"},'
        '{"caption": "Idle in xact", "id": "idle_xact_dur", "class": "table_obj_value", '
          '"title": "Summary time in ''idle in transaction'' state"},'
        '{"caption": "Idle in xact (A)", "id": "idle_xact_a_dur", "class": "table_obj_value", '
          '"title": "Summary time in ''idle in transaction (aborted)'' state"}'
     ']},'
      '{"caption": "Maximal", "columns": ['
        '{"caption": "Active", "id": "max_active_dur", "class": "table_obj_value", '
          '"title": "Longest detected ''active'' state"},'
        '{"caption": "Idle in xact", "id": "max_idle_xact_d", "class": "table_obj_value", '
          '"title": "Longest detected ''idle in xact'' state"},'
        '{"caption": "Idle in xact (A)", "id": "max_idle_xact_a_d", "class": "table_obj_value", '
          '"title": "Longest detected ''idle in transaction (aborted)'' state"},'
        '{"caption": "xact age", "id": "max_xact_age", "class": "table_obj_value", '
          '"title": "Maximal detected backend_xid age"}'
     ']}'
    ']'
    '}]'::jsonb),
(1, 'act_ix', 'actsesshdr', 200, 'Top ''idle in transaction'' states', 'Top ''idle in transaction'' session states by duration', 'act_ix', NULL, 'act_ix',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 1},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(1, 'act_ixa', 'actsesshdr', 300, 'Top ''idle in transaction (aborted)'' states', 'Top ''idle in transaction (aborted)'' session states by duration', 'act_ixa', NULL, 'act_ixa',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 2},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(1, 'act_active', 'actsesshdr', 400, 'Top ''active'' states', 'Top ''active'' session states by duration', 'act_active', NULL, 'act_active',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 3},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(1, 'act_age', 'actsesshdr', 500, 'Top states by transaction age', 'Top states by transaction age', 'act_backend', NULL, 'act_age',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_age",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_age", "value": true},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "Xact duration", "id": "xact_duration_format", "class": "table_obj_value"},'
      '{"caption": "Age", "id": "xmin_age", "class": "table_obj_value", '
        '"title": "Last detected age of backend_xmin"},'
      '{"caption": "State", "id": "state", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(1, 'act_xact_dur', 'actsesshdr', 600, 'Top states by transaction duration', 'Top states by transaction duration', 'act_backend', NULL, 'act_age_dur',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_xact",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_age", "value": true},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "Xact duration", "id": "xact_duration_format", "class": "table_obj_value"},'
      '{"caption": "Age", "id": "xmin_age", "class": "table_obj_value", '
        '"title": "Last detected age of backend_xmin"},'
      '{"caption": "State", "id": "state", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb)
;

-- Query section of regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'sqlela_t', 'sqlsthdr', 100, 'Top SQL by elapsed time', 'Top SQL by elapsed time', 'planning_times', NULL, 'top_ela', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_total_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "total_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "total_time_pct", "class": "table_obj_value", "title": "Elapsed time as a percentage of total cluster elapsed time", "caption": "%Total"}, '
      '{"caption": "Time (s)", "columns": ['
        '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed"}, '
        '{"id": "total_plan_time", "class": "table_obj_value", "title": "Time spent planning statement", "caption": "Plan"}, '
        '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec"}'
        ']}, '
      '{"id": "jit_total_time", "class": "jitTimeCell", "caption": "JIT, time (s)", "condition": "statements_jit_stats"}, '
      '{"class": "table_obj_name", "caption": "I/O time (s)", "condition": "io_times", "columns": ['
        '{"id": "blk_read_time", "class": "table_obj_value", "title": "Time spent reading blocks by statement", "caption": "Read"}, '
        '{"id": "blk_write_time", "class": "table_obj_value", "title": "Time spent writing blocks by statement", "caption": "Write"}'
        ']}, '
      '{"class": "table_obj_name", "caption": "CPU time (s)", "condition": "kcachestatements", "columns": ['
        '{"id": "user_time", "class": "table_obj_value", "caption": "Usr"}, '
        '{"id": "system_time", "class": "table_obj_value", "caption": "Sys"}'
        ']}, '
      '{"id": "plans", "class": "table_obj_value", "caption": "Plans", "title": "Number of times the statement was planned"}, '
      '{"id": "calls", "class": "table_obj_value", "caption": "Executions", "title": "Number of times the statement was executed"} '
    ']'
  '}]'::jsonb),
(1, 'sqlplan_t', 'sqlsthdr', 200, 'Top SQL by planning time', 'Top SQL by planning time', 'planning_times', NULL, 'top_plan', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_plan_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "total_plan_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "total_plan_time", "class": "table_obj_value", "title": "Time spent planning statement", "caption": "Plan elapsed (s)"}, '
      '{"id": "plan_time_pct", "class": "table_obj_value", "title": "Plan elapsed as a percentage of statement elapsed time", "caption": "%Elapsed"}, '
      '{"title": "Planning time statistics", "caption": "Plan times (ms)", "columns": ['
        '{"id": "mean_plan_time", "class": "table_obj_value", "caption": "Mean"}, '
        '{"id": "min_plan_time", "class": "table_obj_value", "caption": "Min"}, '
        '{"id": "max_plan_time", "class": "table_obj_value", "caption": "Max"}, '
        '{"id": "stddev_plan_time", "class": "table_obj_value", "caption": "StdErr"}'
      ']}, '
      '{"id": "plans", "class": "table_obj_value", "title": "Number of times the statement was planned", "caption": "Plans"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
      ']'
    '}]'::jsonb),
(1, 'sqlexec_t', 'sqlsthdr', 300, 'Top SQL by execution time', 'Top SQL by execution time', NULL, NULL, 'top_exec', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_exec_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "total_exec_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "caption": "Query ID", "class": "mono queryId"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec (s)"}, '
      '{"id": "exec_time_pct", "class": "table_obj_value", "title": "Exec time as a percentage of statement elapsed time", "caption": "%Elapsed", "condition": "planning_times"}, '
      '{"id": "total_exec_time_pct", "class": "table_obj_value", "title": "Exec time as a percentage of total cluster elapsed time", "caption": "%Total"}, '
      '{"id": "jit_total_time", "class": "jitTimeCell", "title": "Exec time as a percentage of statement elapsed time", "caption": "JIT time (s)", "condition": "statements_jit_stats"}, '
      '{"caption": "I/O time (s)", "condition": "io_times", "columns": ['
        '{"id": "blk_read_time", "class": "table_obj_value", "caption": "Read"}, '
        '{"id": "blk_write_time", "class": "table_obj_value", "caption": "Write"}'
      ']}, '
      '{"caption": "CPU time (s)", "condition": "kcachestatements", "columns": ['
        '{"id": "user_time", "class": "table_obj_value", "caption": "Usr"}, '
        '{"id": "system_time", "class": "table_obj_value", "caption": "Sys"}'
      ']}, '
      '{"id": "rows", "class": "table_obj_value", "caption": "Rows"}, '
      '{"title": "Execution time statistics", "caption": "Execution times (ms)", "columns": ['
        '{"id": "mean_exec_time", "class": "table_obj_value", "caption": "Mean"}, '
        '{"id": "min_exec_time", "class": "table_obj_value", "caption": "Min"}, '
        '{"id": "max_exec_time", "class": "table_obj_value", "caption": "Max"}, '
        '{"id": "stddev_exec_time", "class": "table_obj_value", "caption": "StdErr"}'
      ']}, '
      '{"id": "calls", "title": "Number of times the statement was executed", "caption": "Executions", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'sqlcalls', 'sqlsthdr', 400, 'Top SQL by executions', 'Top SQL by executions', NULL, NULL, 'top_calls', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"ordering": "ord_calls",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}, '
      '{"id": "calls_pct", "class": "table_obj_value", "title": "Executions of this statement as a percentage of total executions of all statements in a cluster", "caption": "%Total"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "mean_exec_time", "class": "table_obj_value", "caption": "Mean(ms)"}, '
      '{"id": "min_exec_time", "class": "table_obj_value", "caption": "Min(ms)"}, '
      '{"id": "max_exec_time", "class": "table_obj_value", "caption": "Max(ms)"}, '
      '{"id": "stddev_exec_time", "class": "table_obj_value", "caption": "StdErr(ms)"}, '
      '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}'
    ']'
  '}]'::jsonb),
(1, 'sqlio_t', 'sqlsthdr', 500, 'Top SQL by I/O wait time', 'Top SQL by I/O wait time', 'io_times', NULL, 'top_iowait', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"ordering": "ord_io_time",'
    '"filter": {"type": "exists", "field": "io_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "io_time", "class": "table_obj_value", "title": "Time spent by the statement reading and writing blocks", "caption": "IO(s)"}, '
      '{"id": "blk_read_time", "class": "table_obj_value", "title": "Time spent by the statement reading blocks", "caption": "R(s)"}, '
      '{"id": "blk_write_time", "class": "table_obj_value", "title": "Time spent by the statement writing blocks", "caption": "W(s)"}, '
      '{"id": "io_time_pct", "class": "table_obj_value", "title": "I/O time of this statement as a percentage of total I/O time for all statements in a cluster", "caption": "%Total"}, '
      '{"title": "Number of blocks read by the statement", "caption": "Reads", "columns": ['
        '{"id": "shared_blks_read", "title": "Number of shared blocks read by the statement", "caption": "Shr", "class": "table_obj_value"}, '
        '{"id": "local_blks_read", "title": "Number of local blocks read by the statement (usually used for temporary tables)", "caption": "Loc", "class": "table_obj_value"}, '
        '{"id": "temp_blks_read", "title": "Number of temp blocks read by the statement (usually used for operations like sorts and joins)", "caption": "Tmp", "class": "table_obj_value"}'
      ']}, '
      '{"title": "Number of blocks written by the statement", "caption": "Writes", "columns": ['
        '{"id": "shared_blks_written", "title": "Number of shared blocks written by the statement", "caption": "Shr", "class": "table_obj_value"}, '
        '{"id": "local_blks_written", "title": "Number of local blocks written by the statement (usually used for temporary tables)", "caption": "Loc", "class": "table_obj_value"}, '
        '{"id": "temp_blks_written", "title": "Number of temp blocks written by the statement (usually used for operations like sorts and joins)", "caption": "Tmp", "class": "table_obj_value"}'
      ']}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of blocks written by the statement", "caption": "Executions"}'
    ']}]'::jsonb),
(1, 'sqlfetch', 'sqlsthdr', 600, 'Top SQL by shared blocks fetched', 'Top SQL by shared blocks fetched', NULL, NULL, 'top_pgs_fetched', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_shared_blocks_fetched",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "shared_blks_fetched"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "shared_blks_fetched", "class": "table_obj_value", "title": "Shared blocks fetched (read and hit) by the statement", "caption": "blks fetched"}, '
      '{"id": "shared_blks_fetched_pct", "class": "table_obj_value", "title": "Shared blocks fetched by this statement as a percentage of all shared blocks fetched in a cluster", "caption": "%Total"}, '
      '{"id": "shared_hit_pct", "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']}]'::jsonb),
(1, 'sqlshrd', 'sqlsthdr', 700, 'Top SQL by shared blocks read', 'Top SQL by shared blocks read', NULL, NULL, 'top_shared_reads', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "shared_blks_read"},'
    '"ordering": "ord_shared_blocks_read",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "shared_blks_read", "class": "table_obj_value", "title": "Total number of shared blocks read by the statement", "caption": "Reads"}, '
      '{"id": "read_pct", "class": "table_obj_value", "title": "Shared blocks read by this statement as a percentage of all shared blocks read in a cluster", "caption": "%Total"}, '
      '{"id": "shared_hit_pct", "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']}]'::jsonb),
(1, 'sqlshdir', 'sqlsthdr', 800, 'Top SQL by shared blocks dirtied', 'Top SQL by shared blocks dirtied', NULL, NULL, 'top_shared_dirtied', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "shared_blks_dirtied"},'
    '"ordering": "ord_shared_blocks_dirt",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "shared_blks_dirtied", "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Dirtied"}, '
      '{"id": "dirtied_pct", "class": "table_obj_value", "title": "Shared blocks dirtied by this statement as a percentage of all shared blocks dirtied in a cluster", "caption": "%Total"}, '
      '{"id": "shared_blks_hit", "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Hits"}, '
      '{"id": "shared_hit_pct", "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "%Total"}, '
      '{"id": "wal_bytes_fmt", "class": "table_obj_value", "title": "Total amount of WAL bytes generated by the statement", "caption": "WAL", "condition": "statement_wal_bytes"}, '
      '{"id": "wal_bytes_pct", "class": "table_obj_value", "title": "WAL bytes of this statement as a percentage of total WAL bytes generated by a cluster", "caption": "%Total", "condition": "statement_wal_bytes"}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(1, 'sqlshwr', 'sqlsthdr', 900, 'Top SQL by shared blocks written', 'Top SQL by shared blocks written', NULL, NULL, 'top_shared_written', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "shared_blks_written"},'
    '"ordering": "ord_shared_blocks_written",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "shared_blks_written", "class": "table_obj_value", "title": "Total number of shared blocks written by the statement", "caption": "Written"}, '
      '{"id": "tot_written_pct", "class": "table_obj_value", "title": "Shared blocks written by this statement as a percentage of all shared blocks written in a cluster (sum of pg_stat_bgwriter fields buffers_checkpoint, buffers_clean and buffers_backend)", "caption": "%Total"}, '
      '{"id": "backend_written_pct", "class": "table_obj_value", "title": "Shared blocks written by this statement as a percentage total buffers written directly by a backends (buffers_backend of pg_stat_bgwriter view)", "caption": "%BackendW"}, '
      '{"id": "shared_hit_pct", "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(1, 'sqlwalsz', 'sqlsthdr', 1000, 'Top SQL by WAL size', 'Top SQL by WAL size', 'statement_wal_bytes', NULL, 'top_wal_bytes', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "wal_bytes"},'
    '"ordering": "ord_wal",'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "wal_bytes_fmt", "class": "table_obj_value", "title": "Total amount of WAL bytes generated by the statement", "caption": "WAL"}, '
      '{"id": "wal_bytes_pct", "class": "table_obj_value", "title": "WAL bytes of this statement as a percentage of total WAL bytes generated by a cluster", "caption": "%Total"}, '
      '{"id": "shared_blks_dirtied", "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Dirtied"}, '
      '{"id": "wal_fpi", "class": "table_obj_value", "title": "Total number of WAL full page images generated by the statement", "caption": "WAL FPI"}, '
      '{"id": "wal_records", "class": "table_obj_value", "title": "Total number of WAL records generated by the statement", "caption": "WAL records"}'
    ']'
  '}]'::jsonb),
(1, 'sqltmp', 'sqlsthdr', 1100, 'Top SQL by temp usage', 'Top SQL by temp usage', 'statements_top_temp', NULL, 'top_temp', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "sum_tmp_blks"},'
    '"ordering": "ord_temp",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"id": "local_blks_fetched", "class": "table_obj_value", "title": "Number of local blocks fetched (hit + read)", "caption": "Local fetched"}, '
      '{"id": "local_hit_pct", "class": "table_obj_value", "title": "Local blocks hit percentage", "caption": "Hits(%)"}, '
      '{"title": "Number of blocks, used for temporary tables", "caption": "Local (blk)", "columns": ['
        '{"id": "local_blks_written", "class": "table_obj_value", "title": "Number of written local blocks", "caption": "Write"}, '
        '{"id": "local_write_total_pct", "class": "table_obj_value", "title": "Percentage of all local blocks written", "caption": "%Total"}, '
        '{"id": "local_blks_read", "class": "table_obj_value", "title": "Number of read local blocks", "caption": "Read"}, '
        '{"id": "local_read_total_pct", "class": "table_obj_value", "title": "Percentage of all local blocks read", "caption": "%Total"}'
      ']}, '
      '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
        '{"id": "temp_blks_written", "class": "table_obj_value", "title": "Number of written temp blocks", "caption": "Write"}, '
        '{"id": "temp_write_total_pct", "class": "table_obj_value", "title": "Percentage of all temp blocks written", "caption": "%Total"}, '
        '{"id": "temp_blks_read", "class": "table_obj_value", "title": "Number of read temp blocks", "caption": "Read"}, '
        '{"id": "temp_read_total_pct", "class": "table_obj_value", "title": "Percentage of all temp blocks read", "caption": "%Total"}'
      ']}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(1, 'sqltmpiotime', 'sqlsthdr', 1125, 'Top SQL by temp I/O time', 'Top SQL by temp I/O time', 'statements_temp_io_times', NULL, 'top_temp_io_time', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_temp_io_time"},'
    '"ordering": "ord_temp_io_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"title": "Time the statement spent on temporary file blocks I/O", "caption": "Temp I/O time (s)", "columns": ['
        '{"id": "temp_blk_read_time", "class": "table_obj_value", "title": "Time the statement spent reading temporary file blocks, in seconds", "caption": "Read"}, '
        '{"id": "temp_blk_write_time", "class": "table_obj_value", "title": "Time the statement spent reading temporary file blocks, in seconds", "caption": "Write"}, '
        '{"id": "temp_io_time_pct", "class": "table_obj_value", "title": "Time spent on temporary file blocks I/O of this statement as a percentage of total time spent on temporary file blocks I/O by all statements", "caption": "%Total"} '
      ']}, '
      '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
        '{"id": "temp_blks_written", "class": "table_obj_value", "title": "Number of written temp blocks", "caption": "Write"}, '
        '{"id": "temp_write_total_pct", "class": "table_obj_value", "title": "Percentage of all temp blocks written", "caption": "%Total"}, '
        '{"id": "temp_blks_read", "class": "table_obj_value", "title": "Number of read temp blocks", "caption": "Read"}, '
        '{"id": "temp_read_total_pct", "class": "table_obj_value", "title": "Percentage of all temp blocks read", "caption": "%Total"}'
      ']}, '
      '{"id": "total_time", "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": "rows", "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": "calls", "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(1, 'sqljit', 'sqlsthdr', 1150, 'Top SQL by JIT elapsed time', 'Top SQL by JIT elapsed time', 'statements_jit_stats', NULL, 'top_jit', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "sum_jit_time"},'
    '"ordering": "ord_jit",'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "hexqueryid", "class": "mono jitCellId", "caption": "Query ID"}, '
        '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
        '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
        '{"id": "jit_total_time", "class": "table_obj_value", "title": "Time spent on JIT in seconds", "caption": "JIT total (s)"}, '
        '{"caption": "Generation", "columns": ['
          '{"id": "jit_functions", "class": "table_obj_value", "title": "Total number of functions JIT-compiled by the statement.", "caption": "Count"}, '
          '{"id": "jit_generation_time", "class": "table_obj_value", "title": "Time spent by the statement on generating JIT code, in seconds.", "caption": "Time (s)"}'
          ']}, '
        '{"class": "table_obj_name", "caption": "Inlining", "columns": ['
          '{"id": "jit_inlining_count", "class": "table_obj_value", "title": "Number of times functions have been inlined.", "caption": "Count"}, '
          '{"id": "jit_inlining_time", "class": "table_obj_value", "title": "Time spent by the statement on inlining functions, in seconds.", "caption": "Time (s)"}'
          ']}, '
        '{"class": "table_obj_name", "caption": "Optimization", "columns": ['
          '{"id": "jit_optimization_count", "class": "table_obj_value", "title": "Number of times the statement has been optimized.", "caption": "Count"}, '
          '{"id": "jit_optimization_time", "class": "table_obj_value", "title": "Time spent by the statement on optimizing, in seconds.", "caption": "Time (s)"}'
        ']}, '
        '{"class": "table_obj_name", "caption": "Emission", "columns": ['
          '{"id": "jit_emission_count", "class": "table_obj_value", "title": "Number of times code has been emitted.", "caption": "Count"}, '
          '{"id": "jit_emission_time", "class": "table_obj_value", "title": "Total time spent by the statement on emitting code, in seconds.", "caption": "Time (s)"}'
        ']}, '
        '{"class": "table_obj_name", "caption": "Time (s)", "columns": ['
          '{"id": "total_plan_time", "class": "table_obj_value", "title": "Time spent planning statement", "condition": "planning_times", "caption": "Plan"}, '
          '{"id": "total_exec_time", "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec"}'
        ']}, '
        '{"class": "table_obj_name", "caption": "I/O time (s)", "condition": "io_times", "columns": ['
          '{"id": "blk_read_time", "class": "table_obj_value", "title": "Time spent reading blocks by statement", "caption": "Read"}, '
          '{"id": "blk_write_time", "class": "table_obj_value", "title": "Time spent writing blocks by statement", "caption": "Write"}'
        ']}'
    ']}]'::jsonb),
(1, 'sqlkcachehdr', 'sqlsthdr', 1200, 'rusage statistics', 'rusage statistics', 'kcachestatements', NULL, 'kcache_stat', NULL, NULL),
(1, 'sqlrusgcpu_t', 'sqlkcachehdr', 100, 'Top SQL by system and user time', 'Top SQL by system and user time', NULL, NULL, 'kcache_time', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_rusage_statements",'
    '"filter": {"type": "exists", "field": "sum_cpu_time"},'
    '"ordering": "ord_cpu_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"}, '
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
      '{"id": "username", "class": "table_obj_name", "caption": "User"}, '
      '{"title": "Userspace CPU", "caption": "User Time", "columns": ['
        '{"id": "plan_user_time", "class": "table_obj_value", "title": "User CPU time elapsed during planning", "caption": "Plan (s)", "condition": "rusage_planstats"}, '
        '{"id": "exec_user_time", "class": "table_obj_value", "title": "User CPU time elapsed during execution", "caption": "Exec (s)"}, '
        '{"id": "user_time_pct", "class": "table_obj_value", "title": "User CPU time elapsed by this statement as a percentage of total user CPU time", "caption": "%Total"}'
        ']}, '
      '{"title": "Kernelspace CPU", "caption": "System Time", "columns": ['
        '{"id": "plan_system_time", "class": "table_obj_value", "title": "System CPU time elapsed during planning", "caption": "Plan (s)", "condition": "rusage_planstats"}, '
        '{"id": "exec_system_time", "class": "table_obj_value", "title": "System CPU time elapsed during execution", "caption":"Exec (s)"}, '
        '{"id": "system_time_pct", "class": "table_obj_value", "title": "System CPU time elapsed by this statement as a percentage of total system CPU time", "caption": "%Total"}'
        ']}'
      ']'
    '}]'::jsonb),
(1, 'sqlrusgio', 'sqlkcachehdr', 200, 'Top SQL by reads/writes done by filesystem layer', 'Top SQL by reads/writes done by filesystem layer', NULL, NULL, 'kcache_reads_writes', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_rusage_statements",'
    '"filter": {"type": "exists", "field": "sum_io_bytes"},'
    '"ordering": "ord_io_bytes",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryId", "caption": "Query ID"},'
      '{"id": "dbname", "class": "table_obj_name", "caption": "Database"},'
      '{"id": "username", "class": "table_obj_name", "caption": "User"},'
      '{"title": "Filesystem reads", "caption": "Read Bytes", "columns": ['
          '{"id": "plan_reads", "class": "table_obj_value", "title": "Filesystem read amount during planning", "caption": "Plan", "condition": "rusage_planstats"},'
          '{"id": "exec_reads", "class": "table_obj_value", "title": "Filesystem read amount during execution", "caption": "Bytes"},'
          '{"id": "reads_total_pct", "class": "table_obj_value", "title": "Filesystem read amount of this statement as a percentage of all statements FS read amount", "caption": "%Total"}'
      ']},'
      '{"title": "Filesystem writes", "caption": "Write Bytes", "columns": ['
          '{"id": "plan_writes", "class": "table_obj_value", "title": "Filesystem write amount during planning", "caption": "Plan", "condition": "rusage_planstats"},'
          '{"id": "exec_writes", "class": "table_obj_value", "title": "Filesystem write amount during execution", "caption": "Bytes"},'
          '{"id": "writes_total_pct", "class": "table_obj_value", "title": "Filesystem write amount of this statement as a percentage of all statements FS read amount", "caption": "%Total"}'
      ']}'
    ']'
  '}]'::jsonb),
(1, 'sqllist', 'sqlsthdr', 1300, 'Complete list of SQL texts', 'Complete list of SQL texts', NULL, NULL, 'sql_list', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "queries",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0}'
    '],'
    '"columns": ['
      '{"id": "hexqueryid", "class": "mono queryTextId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "query_texts", "class": "mono queryText", "caption": "Query Text"}'
    ']'
  '}]'::jsonb)
;

-- Schema objects section of a regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'tblscan', 'objects', 100, 'Top tables by estimated sequentially scanned volume', 'Top tables by estimated sequentially scanned volume', NULL, NULL, 'scanned_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_seq_scan"},'
    '"ordering": "ord_seq_scan",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": ["relname", "toastrelname"], "class": "table_obj_name"},'
     '{"caption": "~SeqBytes", "id": ["seqscan_bytes_pretty", "t_seqscan_bytes_pretty"], '
       '"class": "table_obj_value", '
       '"title": "Estimated number of bytes, fetched by sequential scans"}, '
     '{"caption": "SeqScan", "id": ["seq_scan", "toastseq_scan"], '
       '"class": "table_obj_value", '
       '"title": "Number of sequential scans initiated on this table"},'
     '{"caption": "IxScan", "id": ["idx_scan", "toastidx_scan"], '
       '"class": "table_obj_value", '
       '"title": "Number of index scans initiated on this table"},'
     '{"caption": "IxFet", "id": ["idx_tup_fetch", "toastidx_tup_fetch"], '
       '"class": "table_obj_value", '
       '"title": "Number of live rows fetched by index scans"},'
     '{"caption": "Ins", "id": ["n_tup_ins", "toastn_tup_ins"], '
       '"class": "table_obj_value", '
       '"title": "Number of rows inserted"},'
     '{"caption": "Upd", "id": ["n_tup_upd", "toastn_tup_upd"], '
       '"class": "table_obj_value", '
       '"title": "Number of rows updated (includes HOT updated rows)"},'
     '{"caption": "Del", "id": ["n_tup_del", "toastn_tup_del"], '
       '"class": "table_obj_value", '
       '"title": "Number of rows deleted"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd", "toastn_tup_hot_upd"], '
       '"class": "table_obj_value", '
       '"title": "Number of rows HOT updated (i.e., with no separate index update required)"}'
    ']'
  '}]'::jsonb),
(1, 'tblfetch', 'objects', 200, 'Top tables by blocks fetched', 'Top tables by blocks fetched', NULL, NULL, 'fetch_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_tables",'
    '"filter": {"type": "exists", "field": "ord_fetch"},'
    '"ordering": "ord_fetch",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
    '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
    '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
    '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
    '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
    '{"caption": "Heap", "columns": ['
         '{"caption": "Blks", "id": "heap_blks_fetch", "title": "Number of blocks fetched (read+hit) from this table", "class": "table_obj_value"},'
         '{"caption": "%Total", "id": "heap_blks_proc_pct", "title": "Heap blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']},'
    '{"caption": "Ix", "columns": ['
         '{"caption": "Blks", "id": "idx_blks_fetch", "title": "Number of blocks fetched (read+hit) from all indexes on this table", "class": "table_obj_value"},'
         '{"caption": "%Total", "id": "idx_blks_fetch_pct", "title": "Indexes of blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']},'
    '{"caption": "TOAST", "columns": ['
         '{"caption": "Blks", "id": "toast_blks_fetch", "title": "Number of blocks fetched (read+hit) from this table''s TOAST table (if any)", "class": "table_obj_value"},'
         '{"caption": "%Total", "id": "toast_blks_fetch_pct", "title": "TOAST blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']},'
    '{"caption": "TOAST-Ix", "columns": ['
         '{"caption": "Blks", "id": "tidx_blks_fetch", "title": "Number of blocks fetched (read+hit) from this table''s TOAST table indexes (if any)", "class": "table_obj_value"},'
         '{"caption": "%Total", "id": "tidx_blks_fetch_pct", "title": "TOAST table index blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']}'
   ']'
  '}]'::jsonb),
(1, 'tblrd', 'objects', 300, 'Top tables by blocks read', 'Top tables by blocks read', NULL, NULL, 'read_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_tables",'
    '"filter": {"type": "exists", "field": "ord_read"},'
    '"ordering": "ord_read",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Heap", "columns": ['
          '{"caption": "Blks", "id": "heap_blks_read", "title": "Number of blocks read from this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": "heap_blks_read_pct", "title": "Heap blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Ix", "columns": ['
          '{"caption": "Blks", "id": "idx_blks_read", "title": "Number of blocks read from all indexes on this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": "idx_blks_read_pct", "title": "Indexes of blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST", "columns": ['
          '{"caption": "Blks", "id": "toast_blks_read", "title": "Number of blocks read from this table''s TOAST table (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": "toast_blks_read_pct", "title": "TOAST blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST-Ix", "columns": ['
          '{"caption": "Blks", "id": "tidx_blks_read", "title": "Number of blocks read from this table''s TOAST table indexes (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": "tidx_blks_read_pct", "title": "TOAST table index blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Hit(%)", "id": "hit_pct", "class": "table_obj_value", "title": "Number of heap, indexes, toast and toast index blocks fetched from shared buffers as a percentage of all their blocks fetched from shared buffers and file system"}'
    ']'
  '}]'::jsonb),
(1, 'tbldml', 'objects', 400, 'Top DML tables', 'Top DML tables', NULL, NULL, 'dml_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_dml"},'
    '"ordering": "ord_dml",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "toastrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": ["relname", "toastrelname"], "class": "table_obj_name"},'
     '{"caption": "Ins", "id": ["n_tup_ins", "toastn_tup_ins"], "title": "Number of rows inserted", "class": "table_obj_value"},'
     '{"caption": "Upd", "id": ["n_tup_upd", "toastn_tup_upd"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Del", "id": ["n_tup_del", "toastn_tup_del"], "title": "Number of rows deleted", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd", "toastn_tup_hot_upd"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"},'
     '{"caption": "SeqScan", "id": ["seq_scan", "toastseq_scan"], "title": "Number of live rows fetched by sequential scans", "class": "table_obj_value"},'
     '{"caption": "SeqFet", "id": ["seq_tup_read", "toastseq_tup_read"], "title": "Number of live rows fetched by sequential scans", "class": "table_obj_value"},'
     '{"caption": "IxScan", "id": ["idx_scan", "toastidx_scan"], "title": "Number of index scans initiated on this table", "class": "table_obj_value"},'
     '{"caption": "IxFet", "id": ["idx_tup_fetch", "toastidx_tup_fetch"], "title": "Number of live rows fetched by index scans", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'tblud', 'objects', 500, 'Top tables by updated/deleted tuples', 'Top tables by updated/deleted tuples', NULL, NULL, 'upd_del_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_upd"},'
    '"ordering": "ord_upd",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "toastrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": ["relname", "toastrelname"], "class": "table_obj_name"},'
     '{"caption": "Upd", "id": ["n_tup_upd", "toastn_tup_upd"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd", "toastn_tup_hot_upd"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"},'
     '{"caption": "Del", "id": ["n_tup_del", "toastn_tup_del"], "title": "Number of rows deleted", "class": "table_obj_value"},'
     '{"caption": "Vacuum", "id": ["vacuum_count", "toastvacuum_count"], "title": "Number of times this table has been manually vacuumed (not counting VACUUM FULL)", "class": "table_obj_value"},'
     '{"caption": "AutoVacuum", "id": ["autovacuum_count", "toastautovacuum_count"], "title": "Number of times this table has been vacuumed by the autovacuum daemon", "class": "table_obj_value"},'
     '{"caption": "Analyze", "id": ["analyze_count", "toastanalyze_count"], "title": "Number of times this table has been manually analyzed", "class": "table_obj_value"},'
     '{"caption": "AutoAnalyze", "id": ["autoanalyze_count", "toastautoanalyze_count"], "title": "Number of times this table has been analyzed by the autovacuum daemon", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'tblupd_np', 'objects', 550, 'Top tables by new-page updated tuples', 'Top tables by new-page updated tuples', 'table_new_page_updates', NULL, 'upd_np_tbl',
  '{"class": "notice", "text": "Top tables by number of rows updated where the successor version goes onto a new heap page, '
  'leaving behind an original version with a t_ctid field that points to a different heap page. '
  'These are always non-HOT updates"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_upd_np"},'
    '"ordering": "ord_upd_np",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "toastrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": ["relname", "toastrelname"], "class": "table_obj_name"},'
     '{"caption": "NP Upd", "id": ["n_tup_newpage_upd", "toastn_tup_newpage_upd"], "title": "Number of rows updated to a new heap page", "class": "table_obj_value"},'
     '{"caption": "%Upd", "id": ["np_upd_pct", "toastnp_upd_pct"], "title": "Number of new-page updated rows as a percentage of all rows updated", "class": "table_obj_value"},'
     '{"caption": "Upd", "id": ["n_tup_upd", "toastn_tup_upd"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd", "toastn_tup_hot_upd"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'tblgrw', 'objects', 600, 'Top growing tables', 'Top growing tables', NULL, NULL, 'growth_tbl',
  '{"class": "notice", "text": "Sizes in square brackets are based on pg_class.relpages data instead of pg_relation_size() function"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_growth"},'
    '"ordering": "ord_growth",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "toastrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": ["relname", "toastrelname"], "class": "table_obj_name"},'
     '{"caption": "Size", "id": ["relsize_pretty", "t_relsize_pretty"], "title": "Table size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
     '{"caption": "Growth", "id": ["growth_pretty", "toastgrowth_pretty"], "title": "Table size increment during report interval", "class": "table_obj_value"},'
     '{"caption": "Ins", "id": ["n_tup_ins", "toastn_tup_ins"], "title": "Number of rows inserted", "class": "table_obj_value"},'
     '{"caption": "Upd", "id": ["n_tup_upd", "toastn_tup_upd"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Del", "id": ["n_tup_del", "toastn_tup_del"], "title": "Number of rows deleted", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd", "toastn_tup_hot_upd"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'ixfetch', 'objects', 700, 'Top indexes by blocks fetched', 'Top indexes by blocks fetched', NULL, NULL, 'fetch_idx', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_indexes",'
    '"filter": {"type": "exists", "field": "ord_fetch"},'
    '"ordering": "ord_fetch",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Index", "id": "indexrelname", "class": "table_obj_name"},'
     '{"caption": "Scans", "id": "idx_scan", "title": "Number of scans performed on index", "class": "table_obj_value"},'
     '{"caption": "Blks", "id": "idx_blks_fetch", "title": "Number of blocks fetched (read+hit) from this index", "class": "table_obj_value"},'
     '{"caption": "%Total", "id": "idx_blks_fetch_pct", "title": "Blocks fetched from this index as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'ixrd', 'objects', 800, 'Top indexes by blocks read', 'Top indexes by blocks read', NULL, NULL, 'read_idx', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_indexes",'
    '"filter": {"type": "exists", "field": "ord_read"},'
    '"ordering": "ord_read",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Index", "id": "indexrelname", "class": "table_obj_name"},'
     '{"caption": "Scans", "id": "idx_scan", "title": "Number of scans performed on index", "class": "table_obj_value"},'
     '{"caption": "Blks Reads", "id": "idx_blks_read", "title": "Number of disk blocks read from this index", "class": "table_obj_value"},'
     '{"caption": "%Total", "id": "idx_blks_read_pct", "title": "Blocks fetched from this index as a percentage of all blocks read in a cluster", "class": "table_obj_value"}, '
     '{"caption": "Hits(%)", "id": "idx_blks_hit_pct", "title": "Index blocks buffer cache hit percentage", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(1, 'ixgrw', 'objects', 900, 'Top growing indexes', 'Top growing indexes', NULL, NULL, 'growth_idx',
  '{"class": "notice", "text": "Sizes in square brackets are based on pg_class.relpages data instead of pg_relation_size() function"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_indexes",'
    '"filter": {"type": "exists", "field": "ord_growth"},'
    '"ordering": "ord_growth",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Index", "id": "indexrelname", "class": "table_obj_name"},'
     '{"caption": "Index", "columns": ['
          '{"id": "indexrelsize_pretty", "caption": "Size", "title": "Index size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
          '{"id": "growth_pretty", "caption": "Growth", "title": "Index size increment during report interval", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Table", "columns": ['
          '{"id": "tbl_n_tup_ins", "caption": "Ins", "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"id": "tbl_n_tup_upd", "caption": "Upd", "title": "Number of rows updated (without HOT updated rows)", "class": "table_obj_value"},'
          '{"id": "tbl_n_tup_del", "caption": "Del", "title": "Number of rows deleted", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb),
(1, 'ixunused', 'objects', 1000, 'Unused indexes', 'Unused indexes', NULL, NULL, 'ix_unused',
  '{"class": "notice", "text": "This table contains non-scanned indexes (during report interval), ordered by number of DML '
  'operations on underlying tables. Constraint indexes are excluded"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_indexes",'
    '"filter": {"type": "exists", "field": "ord_unused"},'
    '"ordering": "ord_unused",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Index", "id": "indexrelname", "class": "table_obj_name"},'
     '{"caption": "Index", "columns": ['
          '{"id": "indexrelsize_pretty", "caption": "Size", "title": "Index size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
          '{"id": "growth_pretty", "caption": "Growth", "title": "Index size increment during report interval", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Table", "columns": ['
          '{"id": "tbl_n_tup_ins", "caption": "Ins", "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"id": "tbl_n_tup_upd", "caption": "Upd", "title": "Number of rows updated (without HOT updated rows)", "class": "table_obj_value"},'
          '{"id": "tbl_n_tup_del", "caption": "Del", "title": "Number of rows deleted", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb)
;

-- Functions section of a regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'func_t', 'funchdr', 100, 'Top functions by total time', 'Top functions by total time', NULL, NULL, 'funcs_time_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_time"},'
    '"ordering": "ord_time",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Function", "id": "funcname", "class": "table_obj_name"},'
     '{"caption": "Executions", "id": "calls", "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": "total_time", "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": "self_time", "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": "m_time", "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": "m_stime", "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb),
(1, 'func_c', 'funchdr', 200, 'Top functions by executions', 'Top functions by executions', NULL, NULL, 'funcs_calls_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_calls"},'
    '"ordering": "ord_calls",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Function", "id": "funcname", "class": "table_obj_name"},'
     '{"caption": "Executions", "id": "calls", "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": "total_time", "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": "self_time", "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": "m_time", "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": "m_stime", "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb),
(1, 'func_trg', 'funchdr', 300, 'Top trigger functions by total time', 'Top trigger functions by total time', 'trigger_function_stats', NULL, 'trg_funcs_time_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_trgtime"},'
    '"ordering": "ord_trgtime",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr"},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr"},'
     '{"caption": "Function", "id": "funcname", "class": "hdr"},'
     '{"caption": "Executions", "id": "calls", "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": "total_time", "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": "self_time", "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": "m_time", "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": "m_stime", "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb)
;

-- Vacuum section of a regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(1, 'vacops', 'vachdr', 100, 'Top tables by vacuum operations', 'Top tables by vacuum operations', NULL, NULL, 'top_vacuum_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_vac"},'
    '"ordering": "ord_vac",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Vacuum count", "id": "vacuum_count", "class": "table_obj_value", "title": "Number of times this table has been manually vacuumed (not counting VACUUM FULL)"},'
     '{"caption": "Autovacuum count", "id": "autovacuum_count", "class": "table_obj_value", "title": "Number of times this table has been vacuumed by the autovacuum daemon"},'
     '{"caption": "Ins", "id": "n_tup_ins", "class": "table_obj_value", "title": "Number of rows inserted"},'
     '{"caption": "Upd", "id": "n_tup_upd", "class": "table_obj_value", "title": "Number of rows updated (includes HOT updated rows)"},'
     '{"caption": "Del", "id": "n_tup_del", "class": "table_obj_value", "title": "Number of rows deleted"},'
     '{"caption": "Upd(HOT)", "id": "n_tup_hot_upd", "class": "table_obj_value", "title": "Number of rows HOT updated (i.e., with no separate index update required)"}'
    ']'
  '}]'::jsonb),
(1, 'anops', 'vachdr', 200, 'Top tables by analyze operations', 'Top tables by analyze operations', NULL, NULL, 'top_analyze_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_anl"},'
    '"ordering": "ord_anl",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Analyze count", "id": "analyze_count", "class": "table_obj_value", "title": "Number of times this table has been manually analyzed"},'
     '{"caption": "Autoanalyze count", "id": "autoanalyze_count", "class": "table_obj_value", "title": "Number of times this table has been analyzed by the autovacuum daemon"},'
     '{"caption": "Ins", "id": "n_tup_ins", "class": "table_obj_value", "title": "Number of rows inserted"},'
     '{"caption": "Upd", "id": "n_tup_upd", "class": "table_obj_value", "title": "Number of rows updated (includes HOT updated rows)"},'
     '{"caption": "Del", "id": "n_tup_del", "class": "table_obj_value", "title": "Number of rows deleted"},'
     '{"caption": "Upd(HOT)", "id": "n_tup_hot_upd", "class": "table_obj_value", "title": "Number of rows HOT updated (i.e., with no separate index update required)"}'
    ']}]'::jsonb),
(1, 'ixvacest', 'vachdr', 300, 'Top indexes by estimated vacuum load', 'Top indexes by estimated vacuum load', NULL, NULL, 'top_ix_vacuum_bytes_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_indexes",'
    '"filter": {"type": "exists", "field": "ord_vac"},'
    '"ordering": "ord_vac",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Index", "id": "indexrelname", "class": "table_obj_name"},'
     '{"id": "vacuum_bytes_pretty", "caption": "Vacuum bytes", "class": "table_obj_value", "title": "Estimated implicit vacuum load caused by table indexes"},'
     '{"id": "vacuum_count", "caption": "Vacuum cnt", "class": "table_obj_value", "title": "Vacuum count on underlying table"},'
     '{"id": "autovacuum_count", "caption": "Autovacuum cnt", "class": "table_obj_value", "title": "Autovacuum count on underlying table"},'
     '{"id": "avg_indexrelsize_pretty", "caption": "IX size", "class": "table_obj_value", "title": "Average index size during report interval"},'
     '{"id": "avg_relsize_pretty", "caption": "Relsize", "class": "table_obj_value", "title": "Average relation size during report interval"}'
    ']'
  '}]'::jsonb),
(1, 'tblbydead', 'vachdr', 400, 'Top tables by dead tuples ratio', 'Top tables by dead tuples ratio', 'top_tables_dead', NULL, 'dead_tbl',
  '{"class": "notice", "text": "Data in this section is not differential. This data is valid for last report sample only"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tbl_last_sample",'
    '"filter": {"type": "exists", "field": "ord_dead"},'
    '"ordering": "ord_dead",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Live", "id": "n_live_tup", "class": "table_obj_value", "title": "Estimated number of live rows"},'
     '{"caption": "Dead", "id": "n_dead_tup", "class": "table_obj_value", "title": "Estimated number of dead rows"},'
     '{"caption": "%Dead", "id": "dead_pct", "class": "table_obj_value", "title": "Dead rows count as a percentage of total rows count"},'
     '{"caption": "Last AV", "id": "last_autovacuum", "class": "table_obj_value", "title": "Last autovacuum ran time"},'
     '{"caption": "Size", "id": "relsize_pretty", "class": "table_obj_value", "title": "Table size without indexes and TOAST"}'
    ']'
  '}]'::jsonb),
(1, 'tblbymod', 'vachdr', 500, 'Top tables by modified tuples ratio', 'Top tables by modified tuples ratio', 'top_tables_mods', NULL, 'mod_tbl',
  '{"class": "notice", "text": "Data in this section is not differential. This data is valid for last report sample only"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tbl_last_sample",'
    '"filter": {"type": "exists", "field": "ord_mod"},'
    '"ordering": "ord_mod",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "table_obj_name"},'
     '{"caption": "Schema", "id": "schemaname", "class": "table_obj_name"},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "table_obj_name"},'
     '{"caption": "Table", "id": "relname", "class": "table_obj_name"},'
     '{"caption": "Live", "id": "n_live_tup", "class": "table_obj_value", "title": "Estimated number of live rows"},'
     '{"caption": "Dead", "id": "n_dead_tup", "class": "table_obj_value", "title": "Estimated number of dead rows"},'
     '{"caption": "Mod", "id": "n_mod_since_analyze", "class": "table_obj_value", "title": "Estimated number of rows modified since this table was last analyzed"},'
     '{"caption": "%Mod", "id": "mods_pct", "class": "table_obj_value", "title": "Modified rows of the table as a percentage of all rows in the table"},'
     '{"caption": "Last AA", "id": "last_autoanalyze", "class": "table_obj_value", "title": "Last autoanalyze ran time"},'
     '{"caption": "Size", "id": "relsize_pretty", "class": "table_obj_value", "title": "Table size without indexes and TOAST"}'
    ']'
  '}]'::jsonb)
;

-- Settings sections
INSERT INTO report_struct(
    report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
    content, sect_struct)
VALUES
(1, 'definedset', 'settings', 800, NULL, NULL, NULL, NULL, NULL, NULL,
  '[{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "defined_val"},'
    '"columns": ['
     '{"caption": "Defined settings", "columns": ['
          '{"caption": "Setting", "id": "name", "class": "table_obj_value"},'
          '{"caption": "reset_val", "id": "reset_val", "class": "table_obj_value switch_bold"},'
          '{"caption": "Unit", "id": "unit", "class": "table_obj_value"},'
          '{"caption": "Source", "id": "source", "class": "table_obj_value"},'
          '{"caption": "Notes", "id": "notes", "class": "table_obj_value switch_bold"}'
     ']}'
    ']'
  '},'
  '{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "default_val"},'
    '"columns": ['
     '{"caption": "Default settings", "columns": ['
          '{"caption": "Setting", "id": "name", "class": "table_obj_value"},'
          '{"caption": "reset_val", "id": "reset_val", "class": "table_obj_value switch_bold"},'
          '{"caption": "Unit", "id": "unit", "class": "table_obj_value"},'
          '{"caption": "Source", "id": "source", "class": "table_obj_value"},'
          '{"caption": "Notes", "id": "notes", "class": "table_obj_value switch_bold"}'
     ']}'
    ']'
  '}]'::jsonb)
;

-- Schema objects section of a differential report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'rep_details', NULL, 10, 'Report details', 'Report details', NULL, NULL, 'rep_details', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "properties",'
    '"columns": ['
        '{"id": "pgprofile_version", "class": "hdr table_obj_value", "caption": "Version", "rowspan": true},'
        '{"id": "server_name", "class": "hdr table_obj_value", "caption": "Server name", "rowspan": true},'
        '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"},'
        '{"caption": "Interval (sample)", "columns": ['
          '{"id": ["start1_id", "start2_id"], "class": "table_obj_value", "caption": "start"},'
          '{"id": ["end1_id", "end2_id"], "class": "table_obj_value", "caption": "end"}'
        ']},'
        '{"caption": "Interval (time)", "columns": ['
          '{"id": ["report_start1", "report_start2"], "class": "table_obj_value", "caption": "start"},'
          '{"id": ["report_end1", "report_end2"], "class": "table_obj_value", "caption": "end"}'
        ']}'
    ']}]'::jsonb),
(2, 'rep_settings', NULL, 11, NULL, NULL, NULL, NULL, 'rep_settings', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "h_ord"},'
    '"ordering": "h_ord",'
    '"columns": ['
      '{"caption": "Setting", "id": "name", "class": "table_obj_name"},'
      '{"caption": "Value", "id": "reset_val", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'stmt_cmt1', NULL, 100, NULL, NULL, 'stmt_cnt_range', NULL, NULL,
  '{"class": "warning", "text": "Report interval contains sample(s) with captured statements count more than 90% of pg_stat_statements.max. '
  'Consider increasing pg_stat_statements.max parameter."}',
  '[{'
    '"type": "row_table", '
    '"source": "stmt_cnt_range",'
    '"ordering": "ord",'
    '"columns": ['
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample ID"}, '
        '{"id": "sample_time", "class": "table_obj_value", "caption": "Sample Time"}, '
        '{"id": "stmt_cnt", "class": "table_obj_value", "caption": "Stmts Captured"}, '
        '{"id": "max_cnt", "class": "table_obj_value", "caption": "pg_stat_statements.max"}'
    ']}]'::jsonb),
(2, 'srvstat', NULL, 300, 'Server statistics', 'Server statistics', NULL, NULL, 'cl_stat', NULL, NULL),
(2, 'actsesshdr', NULL, 350, 'Session states', 'Session states observed by subsamples', 'act_backend', NULL, 'stat_act',
  '{"class": "notice", "text": "Statistics about session states exceeding capturing thresholds."}', NULL),
(2, 'sqlsthdr', NULL, 400, 'SQL query statistics', 'SQL query statistics', 'statstatements', NULL, 'sql_stat', NULL, NULL),
(2, 'objects', NULL, 500, 'Schema object statistics', 'Schema object statistics', NULL, NULL, 'schema_stat', NULL, NULL),
(2, 'funchdr', NULL, 600, 'User function statistics', 'User function statistics', 'function_stats', NULL, 'func_stat', NULL, NULL),
(2, 'vachdr', NULL, 700, 'Vacuum-related statistics', 'Vacuum-related statistics', NULL, NULL, 'vacuum_stats', NULL, NULL),
(2, 'settings', NULL, 800, 'Cluster settings during the report interval', 'Cluster settings during the report interval', NULL, NULL, 'pg_settings', NULL, NULL),
(2, 'stmt_warn', NULL, 900, NULL, 'Warning!', 'stmt_cnt_all', NULL, 'stmt_warn', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "stmt_cnt_all",'
    '"ordering": "sample_id",'
    '"columns": ['
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample ID"}, '
        '{"id": "sample_time", "class": "table_obj_value", "caption": "Sample Time"}, '
        '{"id": "stmt_cnt", "class": "table_obj_value", "caption": "Stmts Captured"}, '
        '{"id": "max_cnt", "class": "table_obj_value", "caption": "pg_stat_statements.max"}'
    ']}]'::jsonb)
;


-- Server section of differential report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href, content,
  sect_struct)
VALUES
(2, 'dbstat', 'srvstat', 100, 'Database statistics', 'Database statistics', NULL, NULL, 'db_stat', NULL, NULL),
(2, 'dbstatreset', 'dbstat', 200, NULL, NULL, 'dbstats_reset', NULL, NULL,
  '{"class": "warning", "text": "Database statistics reset detected during report interval! '
  'Statistics for listed databases and contained objects might be affected."}',
  '[{'
    '"type": "row_table", '
    '"source": "dbstats_reset", '
    '"ordering": "sample_id",'
    '"columns": ['
        '{"id": "interval_num", "class": "table_obj_value", "caption": "I"}, '
        '{"id": "dbname", "class": "table_obj_name", "caption": "Database"}, '
        '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample"}, '
        '{"id": "stats_reset", "class": "table_obj_value", "caption": "Reset time"}'
    ']'
  '}]'::jsonb),
(2, 'dbstatmain', 'dbstat', 300, NULL, NULL, NULL, NULL, NULL, NULL,
'[{'
    '"type": "row_table", '
    '"source": "dbstat", '
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
        '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
        '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"}, '
        '{"caption": "Transactions", "columns": ['
            '{"id": ["xact_commit1", "xact_commit2"], "class": "table_obj_value", "title": "Number of transactions in this database that have been committed", "caption": "Commits"}, '
            '{"id": ["xact_rollback1", "xact_rollback2"], "class": "table_obj_value", "title": "Number of transactions in this database that have been rolled back", "caption": "Rollbacks"}, '
            '{"id": ["deadlocks1", "deadlocks2"], "class": "table_obj_value", "title": "Number of deadlocks detected in this database", "caption": "Deadlocks"}'
        ']}, '
        '{"caption": "Checksums", "condition": "checksum_fail_detected", "columns": ['
            '{"id": ["checksum_failures1", "checksum_failures2"], "class": "table_obj_value", "title": "Number of block checksum failures detected", "caption": "Failures"}, '
            '{"id": ["checksum_last_failure1", "checksum_last_failure2"], "class": "table_obj_value", "title": "Last checksum failure detected", "caption": "Last"}
            ]}, '
        '{"caption": "Block statistics", "columns": ['
            '{"id": ["blks_hit_pct1", "blks_hit_pct2"], "class": "table_obj_value", "title": "Buffer cache hit ratio", "caption": "Hit(%)"}, '
            '{"id": ["blks_read1", "blks_read2"], "class": "table_obj_value", "title": "Number of disk blocks read in this database", "caption": "Read"}, '
            '{"id": ["blks_hit1", "blks_hit2"], "class": "table_obj_value", "title": "Number of times disk blocks were found already in the buffer cache", "caption": "Hit"}'
        ']}, '
        '{"caption": "Block I/O times", "condition": "io_times", "columns": ['
            '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "title": "Time spent reading data file blocks by backends", "caption": "Read"}, '
            '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "title": "Time spent writing data file blocks by backends", "caption": "Write"} '
        ']}, '
        '{"caption": "Tuples", "columns": ['
            '{"id": ["tup_returned1", "tup_returned2"], "class": "table_obj_value", "title": "Number of rows returned by queries in this database", "caption": "Ret"}, '
            '{"id": ["tup_fetched1", "tup_fetched2"], "class": "table_obj_value", "title": "Number of rows fetched by queries in this database", "caption": "Fet"}, '
            '{"id": ["tup_inserted1", "tup_inserted2"], "class": "table_obj_value", "title": "Number of rows inserted by queries in this database", "caption": "Ins"}, '
            '{"id": ["tup_updated1", "tup_updated2"], "class": "table_obj_value", "title": "Number of rows updated by queries in this database", "caption": "Upd"}, '
            '{"id": ["tup_deleted1", "tup_deleted2"], "class": "table_obj_value", "title": "Number of rows deleted", "caption": "Del"}'
        ']}, '
        '{"caption": "Temp files", "columns": ['
            '{"id": ["temp_bytes1", "temp_bytes2"], "class": "table_obj_value", "title": "Total amount of data written to temporary files by queries in this database", "caption": "Size"}, '
            '{"id": ["temp_files1", "temp_files2"], "class": "table_obj_value", "title": "Number of temporary files created by queries in this database", "caption": "Files"}'
        ']}, '
        '{"id": ["datsize1", "datsize2"], "class": "table_obj_value", "title": "Database size as is was at the moment of last sample in report interval", "caption": "Size"}, '
        '{"id": ["datsize_delta1", "datsize_delta2"], "class": "table_obj_value", "title": "Database size increment during report interval", "caption": "Growth"} '
']}]'::jsonb),
(2, 'iostat', 'srvstat', 328, 'Cluster I/O statistics', 'Cluster I/O statistics', 'stat_io', NULL, 'stat_io', NULL, NULL),
(2, 'iostatrst', 'srvstat', 329, NULL, NULL, 'stat_io_reset', NULL, 'stat_io_reset',
'{"class": "warning", "text": "IO stats reset was detected during report interval. Statistic values may be affected"}',
'[{'
    '"type": "row_table",'
    '"source": "stat_io_reset",'
    '"columns": ['
      '{"caption": "Sample ID", "id": "sample_id", "class": "table_obj_name", '
        '"title": "Sample identifier with detected reset"},'
      '{"caption": "Object", "id": "object", "class": "table_obj_name", '
        '"title": "Target object of an I/O operation"},'
      '{"caption": "Backend", "id": "backend_type", "class": "table_obj_name", '
        '"title": "Type of backend (see stat_activity)"},'
      '{"caption": "Context", "id": "context", "class": "table_obj_name", '
        '"title": "The context of an I/O operation"},'
      '{"caption": "Reset time", "id": "stats_reset", "class": "table_obj_value", '
        '"title": "Date and time of the last reset performed in sample"}'
    ']'
    '}]'::jsonb),
(2, 'iostatmain', 'srvstat', 330, NULL, NULL, 'stat_io', NULL, 'stat_io', NULL,
'[{'
    '"type": "row_table",'
    '"source": "stat_io",'
    '"columns": ['
      '{"caption": "Object", "id": "object", "class": "hdr", "rowspan": true, '
        '"title": "Target object of an I/O operation"},'
      '{"caption": "Backend", "id": "backend_type", "class": "hdr", "rowspan": true, '
        '"title": "Type of backend (see stat_activity)"},'
      '{"caption": "Context", "id": "context", "class": "hdr", "rowspan": true, '
        '"title": "The context of an I/O operation"},'
      '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"},'
      '{"caption": "Reads", "columns": ['
        '{"caption": "Count", "id": ["reads1", "reads2"], "class": "table_obj_value", '
          '"title": "Number of read operations"},'
        '{"caption": "Bytes", "id": ["read_sz1", "read_sz2"], "class": "table_obj_value", '
          '"title": "Read data amount"},'
        '{"caption": "Time", "id": ["read_time1", "read_time2"], "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in reading operation (seconds)"}'
     ']},'
      '{"caption": "Writes", "columns": ['
        '{"caption": "Count", "id": ["writes1", "writes2"], "class": "table_obj_value", '
          '"title": "Number of write operations"},'
        '{"caption": "Bytes", "id": ["write_sz1", "write_sz2"], "class": "table_obj_value", '
          '"title": "Written data amount"},'
        '{"caption": "Time", "id": ["write_time1", "write_time2"], "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in writing operations (seconds)"}'
     ']},'
      '{"caption": "Writebacks", "columns": ['
        '{"caption": "Count", "id": ["writebacks1", "writebacks2"], "class": "table_obj_value", '
          '"title": "Number of blocks which the process requested the kernel write out to permanent storage"},'
        '{"caption": "Bytes", "id": ["writeback_sz1", "writeback_sz2"], "class": "table_obj_value", '
          '"title": "The amount of data requested for write out to permanent storage"},'
        '{"caption": "Time", "id": ["writeback_time1", "writeback_time2"], "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in writeback operations (seconds)"}'
     ']},'
      '{"caption": "Extends", "columns": ['
        '{"caption": "Count", "id": ["extends1", "extends2"], "class": "table_obj_value", '
          '"title": "Number of relation extend operations"},'
        '{"caption": "Bytes", "id": ["extend_sz1", "extend_sz2"], "class": "table_obj_value", '
          '"title": "The amount of space used by extend operations"},'
        '{"caption": "Time", "id": ["extend_time1", "extend_time2"], "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in extend operations (seconds)"}'
     ']},'
     '{"caption": "Hits", "id": ["hits1", "hits2"], "class": "table_obj_value", '
       '"title": "The number of times a desired block was found in a shared buffer"},'
     '{"caption": "Evictions", "id": ["evictions1", "evictions2"], "class": "table_obj_value", '
       '"title": "Number of times a block has been written out from a shared or local buffer in order to make it available for another use"},'
     '{"caption": "Reuses", "id": ["reuses1", "reuses2"], "class": "table_obj_value", '
       '"title": "The number of times an existing buffer in a size-limited ring buffer outside of shared buffers was reused as part of an I/O operation in the bulkread, bulkwrite, or vacuum contexts"},'
      '{"caption": "Fsyncs", "columns": ['
        '{"caption": "Count", "id": ["fsyncs1", "fsyncs2"], "class": "table_obj_value", '
          '"title": "Number of fsync calls. These are only tracked in context normal"},'
        '{"caption": "Time", "id": ["fsync_time1", "fsync_time2"], "class": "table_obj_value", '
          '"condition": "io_times", '
          '"title": "Time spent in fsync operations (seconds)"}'
     ']}'
    ']'
    '}]'::jsonb),
(2, 'slrustat', 'srvstat', 358, 'Cluster SLRU statistics', 'Cluster SLRU statistics', 'stat_slru', NULL, 'stat_slru', NULL, NULL),
(2, 'slrustatrst', 'slrustat', 359, NULL, NULL, 'stat_slru_reset', NULL, 'stat_slru_reset',
'{"class": "warning", "text": "SLRU stats reset was detected during report interval. Statistic values may be affected"}',
'[{'
    '"type": "row_table",'
    '"source": "stat_slru_reset",'
    '"columns": ['
      '{"caption": "Sample ID", "id": "sample_id", "class": "table_obj_name", '
        '"title": "Sample identifier with detected reset"},'
      '{"caption": "Name", "id": "name", "class": "table_obj_name", '
        '"title": "Name of the SLRU"},'
      '{"caption": "Reset time", "id": "stats_reset", "class": "table_obj_value", '
        '"title": "Date and time of the last reset performed in sample"}'
    ']'
    '}]'::jsonb),
(2, 'slrustatmain', 'slrustat', 360, NULL, NULL, 'stat_slru', NULL, 'stat_slru', NULL,
'[{'
    '"type": "row_table",'
    '"source": "stat_slru",'
    '"columns": ['
      '{"caption": "Name", "id": "name", "class": "hdr", "rowspan": true, '
        '"title": "Name of the SLRU"},'
      '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"},'
      '{"caption": "Zeroed", "id": ["blks_zeroed1", "blks_zeroed2"], "class": "table_obj_value", '
        '"title": "Number of blocks zeroed during initializations"},'
      '{"caption": "Hits", "id": ["blks_hit1", "blks_hit2"], "class": "table_obj_value", '
        '"title": "Number of times disk blocks were found already in the SLRU, so that a '
        'read was not necessary (this only includes hits in the SLRU, not the operating '
        'system''s file system cache)"},'
      '{"caption": "Reads", "id": ["blks_read1", "blks_read2"], "class": "table_obj_value", '
        '"title": "Number of disk blocks read for this SLRU"},'
      '{"caption": "%Hit", "id": ["hit_pct1", "hit_pct2"], "class": "table_obj_value", '
        '"title": "Number of disk blocks hits for this SLRU as a percentage of reads + hits"},'
      '{"caption": "Writes", "id": ["blks_written1", "blks_written2"], "class": "table_obj_value", '
        '"title": "Number of disk blocks written for this SLRU"},'
      '{"caption": "Checked", "id": ["blks_exists1", "blks_exists2"], "class": "table_obj_value", '
        '"title": "Number of blocks checked for existence for this SLRU (blks_exists field)"},'
      '{"caption": "Flushes", "id": ["flushes1", "flushes2"], "class": "table_obj_value", '
        '"title": "Number of flushes of dirty data for this SLRU"},'
      '{"caption": "Truncates", "id": ["truncates1", "truncates2"], "class": "table_obj_value", '
        '"title": "Number of truncates for this SLRU"}'
    ']'
    '}]'::jsonb),
(2, 'sesstat', 'srvstat', 400, 'Session statistics by database', 'Session statistics by database', 'sess_stats', NULL, 'db_stat_sessions', NULL,
'[{'
    '"type": "row_table",'
    '"source": "dbstat", '
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
        '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true},'
        '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"},'
        '{"caption": "Timings (s)", "columns": ['
          '{"id": ["session_time1", "session_time2"], "class": "table_obj_value", "title": "Time spent by database sessions in this database (note that statistics are only updated when the state of a session changes, so if sessions have been idle for a long time, this idle time wont be included)", "caption": "Total"},'
          '{"id": ["active_time1", "active_time2"], "class": "table_obj_value", "title": "Time spent executing SQL statements in this database (this corresponds to the states active and fastpath function call in pg_stat_activity)", "caption": "Active"},'
          '{"id": ["idle_in_transaction_time1", "idle_in_transaction_time2"], "class": "table_obj_value", "title": "Time spent idling while in a transaction in this database (this corresponds to the states idle in transaction and idle in transaction (aborted) in pg_stat_activity)", "caption": "Idle(T)"}'
        ']},'
        '{"caption": "Sessions", "columns": ['
          '{"id": ["sessions1", "sessions2"], "class": "table_obj_value", "title": "Total number of sessions established to this database", "caption": "Established"},'
          '{"id": ["sessions_abandoned1", "sessions_abandoned2"], "class": "table_obj_value", "title": "Number of database sessions to this database that were terminated because connection to the client was lost", "caption": "Abondoned"},'
          '{"id": ["sessions_fatal1", "sessions_fatal2"], "class": "table_obj_value", "title": "Number of database sessions to this database that were terminated by fatal errors", "caption": "Fatal"},'
          '{"id": ["sessions_killed1", "sessions_killed2"], "class": "table_obj_value", "title": "Number of database sessions to this database that were terminated by operator intervention", "caption": "Killed"}'
        ']}'
    ']}]'::jsonb),
(2, 'stmtstat', 'srvstat', 500, 'Statement statistics by database', 'Statement statistics by database', 'statstatements', NULL, 'st_stat', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "statements_dbstats",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": "true"}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of query executions", "caption": "Calls"}, '
      '{"caption": "Time (s)", "columns": ['
        '{"id": ["total_plan_time1", "total_plan_time2"], "class": "table_obj_value", "title": "Time spent planning queries", "caption": "Plan", "condition": "planning_times"}, '
        '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "TTime spent executing queries", "caption": "Exec"}, '
        '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "title": "Time spent reading blocks", "caption": "Read"}, '
        '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "title": "Time spent writing blocks", "caption": "Write"}, '
        '{"id": ["trg_fn_total_time1", "trg_fn_total_time2"], "class": "table_obj_value", "title": "Time spent in trigger functions", "caption": "Trg"}'
        ']}, '
      '{"caption": "Temp I/O Time", "condition": "statements_temp_io_times", "columns": ['
        '{"id": ["temp_blk_read_time1", "temp_blk_read_time2"], "class": "table_obj_value", "title": "Time spent reading temporary file blocks, in seconds", "caption": "Read"}, '
        '{"id": ["temp_blk_write_time1", "temp_blk_write_time2"], "class": "table_obj_value", "title": "Time spent writing temporary file blocks, in seconds", "caption": "Write"} '
        ']}, '
      '{"title": "Number of blocks fetched (hit + read)", "caption": "Fetched (blk)", "columns": ['
        '{"id": ["shared_gets1", "shared_gets2"], "class": "table_obj_value", "caption": "Shared"}, '
        '{"id": ["local_gets1", "local_gets2"], "class": "table_obj_value", "caption": "Local"}'
        ']}, '
      '{"title": "Number of blocks dirtied", "caption": "Dirtied (blk)", "columns": ['
        '{"id": ["shared_blks_dirtied1", "shared_blks_dirtied2"], "class": "table_obj_value", "caption": "Shared"}, '
        '{"id": ["local_blks_dirtied1", "local_blks_dirtied2"], "class": "table_obj_value", "caption": "Local"}'
        ']}, '
      '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
        '{"id": ["temp_blks_read1", "temp_blks_read2"], "class": "table_obj_value", "caption": "Read"}, '
        '{"id": ["temp_blks_written1", "temp_blks_written2"], "class": "table_obj_value", "caption": "Write"}'
        ']}, '
      '{"title": "Number of blocks, used for temporary tables", "caption": "Local (blk)", "columns": ['
        '{"id": ["local_blks_read1", "local_blks_read2"], "class": "table_obj_value", "caption": "Read"}, '
        '{"id": ["local_blks_written1", "local_blks_written2"], "class": "table_obj_value", "caption": "Write"}'
        ']}, '
      '{"id": ["statements1", "statements2"], "class": "table_obj_value", "caption": "Statements"}, '
      '{"id": ["wal_bytes1", "wal_bytes2"], "class": "table_obj_value", "caption": "WAL size", "condition": "statement_wal_bytes"}'
    ']}]'::jsonb),
(2, 'dbjitstat', 'srvstat', 550, 'JIT statistics by database', 'JIT statistics by database', 'statements_jit_stats', NULL, 'dbagg_jit_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "statements_dbstats",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": "true"},'
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"},'
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of query executions", "caption": "Calls"},'
      '{"caption": "Time (s)", "columns": ['
          '{"id": ["total_plan_time1", "total_plan_time2"], "class": "table_obj_value", "title": "Time spent planning queries", "caption": "Plan", "condition": "planning_times"},'
          '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "Time spent planning queries", "caption": "Exec"}'
          ']},'
      '{"caption": "Generation", "columns": ['
          '{"id": ["jit_functions1", "jit_functions2"], "class": "table_obj_value", "title": "Total number of functions JIT-compiled by the statements", "caption": "Count"},'
          '{"id": ["jit_generation_time1", "jit_generation_time2"], "class": "table_obj_value", "title": "Time spent by the statements on generating JIT code", "caption": "Gen. time"}'
          ']},'
      '{"caption": "Inlining", "columns": ['
          '{"id": ["jit_inlining_count1", "jit_inlining_count2"], "class": "table_obj_value", "title": "Number of times functions have been inlined", "caption": "Count"},'
          '{"id": ["jit_inlining_time1", "jit_inlining_time2"], "class": "table_obj_value", "title": "Time spent by statements on inlining functions", "caption": "Time"}'
          ']},'
      '{"caption": "Optimization", "columns": ['
          '{"id": ["jit_optimization_count1", "jit_optimization_count2"], "class": "table_obj_value", "title": "Number of times statements hasbeen optimized", "caption": "Count"},'
          '{"id": ["jit_optimization_time1", "jit_optimization_time2"], "class": "table_obj_value", "title": "Time spent by statements on optimizing", "caption": "Time"}'
          ']},'
      '{"caption": "Emission", "columns": ['
          '{"id": ["jit_emission_count1", "jit_emission_count2"], "class": "table_obj_value", "title": "Number of times code has been emitted", "caption": "Count"},'
          '{"id": ["jit_emission_time1", "jit_emission_time2"], "class": "table_obj_value", "title": "Time spent executing queries", "caption": "Exec"}'
          ']}'
      ']}]'::jsonb),
(2, 'commonstat', 'srvstat', 600, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(2, 'clusterstat', 'commonstat', 600, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(2, 'clusthdr', 'clusterstat', 700, 'Cluster statistics', 'Cluster statistics', NULL, NULL, 'clu_stat', NULL, NULL),
(2, 'clustrst', 'clusterstat', 800, NULL, NULL, 'cluster_stats_reset', NULL, NULL,
  '{"class": "warning", "text": "Cluster statistics reset detected during report interval! Cluster statistics might be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "cluster_stats_reset",'
    '"columns": ['
      '{"id": "interval_num", "class": "table_obj_value", "caption": "I"},'
      '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample"},'
      '{"id": "bgwriter_stats_reset", "class": "table_obj_value", "caption": "BGWriter reset time"},'
      '{"id": "archiver_stats_reset", "class": "table_obj_value", "caption": "Archiver reset time"}'
    ']}]'::jsonb),
(2, 'clust', 'clusterstat', 900, NULL, NULL, NULL, NULL, NULL, NULL,
  '[{'
    '"type": "column_table", '
    '"source": "cluster_stats",'
    '"columns": ['
        '{"caption": "Metric"}, '
        '{"caption": "Value (1)", "title": "properties.timePeriod1"}, '
        '{"caption": "Value (2)", "title": "properties.timePeriod2"}'
    '],'
    '"rows": ['
        '{"caption": "Scheduled checkpoints", "cells": ['
            '{"id": "checkpoints_timed1", "class": "table_obj_value int1"}, '
            '{"id": "checkpoints_timed2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Requested checkpoints", "cells": ['
            '{"id": "checkpoints_req1", "class": "table_obj_value int1"},'
            '{"id": "checkpoints_req2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Checkpoint write time (s)", "cells": ['
            '{"id": "checkpoint_write_time1", "class": "table_obj_value int1"},'
            '{"id": "checkpoint_write_time2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Checkpoint sync time (s)", "cells": ['
            '{"id": "checkpoint_sync_time1", "class": "table_obj_value int1"},'
            '{"id": "checkpoint_sync_time2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Checkpoint buffers written", "cells": ['
            '{"id": "buffers_checkpoint1", "class": "table_obj_value int1"},'
            '{"id": "buffers_checkpoint2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Background buffers written", "cells": ['
            '{"id": "buffers_clean1", "class": "table_obj_value int1"},'
            '{"id": "buffers_clean2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Backend buffers written", "cells": ['
            '{"id": "buffers_backend1", "class": "table_obj_value int1"},'
            '{"id": "buffers_backend2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Backend fsync count", "cells": ['
            '{"id": "buffers_backend_fsync1", "class": "table_obj_value int1"},'
            '{"id": "buffers_backend_fsync2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Bgwriter interrupts (too many buffers)", "cells": ['
            '{"id": "maxwritten_clean1", "class": "table_obj_value int1"},'
            '{"id": "maxwritten_clean2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Number of buffers allocated", "cells": ['
            '{"id": "buffers_alloc1", "class": "table_obj_value int1"},'
            '{"id": "buffers_alloc2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "WAL generated", "cells": ['
            '{"id": "wal_size_pretty1", "class": "table_obj_value int1"},'
            '{"id": "wal_size_pretty2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "Start LSN", "cells": ['
            '{"id": "start_lsn1", "class": "table_obj_value int1"},'
            '{"id": "start_lsn2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "End LSN", "cells": ['
            '{"id": "end_lsn1", "class": "table_obj_value int1"},'
            '{"id": "end_lsn2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "WAL segments archived", "cells": ['
            '{"id": "archived_count1", "class": "table_obj_value int1"},'
            '{"id": "archived_count2", "class": "table_obj_value int2"}'
        ']}, '
        '{"caption": "WAL segments archive failed", "cells": ['
            '{"id": "failed_count1", "class": "table_obj_value int1"},'
            '{"id": "failed_count2", "class": "table_obj_value int2"}'
        ']}'
    ']'
    '}]'::jsonb),
(2, 'walstat', 'commonstat', 950, NULL, NULL, NULL, NULL, NULL, NULL, NULL),
(2, 'walsthdr', 'walstat', 1000, 'WAL statistics', 'WAL statistics', 'wal_stats', NULL, 'wal_stat', NULL, NULL),
(2, 'walstrst', 'walstat', 1100, NULL, NULL, 'wal_stats_reset', NULL, NULL,
  '{"class": "warning", "text": "WAL statistics reset detected during report interval! WAL statistics might be affected"}',
  '[{'
    '"type": "row_table",'
    '"source": "wal_stats_reset",'
    '"columns": ['
      '{"id": "interval_num", "class": "table_obj_value", "caption": "I"},'
      '{"id": "sample_id", "class": "table_obj_value", "caption": "Sample"},'
      '{"id": "wal_stats_reset", "class": "table_obj_value", "caption": "WAL stats reset time"}'
    ']}]'::jsonb),
(2, 'walst', 'walstat', 1200, NULL, NULL, 'wal_stats', NULL, NULL, NULL,
  '[{'
    '"type": "column_table",'
    '"source": "wal_stats",'
    '"columns": ['
        '{"caption": "Metric"},'
        '{"caption": "Value (1)", "title": "properties.timePeriod1"},'
        '{"caption": "Value (2)", "title": "properties.timePeriod2"}'
    '],'
    '"rows": ['
      '{"caption": "WAL generated", "title": "Total amount of WAL generated", "cells": ['
          '{"id": "wal_bytes1", "class": "table_obj_value int1"},'
          '{"id": "wal_bytes2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL per second", "title": "Average amount of WAL generated per second", "cells": ['
          '{"id": "wal_bytes_per_sec1", "class": "table_obj_value int1"}, '
          '{"id": "wal_bytes_per_sec2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL records", "title": "Total number of WAL records generated", "cells": ['
          '{"id": "wal_records1", "class": "table_obj_value int1"},'
          '{"id": "wal_records2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL FPI", "title": "Total number of WAL full page images generated", "cells": ['
          '{"id": "wal_fpi1", "class": "table_obj_value int1"},'
          '{"id": "wal_fpi2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL buffers full", "title": "Number of times WAL data was written to disk because WAL buffers became full", "cells": ['
          '{"id": "wal_buffers_full1", "class": "table_obj_value int1"},'
          '{"id": "wal_buffers_full2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL writes", "title": "Number of times WAL buffers were written out to disk via XLogWrite request", "cells": ['
          '{"id": "wal_write1", "class": "table_obj_value int1"}, '
          '{"id": "wal_write2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL writes per second", "title": "Average number of times WAL buffers were written out to disk via XLogWrite request per .second", "cells": ['
          '{"id": "wal_write_per_sec1", "class": "table_obj_value int1"}, '
          '{"id": "wal_write_per_sec2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL sync", "title": "Number of times WAL files were synced to disk via issue_xlog_fsync request (if fsync is on and wal_sync_method is either fdatasync, fsync or fsync_writethrough, otherwise zero)", "cells": ['
          '{"id": "wal_sync1", "class": "table_obj_value int1"}, '
          '{"id": "wal_sync2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL syncs per second", "title": "Average number of times WAL files were synced to disk via issue_xlog_fsync request per second", "cells": ['
          '{"id": "wal_sync_per_sec1", "class": "table_obj_value int1"}, '
          '{"id": "wal_sync_per_sec2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL write time (s)", "title": "Total amount of time spent writing WAL buffers to disk via XLogWrite request, in milliseconds (if track_wal_io_timing is enabled, otherwise zero). This includes the sync time when wal_sync_method is either open_datasync or open_sync", "cells": ['
          '{"id": "wal_write_time1", "class": "table_obj_value int1"}, '
          '{"id": "wal_write_time2", "class": "table_obj_value int2"}'
      ']},'
      '{"title": "WAL write time as a percentage of the report duration time", "caption": "WAL write duty", "cells": ['
          '{"id": "wal_write_time_per_sec1", "class": "table_obj_value int1"}, '
          '{"id": "wal_write_time_per_sec2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL sync time (s)", "title": "Total amount of time spent syncing WAL files to disk via issue_xlog_fsync request, in milliseconds (if track_wal_io_timing is enabled, fsync is on, and wal_sync_method is either fdatasync, fsync or fsync_writethrough, otherwise zero)", "cells": ['
          '{"id": "wal_sync_time1", "class": "table_obj_value int1"}, '
          '{"id": "wal_sync_time2", "class": "table_obj_value int2"}'
      ']},'
      '{"caption": "WAL sync duty", "title": "WAL sync time as a percentage of the report duration time", "cells": ['
          '{"id": "wal_sync_time_per_sec1", "class": "table_obj_value int1"},'
          '{"id": "wal_sync_time_per_sec2", "class": "table_obj_value int2"}'
      ']}'
    ']'
  '}]'::jsonb),
(2, 'tbspst', 'srvstat', 1400, 'Tablespace statistics', 'Tablespace statistics', NULL, NULL, 'tablespace_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "tablespace_stats",'
    '"ordering": "tablespacename",'
    '"columns": ['
      '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
      '{"caption": "Path", "id": "tablespacepath", "class": "hdr", "rowspan": true},'
      '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"}, '
      '{"caption": "Size", "id": ["size1", "size2"], "class": "table_obj_value", '
      '"title": "Tablespace size as it was at the moment of last sample in report interval"},'
      '{"caption": "Growth", "id": ["size_delta1", "size_delta2"], "class": "table_obj_value", '
      '"title": "Tablespace size increment during report interval"}'
    ']'
  '}]'::jsonb),
(2, 'wait_sampling_srvstats', 'srvstat', 1500, 'Wait sampling', 'Wait sampling', 'wait_sampling_tot', NULL, 'wait_sampling', NULL, NULL),
(2, 'wait_sampling_total', 'wait_sampling_srvstats', 100, 'Wait events types', 'Wait events types', 'wait_sampling_tot', NULL, 'wait_sampling_total', NULL,
  '[{'
      '"type": "row_table", '
      '"source": "wait_sampling_total_stats",'
      '"ordering": "event_type_order",'
      '"columns": ['
          '{"id": "event_type", "class": "hdr", "caption": "Wait event type", "rowspan": true}, '
          '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"}, '
          '{"id": ["stmt_waited1", "stmt_waited2"], "class": "table_obj_value", "title": "Time, waited in events of wait event type executing statements in seconds", "caption": "Statements Waited (s)"}, '
          '{"id": ["stmt_waited_pct1", "stmt_waited_pct2"], "class": "table_obj_value", "title": "Time, waited in events of wait event type as a percentage of total time waited in a cluster executing statements", "caption": "%Total"}, '
          '{"id": ["tot_waited1", "tot_waited2"], "class": "table_obj_value", "title": "Time, waited in events of wait event type by all backends (including background activity) in seconds", "caption": "All Waited (s)"}, '
          '{"id": ["tot_waited_pct1", "tot_waited_pct2"], "class": "table_obj_value", "title": "Time, waited in events of wait event type as a percentage of total time waited in a cluster by all backends (including background activity)", "caption": "%Total"}'
      ']}]'::jsonb),
(2, 'wait_sampling_statements', 'wait_sampling_srvstats', 200, 'Top wait events (statements)', 'Top wait events (statements)', 'wait_sampling_tot', NULL, 'wt_smp_stmt', '{"class": "notice", "text": "Top wait events detected in statements execution"}',
  '[{'
    '"type": "row_table", '
    '"source": "wait_sampling_events",'
    '"filter": {"type": "exists", "field": "stmt_filter"},'
    '"ordering": "stmt_ord",'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "event_type", "class": "hdr", "caption": "Wait event type", "rowspan": true}, '
        '{"id": "event", "class": "hdr", "caption": "Wait event", "rowspan": true}, '
        '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"}, '
        '{"id": ["stmt_waited1", "stmt_waited2"], "class": "table_obj_value", "title": "Time, waited in event executing statements in seconds", "caption": "Waited (s)"}, '
        '{"id": ["stmt_waited_pct1", "stmt_waited_pct2"], "class": "table_obj_value", "title": "Time, waited in event as a percentage of total time waited in a cluster executing statements", "caption": "%Total"}'
    ']}]'::jsonb),
(2, 'wait_sampling_all', 'wait_sampling_srvstats', 300, 'Top wait events (All)', 'Top wait events (All)', 'wait_sampling_tot', NULL, 'wt_smp_all', '{"class": "notice", "text": "Top wait events detected in all backends"}',
  '[{'
    '"type": "row_table", '
    '"source": "wait_sampling_events",'
    '"filter": {"type": "exists", "field": "total_filter"},'
    '"ordering": "tot_ord",'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "event_type", "class": "hdr", "caption": "Wait event type", "rowspan": true}, '
        '{"id": "event", "class": "hdr", "caption": "Wait event", "rowspan": true}, '
        '{"id": ["1", "2"], "class": "interval", "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I"}, '
        '{"id": ["tot_waited1", "tot_waited2"], "class": "table_obj_value", "title": "Time, waited in event by all backends (including background activity) in seconds", "caption": "Waited (s)"}, '
        '{"id": ["tot_waited_pct1", "tot_waited_pct2"], "class": "table_obj_value", "title": "Time, waited in event by all backends as a percentage of total time waited in a cluster by all backends (including background activity)", "caption": "%Total"}'
    ']}]'::jsonb)
;

-- Subsamples-based section of regular reports
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'db_activity_agg', 'actsesshdr', 100, 'Session state statistics by database', 'Session state statistics by database', 'act_backend', NULL, 'db_activity_agg',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "db_activity_agg",'
    '"ordering": "ord_db",'
    '"highlight": ['
      '{"id": "dbname", "index": 0}'
    '],'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "hdr", "rowspan": "true"},'
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"caption": "Summary", "columns": ['
        '{"caption": "Active", "id": ["active_dur1", "active_dur2"], "class": "table_obj_value", '
          '"title": "Summary time in ''active'' state"},'
        '{"caption": "Idle in xact", "id": ["idle_xact_dur1", "idle_xact_dur2"], "class": "table_obj_value", '
          '"title": "Summary time in ''idle in transaction'' state"},'
        '{"caption": "Idle in xact (A)", "id": ["idle_xact_a_dur1", "idle_xact_a_dur2"], "class": "table_obj_value", '
          '"title": "Summary time in ''idle in transaction (aborted)'' state"}'
     ']},'
      '{"caption": "Maximal", "columns": ['
        '{"caption": "Active", "id": ["max_active_dur1", "max_active_dur2"], "class": "table_obj_value", '
          '"title": "Longest detected ''active'' state"},'
        '{"caption": "Idle in xact", "id": ["max_idle_xact_d1", "max_idle_xact_d2"], "class": "table_obj_value", '
          '"title": "Longest detected ''idle in xact'' state"},'
        '{"caption": "Idle in xact (A)", "id": ["max_idle_xact_a_d1", "max_idle_xact_a_d2"], "class": "table_obj_value", '
          '"title": "Longest detected ''idle in transaction (aborted)'' state"},'
        '{"caption": "xact age", "id": ["max_xact_age1", "max_xact_age2"], "class": "table_obj_value", '
          '"title": "Maximal detected backend_xid age"}'
     ']}'
    ']'
    '}]'::jsonb),
(2, 'act_ix', 'actsesshdr', 200, 'Top ''idle in transaction'' states', 'Top ''idle in transaction'' session states by duration', 'act_ix', NULL, 'act_ix',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 1},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(2, 'act_ixa', 'actsesshdr', 300, 'Top ''idle in transaction (aborted)'' states', 'Top ''idle in transaction (aborted)'' session states by duration', 'act_ixa', NULL, 'act_ixa',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 2},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(2, 'act_active', 'actsesshdr', 400, 'Top ''active'' states', 'Top ''active'' session states by duration', 'act_active', NULL, 'act_active',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_dur",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_state_code", "value": 3},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(2, 'act_age', 'actsesshdr', 500, 'Top states by transaction age', 'Top states by transaction age', 'act_backend', NULL, 'act_age',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_age",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_age", "value": true},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "Xact duration", "id": "xact_duration_format", "class": "table_obj_value"},'
      '{"caption": "Age", "id": "xmin_age", "class": "table_obj_value", '
        '"title": "Last detected age of backend_xmin"},'
      '{"caption": "State", "id": "state", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb),
(2, 'act_xact_dur', 'actsesshdr', 600, 'Top states by transaction duration', 'Top states by transaction duration', 'act_backend', NULL, 'act_age_dur',
  NULL,
'[{'
    '"type": "row_table",'
    '"source": "act_top_states",'
    '"ordering": "ord_xact",'
    '"limit": "topn",'
    '"highlight": ['
      '{"id": "pid", "index": 0},'
      '{"id": "state_change", "index": 0}'
    '],'
    '"preview": ['
      '{"id": "act_query_md5", "dataset": "act_queries"}'
    '],'
    '"filter": {"type": "equal", "field": "flt_age", "value": true},'
    '"columns": ['
      '{"caption": "Database", "id": "dbname", "class": "table_obj_name"},'
      '{"caption": "User", "id": "username", "class": "table_obj_name"},'
      '{"caption": "App", "id": "application_name", "class": "table_obj_name", '
        '"title": "application_name"},'
      '{"caption": "Addr", "id": "client_addr", "class": "table_obj_name", '
        '"title": "client_addr"},'
      '{"caption": "Pid", "id": "pid", "class": "table_obj_name"},'
      '{"caption": "Xact start", "id": "xact_start_format", "class": "table_obj_value"},'
      '{"caption": "Xact duration", "id": "xact_duration_format", "class": "table_obj_value"},'
      '{"caption": "Age", "id": "xmin_age", "class": "table_obj_value", '
        '"title": "Last detected age of backend_xmin"},'
      '{"caption": "State", "id": "state", "class": "table_obj_value"},'
      '{"caption": "State change", "id": "state_change_format", "class": "table_obj_value"},'
      '{"caption": "State duration", "id": "state_duration_format", "class": "table_obj_value"}'
    ']}]'::jsonb)
;

-- Query section of differential report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'sqlela_t', 'sqlsthdr', 100, 'Top SQL by elapsed time', 'Top SQL by elapsed time', 'planning_times', NULL, 'top_ela', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_total_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "ord_total_time"},'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
        '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
        '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
        '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
        '{"id": ["total_time_pct1", "total_time_pct2"], "class": "table_obj_value", "title": "Elapsed time as a percentage of total cluster elapsed time", "caption": "%Total"}, '
        '{"caption": "Time (s)", "columns": ['
            '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed"}, '
            '{"id": ["total_plan_time1", "total_plan_time2"], "class": "table_obj_value", "title": "Time spent planning statement", "caption": "Plan"}, '
            '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec"}'
            ']}, '
        '{"id": ["jit_total_time1", "jit_total_time2"], "class": "jitTimeCell", "caption": "JIT, time (s)", "condition": "statements_jit_stats"}, '
        '{"class": "table_obj_name", "caption": "I/O time (s)", "condition": "io_times", "columns": ['
            '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "title": "Time spent reading blocks by statement", "caption": "Read"}, '
            '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "title": "Time spent writing blocks by statement", "caption": "Write"}'
            ']}, '
        '{"class": "table_obj_name", "caption": "CPU time (s)", "condition": "kcachestatements", "columns": ['
            '{"id": ["user_time1", "user_time2"], "class": "table_obj_value", "caption": "Usr"}, '
            '{"id": ["system_time1", "system_time2"], "class": "table_obj_value", "caption": "Sys"}'
            ']}, '
        '{"id": ["plans1", "plans2"], "class": "table_obj_value", "caption": "Plans", "title": "Number of times the statement was planned"}, '
        '{"id": ["calls1", "calls2"], "class": "table_obj_value", "caption": "Executions", "title": "Number of times the statement was executed"} '
    ']'
  '}]'::jsonb),
(2, 'sqlplan_t', 'sqlsthdr', 200, 'Top SQL by planning time', 'Top SQL by planning time', 'planning_times', NULL, 'top_plan', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_plan_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "ord_plan_time"},'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
        '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
        '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
        '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
        '{"id": ["total_plan_time1", "total_plan_time2"], "class": "table_obj_value", "title": "Time spent planning statement", "caption": "Plan elapsed (s)"}, '
        '{"id": ["plan_time_pct1", "plan_time_pct2"], "class": "table_obj_value", "title": "Plan elapsed as a percentage of statement elapsed time", "caption": "%Elapsed"}, '
        '{"title": "Planning time statistics", "caption": "Plan times (ms)", "columns": ['
            '{"id": ["mean_plan_time1", "mean_plan_time2"], "class": "table_obj_value", "caption": "Mean"}, '
            '{"id": ["min_plan_time1", "min_plan_time2"], "class": "table_obj_value", "caption": "Min"}, '
            '{"id": ["max_plan_time1", "max_plan_time2"], "class": "table_obj_value", "caption": "Max"}, '
            '{"id": ["stddev_plan_time1", "stddev_plan_time2"], "class": "table_obj_value", "caption": "StdErr"}'
        ']}, '
        '{"id": ["plans1", "plans2"], "class": "table_obj_value", "title": "Number of times the statement was planned", "caption": "Plans"}, '
        '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqlexec_t', 'sqlsthdr', 300, 'Top SQL by execution time', 'Top SQL by execution time', NULL, NULL, 'top_exec', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_exec_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "ord_exec_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "caption": "Query ID", "class": "hdr mono queryId", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec (s)"}, '
      '{"id": ["exec_time_pct1", "exec_time_pct2"], "class": "table_obj_value", "title": "Exec time as a percentage of statement elapsed time", "caption": "%Elapsed", "condition": "planning_times"}, '
      '{"id": ["total_exec_time_pct1", "total_exec_time_pct2"], "class": "table_obj_value", "title": "Exec time as a percentage of total cluster elapsed time", "caption": "%Total"}, '
      '{"id": ["jit_total_time1", "jit_total_time2"], "class": "jitTimeCell", "title": "Exec time as a percentage of statement elapsed time", "caption": "JIT time (s)", "condition": "statements_jit_stats"}, '
      '{"caption": "I/O time (s)", "condition": "io_times", "columns": ['
          '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "caption": "Read"}, '
          '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "caption": "Write"}'
      ']}, '
      '{"caption": "CPU time (s)", "condition": "kcachestatements", "columns": ['
          '{"id": ["user_time1", "user_time2"], "class": "table_obj_value", "caption": "Usr"}, '
          '{"id": ["system_time1", "system_time2"], "class": "table_obj_value", "caption": "Sys"}'
      ']}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "caption": "Rows"}, '
      '{"title": "Execution time statistics", "caption": "Execution times (ms)", "columns": ['
          '{"id": ["mean_exec_time1", "mean_exec_time2"], "class": "table_obj_value", "caption": "Mean"}, '
          '{"id": ["min_exec_time1", "min_exec_time2"], "class": "table_obj_value", "caption": "Min"}, '
          '{"id": ["max_exec_time1", "max_exec_time2"], "class": "table_obj_value", "caption": "Max"}, '
          '{"id": ["stddev_exec_time1", "stddev_exec_time2"], "class": "table_obj_value", "caption": "StdErr"}'
      ']}, '
      '{"id": ["calls1", "calls2"], "title": "Number of times the statement was executed", "caption": "Executions", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'sqlcalls', 'sqlsthdr', 400, 'Top SQL by executions', 'Top SQL by executions', NULL, NULL, 'top_calls', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_calls",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}, '
      '{"id": ["calls_pct1", "calls_pct2"], "class": "table_obj_value", "title": "Executions of this statement as a percentage of total executions of all statements in a cluster", "caption": "%Total"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["mean_exec_time1", "mean_exec_time2"], "class": "table_obj_value", "caption": "Mean(ms)"}, '
      '{"id": ["min_exec_time1", "min_exec_time2"], "class": "table_obj_value", "caption": "Min(ms)"}, '
      '{"id": ["max_exec_time1", "max_exec_time2"], "class": "table_obj_value", "caption": "Max(ms)"}, '
      '{"id": ["stddev_exec_time1", "stddev_exec_time2"], "class": "table_obj_value", "caption": "StdErr(ms)"}, '
      '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}'
    ']'
  '}]'::jsonb),
(2, 'sqlio_t', 'sqlsthdr', 500, 'Top SQL by I/O wait time', 'Top SQL by I/O wait time', 'io_times', NULL, 'top_iowait', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_io_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "ord_io_time"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["io_time1", "io_time2"], "class": "table_obj_value", "title": "Time spent by the statement reading and writing blocks", "caption": "IO(s)"}, '
      '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "title": "Time spent by the statement reading blocks", "caption": "R(s)"}, '
      '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "title": "Time spent by the statement writing blocks", "caption": "W(s)"}, '
      '{"id": ["io_time_pct1", "io_time_pct2"], "class": "table_obj_value", "title": "I/O time of this statement as a percentage of total I/O time for all statements in a cluster", "caption": "%Total"}, '
      '{"title": "Number of blocks read by the statement", "caption": "Reads", "columns": ['
          '{"id": ["shared_blks_read1", "shared_blks_read2"], "title": "Number of shared blocks read by the statement", "caption": "Shr", "class": "table_obj_value"}, '
          '{"id": ["local_blks_read1", "local_blks_read2"], "title": "Number of local blocks read by the statement (usually used for temporary tables)", "caption": "Loc", "class": "table_obj_value"}, '
          '{"id": ["temp_blks_read1", "temp_blks_read2"], "title": "Number of temp blocks read by the statement (usually used for operations like sorts and joins)", "caption": "Tmp", "class": "table_obj_value"}'
      ']}, '
      '{"title": "Number of blocks written by the statement", "caption": "Writes", "columns": ['
          '{"id": ["shared_blks_written1", "shared_blks_written2"], "title": "Number of shared blocks written by the statement", "caption": "Shr", "class": "table_obj_value"}, '
          '{"id": ["local_blks_written1", "local_blks_written2"], "title": "Number of local blocks written by the statement (usually used for temporary tables)", "caption": "Loc", "class": "table_obj_value"}, '
          '{"id": ["temp_blks_written1", "temp_blks_written2"], "title": "Number of temp blocks written by the statement (usually used for operations like sorts and joins)", "caption": "Tmp", "class": "table_obj_value"}'
      ']}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of blocks written by the statement", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqlfetch', 'sqlsthdr', 600, 'Top SQL by shared blocks fetched', 'Top SQL by shared blocks fetched', NULL, NULL, 'top_pgs_fetched', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"ordering": "ord_shared_blocks_fetched",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"filter": {"type": "exists", "field": "ord_shared_blocks_fetched"},'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["shared_blks_fetched1", "shared_blks_fetched2"], "class": "table_obj_value", "title": "Shared blocks fetched (read and hit) by the statement", "caption": "blks fetched"}, '
      '{"id": ["shared_blks_fetched_pct1", "shared_blks_fetched_pct2"], "class": "table_obj_value", "title": "Shared blocks fetched by this statement as a percentage of all shared blocks fetched in a cluster", "caption": "%Total"}, '
      '{"id": ["shared_hit_pct1", "shared_hit_pct2"], "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']}]'::jsonb),
(2, 'sqlshrd', 'sqlsthdr', 700, 'Top SQL by shared blocks read', 'Top SQL by shared blocks read', NULL, NULL, 'top_shared_reads', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_shared_blocks_read"},'
    '"ordering": "ord_shared_blocks_read",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["shared_blks_read1", "shared_blks_read2"], "class": "table_obj_value", "title": "Total number of shared blocks read by the statement", "caption": "Reads"}, '
      '{"id": ["read_pct1", "read_pct2"], "class": "table_obj_value", "title": "Shared blocks read by this statement as a percentage of all shared blocks read in a cluster", "caption": "%Total"}, '
      '{"id": ["shared_hit_pct1", "shared_hit_pct2"], "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqlshdir', 'sqlsthdr', 800, 'Top SQL by shared blocks dirtied', 'Top SQL by shared blocks dirtied', NULL, NULL, 'top_shared_dirtied', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_shared_blocks_dirt"},'
    '"ordering": "ord_shared_blocks_dirt",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"],  "class": "interval", "caption": "I"}, '
      '{"id": ["shared_blks_dirtied1", "shared_blks_dirtied2"], "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Dirtied"}, '
      '{"id": ["dirtied_pct1", "dirtied_pct2"], "class": "table_obj_value", "title": "Shared blocks dirtied by this statement as a percentage of all shared blocks dirtied in a cluster", "caption": "%Total"}, '
      '{"id": ["shared_blks_hit1", "shared_blks_hit2"], "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Hits"}, '
      '{"id": ["shared_hit_pct1", "shared_hit_pct2"], "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "%Total"}, '
      '{"id": ["wal_bytes1", "wal_bytes2"], "class": "table_obj_value", "title": "Total amount of WAL bytes generated by the statement", "caption": "WAL", "condition": "statement_wal_bytes"}, '
      '{"id": ["wal_bytes_pct1", "wal_bytes_pct2"], "class": "table_obj_value", "title": "WAL bytes of this statement as a percentage of total WAL bytes generated by a cluster", "caption": "%Total", "condition": "statement_wal_bytes"}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqlshwr', 'sqlsthdr', 900, 'Top SQL by shared blocks written', 'Top SQL by shared blocks written', NULL, NULL, 'top_shared_written', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_shared_blocks_written"},'
    '"ordering": "ord_shared_blocks_written",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"],  "class": "interval", "caption": "I"}, '
      '{"id": ["shared_blks_written1", "shared_blks_written2"], "class": "table_obj_value", "title": "Total number of shared blocks written by the statement", "caption": "Written"}, '
      '{"id": ["tot_written_pct1", "tot_written_pct2"], "class": "table_obj_value", "title": "Shared blocks written by this statement as a percentage of all shared blocks written in a cluster (sum of pg_stat_bgwriter fields buffers_checkpoint, buffers_clean and buffers_backend)", "caption": "%Total"}, '
      '{"id": ["backend_written_pct1", "backend_written_pct2"], "class": "table_obj_value", "title": "Shared blocks written by this statement as a percentage total buffers written directly by a backends (buffers_backend of pg_stat_bgwriter view)", "caption": "%BackendW"}, '
      '{"id": ["shared_hit_pct1", "shared_hit_pct2"], "class": "table_obj_value", "title": "Shared blocks hits as a percentage of shared blocks fetched (read + hit)", "caption": "Hits(%)"}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqlwalsz', 'sqlsthdr', 1000, 'Top SQL by WAL size', 'Top SQL by WAL size', 'statement_wal_bytes', NULL, 'top_wal_bytes', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_wal"},'
    '"ordering": "ord_wal",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"],  "class": "interval", "caption": "I"}, '
      '{"id": ["wal_bytes1", "wal_bytes2"], "class": "table_obj_value", "title": "Total amount of WAL bytes generated by the statement", "caption": "WAL"}, '
      '{"id": ["wal_bytes_pct1", "wal_bytes_pct2"], "class": "table_obj_value", "title": "WAL bytes of this statement as a percentage of total WAL bytes generated by a cluster", "caption": "%Total"}, '
      '{"id": ["shared_blks_dirtied1", "shared_blks_dirtied2"], "class": "table_obj_value", "title": "Total number of shared blocks dirtied by the statement", "caption": "Dirtied"}, '
      '{"id": ["wal_fpi1", "wal_fpi2"], "class": "table_obj_value", "title": "Total number of WAL full page images generated by the statement", "caption": "WAL FPI"}, '
      '{"id": ["wal_records1", "wal_records2"], "class": "table_obj_value", "title": "Total number of WAL records generated by the statement", "caption": "WAL records"}'
    ']'
  '}]'::jsonb),
(2, 'sqltmp', 'sqlsthdr', 1100, 'Top SQL by temp usage', 'Top SQL by temp usage', 'statements_top_temp', NULL, 'top_temp', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_temp"},'
    '"ordering": "ord_temp",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"],  "class": "interval", "caption": "I"}, '
      '{"id": ["local_blks_fetched1", "local_blks_fetched2"], "class": "table_obj_value", "title": "Number of local blocks fetched (hit + read)", "caption": "Local fetched"}, '
      '{"id": ["local_hit_pct1", "local_hit_pct2"], "class": "table_obj_value", "title": "Local blocks hit percentage", "caption": "Hits(%)"}, '
      '{"title": "Number of blocks, used for temporary tables", "caption": "Local (blk)", "columns": ['
          '{"id": ["local_blks_written1", "local_blks_written2"], "class": "table_obj_value", "title": "Number of written local blocks", "caption": "Write"}, '
          '{"id": ["local_write_total_pct1", "local_write_total_pct2"], "class": "table_obj_value", "title": "Percentage of all local blocks written", "caption": "%Total"}, '
          '{"id": ["local_blks_read1", "local_blks_read2"], "class": "table_obj_value", "title": "Number of read local blocks", "caption": "Read"}, '
          '{"id": ["local_read_total_pct1", "local_read_total_pct2"], "class": "table_obj_value", "title": "Percentage of all local blocks read", "caption": "%Total"}'
      ']}, '
      '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
          '{"id": ["temp_blks_written1", "temp_blks_written2"], "class": "table_obj_value", "title": "Number of written temp blocks", "caption": "Write"}, '
          '{"id": ["temp_write_total_pct1", "temp_write_total_pct2"], "class": "table_obj_value", "title": "Percentage of all temp blocks written", "caption": "%Total"}, '
          '{"id": ["temp_blks_read1", "temp_blks_read2"], "class": "table_obj_value", "title": "Number of read temp blocks", "caption": "Read"}, '
          '{"id": ["temp_read_total_pct1", "temp_read_total_pct2"], "class": "table_obj_value", "title": "Percentage of all temp blocks read", "caption": "%Total"}'
      ']}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqltmpiotime', 'sqlsthdr', 1125, 'Top SQL by temp I/O time', 'Top SQL by temp I/O time', 'statements_temp_io_times', NULL, 'top_temp_io_time', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_temp_io_time"},'
    '"ordering": "ord_temp_io_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"],  "class": "interval", "caption": "I"}, '
      '{"title": "Time the statement spent on temporary file blocks I/O", "caption": "Temp I/O time (s)", "columns": ['
          '{"id": ["temp_blk_read_time1", "temp_blk_read_time2"], "class": "table_obj_value", "title": "Time the statement spent reading temporary file blocks, in seconds", "caption": "Read"}, '
          '{"id": ["temp_blk_write_time1", "temp_blk_write_time2"], "class": "table_obj_value", "title": "Time the statement spent reading temporary file blocks, in seconds", "caption": "Write"}, '
          '{"id": ["temp_io_time_pct1", "temp_io_time_pct2"], "class": "table_obj_value", "title": "Time spent on temporary file blocks I/O of this statement as a percentage of total time spent on temporary file blocks I/O by all statements", "caption": "%Total"} '
      ']}, '
      '{"title": "Number of blocks, used in operations (like sorts and joins)", "caption": "Temp (blk)", "columns": ['
          '{"id": ["temp_blks_written1", "temp_blks_written2"], "class": "table_obj_value", "title": "Number of written temp blocks", "caption": "Write"}, '
          '{"id": ["temp_write_total_pct1", "temp_write_total_pct2"], "class": "table_obj_value", "title": "Percentage of all temp blocks written", "caption": "%Total"}, '
          '{"id": ["temp_blks_read1", "temp_blks_read2"], "class": "table_obj_value", "title": "Number of read temp blocks", "caption": "Read"}, '
          '{"id": ["temp_read_total_pct1", "temp_read_total_pct2"], "class": "table_obj_value", "title": "Percentage of all temp blocks read", "caption": "%Total"}'
      ']}, '
      '{"id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent by the statement", "caption": "Elapsed(s)"}, '
      '{"id": ["rows1", "rows2"], "class": "table_obj_value", "title": "Total number of rows retrieved or affected by the statement", "caption": "Rows"}, '
      '{"id": ["calls1", "calls2"], "class": "table_obj_value", "title": "Number of times the statement was executed", "caption": "Executions"}'
    ']'
  '}]'::jsonb),
(2, 'sqljit', 'sqlsthdr', 1150, 'Top SQL by JIT elapsed time', 'Top SQL by JIT elapsed time', 'statements_jit_stats', NULL, 'top_jit', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_statements",'
    '"filter": {"type": "exists", "field": "ord_jit"},'
    '"ordering": "ord_jit",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono jitCellId", "caption": "Query ID", "rowspan": true}, '
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
      '{"id": ["jit_total_time1", "jit_total_time2"], "class": "table_obj_value", "title": "Time spent on JIT in seconds", "caption": "JIT total (s)"}, '
      '{"caption": "Generation", "columns": ['
          '{"id": ["jit_functions1", "jit_functions2"], "class": "table_obj_value", "title": "Total number of functions JIT-compiled by the statement.", "caption": "Count"}, '
          '{"id": ["jit_generation_time1", "jit_generation_time2"], "class": "table_obj_value", "title": "Time spent by the statement on generating JIT code, in seconds.", "caption": "Time (s)"}'
          ']}, '
      '{"class": "table_obj_name", "caption": "Inlining", "columns": ['
          '{"id": ["jit_inlining_count1", "jit_inlining_count2"], "class": "table_obj_value", "title": "Number of times functions have been inlined.", "caption": "Count"}, '
          '{"id": ["jit_inlining_time1", "jit_inlining_time2"], "class": "table_obj_value", "title": "Time spent by the statement on inlining functions, in seconds.", "caption": "Time (s)"}'
          ']}, '
      '{"class": "table_obj_name", "caption": "Optimization", "columns": ['
          '{"id": ["jit_optimization_count1", "jit_optimization_count2"], "class": "table_obj_value", "title": "Number of times the statement has been optimized.", "caption": "Count"}, '
          '{"id": ["jit_optimization_time1", "jit_optimization_time2"], "class": "table_obj_value", "title": "Time spent by the statement on optimizing, in seconds.", "caption": "Time (s)"}'
      ']}, '
      '{"class": "table_obj_name", "caption": "Emission", "columns": ['
          '{"id": ["jit_emission_count1", "jit_emission_count2"], "class": "table_obj_value", "title": "Number of times code has been emitted.", "caption": "Count"}, '
          '{"id": ["jit_emission_time1", "jit_emission_time2"], "class": "table_obj_value", "title": "Time spent by the statement on emitting code, in seconds.", "caption": "Time (s)"}'
      ']}, '
      '{"class": "table_obj_name", "caption": "Time (s)", "columns": ['
          '{"id": ["total_plan_time1", "total_plan_time2"], "class": "table_obj_value", "title": "Time spent planning statement", "condition": "planning_times", "caption": "Plan"}, '
          '{"id": ["total_exec_time1", "total_exec_time2"], "class": "table_obj_value", "title": "Time spent executing statement", "caption": "Exec"}'
      ']}, '
      '{"class": "table_obj_name", "caption": "I/O time (s)", "condition": "io_times", "columns": ['
          '{"id": ["blk_read_time1", "blk_read_time2"], "class": "table_obj_value", "title": "Time spent reading blocks by statement", "caption": "Read"}, '
          '{"id": ["blk_write_time1", "blk_write_time2"], "class": "table_obj_value", "title": "Time spent writing blocks by statement", "caption": "Write"}'
      ']}'
    ']}]'::jsonb),
(2, 'sqlkcachehdr', 'sqlsthdr', 1200, 'rusage statistics', 'rusage statistics', 'kcachestatements', NULL, 'kcache_stat', NULL, NULL),
(2, 'sqlrusgcpu_t', 'sqlkcachehdr', 100, 'Top SQL by system and user time', 'Top SQL by system and user time', NULL, NULL, 'kcache_time', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "top_rusage_statements",'
    '"filter": {"type": "exists", "field": "ord_cpu_time"},'
    '"ordering": "ord_cpu_time",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
        '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true}, '
        '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true}, '
        '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true}, '
        '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"}, '
        '{"title": "Userspace CPU", "caption": "User Time", "columns": ['
            '{"id": ["plan_user_time1", "plan_user_time2"], "class": "table_obj_value", "title": "User CPU time elapsed during planning", "caption": "Plan (s)", "condition": "rusage_planstats"}, '
            '{"id": ["exec_user_time1", "exec_user_time2"], "class": "table_obj_value", "title": "User CPU time elapsed during execution", "caption": "Exec (s)"}, '
            '{"id": ["user_time_pct1", "user_time_pct2"], "class": "table_obj_value", "title": "User CPU time elapsed by this statement as a percentage of total user CPU time", "caption": "%Total"}'
            ']}, '
        '{"title": "Kernelspace CPU", "caption": "System Time", "columns": ['
            '{"id": ["plan_system_time1", "plan_system_time2"], "class": "table_obj_value", "title": "System CPU time elapsed during planning", "caption": "Plan (s)", "condition": "rusage_planstats"}, '
            '{"id": ["exec_system_time1", "exec_system_time2"], "class": "table_obj_value", "title": "System CPU time elapsed during execution", "caption":"Exec (s)"}, '
            '{"id": ["system_time_pct1", "system_time_pct2"], "class": "table_obj_value", "title": "System CPU time elapsed by this statement as a percentage of total system CPU time", "caption": "%Total"}'
            ']}'
        ']'
    '}]'::jsonb),
(2, 'sqlrusgio', 'sqlkcachehdr', 200, 'Top SQL by reads/writes done by filesystem layer', 'Top SQL by reads/writes done by filesystem layer', NULL, NULL, 'kcache_reads_writes', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_rusage_statements",'
    '"filter": {"type": "exists", "field": "ord_io_bytes"},'
    '"ordering": "ord_io_bytes",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0},'
      '{"id": "queryid", "index": 1},'
      '{"id": "dbname", "index": 1},'
      '{"id": "username", "index": 1}'
    '],'
    '"preview": ['
      '{"id": "hexqueryid", "dataset": "queries"}'
    '],'
    '"limit": "topn",'
    '"columns": ['
      '{"id": "hexqueryid", "class": "hdr mono queryId", "caption": "Query ID", "rowspan": true},'
      '{"id": "dbname", "class": "hdr", "caption": "Database", "rowspan": true},'
      '{"id": "username", "class": "hdr", "caption": "User", "rowspan": true},'
      '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval", "caption": "I"},'
      '{"title": "Filesystem reads", "caption": "Read Bytes", "columns": ['
          '{"id": ["plan_reads1", "plan_reads2"], "class": "table_obj_value", "title": "Filesystem read amount during planning", "caption": "Plan", "condition": "rusage_planstats"},'
          '{"id": ["exec_reads1", "exec_reads2"], "class": "table_obj_value", "title": "Filesystem read amount during execution", "caption": "Bytes"},'
          '{"id": ["reads_total_pct1", "reads_total_pct2"], "class": "table_obj_value", "title": "Filesystem read amount of this statement as a percentage of all statements FS read amount", "caption": "%Total"}'
      ']},'
      '{"title": "Filesystem writes", "caption": "Write Bytes", "columns": ['
          '{"id": ["plan_writes1", "plan_writes2"], "class": "table_obj_value", "title": "Filesystem write amount during planning", "caption": "Plan", "condition": "rusage_planstats"},'
         '{"id": ["exec_writes1", "exec_writes2"], "class": "table_obj_value", "title": "Filesystem write amount during execution", "caption": "Bytes"},'
         '{"id": ["writes_total_pct1", "writes_total_pct2"], "class": "table_obj_value", "title": "Filesystem write amount of this statement as a percentage of all statements FS read amount", "caption": "%Total"}'
      ']}'
    ']'
  '}]'::jsonb),
(2, 'sqllist', 'sqlsthdr', 1300, 'Complete list of SQL texts', 'Complete list of SQL texts', NULL, NULL, 'sql_list', NULL,
  '[{'
    '"type": "row_table", '
    '"source": "queries",'
    '"highlight": ['
      '{"id": "hexqueryid", "index": 0}'
    '],'
    '"columns": ['
        '{"id": "hexqueryid", "class": "mono queryTextId", "caption": "Query ID", "rowspan": true}, '
        '{"id": "query_texts", "class": "mono queryText", "caption": "Query Text"}'
    ']'
  '}]'::jsonb)
;

-- Schema objects section of a differential report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'tblscan', 'objects', 100, 'Top tables by estimated sequentially scanned volume', 'Top tables by estimated sequentially scanned volume', NULL, NULL, 'scanned_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_seq_scan"},'
    '"ordering": "ord_seq_scan",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "rowspan":true, "class": "hdr"},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Table", "class": "table_obj_value", "columns": ['
       '{"caption": "~SeqBytes", "id": ["seqscan_bytes_pretty1", "seqscan_bytes_pretty2"], '
         '"class": "table_obj_value", '
         '"title": "Estimated number of bytes, fetched by sequential scans on this table"}, '
       '{"caption": "SeqScan", "id": ["seq_scan1", "seq_scan2"], '
         '"class": "table_obj_value", '
         '"title": "Number of sequential scans initiated on this table"}, '
       '{"caption": "IxScan", "id": ["idx_scan1", "idx_scan2"], '
         '"class": "table_obj_value", '
         '"title": "Number of index scans initiated on this table"}, '
       '{"caption": "IxFet", "id": ["idx_tup_fetch1", "idx_tup_fetch2"], '
         '"class": "table_obj_value", '
         '"title": "Number of live rows fetched by index scans from this table"} '
     ']},'
     '{"caption": "TOAST", "class": "table_obj_value", "columns": ['
       '{"caption": "~SeqBytes", "id": ["t_seqscan_bytes_pretty1", "t_seqscan_bytes_pretty2"], '
         '"class": "table_obj_value", '
         '"title": "Estimated number of bytes, fetched by sequential scans on TOAST table"}, '
       '{"caption": "SeqScan", "id": ["toastseq_scan1", "toastseq_scan2"], '
         '"class": "table_obj_value", '
         '"title": "Number of sequential scans initiated on TOAST table"}, '
       '{"caption": "IxScan", "id": ["toastidx_scan1", "toastidx_scan2"], '
         '"class": "table_obj_value", '
         '"title": "Number of index scans initiated on TOAST table"}, '
       '{"caption": "IxFet", "id": ["toastidx_tup_fetch1", "toastidx_tup_fetch2"], '
         '"class": "table_obj_value", '
         '"title": "Number of live rows fetched by index scans from TOAST table"} '
     ']}'
    ']'
  '}]'::jsonb),
(2, 'tblfetch', 'objects', 200, 'Top tables by blocks fetched', 'Top tables by blocks fetched', NULL, NULL, 'fetch_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_tables",'
    '"filter": {"type": "exists", "field": "ord_fetch"},'
    '"ordering": "ord_fetch",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Heap", "columns": ['
          '{"caption": "Blks", "id": ["heap_blks_fetch1", "heap_blks_fetch2"], "title": "Number of blocks fetched (read+hit) from this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["heap_blks_proc_pct1", "heap_blks_proc_pct2"], "title": "Heap blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Ix", "columns": ['
          '{"caption": "Blks", "id": ["idx_blks_fetch1", "idx_blks_fetch2"], "title": "Number of blocks fetched (read+hit) from all indexes on this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["idx_blks_fetch_pct1", "idx_blks_fetch_pct2"], "title": "Indexes of blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST", "columns": ['
          '{"caption": "Blks", "id": ["toast_blks_fetch1", "toast_blks_fetch2"], "title": "Number of blocks fetched (read+hit) from this table''s TOAST table (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["toast_blks_fetch_pct1", "toast_blks_fetch_pct2"], "title": "TOAST blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST-Ix", "columns": ['
          '{"caption": "Blks", "id": ["tidx_blks_fetch1", "tidx_blks_fetch2"], "title": "Number of blocks fetched (read+hit) from this table''s TOAST table indexes (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["tidx_blks_fetch_pct1", "tidx_blks_fetch_pct2"], "title": "TOAST table index blocks fetched for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb),
(2, 'tblrd', 'objects', 300, 'Top tables by blocks read', 'Top tables by blocks read', NULL, NULL, 'read_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_tables",'
    '"filter": {"type": "exists", "field": "ord_read"},'
    '"ordering": "ord_read",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Heap", "columns": ['
          '{"caption": "Blks", "id": ["heap_blks_read1", "heap_blks_read2"], "title": "Number of blocks read from this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["heap_blks_read_pct1", "heap_blks_read_pct2"], "title": "Heap blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Ix", "columns": ['
          '{"caption": "Blks", "id": ["idx_blks_read1", "idx_blks_read2"], "title": "Number of blocks read from all indexes on this table", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["idx_blks_read_pct1", "idx_blks_read_pct2"], "title": "Indexes of blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST", "columns": ['
          '{"caption": "Blks", "id": ["toast_blks_read1", "toast_blks_read2"], "title": "Number of blocks read from this table''s TOAST table (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["toast_blks_read_pct1", "toast_blks_read_pct2"], "title": "TOAST blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST-Ix", "columns": ['
          '{"caption": "Blks", "id": ["tidx_blks_read1", "tidx_blks_read2"], "title": "Number of blocks read from this table''s TOAST table indexes (if any)", "class": "table_obj_value"},'
          '{"caption": "%Total", "id": ["tidx_blks_read_pct1", "tidx_blks_read_pct2"], "title": "TOAST table index blocks read for this table as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Hit(%)", "id": ["hit_pct1", "hit_pct2"], "class": "table_obj_value", "title": "Number of heap, indexes, toast and toast index blocks fetched from shared buffers as a percentage of all their blocks fetched from shared buffers and file system"}'
    ']'
  '}]'::jsonb),
(2, 'tbldml', 'objects', 400, 'Top DML tables', 'Top DML tables', NULL, NULL, 'dml_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_dml"},'
    '"ordering": "ord_dml",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Table", "columns": ['
          '{"caption": "Ins", "id": ["n_tup_ins1", "n_tup_ins2"], "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
          '{"caption": "Del", "id": ["n_tup_del1", "n_tup_del2"], "title": "Number of rows deleted", "class": "table_obj_value"},'
          '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST", "columns": ['
          '{"caption": "Ins", "id": ["toastn_tup_ins1", "toastn_tup_ins2"], "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"caption": "Upd", "id": ["toastn_tup_upd1", "toastn_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
          '{"caption": "Del", "id": ["toastn_tup_del1", "toastn_tup_del2"], "title": "Number of rows deleted", "class": "table_obj_value"},'
          '{"caption": "Upd(HOT)", "id": ["toastn_tup_hot_upd1", "toastn_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb),
(2, 'tblud', 'objects', 500, 'Top tables by updated/deleted tuples', 'Top tables by updated/deleted tuples', NULL, NULL, 'upd_del_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_upd"},'
    '"ordering": "ord_upd",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"},'
     '{"caption": "Del", "id": ["n_tup_del1", "n_tup_del2"], "title": "Number of rows deleted", "class": "table_obj_value"},'
     '{"caption": "Vacuum", "id": ["vacuum_count1", "vacuum_count2"], "title": "Number of times this table has been manually vacuumed (not counting VACUUM FULL)", "class": "table_obj_value"},'
     '{"caption": "AutoVacuum", "id": ["autovacuum_count1", "autovacuum_count2"], "title": "Number of times this table has been vacuumed by the autovacuum daemon", "class": "table_obj_value"},'
     '{"caption": "Analyze", "id": ["analyze_count1", "analyze_count2"], "title": "Number of times this table has been manually analyzed", "class": "table_obj_value"},'
     '{"caption": "AutoAnalyze", "id": ["autoanalyze_count1", "autoanalyze_count2"], "title": "Number of times this table has been analyzed by the autovacuum daemon", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'tblupd_np', 'objects', 550, 'Top tables by new-page updated tuples', 'Top tables by new-page updated tuples', 'table_new_page_updates', NULL, 'upd_np_tbl',
  '{"class": "notice", "text": "Top tables by number of rows updated where the successor version goes onto a new heap page, '
  'leaving behind an original version with a t_ctid field that points to a different heap page. '
  'These are always non-HOT updates."}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_upd_np"},'
    '"ordering": "ord_upd_np",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "NP Upd", "id": ["n_tup_newpage_upd1", "n_tup_newpage_upd2"], "title": "Number of rows updated to a new heap page", "class": "table_obj_value"},'
     '{"caption": "%Upd", "id": ["np_upd_pct1", "np_upd_pct2"], "title": "Number of new-page updated rows as a percentage of all rows updated", "class": "table_obj_value"},'
     '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'tblgrw', 'objects', 600, 'Top growing tables', 'Top growing tables', NULL, NULL, 'growth_tbl',
  '{"class": "notice", "text": "Sizes in square brackets are based on pg_class.relpages data instead of pg_relation_size() function"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_growth"},'
    '"ordering": "ord_growth",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Table", "columns": ['
          '{"caption": "Size", "id": ["relsize_pretty1", "relsize_pretty2"], "title": "Table size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
          '{"caption": "Growth", "id": ["growth_pretty1", "growth_pretty2"], "title": "Table size increment during report interval", "class": "table_obj_value"},'
          '{"caption": "Ins", "id": ["n_tup_ins1", "n_tup_ins2"], "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
          '{"caption": "Del", "id": ["n_tup_del1", "n_tup_del2"], "title": "Number of rows deleted", "class": "table_obj_value"},'
          '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
     ']},'
     '{"caption": "TOAST", "columns": ['
          '{"caption": "Size", "id": ["t_relsize_pretty1", "t_relsize_pretty2"], "title": "Table size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
          '{"caption": "Growth", "id": ["toastgrowth_pretty1", "toastgrowth_pretty2"], "title": "Table size increment during report interval", "class": "table_obj_value"},'
          '{"caption": "Ins", "id": ["toastn_tup_ins1", "toastn_tup_ins2"], "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"caption": "Upd", "id": ["toastn_tup_upd1", "toastn_tup_upd2"], "title": "Number of rows updated (includes HOT updated rows)", "class": "table_obj_value"},'
          '{"caption": "Del", "id": ["toastn_tup_del1", "toastn_tup_del2"], "title": "Number of rows deleted", "class": "table_obj_value"},'
          '{"caption": "Upd(HOT)", "id": ["toastn_tup_hot_upd1", "toastn_tup_hot_upd2"], "title": "Number of rows HOT updated (i.e., with no separate index update required)", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb),
(2, 'ixfetch', 'objects', 700, 'Top indexes by blocks fetched', 'Top indexes by blocks fetched', NULL, NULL, 'fetch_idx', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_indexes",'
    '"filter": {"type": "exists", "field": "ord_fetch"},'
    '"ordering": "ord_fetch",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "Index", "id": "indexrelname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Scans", "id": ["idx_scan1", "idx_scan2"], "title": "Number of scans performed on index", "class": "table_obj_value"},'
     '{"caption": "Blks", "id": ["idx_blks_fetch1", "idx_blks_fetch2"], "title": "Number of blocks fetched (read+hit) from this index", "class": "table_obj_value"},'
     '{"caption": "%Total", "id": ["idx_blks_fetch_pct1", "idx_blks_fetch_pct2"], "title": "Blocks fetched from this index as a percentage of all blocks fetched in a cluster", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'ixrd', 'objects', 800, 'Top indexes by blocks read', 'Top indexes by blocks read', NULL, NULL, 'read_idx', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_io_indexes",'
    '"filter": {"type": "exists", "field": "ord_read"},'
    '"ordering": "ord_read",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "Index", "id": "indexrelname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Scans", "id": ["idx_scan1", "idx_scan2"], "title": "Number of scans performed on index", "class": "table_obj_value"},'
     '{"caption": "Blks Reads", "id": ["idx_blks_read1", "idx_blks_read2"], "title": "Number of disk blocks read from this index", "class": "table_obj_value"},'
     '{"caption": "%Total", "id": ["idx_blks_read_pct1", "idx_blks_read_pct2"], "title": "Blocks fetched from this index as a percentage of all blocks read in a cluster", "class": "table_obj_value"}, '
     '{"caption": "Hits(%)", "id": ["idx_blks_hit_pct1", "idx_blks_hit_pct2"], "title": "Index blocks buffer cache hit percentage", "class": "table_obj_value"}'
    ']'
  '}]'::jsonb),
(2, 'ixgrw', 'objects', 900, 'Top growing indexes', 'Top growing indexes', NULL, NULL, 'growth_idx',
  '{"class": "notice", "text": "Sizes in square brackets are based on pg_class.relpages data instead of pg_relation_size() function"}',
  '[{'
    '"type": "row_table",'
    '"source": "top_indexes",'
    '"filter": {"type": "exists", "field": "ord_growth"},'
    '"ordering": "ord_growth",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "Index", "id": "indexrelname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Index", "columns": ['
          '{"id": ["indexrelsize_pretty1", "indexrelsize_pretty2"], "caption": "Size", "title": "Index size, as it was at the moment of last sample in report interval", "class": "table_obj_value"},'
          '{"id": ["growth_pretty1", "growth_pretty2"], "caption": "Growth", "title": "Index size increment during report interval", "class": "table_obj_value"}'
     ']},'
     '{"caption": "Table", "columns": ['
          '{"id": ["tbl_n_tup_ins1", "tbl_n_tup_ins2"], "caption": "Ins", "title": "Number of rows inserted", "class": "table_obj_value"},'
          '{"id": ["tbl_n_tup_upd1", "tbl_n_tup_upd2"], "caption": "Upd", "title": "Number of rows updated (without HOT updated rows)", "class": "table_obj_value"},'
          '{"id": ["tbl_n_tup_del1", "tbl_n_tup_del2"], "caption": "Del", "title": "Number of rows deleted", "class": "table_obj_value"}'
     ']}'
    ']'
  '}]'::jsonb)
;

-- Functions section of a differential report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'func_t', 'funchdr', 100, 'Top functions by total time', 'Top functions by total time', NULL, NULL, 'funcs_time_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_time"},'
    '"ordering": "ord_time",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Function", "id": "funcname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Executions", "id": ["calls1", "calls2"], "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": ["self_time1", "self_time2"], "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": ["m_time1", "m_time2"], "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": ["m_stime1", "m_stime2"], "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb),
(2, 'func_c', 'funchdr', 200, 'Top functions by executions', 'Top functions by executions', NULL, NULL, 'funcs_calls_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_calls"},'
    '"ordering": "ord_calls",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Function", "id": "funcname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Executions", "id": ["calls1", "calls2"], "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": ["self_time1", "self_time2"], "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": ["m_time1", "m_time2"], "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": ["m_stime1", "m_stime2"], "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb),
(2, 'func_trg', 'funchdr', 300, 'Top trigger functions by total time', 'Top trigger functions by total time', 'trigger_function_stats', NULL, 'trg_funcs_time_stat', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_functions",'
    '"filter": {"type": "exists", "field": "ord_trgtime"},'
    '"ordering": "ord_trgtime",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "funcname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Function", "id": "funcname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Executions", "id": ["calls1", "calls2"], "title": "Number of times this function has been called", "class": "table_obj_value"},'
     '{"caption": "Time (s)", "title": "Function execution timing statistics", "columns": ['
          '{"caption": "Total", "id": ["total_time1", "total_time2"], "class": "table_obj_value", "title": "Time spent in this function and all other functions called by it"},'
          '{"caption": "Self", "id": ["self_time1", "self_time2"], "class": "table_obj_value", "title": "Time spent in this function itself, not including other functions called by it"},'
          '{"caption": "Mean", "id": ["m_time1", "m_time2"], "class": "table_obj_value", "title": "Mean total time per execution"},'
          '{"caption": "Mean self", "id": ["m_stime1", "m_stime2"], "class": "table_obj_value", "title": "Mean self time per execution"}'
     ']}'
    ']'
  '}]'::jsonb)
;

-- Vacuum section of a regular report
INSERT INTO report_struct (
  report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
  content, sect_struct)
VALUES
(2, 'vacops', 'vachdr', 100, 'Top tables by vacuum operations', 'Top tables by vacuum operations', NULL, NULL, 'top_vacuum_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_vac"},'
    '"ordering": "ord_vac",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Vacuum count", "id": ["vacuum_count1", "vacuum_count2"], "class": "table_obj_value", "title": "Number of times this table has been manually vacuumed (not counting VACUUM FULL)"},'
     '{"caption": "Autovacuum count", "id": ["autovacuum_count1", "autovacuum_count2"], "class": "table_obj_value", "title": "Number of times this table has been vacuumed by the autovacuum daemon"},'
     '{"caption": "Ins", "id": ["n_tup_ins1", "n_tup_ins2"], "class": "table_obj_value", "title": "Number of rows inserted"},'
     '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "class": "table_obj_value", "title": "Number of rows updated (includes HOT updated rows)"},'
     '{"caption": "Del", "id": ["n_tup_del1", "n_tup_del2"], "class": "table_obj_value", "title": "Number of rows deleted"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "class": "table_obj_value", "title": "Number of rows HOT updated (i.e., with no separate index update required)"}'
    ']'
  '}]'::jsonb),
(2, 'anops', 'vachdr', 200, 'Top tables by analyze operations', 'Top tables by analyze operations', NULL, NULL, 'top_analyze_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_tables",'
    '"filter": {"type": "exists", "field": "ord_anl"},'
    '"ordering": "ord_anl",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "I", "id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "class": "interval"},'
     '{"caption": "Analyze count", "id": ["analyze_count1", "analyze_count2"], "class": "table_obj_value", "title": "Number of times this table has been manually analyzed"},'
     '{"caption": "Autoanalyze count", "id": ["autoanalyze_count1", "autoanalyze_count2"], "class": "table_obj_value", "title": "Number of times this table has been analyzed by the autovacuum daemon"},'
     '{"caption": "Ins", "id": ["n_tup_ins1", "n_tup_ins2"], "class": "table_obj_value", "title": "Number of rows inserted"},'
     '{"caption": "Upd", "id": ["n_tup_upd1", "n_tup_upd2"], "class": "table_obj_value", "title": "Number of rows updated (includes HOT updated rows)"},'
     '{"caption": "Del", "id": ["n_tup_del1", "n_tup_del2"], "class": "table_obj_value", "title": "Number of rows deleted"},'
     '{"caption": "Upd(HOT)", "id": ["n_tup_hot_upd1", "n_tup_hot_upd2"], "class": "table_obj_value", "title": "Number of rows HOT updated (i.e., with no separate index update required)"}'
    ']'
  '}]'::jsonb),
(2, 'ixvacest', 'vachdr', 300, 'Top indexes by estimated vacuum load', 'Top indexes by estimated vacuum load', NULL, NULL, 'top_ix_vacuum_bytes_cnt_tbl', NULL,
  '[{'
    '"type": "row_table",'
    '"source": "top_indexes",'
    '"filter": {"type": "exists", "field": "ord_vac"},'
    '"ordering": "ord_vac",'
    '"highlight": ['
      '{"id": "dbname", "index": 0},'
      '{"id": "tablespacename", "index": 1},'
      '{"id": "schemaname", "index": 1},'
      '{"id": "relname", "index": 1},'
      '{"id": "indexrelname", "index": 1}'
    '],'
    '"limit": "topn",'
    '"columns": ['
     '{"caption": "DB", "id": "dbname", "class": "hdr", "rowspan": true},'
     '{"caption": "Tablespace", "id": "tablespacename", "class": "hdr", "rowspan": true},'
     '{"caption": "Schema", "id": "schemaname", "class": "hdr", "rowspan": true},'
     '{"caption": "Table", "id": "relname", "class": "hdr", "rowspan": true},'
     '{"caption": "Index", "id": "indexrelname", "class": "hdr", "rowspan": true},'
     '{"id": ["1", "2"], "title":["properties.timePeriod1", "properties.timePeriod2"], "caption": "I", "class": "interval"},'
     '{"id": ["vacuum_bytes_pretty1", "vacuum_bytes_pretty2"], "caption": "Vacuum bytes", "class": "table_obj_value", "title": "Estimated implicit vacuum load caused by table indexes"},'
     '{"id": ["vacuum_count1", "vacuum_count2"], "caption": "Vacuum cnt", "class": "table_obj_value", "title": "Vacuum count on underlying table"},'
     '{"id": ["autovacuum_count1", "autovacuum_count2"], "caption": "Autovacuum cnt", "class": "table_obj_value", "title": "Autovacuum count on underlying table"},'
     '{"id": ["avg_indexrelsize_pretty1", "avg_indexrelsize_pretty2"], "caption": "IX size", "class": "table_obj_value", "title": "Average index size during report interval"},'
     '{"id": ["avg_relsize_pretty1", "avg_relsize_pretty2"], "caption": "Relsize", "class": "table_obj_value", "title": "Average relation size during report interval"}'
    ']'
  '}]'::jsonb)
;

-- Settings sections
INSERT INTO report_struct(
    report_id, sect_id, parent_sect_id, s_ord, toc_cap, tbl_cap, feature, function_name, href,
    content, sect_struct)
VALUES
(2, 'definedset', 'settings', 800, NULL, NULL, NULL, NULL, NULL, NULL,
  '[{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "defined_val"},'
    '"columns": ['
     '{"caption": "Defined settings", "columns": ['
          '{"caption": "Setting", "id": "name", "class": "table_obj_value"},'
          '{"caption": "reset_val", "id": "reset_val", "class": "table_obj_value switch_bold"},'
          '{"caption": "Unit", "id": "unit", "class": "table_obj_value"},'
          '{"caption": "Source", "id": "source", "class": "table_obj_value"},'
          '{"caption": "Notes", "id": "notes", "class": "table_obj_value switch_bold"}'
     ']}'
    ']'
  '},'
  '{'
    '"type": "row_table",'
    '"source": "settings",'
    '"filter": {"type": "exists", "field": "default_val"},'
    '"columns": ['
     '{"caption": "Default settings", "columns": ['
          '{"caption": "Setting", "id": "name", "class": "table_obj_value"},'
          '{"caption": "reset_val", "id": "reset_val", "class": "table_obj_value switch_bold"},'
          '{"caption": "Unit", "id": "unit", "class": "table_obj_value"},'
          '{"caption": "Source", "id": "source", "class": "table_obj_value"},'
          '{"caption": "Notes", "id": "notes", "class": "table_obj_value switch_bold"}'
     ']}'
    ']'
  '}]'::jsonb)
;
